/* -------------------------------------------------------
 *
 * (C)Copyright 2001,2007
 * International Business Machines Corporation,
 *
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the
 * following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *
 * - Neither the name of IBM Corporation nor the names of its
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Filename : dma_tag_trace.c
 *
 * Purpose  : Simple reader to collect transaction traces
 *
 * ------------------------------------------------------- */

#define MCONFIG_SPU                            1
#define MCONFIG_STI_EMITTER                    1

#include "common/mambo.h"
#include "emitter/emit_collector.h"
#include "trans_id.h"
#include <memory.h>

EMITTER_CONTROLS eControl;
int reader;

int main (int argc, char **argv)
{
    int         pid = 0;
    EMIT_DATA   *ePtr = NULL;
    FILE        *fp = NULL;

    printf("Reader Started\n");

    if (argc < 3)
        {
            printf("ERROR: Usage: %s simulator-process-id csv-dump-file\n", argv[0]);
            exit(-1);
        }

    pid = atoi(argv[1]);

    /* Open the output file */
    if ((fp = fopen(argv[2], "w")) == NULL)
        {
            fprintf(stderr, "Cannot open file '%s'\n", argv[2]);
            exit(-1);
        }

    /* Attach Reader to simulator */
    reader = EMIT_AttachReader(pid, &eControl, 0, TRUE);
    if (reader == -1)
        {
            fprintf(stderr, "ERROR: Failed to attach reader\n");
            return(0);
        }

    EMIT_ReadyWait(&eControl, reader, TRUE);

    do
        {
            EMIT_AdvanceCurrent(&eControl, reader);
            ePtr = EMIT_CurrentEntry(&eControl, reader);
            /* only print the test cases */
            switch (ePtr->header.head.tag)
                {
                    /*
                     * This record is produced as soon as a tag group status request
                     * is made by the SPU (writing to channel 23, 0x17)
                     */
                case TAG_APU_DMA_TG_STATUS_RQ:
                    {
                        fprintf(fp, "[%lld]SPU(%d):: DMA tag-group status request ",
                                ePtr->apu_dma_tg_status_rq.head.timestamp,
                                ePtr->apu_dma_tg_status_rq.apu);
                        fprintf(fp, "(type=%d, mask=0x%X)\n",
                                ePtr->apu_dma_tg_status_rq.type,
                                ePtr->apu_dma_tg_status_rq.mask);
                    }
                    break;

                    /*
                     * This record is produced as soon as all pending commands for a given
                     * unmasked SPU tag group are completed at the DMA controller
                     */
                case TAG_APU_DMA_TG_COMPLETE:
                    {
                        fprintf(fp, "[%lld]SPU(%d):: DMA tag-group complete ",
                                ePtr->apu_dma_tg_complete.head.timestamp,
                                ePtr->apu_dma_tg_complete.apu);
                        fprintf(fp, "(id=%d)\n",
                                ePtr->apu_dma_tg_complete.gid);
                    }
                    break;

                    /*
                     * This record is produced under two circumstances:
                     *
                     * 1. Whenever an SPU-initiated tag status query is explicitly
                     *    performed (via read of the tag status channel, 24 (0x18)). In
                     *    this case, the "pushed" flag will be 0
                     *
                     * 2. Whenever the SPU tag status is written to the status channel (24,
                     *    0x18) by the DMA controller. Status results will be written
                     *    to this channel to free a potentially blocked read request
                     *    which was issued after requesting a conditional update. In
                     *    this case the "pushed" flag will be 1. Circumstances triggering
                     *    this condition are:
                     *
                     *    A. When the request type made to channel 23 (0x17) is IMMEDIATE
                     *    B. When the request is conditional, but satisfied immediately
                     *       at the time the write was made to the request channel (23, 0x17)
                     *    C. When the request is conditional, and satisfied during DMA
                     *       command completion as SPU DMAQ entries are retired
                     */
                case TAG_APU_DMA_TG_STATUS_QRY:
                    {
                        fprintf(fp, "[%lld]SPU(%d):: DMA tag-group status query ",
                                ePtr->apu_dma_tg_status_qry.head.timestamp,
                                ePtr->apu_dma_tg_status_qry.apu);
                        fprintf(fp, "(tags=0x%X, mask=0x%X, pushed=%s)\n",
                                ePtr->apu_dma_tg_status_qry.tags,
                                ePtr->apu_dma_tg_status_qry.mask,
                                (ePtr->apu_dma_tg_status_qry.pushed ? "YES" : "no"));
                    }
                    break;

                    /*
                     * This record is produced as soon as the PPU writes a request
                     * to the query type status register mapped to the SPU problem
                     * memory space
                     */
                case TAG_PPU_DMA_TG_STATUS_RQ:
                    {
                        fprintf(fp, "[%lld]PPU:: DMA tag-group status request ",
                                ePtr->ppu_dma_tg_status_rq.head.timestamp);
                        fprintf(fp, "(type=%d, mask=0x%X)\n",
                                ePtr->ppu_dma_tg_status_rq.type,
                                ePtr->ppu_dma_tg_status_rq.mask);
                    }
                    break;

                    /*
                     * This record is produced as soon as all pending commands for a given
                     * unmasked PPU tag group are completed at the DMA controller
                     */
                case TAG_PPU_DMA_TG_COMPLETE:
                    {
                        fprintf(fp, "[%lld]PPU:: DMA tag-group complete ",
                                ePtr->ppu_dma_tg_complete.head.timestamp);
                        fprintf(fp, "(id=%d)\n",
                                ePtr->ppu_dma_tg_complete.gid);
                    }
                    break;

                    /*
                     * This record is produced as soon as an unmasked DMA tag group completion
                     * interrupt is raised for the PPU
                     */
                case TAG_PPU_DMA_TG_INTR:
                    {
                        fprintf(fp, "[%lld]PPU:: DMA tag-group interrupt ",
                                ePtr->ppu_dma_tg_intr.head.timestamp);
                        fprintf(fp, "(type=%d, tags=0x%X, mask=0x%X)\n",
                                ePtr->ppu_dma_tg_intr.type,
                                ePtr->ppu_dma_tg_intr.tags,
                                ePtr->ppu_dma_tg_intr.mask);
                    }
                    break;

                    /*
                     * This record is produced as soon as the PPU reads from
                     * the DMA tag status register in the SPU problem state
                     * memory map area
                     */
                case TAG_PPU_DMA_TG_STATUS_QRY:
                    {
                        fprintf(fp, "[%lld]PPU:: DMA tag-group status query ",
                                ePtr->ppu_dma_tg_status_qry.head.timestamp);
                        fprintf(fp, "(tags=0x%X, mask=0x%X)\n",
                                ePtr->ppu_dma_tg_status_qry.tags,
                                ePtr->ppu_dma_tg_status_qry.mask);
                    }
                    break;

                default:
                    /* Ignore */
                    break;
                }
        }
    while (ePtr->header.head.tag != TAG_FOOTER);

    EMIT_Quit(&eControl, reader, TRUE);

    printf("Reader Terminating\n");
    fclose(fp);

    return(1);
}
