/* -------------------------------------------------------
 *
 * (C)Copyright 2001,2007
 * International Business Machines Corporation,
 *
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the
 * following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *
 * - Neither the name of IBM Corporation nor the names of its
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Filename : emitter_reader.c
 *
 * Purpose  :
 *
 *  This contains all the common routines used by
 *  emitter readers.
 *
 * ------------------------------------------------------- */

#include "common/mambo.h"
#include "emitter/atomic.h"
#include "emitter/emitter_data.h"
#include "emitter/emitter_output_t.h"

#ifdef __CYGWIN32__
#include <sys/types.h>
#include <cygwin/shm.h>
#else /* #ifdef __CYGWIN32__ */
#include <sys/shm.h>
#endif /* #ifdef __CYGWIN32__ */

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>

#define DEBUG_EMITTER 0

/* EMIT_AttachReader
 *  Called by reader to attach to the simulator shared memory region
 *  specified by the simulator process ID of pid
 */
int EMIT_AttachReader (int pid, EMITTER_CONTROLS *eControl, int history, boolean verbose)
{
    int reader;
    int countdown;

    if (!eControl)
        {
            fprintf(stderr, "ERROR: eControl is NULL\n");
            return(- 1);
        }

    if (history >= EMIT_DATA_ENTRIES)
        {
            fprintf(stderr, "ERROR: requested history size of %d is greater than the EMITTER data size of %d\n",
                    history, EMIT_DATA_ENTRIES);
            return(- 1);
        }

    eControl->pid = pid;
    eControl->key = pid << 2;

    if (verbose)
        fprintf(stderr,"READER: Shared Memory Key is %d\n", eControl->key);

    /* Sometimes there is a timing issue here waiting for the
     * emitter to be setup.  We will try to attach a few times
     * before we give up */
    countdown = 5;
    while (countdown-- > 0 &&
           ((eControl->eShmid = shmget(eControl->key | KEY_WRITER_DATA, 0, 0)) == -1))
        {
            sleep(1);
        }
    if (eControl->eShmid == -1)
        {
            /* OK we tried, there must be bigger problems */
            fprintf(stderr, "ERROR: shmget returned -1\n");
            return(-1);
        }

    {
        /* get shared memory area for emitter data */
        VOID *rc;
        rc = shmat(eControl->eShmid, 0, SHM_RDONLY);
        if (rc == CAST(VOID *,-1))
            {
                fprintf(stderr, "ERROR: shmat returned -1\n");
                return(-1);
            }
        eControl->edata = CAST(EMIT_DATA *,rc);
    }

    if (verbose)
        fprintf(stderr,"READER: Writer Data found at 0x%08lX\n",(long) eControl->edata);

    eControl->rShmid = shmget(eControl->key | KEY_READER_DATA, 0, 0);
    if (eControl->rShmid == -1)
        {
            fprintf(stderr, "ERROR: shmget returned -1\n");
            return(-1);
        }

    {
        VOID *rc;
        rc = shmat(eControl->rShmid, 0, 0);
        if (rc == CAST(VOID *,-1))
            {
                fprintf(stderr, "ERROR: shmat returned -1\n");
                return(- 1);
            }
        eControl->rdata = CAST(READER_DATA *,rc);
    }

    if (verbose)
        fprintf(stderr,"READER: Reader Data found at 0x%08lX\n",(long) eControl->rdata);

    /* Get the index of this reader */
    ATOMIC_ADD(1, eControl->rdata->readers_attached);
    reader = ATOMIC_READ(eControl->rdata->readers_attached) - 1;
    if (reader >= MAX_READERS)
        {
            fprintf(stderr, "ERROR: MAXIMUM readers of %d exceeded\n", MAX_READERS);
            return(- 1);
        }

    eControl->rdata->reader[reader].first = 0;
    eControl->rdata->reader[reader].current = 0;
    eControl->rdata->reader[reader].history = history;
    eControl->rdata->reader[reader].ready = TRUE;

    fprintf(stderr,"READER: Reader #%d attached\n", reader);
    return(reader);
}

/* EMIT_SetReadersExpected
 *  Called by a reader to specify how many readers are expected to
 *  attach to the running simulator
 *  Returns 1 on success and 0 on failure
 */
int EMIT_SetReadersExpected (int pid, int readers, boolean verbose)
{
    int key;
    int rShmid;
    READER_DATA *rdata;

    key = pid << 2;

    if (verbose)
        fprintf(stderr,"READER: Shared Memory Key is %d\n", key);

    rShmid = shmget(key | KEY_READER_DATA, 0, 0);
    if (rShmid == -1)
        {
            fprintf(stderr, "ERROR: shmget returned -1\n");
            return(0);
        }

    {
        VOID *rc;
        rc = shmat(rShmid, 0, 0);
        if (rc == CAST(VOID *,-1))
            {
                fprintf(stderr, "ERROR: shmat returned -1\n");
                return(0);
            }
        rdata = CAST(READER_DATA *,rc);
    }

    if (readers >= MAX_READERS)
        {
            fprintf(stderr, "ERROR: readers_expected %d exceeded MAXIMUM readers %d\n",
                    readers, MAX_READERS);
            return(0);
        }

    rdata->readers_expected = readers;

    shmdt(rdata);            /* We are done so detach from shared memory */

    return(1);
}


/* EMIT_AdvanceCurrent
 *  Called by reader to advance the active reader pointer by one.
 */
void EMIT_AdvanceCurrent (EMITTER_CONTROLS *eControl, int rdr)
{
    int min;
    sint64 tmp_min_first = -1;

#if DEBUG_EMITTER
    fprintf(stdout, "Advance current for rdr=%d\n", rdr); fflush(stdout);
#endif /* #if DEBUG_EMITTER */

    /* We need to wait before we bump the readers first entry until the
     * writer has written something there */
    while (EINDEX(eControl->rdata->reader[rdr].current + 1) == EINDEX(eControl->rdata->next_write))
        {
#if DEBUG_EMITTER
            fprintf(stderr, "Reader waiting for advance of writer rdr=%d\n", rdr);
            fprintf(stderr, "Next-Write=%ld, Reader-Current+1=%ld",
                    (long)EINDEX(eControl->rdata->next_write),
                    (long)EINDEX(eControl->rdata->reader[rdr].current + 1));
            fflush(stderr);
#endif /* #if DEBUG_EMITTER */
            usleep(50);
        }
    eControl->rdata->reader[rdr].current++;

#if DEBUG_EMITTER
    if (eControl->edata[EINDEX(eControl->rdata->reader[rdr].current)].header.head.seqid !=
        eControl->rdata->reader[rdr].current)
        {
            fprintf(stderr, "ERROR: seqid %lld is not equal to current index %lld\n",
                    eControl->edata[EINDEX(eControl->rdata->reader[rdr].current)].header.head.seqid,
                    eControl->rdata->reader[rdr].current);
            exit(-1);

        }
#endif /* #if DEBUG_EMITTER */

    /*
     * Maintain requested number history entries
     */
    if ((eControl->rdata->reader[rdr].current - eControl->rdata->reader[rdr].first) >
        eControl->rdata->reader[rdr].history)
        eControl->rdata->reader[rdr].first++;

    /*
     * Iterate through all the readers and compute the minimum
     * first entry (oldest entry still needed by a reader)
     */
    for (min = 0; min < MAX_READERS; min++)
        {
            /* Check if this reader is still attached (ready) and if its
             * first entry is the minimum of all the readers.  If so,
             * set the global minimum field to this value. */
            if (eControl->rdata->reader[min].ready &&
                ((tmp_min_first == -1) ||
                 (eControl->rdata->reader[min].first < tmp_min_first)))
                tmp_min_first = eControl->rdata->reader[min].first;
        }
    eControl->rdata->min_first = tmp_min_first;

    if (tmp_min_first == -1)
        {
            fprintf(stderr, "ERROR: tmp_min_first == -1, readers:%d, first=%lld, ready=%d\n",
                    eControl->rdata->readers_expected,
                    eControl->rdata->reader[0].first,
                    eControl->rdata->reader[0].ready);
            exit(-1);
        }
}

/* EMIT_ReadyWait
 *  Called by reader to wait for the header to be written out
 *  This indicates the start of the emitter stream
 */

#define MAX_WAIT_TIME_IN_SECONDS 90

void EMIT_ReadyWait (EMITTER_CONTROLS *eControl, int rdr, boolean verbose)
{
    READER_DATA *sr = eControl->rdata;
    int countdown;

    if (verbose)
        fprintf(stderr,"READER(%d): Reader Waiting for Header %d in Emitter Data\n", rdr, TAG_HEADER);

    /* First wait for anything to be emitted. */
    /* time out after some number of seconds. Convert seconds to microseconds */
    countdown = MAX_WAIT_TIME_IN_SECONDS * 1000000  / 50;
    while ((eControl->rdata->next_write == 0) && (countdown-- > 0))
        {
            usleep(50);
        }

    /* check if we found something, or if we timed-out */
    if (countdown <= 0)
        {
            fprintf(stderr,"READER(%d): waited %d seconds; no emitter data ... Exiting\n",
                    rdr, MAX_WAIT_TIME_IN_SECONDS);

            EMIT_Quit(eControl, rdr, verbose);
        }

    /* The first thing emitted should be a header record */
    if (EMIT_CurrentEntry(eControl, rdr)->header.head.tag != TAG_HEADER)
        {
            fprintf(stderr,"READER(%d): Did not find TAG_HEADER (%d) found %d ... Exiting\n",
                    rdr, TAG_HEADER,
                    EMIT_CurrentEntry(eControl, rdr)->header.head.tag);

            EMIT_Quit(eControl, rdr, verbose);
        }

    if (verbose)
        fprintf(stderr,"READER(%d): Reader Found the Header in Emitter Data\n", rdr);

    /* Wait number of readers to be greater than 0 and equal to
     * the number of attached readers */
    countdown = 20;
    while ((countdown-- > 0) &&
           ((sr->readers_expected == 0) ||
            ((sr->readers_attached != sr->readers_expected))))
        {
            sleep(1);
            if (verbose)
                fprintf(stderr,"READER(%d): ReadyWait still waiting for all readers to attach (countdown=%d)\n",
                        rdr, countdown);
        }
    if (countdown <= 0)
        {
            fprintf(stderr,"READER(%d): ReadyWait failed waiting 20 seconds for all readers to attach. Exiting...\n", rdr);
            EMIT_Quit(eControl, rdr, verbose);
        }
}

/* EMIT_Quit
 *  Called by reader to detach and clean up the shared memory
 *  buffers
 */
void EMIT_Quit (EMITTER_CONTROLS *eControl, int rdr, boolean verbose)
{
    if (verbose)
        fprintf(stderr,"READER: Emitter #%d exiting\n", rdr);

    eControl->rdata->reader[rdr].ready = FALSE;/* Mark this reader as DEAD */

    ATOMIC_SUB(1, eControl->rdata->readers_expected);
    ATOMIC_SUB(1, eControl->rdata->readers_attached);

    /* Detach from the shared memory */
    shmdt(eControl->rdata);
    shmdt(eControl->edata);
}
