/* -------------------------------------------------------
 *
 * (C)Copyright 2001,2007
 * International Business Machines Corporation,
 *
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the
 * following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *
 * - Neither the name of IBM Corporation nor the names of its
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Filename : memtrace.c
 *
 * Purpose  : emitter reader that dumps a memory access
 *            trace
 *
 * ------------------------------------------------------- */

#include "common/mambo.h"
#include "emitter/emit_collector.h"

#include <string.h>

void finish_signal(int sig);

unsigned long long inst = 0;
unsigned long long group_index;
unsigned long long inst_per_sample = 1024;
unsigned long long idle_start = 0;
unsigned long long idle_end = 0;

struct cpu_mode
{
    int kernel_mode;
    int bit_64_mode;
    int inst_relo_on;
};
typedef struct cpu_mode CPU_MODES;

CPU_MODES *cpu_mode;

struct Config
{
    int cpus;
    unsigned long long mem_size;
    unsigned long long clock_frequency;
    int disk_latency;
    float cpi;
    int l1d_block_size;
    int l1i_block_size;
} config;

EMITTER_CONTROLS eControl;
int reader;

FILE *outfile;


struct outrecord
{
    uint64 eaddr;
    char length;
};
typedef struct outrecord outrecord;

int main (int argc, char **argv)
{
    int pid;
    int end = 0;
    int i;
    EMIT_DATA *eEntry;          /* Used in loop */
    char today[50];
    time_t runtime;
    outrecord *outrec;
    int lines_written = 0;

    outrec = (outrecord *) malloc(sizeof(outrecord));


#define MAX_LINES_TO_WRITE 800000000

    printf("Reader Started\n");

    if (argc < 3)
        {
            printf("ERROR: Usage: %s simulator-process-id tracefilename\n", argv[0]);
            exit(-1);
        }

    pid = atoi(argv[1]);
    outfile = fopen(argv[2], "w");
    printf("Reader opened file %s for output\n",argv[2]);fflush(stdout);
    ASSERT(outfile != NULL);

    /* Attach Reader to simulator */
    reader = EMIT_AttachReader(pid, &eControl, 0, TRUE);
    if (reader == -1)
        {
            fprintf(stderr, "ERROR: Failed to attach reader\n");
            finish_signal(0);
        }

    /* if (signal(SIGHUP, finish_signal) == SIG_ERR) perror("signal(SIGHUP)"); */
    if (signal(SIGINT, finish_signal) == SIG_ERR)
        perror("signal(SIGINT)");
    if (signal(SIGQUIT, finish_signal) == SIG_ERR)
        perror("signal(SIGQUIT)");
    if (signal(SIGILL, finish_signal) == SIG_ERR)
        perror("signal(SIGILL)");
    if (signal(SIGTRAP, finish_signal) == SIG_ERR)
        perror("signal(SIGTRAP)");
    if (signal(SIGABRT, finish_signal) == SIG_ERR)
        perror("signal(SIGABRT)");

    EMIT_ReadyWait(&eControl, reader, TRUE);

    /* First read in the header and config records */
    end = 0;
    while (!end)
        {
            EMIT_AdvanceCurrent(&eControl, reader);
            eEntry = EMIT_CurrentEntry(&eControl, reader);
            switch (eEntry->inst.head.tag)
                {
                case TAG_HEADER:
                    break;
                case TAG_CONFIG:
                    config.cpus = eEntry->config.cpus;
                    config.mem_size = eEntry->config.mem_size;
                    config.clock_frequency = eEntry->config.clock_frequency;
                    config.disk_latency = eEntry->config.disk_latency;
                    config.cpi = eEntry->config.cpi;
                    config.l1d_block_size = eEntry->config.l1d_block_size;
                    config.l1i_block_size = eEntry->config.l1i_block_size;

                    time(&runtime);
                    strftime(today, 50, "(%m/%d/%Y @ %X)", localtime(&runtime));
                    printf("Stripstats Creation Time: %s\n"
                           "      Simulator Configuration:\n"
                           "------------------------------\n"
                           "          Num CPUS: %d\n"
                           "          Mem Size: %lld\n"
                           "   Clock Frequency: %lld\n"
                           "      Disk Latency: %d\n"
                           "               CPI: %f\n"
                           "    L1D Block Size: %d\n"
                           "    L1I Block Size: %d\n",
                           today,
                           config.cpus,
                           config.mem_size,
                           config.clock_frequency,
                           config.disk_latency,
                           config.cpi,
                           config.l1d_block_size,
                           config.l1i_block_size);

                    /* Allocate this based on the number of cpus */
                    ASSERT(cpu_mode == NULL);
                    cpu_mode = (CPU_MODES*) malloc(sizeof(CPU_MODES)*config.cpus);
                    if (cpu_mode == NULL)
                        {
                            printf("ERROR: failed to allocate the cpu_mode array, cpus=%d\n", config.cpus);
                            exit(-1);
                        }
                    for (i = 0; i < config.cpus; i++)
                        {
                            cpu_mode[i].kernel_mode = 0;
                            cpu_mode[i].bit_64_mode = 0;
                            cpu_mode[i].inst_relo_on = 0;
                        }

                    end = 1;    /* once we get the config record we are done */
                    break;
                default:
                    fprintf(stderr, "The first records should be the header and config but got tag=%d (0x%X)\n",
                            eEntry->inst.head.tag, eEntry->inst.head.tag);
                    exit (-1);
                    break;
                }
        }
    end = 0;

    printf("Emitter beginning to read records!\n");fflush(stdout);
    while (!end)
        {
            EMIT_AdvanceCurrent(&eControl, reader);

            eEntry = EMIT_CurrentEntry(&eControl, reader);

            if (lines_written > MAX_LINES_TO_WRITE)
                {
                    printf("Emitter Exiting because maximum number of records were written!\n");
                    finish_signal(0);
                }

            /* Count this event */
            switch (eEntry->header.head.tag)
                {
                case TAG_MEMWRITE:
                    memcpy(outrec, &eEntry->io.eaddr, sizeof(outrecord));
                    outrec->length |= 0x80;
                    if ((i = fwrite(outrec, sizeof(uint64)+sizeof(char), 1, outfile)) != 1)
                        {
                            printf("Error writing store record!\n");
                            finish_signal(0);
                        }
                    break;
                case TAG_MEMREAD:
                    if ((i = fwrite(&eEntry->io.eaddr, sizeof(uint64)+sizeof(char), 1, outfile)) != 1)
                        {
                            printf("Error writing load record!\n");
                            finish_signal(0);
                        }
                    break;

                default:
                    break;
                }
        }

    finish_signal(0);
    return(- 1);
}

void finish_signal(int sig)
{

    fclose(outfile);

    free(cpu_mode);

    EMIT_Quit(&eControl, reader, TRUE);

    printf("Reader Terminating\n");

    exit(0);
}
