/* -------------------------------------------------------
 *
 * (C)Copyright 2001,2006
 * International Business Machines Corporation,
 *
 * All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the
 * following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *
 * - Neither the name of IBM Corporation nor the names of its
 *   contributors may be used to endorse or promote products
 *   derived from this software without specific prior written
 *   permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Filename : trans_id.c
 *
 * Purpose    : Operations on transaction ID's
 *
 * ------------------------------------------------------- */

#include "common/mambo.h"

#include <stdio.h>
#include <string.h>                               /* strncpy */
#include <stdlib.h>
#include <ctype.h>
#include "trans_id.h"

#ifndef ASSERT
#include <assert.h>
#define ASSERT  assert
#endif /* #ifndef ASSERT */

#define MAX_PRINT_BUF_LEN 256

static TRID_Major   NextMajorID = 0;

/*___________________________________________________________
 * MAJOR ID Methods
 */

/**
 * TRID_SetDesc: Set the fragment description
 *
 * \param   t        Pointer to a TRID
 * \param   desc     Fragment Description (or NULL)
 */
void TRID_SetDesc(TRID *t, const char *desc)
{
    ASSERT(t != NULL);
    if (desc != NULL)
        {
            strncpy(t->desc, desc, sizeof(t->desc));
            t->desc[sizeof(t->desc) - 1] = '\0';
        }
    else
        {
            t->desc[0] = '\0';
        }
}

/**
 * TRID_GetMajor: Retrieve the Major ID
 *
 * \param   t        Pointer to a TRID
 * \return  Major ID of TRID
 */
TRID_Major TRID_GetMajor(TRID* t)
{
    ASSERT(t != NULL);
    return(t->mid);
}

/**
 * TRID_SetDone: Set/clear the TRID done flag
 *
 * \param   t        Pointer to a TRID
 * \param   done     TRUE if done, FALSE otherwise
 */
void TRID_SetDone(TRID *t, boolean done)
{
    ASSERT(t != NULL);
    t->done = (done ? 0xFF : 0);
}

/**
 * TRID_IsDone: Return TRUE if TRID has been marked completed
 *
 * \param   t        Pointer to a TRID
 * \return TRUE if done, FALSE otherwise
 */
boolean TRID_IsDone(TRID *t)
{
    ASSERT(t != NULL);
    return(t->done == 0 ? FALSE : TRUE);
}

/*___________________________________________________________
 * FRAGMENT Methods
 */

/**
 * TRID_SetFragDesc: Set the fragment description
 *
 * \param   t        Pointer to a TRID
 * \param   desc     Fragment Description (or NULL)
 */
void TRID_SetFragDesc(TRID *t, const char *desc)
{
    ASSERT(t != NULL);
    ASSERT(t->level > 0);
    if (desc != NULL)
        {
            strncpy(t->frags[t->level - 1].fdes.c, desc, sizeof(TRID_FragDesc));
            t->frags[t->level - 1].fdes.c[sizeof(TRID_FragDesc) - 1] = '\0';
        }
    else
        {
            t->frags[t->level - 1].fdes.c[0] = '\0';
        }
}

/**
 * TRID_SetFrag: Set the fragment ID in the current branch
 *
 * \param   t        Pointer to a TRID
 * \param   fid      Fragment ID
 * \param   desc     Fragment Description (or NULL)
 */
void TRID_SetFrag(TRID *t, TRID_Frag fid, const char *desc)
{
    ASSERT(t != NULL);
    ASSERT(t->level > 0);
    t->frags[t->level - 1].fid = fid;
    TRID_SetFragDesc(t, desc);
}

/**
 * TRID_NewFrag: Increment the fragment ID in the current branch
 *
 * \param   tp        Pointer to a TRID parent
 * \param   tc        Pointer to a TRID child (initialized by this method)
 * \param   desc      Description (or NULL)
 */
void TRID_NewFrag(TRID *tp, TRID *tc, const char *desc)
{
    ASSERT(tp != NULL);
    ASSERT(tp->level < TRID_MAX_BRANCHES);

    if (tp->frags[tp->level].fid == TRID_INVALID_FRAG)
        {
            tp->frags[tp->level].fid = 0;
        }
    else
        {
            ++(tp->frags[tp->level].fid);
        }

    *tc = *tp;
    tc->level = tp->level + 1;
    TRID_SetFragDesc(tc, desc);
}

/**
 * TRID_GetFragID: Get the Fragment ID at the tail of the list
 *
 * \param   t        Pointer to a TRID
 * \return Fragment ID in current branch
 */
TRID_Frag TRID_GetFragID(TRID* t)
{
    ASSERT(t != NULL);
    return(t->level == 0 ? TRID_INVALID_FRAG : t->frags[t->level - 1].fid);
}

/**
 * TRID_GetFragDesc: Get the Fragment Description at the tail of the list
 *
 * \param   t        Pointer to a TRID
 * \param   desc     (OUT) Pointer to description string to fill
 * \param   n        Number of bytes in description string storage
 */
void TRID_GetFragDesc(TRID *t, char *desc, unsigned int n)
{
    ASSERT(t != NULL);
    ASSERT(t->level > 0);
    ASSERT(desc != NULL);
    ASSERT(n != 0);

    strncpy(desc, t->frags[t->level - 1].fdes.c, n);
    desc[n - 1] = '\0';
}

/*___________________________________________________________
 * BRANCH Methods
 */

/**
 * TRID_DeleteAllBranches: Delete the most recent fragment branch
 *
 * \param   t        Pointer to a TRID
 */
void TRID_DeleteAllBranches(TRID *t)
{
    uint32  i = 0;  /* Current branch loop index */
    ASSERT(t != NULL);

    for (i = t->level; i < TRID_MAX_BRANCHES; ++i)
        {
            t->frags[i].fid = TRID_INVALID_FRAG;
            t->frags[i].fdes.c[0] = '\0';
        }
}

/**
 * TRID_GetBranchLevel: Determine the level of the given TRID
 *
 * \param   t        Pointer to a TRID
 * \return  Level of the TRID
 */
uint32 TRID_GetBranchLevel(TRID *t)
{
    ASSERT(t != NULL);
    return(t->level);
}

/**
 * TRID_IsFragmented: Return TRUE if the transaction has branches
 *
 * \param   t        Pointer to a TRID
 * \return TRUE if the transaction has branches
 */
boolean TRID_IsFragmented(TRID *t)
{
    ASSERT(t != NULL);
    return(t->frags[t->level].fid != TRID_INVALID_FRAG);
}

/*___________________________________________________________
 * TRID Methods
 */

/**
 * TRID_Initialize: Initialze a TRID
 *
 * \param   t        Pointer to a TRID to be initialized
 * \param   desc     Description (or NULL)
 */
void TRID_Initialize(TRID* t, const char *desc)
{
    ASSERT(t != NULL);
    t->mid     = NextMajorID++;
    t->level   = 0;
    t->done    = 0;
    TRID_SetDesc(t, desc);
    TRID_DeleteAllBranches(t);
}

/**
 * TRID_Invalidate: Invalidate a TRID
 *
 * \param   t        Pointer to a TRID to be invalidated
 */
void TRID_Invalidate(TRID* t)
{
    ASSERT(t != NULL);
    t->mid     = TRID_INVALID_MAJOR;
    t->level   = 0;
    t->done    = 0;
    t->desc[0] = '\0';
    TRID_DeleteAllBranches(t);
}

/**
 * TRID_IsValid: Query the validity of a TRID
 *
 * \param   t        Pointer to a TRID to be checked
 *
 * \return TRUE if TRID is valid
 */
boolean TRID_IsValid(TRID* t)
{
    ASSERT(t != NULL);
    return(t->mid == TRID_INVALID_MAJOR ? FALSE : TRUE);
}

/**
 * TRID_Copy: Copy a TRID from source to destination
 *
 * \param   dst     Pointer to Destination TRID
 * \param   src     Pointer to Source TRID
 */
void TRID_Copy(TRID *dst, TRID *src)
{
    ASSERT(dst != NULL);
    ASSERT(src != NULL);
    *dst = *src;
}

/**
 * TRID_Match: Determine whether two TRID's match
 *
 * \param   t1      First TRID
 * \param   t2      Second TRID
 * \param   levels  Resolution (number of levels to match, 0 == FULL)
 * \return  TRUE if match, FALSE otherwise
 */
boolean TRID_Match(TRID *t1, TRID *t2, uint32 levels)
{
    ASSERT(t1 != NULL);
    ASSERT(t2 != NULL);

    /* Major ID's must always compare equal */
    if (t1->mid != t2->mid)
        {
            return(FALSE);
        }

    /* Fail unequal levels in FULL comparison case */
    if ((levels == 0) && (t1->level != t2->level))
        {
            return(FALSE);
        }

    /* Perform match check */
    {
        uint32  chk_level;
        uint32  i;

        /* Determine the number of levels to check */
        if (levels == 0)
            {
                ASSERT(t1->level == t2->level);
                chk_level = t1->level;
            }
        else
            {
                chk_level = (t1->level < (levels - 1) ? t1->level : (levels - 1));
                chk_level = (t2->level < chk_level ? t2->level : chk_level);

                /* Fail if we're checking fewer levels than requested and the TRID levels differ */
                if ((chk_level < (levels - 1)) && (t1->level != t2->level))
                    {
                        return(FALSE);
                    }
            }

        /* Check the fragment id's at each level */
        for (i = 0; (i < chk_level); ++i)
            {
                if (t1->frags[i].fid != t2->frags[i].fid)
                    {
                        return(FALSE);
                    }
            }
    }

    return(TRUE);
}

/**
 * TRID_LessThan: Determine whether t1 < t2
 *
 * \param   t1      First TRID
 * \param   t2      Second TRID
 * \return  TRUE if t1 < t2, FALSE otherwise
 */
boolean TRID_LessThan(TRID *t1, TRID *t2)
{
    ASSERT(t1 != NULL);
    ASSERT(t2 != NULL);

    /* Simple MID check */
    if (t1->mid != t2->mid)
        {
            return(t1->mid < t2->mid ? TRUE : FALSE);
        }
    else
        {
            uint32  chk_level;
            uint32  i;

            /* Determine the number of levels to check */
            chk_level = (t1->level < t2->level ? t1->level : t2->level);

            /* Check the fragment id's at each level */
            for (i = 0; i < chk_level; ++i)
                {
                    if (t1->frags[i].fid != t2->frags[i].fid)
                        {
                            return(t1->frags[i].fid < t2->frags[i].fid ? TRUE : FALSE);
                        }
                }

            /* If they've matched up to this point, we return t1 < t2 if t1 has
             * fewer levels than t2.
             */
            return(t1->level < t2->level ? TRUE : FALSE);
        }
}

/**
 * TRID_ToString: Convert a TRID to a string
 *
 * \param   t   Pointer to a TRID
 * \param   s   String to fill with TRID representation
 * \param   len Maximum length of s
 */
void TRID_ToString(TRID* t, char* s, uint32 len)
{
    uint32  i = 0;  /* Current branch loop index */
    uint32  p = 0;  /* Current position in s */

    ASSERT(t != NULL);
    ASSERT(s != NULL);
    ASSERT(len != 0);

    if (t->mid == TRID_INVALID_MAJOR)
        {
            p = snprintf(s, len, "?");
        }
    else
        {
            p = snprintf(s, len, "%lld", t->mid);
        }

    for (i = 0; (i < t->level) && (p < len); ++i)
        {
            TRID_Frag   fid = t->frags[i].fid;

            if (fid == TRID_INVALID_FRAG)
                {
                    p += snprintf(&s[p], len - p, ".?");
                }
            else
                {
                    p += snprintf(&s[p], len - p, ".%d", fid);
                }
        }
    s[len - 1] = '\0';
}

/**
 * TRID_DescToString: Convert a TRID description to a string
 *
 * \param   t       Pointer to a TRID
 * \param   s       String to fill with TRID representation
 * \param   len     Maximum length of s
 * \param   delim   Delimiter character to use
 */
void TRID_DescToString(TRID* t, char* s, uint32 len, char delim)
{
    uint32  i = 0;  /* Current branch loop index */
    uint32  p = 0;  /* Current position in s */

    ASSERT(t != NULL);
    ASSERT(s != NULL);
    ASSERT(len != 0);

    if (t->desc[0] == '\0')
        {
            p = snprintf(s, len, "?");
        }
    else
        {
            p = snprintf(s, len, "%s", t->desc);
        }

    for (i = 0; (i < t->level) && (p < len); ++i)
        {
            TRID_FragDesc *fdes = &(t->frags[i].fdes);

            if (fdes->c[0] == '\0')
                {
                    p += snprintf(&s[p], len - p, "%c?", delim);
                }
            else
                {
                    p += snprintf(&s[p], len - p, "%c%s", delim, fdes->c);
                }
        }
    s[len - 1] = '\0';
}

/**
 * TRID_FromString: Convert a string to a TRID
 *
 * \param   t   Pointer to a TRID
 * \param   s   String to parse
 */
void TRID_FromString(TRID* t, char* s)
{
    uint32   i;      /* Current branch loop index */
    char     *p = s; /* Current position in s */

    ASSERT(t != NULL);
    ASSERT(s != NULL);

    t->mid   = strtoull(p, &p, 10);
    t->level = 0;
    for (i = 0; i < TRID_MAX_BRANCHES; ++i)
        {
            /* Skip to next run of digits */
            while (*p && (!isdigit(*p)))
                {
                    ++p;
                }

            /* If we're at a digit sequence, parse to the frag id */
            if (*p)
                {
                    t->frags[t->level].fid = strtoul(p, &p, 10);
                    ++(t->level);
                }
            /* Otherwise, we initialize with the invalid fragment id */
            else
                {
                    t->frags[t->level].fid = TRID_INVALID_FRAG;
                }
        }
}

/**
 * TRID_DescFromString: Convert a string to a TRID description
 *
 * \param   t       Pointer to a TRID
 * \param   s       String to parse
 * \param   delim   Character to interpret as a delimiter
 */
void TRID_DescFromString(TRID* t, char* s, char delim)
{
}

/**
 * TRID_Print: Send the string representation to the output stream
 *
 * \param   t   Pointer to a TRID
 * \param   fp  Output stream pointer
 */
void TRID_Print(FILE* fp, TRID* t)
{
    char    id_buf[MAX_PRINT_BUF_LEN];
    char    desc_buf[MAX_PRINT_BUF_LEN];

    ASSERT(fp != NULL);
    ASSERT(t != NULL);

    TRID_ToString(t, id_buf, sizeof(id_buf));
    TRID_DescToString(t, desc_buf, sizeof(desc_buf), '.');
    fprintf(fp, "%s [%s]", id_buf, desc_buf);
}

