/* -------------------------------------------------------
 *
 * Licensed Materials - Property of IBM.
 * (C) Copyright IBM Corporation 2001, 2007
 * All Rights Reserved.
 *
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corporation.
 *
 * Filename : basictypes_t.h
 *
 * Purpose  : Generic Types used in the Mambo Simulator
 *
 * ------------------------------------------------------- */

#ifndef _BASICTYPES_T_H_
#define _BASICTYPES_T_H_

/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

/* We want to use boolean (for boolean types) wherever it is
 * appropriate.  This should make the intent more clear, even if
 * it doesn't save memory or time (tough it might).
 *
 * If boolean has not been defined as a type, define it.
 *
 * Normally, we are compiling in C, so we define boolean to be an
 * int (the most efficient representation).  We could make it a
 * char or short, but that might be less efficient in run time
 * access.
 *
 * The alternative is that we are compiling with C++.  In this
 * case, we take advantage of the tighter data type rules for C++
 * and define boolean as its own class.  We want to make sure
 * that we do not intermix booleans and ints or other types (we
 * found cases where FALSE was being used instead of NULL).  To
 * allow booleans to be tested and mixed with expressions that
 * the C++ compiler knows as "bool", we allow booleans to be
 * converted to bools.
 */

#if !defined(__cplusplus) || !defined(TYPE_TEST)

typedef unsigned int boolean;

/* AIX will have defined TRUE and FALSE by now; most systems have not. */
#ifndef FALSE
#define FALSE           (0)
#endif /* #ifndef FALSE */
#ifndef TRUE
#define TRUE            (1)
#endif /* #ifndef TRUE */

#else /* #if !defined(__cplusplus) || !defined(TYPE_TEST) */

class boolean
{
 private:
    /* a boolean has a private bool value */
    int value;

 public:
    /* constructor to allow TRUE and FALSE to be defined */
    boolean(int i) { value = i; }

    /* constructor for boolean variables to be defined */
    boolean() { value = 0; }

    /* allow a boolean to be converted to a bool */
    /* this allows booleans to be tested, and to be
       combined with bools */
    operator bool() { return(value); }
    operator bool() const { return(value); }
    operator bool() volatile { return(value); }

    bool operator!() { return(!value); }
    bool operator||(const boolean x) { return(value || x.value); }
    bool operator&&(const boolean x) { return(value && x.value); }
    bool operator||(const bool x) { return(value || x); }
    bool operator&&(const bool x) { return(value && x); }
};

/* if TRUE/FALSE are already defined, undefine them, so we can
   give our own defintion. */
#ifdef TRUE
#undef TRUE
#endif /* #ifdef TRUE */
#ifdef FALSE
#undef FALSE
#endif /* #ifdef FALSE */
const boolean FALSE(0);
const boolean TRUE(1);
#endif /* #if !defined(__cplusplus) || !defined(TYPE_TEST) */


/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */


#if !defined(_AIX)
#if !defined(_SIMICS_API_H)
#ifndef HAVE_INT8
typedef signed char       int8;
#define HAVE_INT8
#endif /* #ifndef HAVE_INT8 */

typedef unsigned char     uchar;

#ifndef HAVE_INT16
typedef signed short      int16;
#define HAVE_INT16
#endif /* #ifndef HAVE_INT16 */

#ifndef HAVE_INT32
typedef signed int        int32;
#define HAVE_INT32
#endif /* #ifndef HAVE_INT32 */
#endif /* #if !defined(_SIMICS_API_H) */
#endif /* #if !defined(_AIX) */

#if !defined(_SIMICS_API_H)
#ifndef HAVE_UINT8
typedef unsigned char      uint8;
#define HAVE_UINT8
#endif /* #ifndef HAVE_UINT8 */

#ifndef HAVE_UINT16
typedef unsigned short     uint16;
#define HAVE_UINT16
#endif /* #ifndef HAVE_UINT16 */

#ifndef HAVE_UINT32
typedef unsigned int       uint32;
#define HAVE_UINT32
#endif /* #ifndef HAVE_UINT32 */

#ifndef HAVE_UINT64
typedef unsigned long long uint64;
#define HAVE_UINT64
#endif /* #ifndef HAVE_UINT64 */

#endif /* #if !defined(_SIMICS_API_H) */


#define int64   USE_SINT64_INSTEAD
typedef long long         sint64;


/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

#define Format32        "0x%08X"
#define FormatRaw32     "%08x"      /* Used by gdb: must be lower case x */
#define Format32Dec     "%d"
#define Format32Read    "%x"        /* Do not specify field width for restore from checkpoint. Do not specify 0x because scanf checks for it */
#define Format32Write   "0x%X"
#define FormatRaw32Len  (8)         /* Number of hexadecimal digits */

#define MAX_INT32       (0x7FFFFFFFU)
#define MIN_INT32       (0x80000000U)
#define MAX_UINT32      (0xFFFFFFFFU)


/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

#define Format64        "0x%016llX"
#define FormatRaw64     "%016llx"   /* Used by gdb: must be lower case x */
#define FormatRaw64Up   "%016llX"
#define Format64Dec     "%lld"
#define Format64Read    "%llX"      /* Do not specify field width for restore from checkpoint. Do not specify 0x because scanf checks for it */
#define Format64Write   "0x%llX"
#define FormatRaw64Len  (16)        /* Number of hexadecimal digits */

#define MAX_INT64       (0x7FFFFFFFFFFFFFFFULL)
#define MIN_INT64       (0x8000000000000000ULL)
#define MAX_UINT64      (0xFFFFFFFFFFFFFFFFULL)


/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

#ifndef CAST
#define CAST(t,e) ((t)(e))
#endif /* #ifndef CAST */

#define HI(_v)  CAST(uint32, ((CAST(uint64,_v) >> 32) & M_32_BITS))
#define LO(_v)  CAST(uint32, (_v & M_32_BITS))


/* 32 or 64 bit address printed in hex */
/* how to print a VOID  **/
#define FormatAddress "<%p>"


/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

/* For computations of Power */
typedef double Power_t;
#define FormatPower  "%8.4f"


/* Time domains return time in picoseconds */
typedef uint64 Picosecond;
#define FormatPico  Format64


#ifdef ALLOW_ENUM_BIT_FIELD
#define BITS(x) :x
#else /* #ifdef ALLOW_ENUM_BIT_FIELD */
#define BITS(x)
#endif /* #ifdef ALLOW_ENUM_BIT_FIELD */

#endif /* #ifndef _BASICTYPES_T_H_ */
