/* -------------------------------------------------------
 *
 * Licensed Materials - Property of IBM.
 * (C) Copyright IBM Corporation 2001, 2007
 * All Rights Reserved.
 *
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corporation.
 *
 * Filename : apu_callthru.h
 *
 * Purpose  :   */
/** \file
 *
 *   \brief
 *   Mambo callthru functions that can be invoked by the simulated program.
 *
 *   This file is compiled with SPU programs that are run under the Mambo
 *   simulator.  For example, stand-alone programs or operating
 *   systems.  (This file is not compiled into Mambo!)  The programs
 *   call the callthru wrapper functions which use an MTSPR instruction
 *   (with RB != 0) to signal the Mambo simulator to emulate special
 *   support.
 *
 *   NOTE: The Mambo* routines must not be inlined for parameter passing
 *   under the SPU ABI to work.
 */
/* ------------------------------------------------------- */

#ifndef _APU_CALLTHRU_H_
#define _APU_CALLTHRU_H_

#define INLINE      static inline

#ifndef __xlc__

INLINE void cthru_pviz_run(void)
{
    asm volatile(".long 0x21940000" : : : "memory");  /* mtspr 0,0 (rb=80), syscall code 0x50 */
}

INLINE void cthru_pviz_stop(void)
{
    asm volatile(".long 0x21944000" : : : "memory");  /* mtspr 0,0 (rb=81), syscall code 0x51 */
}

INLINE void cthru_pviz_set_delta(unsigned int delta)
{
    register unsigned int d asm ("$3") = delta;
    asm volatile(".long 0x21948000" : : "r" (d) : "memory");  /* mtspr 0,0 (rb=82), syscall code 0x52 */
}

INLINE int cthru_emit_perf_start(void)
{
    register int r asm ("$3") = 0;
    asm volatile(".long 0x21980000" : "=r" (r) : : "memory");  /* mtspr 0,0 (rb=96), syscall code 0x60 */
    return(r);
}

INLINE int cthru_emit_perf_stop(void)
{
    register int r asm ("$3") = 0;
    asm volatile(".long 0x21984000" : "=r" (r) : : "memory");  /* mtspr 0,0 (rb=97), syscall code 0x61 */
    return(r);
}

INLINE int cthru_emit_perf_clear(void)
{
    register int r asm ("$3") = 0;
    asm volatile(".long 0x21988000" : "=r" (r) : : "memory");  /* mtspr 0,0 (rb=98), syscall code 0x62 */
    return(r);
}

INLINE int cthru_emit_perf_stats(void)
{
    register int r asm ("$3") = 0;
    asm volatile(".long 0x2198C000" : "=r" (r) : : "memory");  /* mtspr 0,0 (rb=99), syscall code 0x63 */
    return(r);
}

INLINE int cthru_emit_perf_pipe(char *pipe)
{
    register char* p asm ("$3") = pipe;
    asm volatile(".long 0x21990000" : "=r" (p) : "r" (p) : "memory");  /* mtspr 0,0 (rb=100), syscall code 0x64 */
    return((int)(p));
}

INLINE void cthru_emit_start(void)
{
    asm volatile(".long 0x219C0000" : : : "memory");  /* mtspr 0,0 (rb=112), syscall code 0x70 */
}

INLINE void cthru_emit_stop(void)
{
    asm volatile(".long 0x219C4000" : : : "memory");  /* mtspr 0,0 (rb=113), syscall code 0x71 */
}

INLINE int cthru_emit_anno(char *string)
{
    register char* s asm ("$3") = string;
    asm volatile(".long 0x219C8000" : "=r" (s) : "r" (s) : "memory");  /* mtspr 0,0 (rb=114), syscall code 0x72 */
    return((int)(s));
}

INLINE int cthru_emit_new_stk(char *name)
{
    register char* n asm ("$3") = name;
    asm volatile(".long 0x219CC000" : "=r" (n) : "r" (n) : "memory");  /* mtspr 0,0 (rb=115), syscall code 0x73 */
    return((int)(n));
}

INLINE int cthru_emit_push_tag(unsigned int stack, char *tag)
{
    register unsigned int s asm ("$3") = stack;
    register char* t asm ("$4") = tag;
    asm volatile(".long 0x219D0000" : "=r" (s) : "r" (s), "r" (t): "memory");  /* mtspr 0,0 (rb=116), syscall code 0x74 */
    return((int)(s));
}

INLINE int cthru_emit_pop_tag(unsigned int stack, char *tag, unsigned int max_tag_len)
{
    register unsigned int s asm ("$3") = stack;
    register char* t asm ("$4") = tag;
    register unsigned int m asm ("$5") = max_tag_len;
    asm volatile(".long 0x219D4000" : "=r" (s) : "r" (s), "r" (t), "r" (m) : "memory");  /* mtspr 0,0 (rb=117), syscall code 0x75 */
    return((int)(s));
}

INLINE int cthru_emit_tos_tag(unsigned int stack, char *tag, unsigned int max_tag_len)
{
    register unsigned int s asm ("$3") = stack;
    register char* t asm ("$4") = tag;
    register unsigned int m asm ("$5") = max_tag_len;
    asm volatile(".long 0x219D8000" : "=r" (s) : "r" (s), "r" (t), "r" (m) : "memory");  /* mtspr 0,0 (rb=118), syscall code 0x76 */
    return((int)(s));
}

INLINE int cthru_emit_del_stk(unsigned int stack)
{
    register unsigned int s asm ("$3") = stack;
    asm volatile(".long 0x219DC000" : "=r" (s) : "r" (s) : "memory");  /* mtspr 0,0 (rb=119), syscall code 0x77 */
    return((int)(s));
}

INLINE int cthru_emit_del_all_stk(void)
{
    register int r asm ("$3") = 0;
    asm volatile(".long 0x219E0000" : "=r" (r) : : "memory");  /* mtspr 0,0 (rb=120), syscall code 0x78 */
    return(r);
}

INLINE int cthru_call_tcl(char *string)
{
    register char* s asm ("$3") = string;
    asm volatile(".long 0x219E4000" : "=r" (s) : "r" (s) : "memory");  /* mtspr 0,0 (rb=121), syscall code 0x79 */
    return((int)(s));
}

INLINE int cthru_emit_get_stk_hnd(void)
{
    register int h asm ("$3") = 0;
    asm volatile(".long 0x219E8000" : "=r" (h) : : "memory");  /* mtspr 0,0 (rb=122), syscall code 0x7A */
    return(h);
}

INLINE int cthru_emit_get_stk_dpt(unsigned int stack)
{
    register unsigned int s asm ("$3") = stack;
    asm volatile(".long 0x219EC000" : "=r" (s) : "r" (s) : "memory");  /* mtspr 0,0 (rb=123), syscall code 0x7B */
    return((int)(s));
}

#else /* #ifndef __xlc__ */

#pragma mc_func cthru_pviz_run          { "21940000" } /* mtspr 0,0 (rb=80), syscall code 0x50 */
#pragma mc_func cthru_pviz_stop         { "21944000" } /* mtspr 0,0 (rb=81), syscall code 0x51 */
#pragma mc_func cthru_pviz_set_delta    { "21948000" } /* mtspr 0,0 (rb=82), syscall code 0x52 */

#pragma mc_func cthru_emit_perf_start   { "21980000" } /* mtspr 0,0 (rb=96), syscall code 0x60 */
#pragma mc_func cthru_emit_perf_stop    { "21984000" } /* mtspr 0,0 (rb=97), syscall code 0x61 */
#pragma mc_func cthru_emit_perf_clear   { "21988000" } /* mtspr 0,0 (rb=98), syscall code 0x62 */
#pragma mc_func cthru_emit_perf_stats   { "2198C000" } /* mtspr 0,0 (rb=99), syscall code 0x63 */
#pragma mc_func cthru_emit_perf_pipe    { "21990000" } /* mtspr 0,0 (rb=100), syscall code 0x64 */

#pragma mc_func cthru_emit_start        { "219C0000" } /* mtspr 0,0 (rb=112), syscall code 0x70 */
#pragma mc_func cthru_emit_stop         { "219C4000" } /* mtspr 0,0 (rb=113), syscall code 0x71 */
#pragma mc_func cthru_emit_anno         { "219C8000" } /* mtspr 0,0 (rb=114), syscall code 0x72 */
#pragma mc_func cthru_emit_new_stk      { "219CC000" } /* mtspr 0,0 (rb=115), syscall code 0x73 */
#pragma mc_func cthru_emit_push_tag     { "219D0000" } /* mtspr 0,0 (rb=116), syscall code 0x74 */
#pragma mc_func cthru_emit_pop_tag      { "219D4000" } /* mtspr 0,0 (rb=117), syscall code 0x75 */
#pragma mc_func cthru_emit_tos_tag      { "219D8000" } /* mtspr 0,0 (rb=118), syscall code 0x76 */
#pragma mc_func cthru_emit_del_stk      { "219DC000" } /* mtspr 0,0 (rb=119), syscall code 0x77 */
#pragma mc_func cthru_emit_del_all_stk  { "219E0000" } /* mtspr 0,0 (rb=120), syscall code 0x78 */
#pragma mc_func cthru_call_tcl          { "219E4000" } /* mtspr 0,0 (rb=121), syscall code 0x79 */
#pragma mc_func cthru_emit_get_stk_hnd  { "219E8000" } /* mtspr 0,0 (rb=122), syscall code 0x7A */
#pragma mc_func cthru_emit_get_stk_dpt  { "219EC000" } /* mtspr 0,0 (rb=123), syscall code 0x7B */

#endif /* #ifndef __xlc__ */

/**
 * Begin/resume producing emitter records (TAG_START_STATS)
 *
 * Causes Mambo to begin or resume producing emitter records
 */
INLINE void MamboStartStats(void)
{
    cthru_emit_start();
}

/**
 * Cease producing emitter records (TAG_STOP_STATS)
 *
 * Causes Mambo to stop producing emitter records
 *
 * \return  undefined
 */
INLINE void MamboStopStats(void)
{
    cthru_emit_stop();
}

/**
 * Emit an ANNOTATION emitter record (TAG_ANNOTATE_STATS)
 *
 * Causes Mambo to emit a STATS_ANNOTATION_INFO structure whose
 * text component is the string given (truncated to 60 characters)
 *
 * \param   string  Annotation text string
 * \return  (boolean semantics) 0 if failure, 1 if success
 */
INLINE int MamboAnnotateStats(char *string)
{
    return(cthru_emit_anno(string));
}

/**
 * Create an emitter annotation tag stack
 *
 * Creates a new emitter annotation tag stack. The stack
 * handle (an integer) is returned for use in subsequent
 * tag stack calls. The stack name is truncated to 20
 * characters, if provided
 *
 * \param   name    Name of new stack (is set to "<handle>" if NULL)
 *
 * \return  New tag stack handle (integer)
 */
INLINE int MamboStatsCreateTagStack(char *name)
{
    return(cthru_emit_new_stk(name));
}

/**
 * Retrieve the last referenced tag stack handle for the current APU
 *
 * \return  Last referenced tag stack handle for the current APU
 */
INLINE int MamboStatsLastTagStackHandle(void)
{
    return(cthru_emit_get_stk_hnd());
}

/**
 * Emit an ANNOTATION emitter record (TAG_ANNOTATE_STATS) and push onto given stack
 *
 * Causes Mambo to emit a STATS_ANNOTATION_INFO structure whose
 * text component is the tag string given (truncated to 60 characters).
 * Pushes this tag onto the given tag stack
 *
 * \param   stack   Annotation stack handle (see MamboStatsCreateTagStack)
 * \param   tag     Annotation text string
 *
 * \return  (boolean semantics) 0 if failure, 1 if success
 */
INLINE int MamboStatsPushTag(unsigned int stack, char *tag)
{
    return(cthru_emit_push_tag(stack, tag));
}

/**
 * Emit the top of stack ANNOTATION emitter record (TAG_ANNOTATE_STATS)
 *
 * Causes Mambo to emit a STATS_ANNOTATION_INFO structure whose
 * text component is the top of the given stack. If the tag pointer
 * given is non-zero, the top of stack tag will be copied to the
 * address given (to the maximum length specified). The top of stack
 * is popped from the given tag stack. If the stack is empty, 0 (failure)
 * is returned.
 *
 * \param   stack       Annotation stack handle (see MamboStatsCreateTagStack)
 * \param   tag         RETURNED tag if non-zero
 * \param   max_tag_len Maximum number of bytes of the TOS tag to copy to tag (including 0-byte)
 *
 * \return  (boolean semantics) 0 if failure, 1 if success
 */
INLINE int MamboStatsPopTag(unsigned int stack, char *tag, unsigned int max_tag_len)
{
    return(cthru_emit_pop_tag(stack, tag, max_tag_len));
}

/**
 * Retrieve the top of stack ANNOTATION
 *
 * If the tag pointer given is non-zero, the top of stack tag text will be
 * copied to the address given (to the maximum length specified). If the stack is
 * empty, 0 is returned (failure).
 *
 * \param   stack       Annotation stack handle (see MamboStatsCreateTagStack)
 * \param   tag         RETURNED tag if non-zero
 * \param   max_tag_len Maximum number of bytes of the TOS tag to copy to tag (including 0-byte)
 *
 * \return  (boolean semantics) 0 if failure, 1 if success
 */
INLINE int MamboStatsTOSTag(unsigned int stack, char *tag, unsigned int max_tag_len)
{
    return(cthru_emit_tos_tag(stack, tag, max_tag_len));
}

/**
 * Retrieve the current depth for the given annotation stack
 *
 * \param   stack       Annotation stack handle (see MamboStatsCreateTagStack)
 *
 * \return  The depth of the given stack
 */
INLINE int MamboStatsTagStackDepth(unsigned int stack)
{
    return(cthru_emit_get_stk_dpt(stack));
}

/**
 * Destroy a given emitter annotation tag stack, releasing its resources
 *
 * \param   stack       Annotation stack handle (see MamboStatsCreateTagStack)
 *
 * \return  (boolean semantics) 1 (always succeeds)
 */
INLINE int MamboStatsDeleteTagStack(unsigned int stack)
{
    return(cthru_emit_del_stk(stack));
}

/**
 * Destroy all emitter annotation tag stacks, releasing all associated resources
 *
 * \return  (boolean semantics) 1 (always succeeds)
 */
INLINE int MamboStatsDeleteAllTagStacks(void)
{
    return(cthru_emit_del_all_stk());
}

/**
 * Execute the string as a TCL command
 *
 *  \param  string   pointer to Tcl command string
 */
INLINE int MamboCallTCL(char *string)
{
    return(cthru_call_tcl(string));
}

/**
 * Start recording SPU statistics
 *
 * \return  undefined
 */
INLINE int MamboSPUPerfStart(void)
{
    return(cthru_emit_perf_start());
}

/**
 * Stop recording SPU statistics
 *
 * \return undefined
 */
INLINE int MamboSPUPerfStop(void)
{
    return(cthru_emit_perf_stop());
}

/**
 * Clear SPU statistics
 *
 * \return undefined
 */
INLINE int MamboSPUPerfClear(void)
{
    return(cthru_emit_perf_clear());
}

/**
 * EMIT SPU statistics
 *
 * \return undefined
 */
INLINE int MamboSPUPerfStats(void)
{
    return(cthru_emit_perf_stats());
}

/**
 * EMIT SPU individual pipeline statistics
 *
 * \param   pipe    SPU execution pipe name
 *
 * \return undefined
 */
INLINE int MamboSPUPerfPipeStats(char *pipe)
{
    return(cthru_emit_perf_pipe(pipe));
}

#ifdef MCONFIG_PVIZ
/**
 * Begin/resume PVIZ monitoring
 *
 * Causes Mambo to begin or resume SPU pipeline visualization monitoring
 */
INLINE void MamboPVIZRun(void)
{
    cthru_pviz_run();
}

/**
 * Cease PVIZ monitoring
 *
 * Causes Mambo to stop SPU pipeline visualizaation monitoring
 *
 * \return  undefined
 */
INLINE void MamboPVIZStop(void)
{
    cthru_pviz_stop();
}

/**
 * Set PVIZ sampling time delta
 *
 * \param       delta   Sampling delta (in SPU cycles)
 */
INLINE void MamboPVIZSetDelta(unsigned int delta)
{
    cthru_pviz_set_delta(delta);
}
#endif /* #ifdef MCONFIG_PVIZ */

#endif /* #ifndef _APU_CALLTHRU_H_ */
