/* -------------------------------------------------------
 *
 * Licensed Materials - Property of IBM.
 * (C) Copyright IBM Corporation 2005, 2006
 * All Rights Reserved.
 *
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corporation.
 *
 * Filename : mambo_iss.h
 *
 * Purpose  : External interfaces of the Mambo Instruction Set Simulator
 */

/**
 * @file mambo_iss.h
 */

#ifndef _MAMBO_ISS_H_
#define _MAMBO_ISS_H_

#ifdef __cplusplus
extern "C" {
#endif /* #ifdef __cplusplus */

    enum mambo_iss_rc
        {
            MAMBO_ISS_SUCCESS  = 0,     /*!< Successful operation */
            MAMBO_ISS_ERROR    = 1      /*!< Failure */
        };
    typedef enum mambo_iss_rc mambo_iss_rc;

    enum mambo_iss_state
        {
            MAMBO_ISS_STOPPED  = 0,     /*!< Processor is stopped */
            MAMBO_ISS_RUNNING  = 1,     /*!< Processor is running */
            MAMBO_ISS_STALLED  = 2      /*!< Processor is stalled */
        };
    typedef enum mambo_iss_state mambo_iss_state;

    typedef void* mambo_core;

    typedef unsigned long long mambo_cycle;
#define mambo_cycle_format "%08llu"

#ifdef MAMBO_ISS_64BIT
    typedef unsigned long long mambo_core_addr;
    typedef unsigned long long xbus_addr;
#define mambo_core_addr_format "0x%016llX"
#define xbus_addr_format "0x%016llX"
#else /* #ifdef MAMBO_ISS_64BIT */
    typedef unsigned long mambo_core_addr;
    typedef unsigned long xbus_addr;
#define mambo_core_addr_format "0x%08lX"
#define xbus_addr_format "0x%08lX"
#endif /* #ifdef MAMBO_ISS_64BIT */

    /** core interrupt type */
    enum mambo_core_interrupt
        {
            MAMBO_CLEAR_INTERRUPTS  = 0x00,     /*!< Clear interrupt signals */
            MAMBO_MACH_CHECK        = 0x01,     /*!< Machine Check */
            MAMBO_EXT_INTERRUPT     = 0x02      /*!< External Interrupt */
        };
    typedef enum mambo_core_interrupt mambo_core_interrupt;

    /** ELF Loader option flags */
    enum mambo_loader_options
        {
            MAMBO_LOAD_STD = 0,             /*!< ISS Standalone mode */
            MAMBO_LOAD_RELOCATE = 1         /*!< Relocate ELF object */
        };
    typedef enum mambo_loader_options mambo_loader_options;

    typedef unsigned char xbus_data;

    /**
     * @internal
     * @brief xbus_ttype -- bus transfer type
     *
     * These encodings are intended to match the 60x/6xx ttype encoding.
     */
    enum xbus_ttype
        {
            XBUS_UNDEFINED = 0xFF,      /* Undefined xbus type */
            XBUS_CLEAN     = 0x00,      /* Clean block (dcbst) */
            XBUS_FLUSH     = 0x04,      /* Flush block (dcbf) */
            XBUS_SYNC      = 0x08,      /* Synchronize */
            XBUS_KILL      = 0x0C,      /* Kill block */
            XBUS_IKILL     = 0x0D,      /* Instruction-line kill block (icbi) */
            XBUS_EIEIO     = 0x10,      /* Ordered I/O operation (eieio) */
            XBUS_TLBIE     = 0x18,      /* TLB invalidate */
            XBUS_TLBSYNC   = 0x09,      /* TLB synchronize */

            XBUS_WR_FLUSH  = 0x02,      /* Write-with-flush */
            XBUS_WR_KILL   = 0x06,      /* Write-with-kill (G=0) */
            XBUS_WR_CLEAN  = 0x06,      /* Write-with-clean (G=1) */

            XBUS_READ      = 0x0A,      /* Read */
            XBUS_RWITM     = 0x0E,      /* Read-with-intent-to-modify */
            XBUS_DCLAIM    = 0x03,      /* Data line claim */

            XBUS_WR_FLUSHA = 0x12,      /* Write-with-flush-atomic (stwcx) */
            XBUS_READA     = 0x1A,      /* Read-atomic (lwarx) */
            XBUS_RWITMA    = 0x1E,      /* Read-with-intent-to-modify-atomic */
            XBUS_DCLAIMA   = 0x13,      /* Data line claim atomic */
            XBUS_RWNITC    = 0x0B,      /* Read-with-no-intent-to-cache */

            XBUS_LAR       = 0x01       /* lwarx reservation set */
        };
    typedef enum xbus_ttype xbus_ttype;

    /**
     * @internal
     * @brief xbus_request -- XBus Request type
     */
    struct xbus_request
    {
        xbus_ttype   ttype;          /* Transfer Type */
        xbus_addr    addr;           /* The physical address */
        int          length;         /* Number of bytes to read/write */
        int          srcid;          /* ID of initiator */
        int          dstid;          /* ID of destination */

        int          bus_data;       /* Bus-defined data value */
        void*        bus_ptr;        /* Bus-defined ptr value */

        int          src_data;       /* Source-defined data value */
        void*        src_ptr;        /* Source-defined ptr value */

        int          dst_data;       /* Destination-defined data value */
        void*        dst_ptr;        /* Destination-defined ptr value */
    };
    typedef struct xbus_request xbus_request;

    /**
     * @internal
     * @brief xbus_response -- XBus Response type
     */
    enum xbus_response
        {
            XBUS_RESP_NULL      = 0x00,      /* NULL */
            XBUS_RESP_ACK       = 0x01,      /* Acknowledge */
            XBUS_RESP_SHR       = 0x03,      /* Data will be provided shared */
            XBUS_RESP_RETRY     = 0x0C       /* Src must retry command */
        };
    typedef enum xbus_response xbus_response;

    struct xbus_methods
    {
        int (*xbus_initialize) (void **xbus_state, int *xbus_agent_id);
        int (*xbus_terminate) (void *xbus_state, int xbus_agent_id);
        int (*xbus_send_command) (void *xbus_state, xbus_request *request);
        int (*xbus_snoop_response) (void *xbus_state, xbus_request *request, xbus_response resp);
        int (*xbus_send_data)  (void *xbus_state, xbus_request *request, xbus_data *data);
        int (*xbus_debug_read)  (void *xbus_state, xbus_addr addr, int length, xbus_data *data, int *retlen);
        int (*xbus_debug_write)  (void *xbus_state, xbus_addr addr, int length, xbus_data *data);
    };
    typedef struct xbus_methods xbus_methods;

    /**
     * @brief Mambo register names
     */
    typedef const char *mambo_reg_name;
#define MAMBO_REG_NAME_GPR  "GPR"
#define MAMBO_REG_NAME_FPR  "FPR"
#define MAMBO_REG_NAME_SPR  "SPR"
#define MAMBO_REG_NAME_SR   "SR"

    /**
     * @brief Mambo register operations (meant to be treated as a bit set of flags)
     */
    typedef unsigned long int mambo_reg_op;
#define MAMBO_REG_OP_NONE   ((mambo_reg_op)(0x0))
#define MAMBO_REG_OP_READ   ((mambo_reg_op)(0x1))
#define MAMBO_REG_OP_WRITE  ((mambo_reg_op)(0x2))
#define MAMBO_REG_OP_MODIFY ((mambo_reg_op)(0x4))
#define MAMBO_REG_OP_ANY    (MAMBO_REG_OP_READ | MAMBO_REG_OP_WRITE | MAMBO_REG_OP_MODIFY)

    /**
     * @brief Mambo register number
     */
    typedef unsigned long int mambo_reg_num;

    /**
     * @brief Mambo register value (large enough to hold FPR values)
     */
    typedef unsigned long long int mambo_reg_value;

    /**
     * @brief Mambo register observation tag number
     */
    typedef unsigned int mambo_obs_tag;

    /**
     * @brief Mambo register observation tag set (bit-set)
     */
    typedef unsigned long long int mambo_tag_set;

    /**
     * @brief Initialize the Mambo ISS library
     *
     * @param[in]  proc_model      Name of the processor model to be simulated
     * @param[in]  init_file       Name of the TCL startup file to use (defaults to ".systemsim.tcl")
     * @param[in]  bus_methods     Pointer to block of XBus methods provides by the bus model
     *
     * @return     Mambo core handle. Asserts if errors are encountered
     *
     * @par Example:
     * @code
     *  // To initialize a PowerPC 750GX model ISS with default TCL starupt file and without XBus support:
     *  mambo_core handle = mambo_iss_init("PPC750GX",NULL, NULL);
     *
     *  // To initialize a PowerPC 750CL model ISS with TCL starupt file "foo.tcl" and WITH XBus support:
     *  xbus_methods    bus_methods;
     *  // Initialize bus_methods
     *  ...
     *  mambo_core handle = mambo_iss_init("PPC750CL","foo.tcl", &bus_methods);
     * @endcode
     */
    extern mambo_core mambo_iss_init(const char *proc_model, const char *init_file, xbus_methods *bus_methods);

    /**
     * @brief  Terminates the Mambo ISS, performs cleanup
     *
     * @param[in] coreobj  Handle to Mambo ISS library (from mambo_iss_init)
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *    mambo_core handle = mambo_iss_init(NULL, NULL);
     *    ...
     *    // ISS operations
     *    ...
     *    mambo_iss_term(handle);
     * @endcode
     */
    extern int mambo_iss_term(mambo_core coreobj);

    /**
     * @brief Run the ISS
     *
     * @param[in]  coreobj         Handle to Mambo ISS library (from mambo_iss_init)
     * @param[in]  cycles_to_run   Number of core cycles to execute
     * @param[out] cycles_run      Number of core cycles executed (optionally NULL)
     * @param[out] next_cycle      Next core cycle at which core has work to perform (optionally NULL)
     * @param[out] pc              Address of program counter at the end of the last cycle executed (optionally NULL)
     * @param[out] reg_obs_set     Set of tag bits (see mambo_iss_observe_reg)
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *    mambo_core        handle;
     *    mambo_cycle       cycles_actually_executed;
     *    mambo_cycle       next_cycle_of_execution;
     *    mambo_core_addr   program_counter;
     *    mambo_tag_set     reg_obs_set;
     *    int               return_code;
     *    ...
     *    // Run for 100 core cycles
     *    return_code = mambo_iss_run(handle, 100, &cycles_actually_executed,
     *                                &next_cycle_of_execution, &program_counter
     *                                &reg_obs_set);
     * @endcode
     *
     * @par Notes:
     * This interface will direct mambo to simulate core functions for one or more
     * clock cycles.  The caller specifies the maximum number of cycles to
     * simulate. The Mambo ISS will simulate at least one cycle, and at most the
     * number of cycles specified by the caller. For functional simulation, the
     * work done within a "cycle" may correspond to less or more than would
     * actually occur in hardware. A fundamental assumption of this interface is
     * that all inputs to the core model (e.g. bus requests, interrupts) that
     * should affect operation of the cycles to be executed have been previously
     * provided. In addition, outputs of the core model (e.g. bus requests) should
     * be considered to be produced at the end of the cycle. (In particular,
     * outputs of the ISS should not generate new events for the core model to
     * handle in the same cycle.)
     */
    extern int mambo_iss_run(mambo_core coreobj, mambo_cycle cycles_to_run, mambo_cycle *cycles_run,
        mambo_cycle *next_cycle, mambo_core_addr *pc, mambo_tag_set *reg_obs_set);


    /**
     * @brief Return the ISS simulator state
     *
     * @param[in]  coreobj         Handle to Mambo ISS (from mambo_iss_init)
     * @param[out] state           State of the ISS core model
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *      mambo_iss_state state;
     *      int             return_code = mambo_iss_query_state(handle, &state);
     * @endcode
     */
    extern int mambo_iss_query_state(mambo_core coreobj, mambo_iss_state *state);


    /**
     * @brief Perform a system reset of the ISS core
     *
     * @param[in]  coreobj         Handle to Mambo ISS (from mambo_iss_init)
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *      int     return_code = mambo_iss_reset(handle);
     * @endcode
     *
     * This interface will request that a system reset interrupt be presented to the
     * core model. Any instructions or bus transactions that were in process are
     * terminated, and the core model will begin processing from the reset state in
     * the next cycle, i.e., in the next call to mambo_iss_run. The system model may
     * call mambo_iss_interrupt at most once for any cycle.
     */
    extern int mambo_iss_reset(mambo_core coreobj);

    /**
     * @brief Set the status of interrupt signals to the ISS core
     *
     * @param[in]  coreobj             Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  interrupt_status    Status of interrupt signals to the ISS core
     * @param[out] next_cycle          Next core cycle at which core has work to perform (optionally NULL)
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *      int     return_code = mambo_iss_interrupt(handle, MAMBO_EXT_INTERRUPT, NULL);
     * @endcode
     *
     * @par Notes:
     * Interrupt_status is a bit mask constructed by ORing values of type mambo_core_interrupt.
     * Interrupt types currently supported are:
     *
     * @li MAMBO_MACH_CHECK      -- Machine check interrupt
     * @li MAMBO_EXT_INTERRUPT   -- External interrupt
     *
     * This interface can be used to request that an designated interrupt be presented to the
     * core model. The ISS will present this interrupt in the next core model cycle executed,
     * i.e., in the next call to mambo_iss_run. The system model may call mambo_iss_interrupt at
     * most once for any cycle.
     *
     * The system model is responsible for clearing the interrupt signal as required by the
     * architecture.  The system model can clear all interrupt signals by calling mambo_iss_interrupt
     * with an interrupt_status value of MAMBO_CLEAR_INTERRUPTS.
     */
    extern int mambo_iss_interrupt(mambo_core coreobj, mambo_core_interrupt interrupt_status, mambo_cycle *next_cycle);

    /**
     * @brief Pass a command directly to the Mambo ISS
     *
     * @param[in]  coreobj          Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  cmd              A command string to be interpreted and executed by the ISS
     * @param[out] result_buf       A buffer to hold the command result
     * @param[in]  buf_length       The size of result_buf provided
     * @param[out] result_length    The actual length of the result value of the command
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *      char buf[256];
     *      int  actual_len;
     *      int  return_code = mambo_iss_command(handle, "mambo_iss_sim version",
     *                                           buf, sizeof(buf), &actual_len);
     * @endcode
     *
     * @par Notes:
     * The system model calls this interface to pass a command to the Mambo ISS. Upon
     * return, the result value contains a formatted text result containing the
     * information requested.
     */
    extern int mambo_iss_command(mambo_core coreobj, const char *cmd, char *result_buf, int buf_length, int *result_length);

    /**
     * @brief Load an ELF object file into simulated memory
     *
     * @param[in]   elf_file        The name of a file containing the ELF object to be loaded
     * @param[in]   bus_methods     Pointer to block of XBus methods provides by the bus model
     * @param[in]   opts            A set of mambo_loader_options indicating the type of loading to be performed
     * @param[in]   reloc_addr      Relocation address (if opts specifies that the object should be relocated)
     * @param[out]  start_addr      Value indicating the start address of the ELF executable (optionally NULL)
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     *      int             return_code;
     *      xbus_addr       start_addr;
     *      xbus_addr       reloc_addr = 64*1024;
     *      xbus_methods    bus_methods;
     *
     *      // Load and relocate "a.out" to 0x10000 (64K, 64 * 1024)
     *      return_code = mambo_load_elf("a.out", &bus_methods, MAMBO_LOAD_RELOCATE, reloc_addr, &start_addr);
     * @endcode
     *
     * @par Notes:
     * The system model calls this interface to load an ELF executable into memory.
     * Typically, this will be done to initialize system memory regions prior to
     * the start of simulation.  The loader will use the bus_debug_write interface
     * of the Bus model to write the file contents to memory.  The loader does not
     * interact with Mambo ISS in any manner. It is the responsibility of the
     * system model to prepare the core to execute an ELF image after it is loaded
     * into memory. When this is done at the start of a simulation, the following
     * steps are needed:
     *
     * @li Reset the core to its power-on state.
     * @li Set the PC to the entry point of the loaded executable
     * @li Set GPR 1 to point to an area to be used for the stack
     */
    extern int mambo_load_elf(const char *elf_file, xbus_methods *bus_methods, mambo_loader_options opts, xbus_addr reloc_addr, xbus_addr *start_addr);

    /**
     * @brief Deliver a bus request to the ISS for snooping
     *
     * @param[in]  coreobj          Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  request          Bus request to be snooped by the ISS
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     * @endcode
     *
     * @par Notes:
     * This interface is invoked by the bus model to deliver a bus transaction to the
     * Mambo ISS to be snooped. Snoop responses are not returned on this interface;
     * instead, they are sent to the bus through bus_snoop response. The ISS will
     * recognize and begin processing the snoop request in the next core model cycle
     * executed, i.e., in the next call to mambo_iss_run. The bus may call
     * mambo_iss_bus_snoop at most once for any cycle.
     */
    extern int mambo_iss_bus_snoop(mambo_core coreobj, xbus_request *request);

    /**
     * @brief Deliver a bus response to the ISS
     *
     * @param[in]  coreobj          Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  request          Associated bus request
     * @param[in]  resp             Response to deliver for the request
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     * @endcode
     *
     * @par Notes:
     * This interface is invoked by the bus model to deliver a response to the Mambo
     * ISS for a bus command it previously issued. The ISS will recognize and begin
     * processing the command response in the next core model cycle executed, i.e., in
     * the next call to mambo_iss_run. The bus may call mambo_iss_bus_command_response
     * at most once for any cycle.
     */
    extern int mambo_iss_bus_command_response(mambo_core coreobj, xbus_request *request, xbus_response resp);

    /**
     * @brief Deliver bus data to the ISS
     *
     * @param[in]  coreobj          Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  request          Associated bus request
     * @param[in]  data             Data to deliver for request
     *
     * @returns MAMBO_ISS_SUCCESS upon successful operation, MAMBO_ISS_ERROR otherwise
     *
     * @par Example:
     * @code
     * @endcode
     *
     * @par Notes:
     * This interface is invoked by the system model to deliver data from the bus to
     * the Mambo ISS. The ISS will recognize and begin processing the bus data in the
     * next core model cycle executed, i.e., in the next call to mambo_iss_run. The
     * bus may call mambo_iss_bus_data at most once for any cycle.
     */
    extern int mambo_iss_bus_data(mambo_core coreobj, xbus_request *request, xbus_data *data);

    /**
     * @brief Associate a tag for any of the
     * indicated operations performed upon the given register
     *
     * @param[in]  coreobj      Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  reg_name     Register name
     * @param[in]  reg_num      Register number
     * @param[in]  reg_op       OR of register operations to observe
     * @param[in]  tag          Observation tag (<64)
     *
     * @returns  MAMBO_ISS_SUCCESS if the callback was associated, MAMBO_ISS_ERROR otherwise.
     *
     * @par Notes:
     * This interface sets up a tag to be associated with operations on the indicated register.
     * If any of those operations occur for the given register as a result of mambo_iss_run,
     * a bit (1 << tag) will be set in the reg_obs_set parameter returned by mambo_iss_run.
     */
    extern int mambo_iss_observe_reg(mambo_core coreobj,
                                     mambo_reg_name reg_name,
                                     mambo_reg_num reg_num,
                                     mambo_reg_op reg_op,
                                     mambo_obs_tag tag);

    /**
     * @brief Disassociate all tags from a given register
     *
     * @param[in]  coreobj      Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  reg_name     Register name
     * @param[in]  reg_num      Register number
     *
     * @returns  MAMBO_ISS_SUCCESS if the callback was disassociated, MAMBO_ISS_ERROR otherwise.
     */
    extern int mambo_iss_forget_reg(mambo_core coreobj,
                                    mambo_reg_name reg_name,
                                    mambo_reg_num reg_num);

    /**
     * @brief Disassociate the given tag from all register operations
     *
     * @param[in]  coreobj      Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  tag          Observation tag (<64)
     *
     * @returns  MAMBO_ISS_SUCCESS if the tag was disassociated, MAMBO_ISS_ERROR otherwise.
     */
    extern int mambo_iss_forget_tag(mambo_core coreobj, mambo_obs_tag tag);

    /**
     * @brief Disassociate all the given tags from all register operations
     *
     * @param[in]  coreobj      Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  tag_set      Set of tags to forget (as a bit-set)
     *
     * @returns  MAMBO_ISS_SUCCESS if the tags were disassociated, MAMBO_ISS_ERROR otherwise.
     */
    extern int mambo_iss_forget_tags(mambo_core coreobj, mambo_tag_set tag_set);

    /**
     * @brief Retrieve a register's value and operation indicators
     *
     * @param[in]  coreobj      Handle to Mambo ISS (from mambo_iss_init)
     * @param[in]  reg_name     Register name
     * @param[in]  reg_num      Register number
     * @param[out] reg_op       Operations performed
     * @param[out] reg_value    Current register value
     *
     * @returns  MAMBO_ISS_SUCCESS if the callback was disassociated, MAMBO_ISS_ERROR otherwise.
     */
    extern int mambo_iss_get_reg(mambo_core coreobj,
                                 mambo_reg_name reg_name,
                                 mambo_reg_num reg_num,
                                 mambo_reg_op *reg_op,
                                 mambo_reg_value *reg_value);

#ifdef __cplusplus
}
#endif /* #ifdef __cplusplus */

#endif /* #ifndef _MAMBO_ISS_H_ */
