/* -------------------------------------------------------
 *
 * Licensed Materials - Property of IBM.
 * (C) Copyright IBM Corporation 2006, 2007
 * All Rights Reserved.
 *
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corporation.
 *
 * Filename : plugin_api.h
 *
 * Purpose  : Plugin APIs that receive XXX_handler and
 *            return what you want.
 *
 * ------------------------------------------------------- */

#ifndef _PLUGIN_API_H_
#define _PLUGIN_API_H_

#ifdef CONFIG_PLUGIN

#include "common/plugin_basic.h"

#define PLUGIN_API

#ifdef ENABLE_EYCOOL

/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

/* Mambo_Handler */

extern PLUGIN_API const char *Get_Mambo_Machine_Name(Mambo_Handler mhdlr);

/* Instruction_Handler */

/*
 * Every instruction has global uniqe number.
 */
extern PLUGIN_API uint64 Get_Inst_SeqNumber(Instruction_Handler ihdlr);

/*
 * Get the instructon's 32-bit binary code.
 */
extern PLUGIN_API uint32 Get_Inst_Instbody(Instruction_Handler ihdlr);

/*
 * Print format is predefined printed format by mambo.
 * Don't use it if you don't know it very much.
 */
extern PLUGIN_API const char * Get_Inst_PrintFormat(Instruction_Handler ihdlr);

/*
 * Name and full name, like "add" and "add r0, r1, r2".
 */
extern PLUGIN_API const char * Get_Inst_Name(Instruction_Handler ihdlr);
extern PLUGIN_API void Get_Inst_Name_And_Operand(Instruction_Handler ihdlr, char *buf, int len);

/*
 * Get the instruction's address.
 */
extern PLUGIN_API uint64 Get_Inst_PC(Instruction_Handler ihdlr);

/*
 * Determine whether instruction relocation is on (MSR[IR] == 1)
 */
extern PLUGIN_API boolean Is_Inst_Relocated(Instruction_Handler ihdlr);

/*
 * Determine whether instruction caused an exception
 */
extern PLUGIN_API boolean Is_Inst_Exception(Instruction_Handler ihdlr);

/*
 * Get the instruction's VSID (unshifted)
 */
extern PLUGIN_API uint64 Get_Inst_PC_VSID(Instruction_Handler ihdlr);

/*
 * Get the instruction's RPN
 */
extern PLUGIN_API uint64 Get_Inst_PC_RPN(Instruction_Handler ihdlr);

/*
 * Get the instruction's page size (in bits)
 */
extern PLUGIN_API uint8 Get_Inst_PC_PageSize(Instruction_Handler ihdlr);

/*
 * Get_Inst_NPC return the correct NPC only if the instruction is
 * executed. If the instruction is in the pipeline stage before
 * execution, this function returns the predicted value (same
 * with Get_Inst_NPC_Predicted). You can use Is_Inst_Done to
 * judge if the instruction is executed, and after that use
 * Get_Inst_NPC to get correct NPC.
 *
 * If NPC = PC, this instruction is a cracked instruction.
 */
extern PLUGIN_API uint64 Get_Inst_NPC(Instruction_Handler ihdlr);

/*
 * For normal instructions, NPC_Predicted is the PC + 4.
 * For branch instructions, NPC_Predicted is the predicted target address.
 * For cracked instructions, NPC_Predicted = PC.
 */
extern PLUGIN_API uint64 Get_Inst_NPC_Predicted(Instruction_Handler ihdlr);

/*
 * Get the instruction's ASR
 */
extern PLUGIN_API uint64 Get_Inst_ASR(Instruction_Handler ihdlr);

/*
 * Get the instruction's SDR1
 */
extern PLUGIN_API uint64 Get_Inst_SDR1(Instruction_Handler ihdlr);

/*
 * In SMT environment, every instruction belongs to a thread.
 */
extern PLUGIN_API uint32 Get_Inst_ThreadId(Instruction_Handler ihdlr);

/*
 * Is this instruction executed. (After execution stage in pipeline).
 */
extern PLUGIN_API boolean Is_Inst_Done(Instruction_Handler ihdlr);

/*
 * Is this a sim support pseudo-instruction
 */
extern PLUGIN_API boolean Is_Inst_SimSupport(Instruction_Handler ihdlr);

/*
 * Is this instruction a branch instruction.
 */
extern PLUGIN_API boolean Is_Inst_Branch(Instruction_Handler ihdlr);

/*
 * Is this instruction a unconditional branch.
 */
extern PLUGIN_API boolean Is_Inst_Unconditional_Branch(Instruction_Handler ihdlr);

/*
 * Is this branch instruction taken.
 *
 * Return TRUE only when two situations are both satisfied:
 * 1) The instruction is executed.
 * 2) It is a branch instruction and really taken.
 * That is, if a instruction is not executed, even if it is
 * unconditional branch, this function will return FALSE.
 * You can use Is_Inst_Done to judge if the instruction is
 * executed.
 */
extern PLUGIN_API boolean Is_Inst_Branch_Taken(Instruction_Handler ihdlr);

/*
 * Is this a VMX instruction.
 */
extern PLUGIN_API boolean Is_Inst_VMX(Instruction_Handler ihdlr);

/*
 * Is this a load instruction.
 */
extern PLUGIN_API boolean Is_Inst_Load(Instruction_Handler ihdlr);

/*
 * Is this a store instruction.
 */
extern PLUGIN_API boolean Is_Inst_Store(Instruction_Handler ihdlr);

/*
 * Is the data effective address valid?
 */
extern PLUGIN_API boolean Is_Inst_Data_EA_Valid(Instruction_Handler ihdlr);

/*
 * Is the data physical address valid?
 */
extern PLUGIN_API boolean Is_Inst_Data_PA_Valid(Instruction_Handler ihdlr);

/*
 * Get the data effective address
 */
extern PLUGIN_API uint64 Get_Inst_Data_EA(Instruction_Handler ihdlr);

/*
 * Is data relocation on? (MSR[DR} == 1)
 */
PLUGIN_API boolean Is_Data_Relocated(Instruction_Handler ihdlr);

/*
 * Get the data's VSID (unshifted)
 */
extern PLUGIN_API uint64 Get_Inst_Data_VSID(Instruction_Handler ihdlr);

/*
 * Get the data's RPN
 */
extern PLUGIN_API uint64 Get_Inst_Data_RPN(Instruction_Handler ihdlr);

/*
 * Get the data's page size (in bits)
 */
extern PLUGIN_API uint8 Get_Inst_Data_PageSize(Instruction_Handler ihdlr);

/*
 * Structure used to save register values.
 */
#define MAX_GPRS 32
#define MAX_FPRS 4
#define MAX_SPRS 4
#define MAX_VMXRS 4
struct Plugin_Inst_Registers
{
    // GPR
    int gpr_num;
    int gpr_index[MAX_GPRS];
    uint64 gpr_value[MAX_GPRS];

    // FPR
    int fpr_num;
    int fpr_index[MAX_FPRS];
    uint64 fpr_value[MAX_FPRS];

    // SPR
    int spr_num;
    int spr_index[MAX_SPRS];
    uint64 spr_value[MAX_SPRS];

    // VMX
    int vmx_num;
    int vmx_index[MAX_VMXRS];
    QuadWord vmx_value[MAX_VMXRS];
};
typedef struct Plugin_Inst_Registers Plugin_Inst_Registers;

/*
 * Get_Inst_Input_Registers :
 *
 * Get the instruction's gpr/fpr/spr/mx input register (value,
 * and index) if there is.
 *
 * For VMX instructions, the input may be gpr or vmx registers.
 * Since the current implementation in mambo for VMX instructions
 * does not use register renaming, this function will return
 * incorrect value when output register is the same one with the
 * input. The solution is to save the input register values
 * twice : one is before execution (the original value), another
 * is after execution (the modified value). Caller should do that.
 *
 * For non-VMX instructions, the input may be gpr, fpr or spr.
 * The code to support fpr is written but not tested. And the
 * support for spr is not available now. This function relies on
 * the register renaming facility in eycool. Make sure that your
 * mambo enables eycool and you have used
 * "./configure --enable-qtdump" before compiling mambo.
 *
 * Return values :
 * 1) gpr/fpr/spr/vmx_num (0~MAX_GPRS/FPRS/SPRS/VMXRS) :
 * 0 : no input register.
 * 1~MAX_GRPS/FPRS/SPRS/VMXRS : the number of input registers.
 *
 * 2) gpr/fpr/spr/vmx_index :
 * The register index.
 *
 * 3) gpr/fpr/spr/vmx_value :
 * The register value.
 */

extern PLUGIN_API void Get_Inst_Input_Registers(Instruction_Handler ihdlr, Plugin_Inst_Registers * Reg);

/*
 * Get_Inst_Output_Registers :
 *
 * Get the instruction's output registers. Output register may
 * be gpr, spr, fpr and vmx registers. The code for fpr is
 * written but not tested.
 *
 * Return values :
 * 1) gpr/fpr/spr/vmx_num :
 * 0 : no output register.
 * 1~MAX_GPRS/FPRS/SPRS/VMXRS : The number of output registers.
 *
 * 2) gpr/fpr/spr/vmx_index :
 * The register index.
 *
 * 3) gpr/fpr/spr/vmx_value :
 * The register value.
 */

extern PLUGIN_API void Get_Inst_Output_Registers(Instruction_Handler ihdr, Plugin_Inst_Registers * Reg);

/* ************************************************************ */
/*                                                              */
/*                                                              */
/* ************************************************************ */

/* Thread_Handler */
extern PLUGIN_API uint32 Get_Thread_Id(Thread_Handler thdlr);

#endif /* #ifdef ENABLE_EYCOOL */
#endif /* #ifdef CONFIG_PLUGIN */
#endif /* #ifndef _PLUGIN_API_H_ */
