/* -------------------------------------------------------
 *
 * Licensed Materials - Property of IBM.
 * (C) Copyright IBM Corporation 2004, 2006
 * All Rights Reserved.
 *
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corporation.
 *
 * Filename : trans_id.h
 *
 * Purpose    : Operations on transaction ID's
 *
 * ------------------------------------------------------- */

#ifndef _TRANS_ID_H_
#define _TRANS_ID_H_

#include "trans_id_t.h"

#ifndef MCONFIG_NO_TRID

/*___________________________________________________________
 * MAJOR ID Methods
 */

/**
 * TRID_GetMajor: Retrieve the Major ID
 *
 * \param   t        Pointer to a TRID
 * \return  Major ID of TRID
 */
extern TRID_Major TRID_GetMajor(TRID* t);

/**
 * TRID_SetDesc: Set the fragment description
 *
 * \param   t        Pointer to a TRID
 * \param   desc     Fragment Description (or NULL)
 */
extern void TRID_SetDesc(TRID *t, const char *desc);

/**
 * TRID_SetDone: Set/clear the TRID done flag
 *
 * \param   t        Pointer to a TRID
 * \param   done     TRUE if done, FALSE otherwise
 */
extern void TRID_SetDone(TRID *t, boolean done);

/**
 * TRID_IsDone: Return TRUE if TRID has been marked completed
 *
 * \param   t        Pointer to a TRID
 * \return TRUE if done, FALSE otherwise
 */
extern boolean TRID_IsDone(TRID *t);

/*___________________________________________________________
 * FRAGMENT Methods
 */

/**
 * TRID_SetFragDesc: Set the fragment description
 *
 * \param   t        Pointer to a TRID
 * \param   desc     Fragment Description (or NULL)
 */
extern void TRID_SetFragDesc(TRID *t, const char *desc);

/**
 * TRID_SetFrag: Set the fragment ID in the current branch
 *
 * \param   t        Pointer to a TRID
 * \param   fid      Fragment ID
 * \param   desc     Fragment Description (or NULL)
 */
extern void TRID_SetFrag(TRID *t, TRID_Frag fid, const char *desc);

/**
 * TRID_NewFrag: Increment the fragment ID in the current branch
 *
 * \param   tp        Pointer to a TRID parent
 * \param   tc        Pointer to a TRID child (initialized by this method)
 * \param   desc      Description (or NULL)
 */
extern void TRID_NewFrag(TRID *tp, TRID *tc, const char *desc);

/**
 * TRID_GetFragID: Get the Fragment ID at the tail of the list
 *
 * \param   t        Pointer to a TRID
 * \return Fragment ID in current branch
 */
extern TRID_Frag TRID_GetFragID(TRID* t);

/**
 * TRID_GetFragDesc: Get the Fragment Description at the tail of the list
 *
 * \param   t        Pointer to a TRID
 * \param   desc     (OUT) Pointer to description string to fill
 * \param   n        Number of bytes in description string storage
 */
extern void TRID_GetFragDesc(TRID *t, char *desc, unsigned int n);

/*___________________________________________________________
 * BRANCH Methods
 */

/**
 * TRID_DeleteAllBranches: Delete the most recent fragment branch
 *
 * \param   t        Pointer to a TRID
 */
extern void TRID_DeleteAllBranches(TRID *t);

/**
 * TRID_GetBranchLevel: Determine the branch level of the given TRID
 *
 * \param   t        Pointer to a TRID
 * \return Level number of this TRID
 */
extern uint32 TRID_GetBranchLevel(TRID *t);

/**
 * TRID_IsFragmented: Return TRUE if the transaction has branches
 *
 * \param   t        Pointer to a TRID
 * \return TRUE if the transaction has branches
 */
extern boolean TRID_IsFragmented(TRID *t);

/*___________________________________________________________
 * TRID Methods
 */

/**
 * TRID_Initialize: Initialze a TRID
 *
 * \param   t        Pointer to a TRID to be initialized
 * \param   desc     Description (or NULL)
 */
extern void TRID_Initialize(TRID* t, const char *desc);

/**
 * TRID_Invalidate: Invalidate a TRID
 *
 * \param   t        Pointer to a TRID to be invalidated
 */
extern void TRID_Invalidate(TRID* t);

/**
 * TRID_IsValid: Query the validity of a TRID
 *
 * \param   t        Pointer to a TRID to be checked
 *
 * \return TRUE if TRID is valid
 */
extern boolean TRID_IsValid(TRID* t);

/**
 * TRID_Copy: Copy a TRID from source to destination
 *
 * \param   dst     Pointer to Destination TRID
 * \param   src     Pointer to Source TRID
 */
extern void TRID_Copy(TRID *dst, TRID *src);

/**
 * TRID_Match: Determine whether two TRID's match
 *
 * \param   t1      First TRID
 * \param   t2      Second TRID
 * \param   levels  Resolution (number of levels to match, 0 == FULL)
 * \return  TRUE if match, FALSE otherwise
 */
extern boolean TRID_Match(TRID *t1, TRID *t2, uint32 levels);

/**
 * TRID_LessThan: Determine whether t1 < t2
 *
 * \param   t1      First TRID
 * \param   t2      Second TRID
 * \return  TRUE if t1 < t2, FALSE otherwise
 */
extern boolean TRID_LessThan(TRID *t1, TRID *t2);

/**
 * TRID_ToString: Convert a TRID to a string
 *
 * \param   t   Pointer to a TRID
 * \param   s   String to fill with TRID representation
 * \param   len Maximum length of s
 */
extern void TRID_ToString(TRID* t, char* s, uint32 len);

/**
 * TRID_DescToString: Convert a TRID description to a string
 *
 * \param   t       Pointer to a TRID
 * \param   s       String to fill with TRID representation
 * \param   len     Maximum length of s
 * \param   delim   Delimiter character to use
 */
extern void TRID_DescToString(TRID* t, char* s, uint32 len, char delim);

/**
 * TRID_FromString: Convert a string to a TRID
 *
 * \param   t   Pointer to a TRID
 * \param   s   String to parse
 */
extern void TRID_FromString(TRID* t, char* s);

/**
 * TRID_DescFromString: Convert a string to a TRID description
 *
 * \param   t       Pointer to a TRID
 * \param   s       String to parse
 * \param   delim   Character to interpret as a delimiter
 */
extern void TRID_DescFromString(TRID* t, char* s, char delim);

/**
 * TRID_Print: Send the string representation to the output stream
 *
 * \param   fp  Output stream pointer
 * \param   t   Pointer to a TRID
 */
extern void TRID_Print(FILE *fp, TRID* t);

#else /* #ifndef MCONFIG_NO_TRID */

#define TRID_GetMajor(t) 0
#define TRID_SetDesc(t, desc)
#define TRID_SetDone(t, done)
#define TRID_IsDone(t) FALSE
#define TRID_SetFragDesc(t, desc)
#define TRID_SetFrag(t, fid, desc)
#define TRID_NewFrag(tp, tc, desc)
#define TRID_GetFragID(t) 0
#define TRID_GetFragDesc(t, desc, n)
#define TRID_DeleteAllBranches(t)
#define TRID_GetBranchLevel(t) 0
#define TRID_IsFragmented(t) FALSE
#define TRID_Initialize(t, desc)
#define TRID_Invalidate(t)
#define TRID_IsValid(t) FALSE
#define TRID_Copy(dst, src)
#define TRID_Match(t1, t2, levels) FALSE
#define TRID_LessThan(t1, t2) FALSE
#define TRID_ToString(t, s, len)
#define TRID_DescToString(t, s, len, delim)
#define TRID_FromString(t, s)
#define TRID_DescFromString(t, s, delim)
#define TRID_Print(fp, t)

#endif /* #ifndef MCONFIG_NO_TRID */

#endif /* #ifndef _TRANS_ID_H_ */
