#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2005, 2006
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#   cell_util.tcl
#
#   TCL procs for various utility functions.
#------------------------------------------------------------------------

proc set_spu_model_pipeline { {sim mysim} } {
    puts "***** SETTING SPU CYCLE ACCURATE (PIPELINE) MODE *****"
    for {set mcm 0} {$mcm < [$sim display number_of_MCMs]} {incr mcm} {
        for {set spunum 0} {$spunum < 8} {incr spunum} {
            $sim mcm $mcm spu $spunum set model pipeline
        }
    }
}

proc set_spu_model_fast { {sim mysim} } {
    puts "***** SETTING SPU FAST MODE *****"
    for {set mcm 0} {$mcm < [$sim display number_of_MCMs]} {incr mcm} {
        for {set spunum 0} {$spunum < 8} {incr spunum} {
            $sim mcm $mcm spu $spunum set model fast
        }
    }
}

proc set_spu_model_instruction { {sim mysim} } {
    puts "***** SETTING SPU INSTRUCTION MODE *****"
    for {set mcm 0} {$mcm < [$sim display number_of_MCMs]} {incr mcm} {
        for {set spunum 0} {$spunum < 8} {incr spunum} {
            $sim mcm $mcm spu $spunum set model instruction
        }
    }
}

proc set_timing_mode { {sim mysim} } {
    puts "Timing mode deprecated.  Use cycle mode"
}

proc set_caches_and_bus_on { {sim mysim} } {

    for {set mcm 0} {$mcm < [$sim display number_of_MCMs]} {incr mcm} {
        if {[catch {$sim mcm $mcm caches set on}]} {
            return
        }
        if {[catch {$sim mcm $mcm bus on}]} {
            return
        }
    }

    puts "***** SETTING CACHES AND BUS ON MODE *****"
}

proc turn_perf_models_on { {sim mysim} } {
    # Enable all the appropriate performance models
    $sim mode cycle
}

proc set_mode_from_env { {sim mysim} } {
    global env

    # First check for the system modes
    if {[info exists env(FASTEST)] && $env(FASTEST)==1} {
        $sim mode fastest
    } elseif {[info exists env(FULL)] && $env(FULL)==1} {
        $sim mode full
    } elseif {[info exists env(CYCLE)] && $env(CYCLE)==1} {
        $sim mode cycle
    } else {

	# Now start with simple mode and ...
        $sim mode simple

        # turn on individual models if specified
        if {[info exists env(CACHES)] && $env(CACHES)==1} {
            set_caches_and_bus_on $sim
        }  
        if {[info exists env(PIPELINE)] && $env(PIPELINE)==1} {
            set_spu_model_pipeline $sim
        }
    }
}

# Set the memory size for a configuration.  The sz parameter specifies
# the amount of memory for each Cell BE processor.
proc config_memory_size { conf sz } {
    set size [memsize_toint $sz]
    set onegig [expr 1024*1024*1024]
    if {[compare64 $size $onegig] > 0} {
        set num_regions [div64 $size $onegig]
        set sz 1024M
    } else {
        set num_regions 1
    }
    $conf config memory_regions $num_regions
    $conf config memory_size $sz
    if {[$conf query number_of_MCMs] > 1} {
        $conf config mcm/1/memory_size $sz
        $conf config mcm/1/memory_start [format64 "0x%llx" [memsize_toint $sz]]
    }
}

# Display the memory size for a configuration.
proc display_memory_size { conf } {
    set onemeg      [expr 1024*1024]
    set region_size [memsize_toint [$conf query memory_size]]
    set num_regions [$conf query memory_regions]
    set size        [expr $region_size * $num_regions]
    # Display memory per BE
    puts "Memory size per BE is [format "%dM" [div64 $size $onemeg]]"
    set num_bes     [$conf query number_of_MCMs]
    if {$num_bes > 1} {
        puts "Memory size for the system is [format "%dM" [expr $num_bes*[div64 $size $onemeg]]]"
    }
}

proc config_dual_be { conf { version cell } } {

    $conf config number_of_MCMs 2
    $conf config mcm_attachment shared
    $conf config bus model_2be_system on
    $conf config bus be0_as_ac0 on

    $conf config mcm/1/ROM/start 0xF0000080

    $conf config mu/read_early_start "predict" ;# predict based on BE configuration and MIC Cfg Cntl register setting
    $conf config mu/no_read_reissue on

    set sz [$conf query memory_size] 
    $conf config mcm/1/memory_size $sz
    $conf config mcm/1/memory_start [format64 "0x%llx" [memsize_toint $sz]]

    switch -- $version {
        cell { $conf config mcm/1/be/bp_base 0x20100000000 }
        cpbw { }
        default {
            error "Unknown dual_be version requested"
        }
    }
}

proc enable_isolation_mode { {sim mysim} } {
    global env

    if { ! [ Config::is_defined "MCONFIG_SPU_ISOCNTL" ] } {
        error "Isolation mode support is not available"
    }
     
    # Find the spe_isolation_loader to use
    if {[file exists ./spe_isolation_loader]} {
        set loader_file ./spe_isolation_loader
    }  elseif {[file exists $env(IMAGES_DIR)/cell/spe_isolation_loader]} {
        set loader_file $env(IMAGES_DIR)/cell/spe_isolation_loader
    } else {
        error "could not find spe_isolation_loader file"
    }

    for {set mcm 0} {$mcm < [$sim display number_of_MCMs]} {incr mcm} {
        set root_node [$sim mcm $mcm of peer 0]
        set isolate_node [$sim mcm $mcm of addchild $root_node "spu-isolation" ""]
        set fhandle [open $loader_file]
        mysim mcm $mcm of addprop $isolate_node byte_array "loader" [read $fhandle]
        close $fhandle
    }
}

proc spu_load_tnk { sim fname } {

     if { ![ file exists $fname ] } {
         puts "spu_load_tnk: file not found"
         return
     }
     set file_size    [ file size $fname ]
     set ls_size [ expr 1024*256 ]
     set offset 0

     for { set spu_num 0 } { $spu_num < 8 } { incr spu_num } {
         set size $ls_size
         if { $size > $file_size } {
             set size $file_size
         }
         puts "loadimage spu:$spu_num, size= $size, from file offset $offset\n"
         $sim spu $spu_num loadimage $fname $offset $size
         set offset [ expr $offset + $size ]
         set file_size [ expr $file_size - $size ]
         if { $file_size == 0 } {
             puts "loadimage: end of file reached\n"
             return
         }
     }
}

proc init_instruction_monitor { args } {

     global stop_iim_flag run_time

     set stop_iim_flag 0
     set run_time 0
     eval instruction_monitor_loop $args
}

proc stop_instruction_monitor { args } {

     global stop_iim_flag
     set stop_iim_flag 1
     puts "Stopping Tcl Instruction Count Monitor"
}

proc iim    { args } { eval init_instruction_monitor $args }
proc stopim { args } { eval stop_instruction_monitor $args }

proc instruction_monitor_loop { {sim mysim} {msec 1000} {ic0p 0} {ic1p 0} {mcp 0} } {

     global stop_iim_flag run_time

     if { $stop_iim_flag } then {
         return
     }

     if { $msec < 1000  } then { set msec 1000 }
     #if { $msec > 10000 } then { set msec 10000 }

     set ic0  [ $sim cpu 0:0 display instruction_count ]
     set ic1  [ $sim cpu 0:1 display instruction_count ]
     set mc   [ $sim display cycles ]
     set ic0d [ sub64 $ic0 $ic0p ]
     set ic1d [ sub64 $ic1 $ic1p ]
     set mcd  [ sub64 $mc  $mcp  ]

     puts "\n==========================================================================="
     puts "   Tcl Instruction        CPU 0:0 total = [ Controls::CommaInsert $ic0  ]"
     puts "   Count Monitor                  delta = [ Controls::CommaInsert $ic0d ]"
     puts " "
     puts "                          CPU 0:1 total = [ Controls::CommaInsert $ic1  ]"
     puts "                                  delta = [ Controls::CommaInsert $ic1d ]"
     puts " "
     puts "                          Cycles  total = [ Controls::CommaInsert $mc   ]"
     puts "                                  delta = [ Controls::CommaInsert $mcd  ]"
     puts " "
     puts "                          Host Time: [ exec date ]"
     puts "                           Run Time: $run_time seconds"
     puts "===========================================================================\n"
     after $msec "instruction_monitor_loop $sim $msec $ic0 $ic1 $mc"
     incr run_time [ expr $msec / 1000 ]
}


namespace eval spu_map {
     variable ls_area_start
     set      ls_area_start(0)  0
     variable num_spc_areas                 8
     variable ls_size                 0x40000
     variable spu_problem_area_size   0x20000
     variable spu_priv1_area_size      0x2000
     variable spc_area_size   [expr (1 << (18 + 1))]        ;# (2 * LS_SIZE) rounded up to a power of 2 boundary.
}

proc spu_map::ls_area {spunum {be 0}} {
     variable ls_area_start
     variable spc_area_size
     return [expr $ls_area_start($be) + ($spunum * $spc_area_size) ]
}

proc spu_map::spu_problem_area {spunum {be 0}} {
     variable spc_area_size

     return [expr [ls_area $spunum $be] + ($spc_area_size >> 1) ]
}

proc spu_map::spu_priv2_area {spunum {be 0}} {
     variable spu_problem_area_size

     return [expr [spu_problem_area $spunum $be] + $spu_problem_area_size ]
}

proc spu_map::spu_priv1_area {spunum {be 0}} {
     variable spu_priv1_area_size
     variable num_spc_areas
     variable spu_priv1_area_start [ ls_area $num_spc_areas ]

     return [expr $spu_priv1_area_start + ($spu_priv1_area_size * $spunum) ]
}

proc spu_map::locate_spu {spunum {be 0} } {
  set lsarea [format %x [spu_map::ls_area $spunum $be] ]
  set prob   [format %x [spu_map::spu_problem_area $spunum $be] ]
  set priv2  [format %x [spu_map::spu_priv2_area $spunum $be] ]
  set priv1  [format %x [spu_map::spu_priv1_area $spunum $be] ]
  return "LS $lsarea Problem $prob Priv2 $priv2 Priv1 $priv1"
}

proc memmap_spu {spunum {be 0} } {
    puts "spu $spunum: [spu_map::locate_spu $spunum $be]"
}

