#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : pviz.tcl
##
## Purpose    : GUI for SPU pipeline statistics visualization
##
## CVS INFO:
##
## $Id$
##
#####################################################
#####################################################

package require BLT

namespace eval PVIZ {
    variable version        "2.0"
    variable initialized    0
}

#####################################################
#####################################################
##
## ELEMENT HIGHLIGHT/DISPLAY OPERATIONS
##
#####################################################
#####################################################

####
# Highlight a chart element. The element will be drawn in
# its active color and the associated toggle label will be
# shown in red.
#
# \param    parent  Parent widget
# \param    chart   Chart widget
# \param    args    If supplied, the chart element to highlight, otherwise
#                   we highlight the current chart element
#
proc PVIZ::Highlight_Chart_Element {parent chart args} {
    if {[llength $args] > 0} {
        set elem [lindex $args 0]
    } else {
        set elem [$chart element get current]
    }
    $chart element activate $elem
    ${parent}.${elem} configure -fg red
}

####
# Remove highlighting from a chart element. The element will be
# drawn in its normal color and the associated toggle will be
# shown in black.
#
# \param    parent  Parent widget
# \param    chart   Chart widget
# \param    args    If supplied, the chart element to remove highlighting from,
#                   otherwise, we de-highlight the current chart element
#
proc PVIZ::Lowlight_Chart_Element {parent chart args} {
    if {[llength $args] > 0} {
        set elem [lindex $args 0]
    } else {
        set elem [$chart element get current]
    }
    $chart element deactivate $elem
    ${parent}.${elem} configure -fg black
}

####
# Display/Hide a chart element. If the linked variable
# is set, then we append the element name to the chart's
# show list (displaying the element). If the linked variable
# is not set, then we remove the element name from the
# chart's show list (hiding the element).
#
# \param    chart   Chart widget
# \param    var     Variable linked to element
# \param    elem    Element name
#
proc PVIZ::Show_Element {chart var elem} {
    upvar #0 $var toggle_on

    set elem_list [$chart element show]
    if {$toggle_on} {
        lappend elem_list $elem
    } else {
        set idx [lsearch -exact $elem_list $elem]
        if {$idx >= 0} {
            set elem_list [lreplace $elem_list $idx $idx]
        }
    }

    $chart element show $elem_list
}

#####################################################
#####################################################
##
## MARKER HIGHLIGHT/DISPLAY OPERATIONS
##
#####################################################
#####################################################

####
# Show the chart marker at the indicated cursor location
#
# \param    sim     Simulator name
# \param    chart   Chart widget name
# \param    x       Cursor x location
# \param    y       Cursor y location
# \param    group   Stats group name (perf, fetch, issue, exec)
# \param    args    If present, the spu number to use, otherwise
#                   we use the currently selected spu for the detail view
#
proc PVIZ::Show_Mark {sim chart x y group args} {
    set elem [$chart element get current]

    # Check for optional spu number
    if {[llength $args] > 0} {
        set spu [lindex $args 0]
    } else {
        set spu [$sim pviz get spu]
    }

    # Find the element closest to the cursor position
    if {[$chart element closest $x $y search $elem]} {

        # If the bin index selected is within range...
        set max_idx [$sim pviz get bin]
        if {$search(index) <= $max_idx} {

            # Transform cursor coordinates to graph coordinates
            set gx [$chart axis invtransform x $x]
            set gy [$chart axis invtransform y $y]

            # Find the time corresponding to the selected bin
            # set t_off [expr {$search(index) * [$sim pviz get time delta]}]
            # set now [expr {[$sim pviz get time start] + $t_off}]
            set t_list [$sim pviz get time index $search(index)]
            set now "[lindex $t_list 0][lindex $t_list 2]"

            # Get the mark data for the selected bin
            array set data [$sim pviz get mark $group $search(index) $spu]

            # Format the mark text
            if {[string equal $group "perf"]} {
                set txt "at:$now\npc- $data(addr):$data(asm)\ncpn:$data(cpn), mode:$data(mode), count:$data(counting)"
            } else {
                array set p0_data $data(0)
                array set p1_data $data(1)
                set txt "at:$now\n0- $p0_data(addr):$p0_data(asm)\n1- $p1_data(addr):$p1_data(asm)"
            }

            # Determine where to position the marker text (up/right if selected bin is
            # in the first half of the graph, up/left if in the second half)
            set anch sw
            if {$search(index) > 50} {
                set anch se
            }
            $chart marker configure mrk -hide 0 -text $txt -coords {$gx $gy} -anchor $anch
        }
    }
}

####
# Hide a marker for the indicated chart
#
# \param    chart   Chart widget name
# \param    m       Marker name
#
proc PVIZ::Hide_Mark {chart m} {
    $chart marker configure $m -hide 1 -coords {}
}

#####################################################
#####################################################
##
## CROSSHAIRS TRACKING OPERATIONS
##
#####################################################
#####################################################

####
# Display the crosshairs
#
# \param    graph   Name of the graph widget
# \param    x       Cursor X position
# \param    y       Cursor Y position
#
proc PVIZ::Show_Crosshairs {graph x y} {
    $graph crosshairs on
    $graph crosshairs config -pos @$x,$y
}

####
# Hide the crosshairs
#
# \param    graph   Graph widget name
#
proc PVIZ::Hide_Crosshairs {graph} {
    $graph crosshairs off
}

####
# Track crosshairs with cursor movement
#
# \param    graph   Graph widget name
# \param    X       Cursor X position
# \param    Y       Cursor Y position
#
proc PVIZ::Track_Crosshairs {graph x y} {
    $graph crosshairs config -pos @$x,$y
}

####
# Set up the crosshair bindings
#
# \param    graph   Graph widget name
#
proc PVIZ::Bind_Crosshairs {graph} {
    bind $graph <Enter> {
        PVIZ::Show_Crosshairs %W %x %y
    }
    bind $graph <Leave> {
        PVIZ::Hide_Crosshairs %W
    }
    bind $graph <Motion> {
        PVIZ::Track_Crosshairs %W %x %y
    }
}

#####################################################
#####################################################
##
## BARCHART/GRAPH DECLARATION FUNCTIONS
##
#####################################################
#####################################################

####
# Instance a stacked bar chart. The legend is hidden, since
# we'll use a set of toggles instead.
#
# \param    sim     Simulator name
# \param    parent  Name of parent widget
# \param    title   Title of bar chart
#
proc PVIZ::Declare_Chart {sim parent title} {
    set delta [$sim pviz get time delta]
    set chart [blt::barchart ${parent}.chart -plotbackground gray32]
    $chart configure -title $title -barmode stacked -barwidth $delta -height 2.25i -width 7i
    #$chart axis configure x -stepsize [expr {$delta * 10}] -subdivisions 5 -title "Cycles"
    $chart axis configure x -stepsize 0 -subdivisions 5 -title "Cycles"
    $chart crosshairs configure -color magenta
    $chart pen configure "activeBar" -borderwidth 0 -fg cyan
    $chart legend configure -hide yes
    return $chart
}

####
# Instance a barchart plotting element. A toggle button will also be
# associated with the element to hide/show the element. Right-clicking
# on the toggle will cause the element to be highlighted.
#
# \param    parent      Name of parent widget
# \param    chart       Name of chart widget
# \param    el_name     Name of the element to create
# \param    v_name      Name of the variable to link to the element
# \param    title       Title for the element
# \param    color       Color of the element
#
proc PVIZ::Declare_Chart_Element {parent chart el_name v_name title color} {
    $chart element create $el_name -xdata x_axis -ydata $v_name
    $chart element configure $el_name -fg $color -borderwidth 0 -label $title

    # Set up a toggle button for the element
    set w [checkbutton ${parent}.${el_name} -text $title -selectcolor $color \
        -variable PVIZ::show_$v_name -command \
        [list PVIZ::Show_Element $chart PVIZ::show_$v_name $el_name]]
    $w select

    # Set up right-click bindings
    bind $w <ButtonPress-3> [list                             \
        PVIZ::Highlight_Chart_Element $parent $chart $el_name \
    ]
    bind $w <ButtonRelease-3> [list                           \
        PVIZ::Lowlight_Chart_Element $parent $chart $el_name  \
    ]

    # Return the name of the toggle button associated with the element
    return $w
}

####
# Instance a graph. The legend will be replaced by a set of toggles
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
# \param    title   Graph title
#
proc PVIZ::Declare_Plot {sim parent title} {
    set delta [$sim pviz get time delta]
    set plot [blt::graph ${parent}.plot -plotbackground gray32]
    $plot configure -title $title -height 2.25i -width 7i
    #$plot axis configure x -stepsize [expr {$delta * 10}] -subdivisions 1 -title "Cycles"
    $plot axis configure x -subdivisions 5 -title "Cycles"
    $plot grid configure -hide no -mapx ""
    $plot crosshairs configure -color cyan
    $plot pen configure "activeLine" -color magenta -symbol ""
    $plot legend configure -hide yes
    return $plot
}

####
# Instance a graph element for plotting. A toggle button will be associated
# with the element to show/hide. Right-clicking on the toggle will highlight
# the plotting element.
#
# \param    parent      Parent widget name
# \param    plot        Graph widget name
# \param    el_name     Name of element to create
# \param    v_name      Name of variable to link to the element
# \param    title       Title of the element
# \param    color       Color of the element
#
proc PVIZ::Declare_Plot_Element {parent plot el_name v_name title color} {
    $plot element create $el_name -xdata x_axis -ydata $v_name
    $plot element configure $el_name -color $color -label $title -symbol "" -smooth linear

    # Create a toggle button to show/hide the element
    set w [checkbutton ${parent}.${el_name} -text $title -selectcolor $color \
        -variable PVIZ::show_$v_name -command                               \
        [list PVIZ::Show_Element $plot PVIZ::show_$v_name $el_name]]
    $w select

    # Set up the right-click bindings
    bind $w <ButtonPress-3> [list                            \
        PVIZ::Highlight_Chart_Element $parent $plot $el_name \
    ]
    bind $w <ButtonRelease-3> [list                          \
        PVIZ::Lowlight_Chart_Element $parent $plot $el_name  \
    ]

    # Return the name of the associated toggle button
    return $w
}

####
# Set up bindings for the right-clicking on the element. The associated
# toggle will be highlighted
#
# \param    parent      Name of Parent widget
# \param    chart       Name of chart/graph widget
#
proc PVIZ::Bind_Legends {parent chart} {
    $chart element bind all <ButtonPress-3> [list   \
        PVIZ::Highlight_Chart_Element $parent %W    \
    ]
    $chart element bind all <ButtonRelease-3> [list \
        PVIZ::Lowlight_Chart_Element $parent %W     \
    ]
}

#####################################################
#####################################################
##
## DETAIL PLOTS
##
#####################################################
#####################################################

####
# Summary Performance Plot
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Perf_Plot {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.perf -relief groove -borderwidth 2]

    # Instance the plot
    set plot [PVIZ::Declare_Plot $sim $frm "Summary"]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Plot_Element $frm $plot sflop perf_sp_flop "Sflop" orange]
    set t1 [PVIZ::Declare_Plot_Element $frm $plot dflop perf_dp_flop "Dflop" purple]
    set t2 [PVIZ::Declare_Plot_Element $frm $plot inst perf_inst "Inst" green]
    set t3 [PVIZ::Declare_Plot_Element $frm $plot cycle perf_cycle "Cycle" red]
    $plot element show {sflop dflop inst cycle}

    # Create and bind the markers
    $plot marker create text -name mrk -foreground yellow -background "" -justify left
    $plot element bind all <ButtonPress-1> [list    \
        PVIZ::Show_Mark $sim %W %x %y perf          \
    ]
    $plot marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $plot
    PVIZ::Bind_Legends $frm $plot

    # Create a layout for the frame
    blt::table $frm                 \
       $plot         0,0  -cspan 4  \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2            \
       $t3           1,3

    # Return the frame name
    return $frm
}

####
# DMA Detail Chart
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::DMA_Chart {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.dma -relief groove -borderwidth 2]

    # Instance the plot
    set chart [PVIZ::Declare_Chart $sim $frm "DMA"]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Chart_Element $frm $chart get dma_get "Get" darkblue]
    set t1 [PVIZ::Declare_Chart_Element $frm $chart put dma_put "Put" purple]
    set t2 [PVIZ::Declare_Chart_Element $frm $chart slb dma_slb_miss "SLB" darkgreen]
    set t3 [PVIZ::Declare_Chart_Element $frm $chart tlb dma_tlb_miss "TLB" darkred]
    set t4 [PVIZ::Declare_Chart_Element $frm $chart ato_get dma_ato_get "AGet" darkorange]
    set t5 [PVIZ::Declare_Chart_Element $frm $chart ato_put dma_ato_put "APut" gray]
    $chart element show {get put slb tlb ato_get ato_put}

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $chart
    PVIZ::Bind_Legends $frm $chart

    # Create a layout for the frame
    blt::table $frm                 \
       $chart        0,0 -cspan 6   \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2            \
       $t3           1,3            \
       $t4           1,4            \
       $t5           1,5

    # Return the frame name
    return $frm
}

####
# Fetch Detail Chart
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Fetch_Chart {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.fetch -relief groove -borderwidth 2]

    # Instance the plot
    set chart [PVIZ::Declare_Chart $sim $frm "Fetch"]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Chart_Element $frm $chart branch_stall fetch_branch_stall "Misp" darkblue]
    set t1 [PVIZ::Declare_Chart_Element $frm $chart hint_stall fetch_hint_stall "Hint" purple]
    set t2 [PVIZ::Declare_Chart_Element $frm $chart pref_miss fetch_prefetch_miss "Pref" darkgreen]
    set t3 [PVIZ::Declare_Chart_Element $frm $chart core_flush fetch_core_flush "Flush" darkred]
    set t4 [PVIZ::Declare_Chart_Element $frm $chart chan_stall fetch_channel_stall "Chan" darkorange]
    $chart element show {branch_stall hint_stall pref_miss core_flush chan_stall}

    # Create and bind the markers
    $chart marker create text -name mrk -foreground yellow -background "" -justify left
    $chart element bind all <ButtonPress-1> [list   \
        PVIZ::Show_Mark $sim %W %x %y fetch         \
    ]
    $chart marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $chart
    PVIZ::Bind_Legends $frm $chart

    # Create a layout for the frame
    blt::table $frm                 \
       $chart        0,0 -cspan 5   \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2            \
       $t3           1,3            \
       $t4           1,4

    # Return the frame name
    return $frm
}

####
# Issue Detail Chart
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Issue_Chart {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.issue -relief groove -borderwidth 2]

    # Instance the plot
    set chart [PVIZ::Declare_Chart $sim $frm "Issue"]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Chart_Element $frm $chart issue issue_issue "Issu" darkgreen]
    set t1 [PVIZ::Declare_Chart_Element $frm $chart op_dep issue_op_dep "Dpnd" darkred]
    set t2 [PVIZ::Declare_Chart_Element $frm $chart dp_block issue_dp_block "DBlk" darkorange]
    set t3 [PVIZ::Declare_Chart_Element $frm $chart sp_block issue_sp_block "SBlk" purple]
    set t4 [PVIZ::Declare_Chart_Element $frm $chart stall issue_iss_stall "Stll" darkblue]
    set t5 [PVIZ::Declare_Chart_Element $frm $chart sbi issue_sbi_conflict "SBI" gray]
    $chart element show {issue op_dep dp_block sp_block stall}

    # Create and bind the markers
    $chart marker create text -name mrk -foreground yellow -background "" -justify left
    $chart element bind all <ButtonPress-1> [list   \
        PVIZ::Show_Mark $sim %W %x %y issue         \
    ]
    $chart marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $chart
    PVIZ::Bind_Legends $frm $chart

    # Create a layout for the frame
    blt::table $frm                 \
       $chart        0,0 -cspan 6   \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2            \
       $t3           1,3            \
       $t4           1,4            \
       $t5           1,5

    # Return the frame name
    return $frm
}

####
# General Execution Detail Chart
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Exec_Chart {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.exec -relief groove -borderwidth 2]

    # Instance the plot
    set chart [PVIZ::Declare_Chart $sim $frm "Exec"]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Chart_Element $frm $chart nop exec_NOP "NOP" darkblue]
    set t1 [PVIZ::Declare_Chart_Element $frm $chart lnop exec_LNOP "LNOP" darkorange]
    set t2 [PVIZ::Declare_Chart_Element $frm $chart ls exec_LS "LS" darkgreen]
    set t3 [PVIZ::Declare_Chart_Element $frm $chart br exec_BR "BR" darkred]
    $chart element show {nop lnop ls br}

    # Create and bind the markers
    $chart marker create text -name mrk -foreground yellow -background "" -justify left
    $chart element bind all <ButtonPress-1> [list   \
        PVIZ::Show_Mark $sim %W %x %y exec          \
    ]
    $chart marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $chart
    PVIZ::Bind_Legends $frm $chart

    # Create a layout for the frame
    blt::table $frm                 \
       $chart        0,0 -cspan 4   \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2            \
       $t3           1,3

    # Return the frame name
    return $frm
}

####
# Integer Execution Detail Chart
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Exec_Int_Chart {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.exec_int -relief groove -borderwidth 2]

    # Instance the plot
    set chart [PVIZ::Declare_Chart $sim $frm "Exec Int"]

    set t0 [PVIZ::Declare_Chart_Element $frm $chart fx2 exec_FX2 "FX2" darkorange]
    set t1 [PVIZ::Declare_Chart_Element $frm $chart fx3 exec_FX3 "FX3" darkred]
    set t2 [PVIZ::Declare_Chart_Element $frm $chart fxb exec_FXB "FXB" darkgreen]
    set t3 [PVIZ::Declare_Chart_Element $frm $chart shuf exec_SHUF "SHUF" darkblue]
    $chart element show {fx2 fx3 fxb shuf}

    # Create and bind the markers
    $chart marker create text -name mrk -foreground yellow -background "" -justify left
    $chart element bind all <ButtonPress-1> [list   \
        PVIZ::Show_Mark $sim %W %x %y exec          \
    ]
    $chart marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $chart
    PVIZ::Bind_Legends $frm $chart

    # Create a layout for the frame
    blt::table $frm                 \
       $chart        0,0 -cspan 4   \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2            \
       $t3           1,3

    return $frm
}

####
# Floating Point Execution Detail Chart
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Exec_Float_Chart {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.exec_float -relief groove -borderwidth 2]

    # Instance the plot
    set chart [PVIZ::Declare_Chart $sim $frm "Exec Float"]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Chart_Element $frm $chart fp6 exec_FP6 "FP6" darkorange]
    set t1 [PVIZ::Declare_Chart_Element $frm $chart fp7 exec_FP7 "FP7" darkgreen]
    set t2 [PVIZ::Declare_Chart_Element $frm $chart fpd exec_FPD "FPD" darkred]
    $chart element show {fp6 fp7 fpd}

    # Create and bind the markers
    $chart marker create text -name mrk -foreground yellow -background "" -justify left
    $chart element bind all <ButtonPress-1> [list   \
        PVIZ::Show_Mark $sim %W %x %y exec          \
    ]
    $chart marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $chart
    PVIZ::Bind_Legends $frm $chart

    # Create a layout for the frame
    blt::table $frm                 \
       $chart        0,0 -cspan 3   \
       $t0           1,0            \
       $t1           1,1            \
       $t2           1,2

    # Return the frame name
    return $frm
}

#####################################################
#####################################################
##
## SUMMARY PLOTS
##
#####################################################
#####################################################

####
# SPU Summary Chart
#
# \param    sim     Simulator name
# \param    spu     SPU number
# \param    parent  Parent widget name
#
proc PVIZ::Summary_Plot {sim spu parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.summary_spu${spu} -relief groove -borderwidth 2]

    # Plot title
    set title [button ${frm}.title_spu${spu} -text "SPE:$spu" -command [list \
        PVIZ::Tab_Detail_Focus $sim $spu                                     \
        ]]

    # Instance the plot
    set plot [PVIZ::Declare_Plot $sim $frm ""]

    # Instance the plot elements
    set t0 [PVIZ::Declare_Plot_Element $frm $plot get spu_${spu}_get "Get" green]
    set t1 [PVIZ::Declare_Plot_Element $frm $plot put spu_${spu}_put "Put" red]
    set t2 [PVIZ::Declare_Plot_Element $frm $plot cycle spu_${spu}_cycle "Cyc" gray]
    set t3 [PVIZ::Declare_Plot_Element $frm $plot inst  spu_${spu}_inst "Inst" yellow]
    set t4 [PVIZ::Declare_Plot_Element $frm $plot chan spu_${spu}_chan "Chan" orange]
    set t5 [PVIZ::Declare_Plot_Element $frm $plot sflop spu_${spu}_sflop "SFlp" blue]
    set t6 [PVIZ::Declare_Plot_Element $frm $plot dflop spu_${spu}_dflop "DFlp" purple]
    $plot element show {get put cycle inst chan sflop dflop}

    # Create and bind the markers
    $plot marker create text -name mrk -foreground yellow -background "" -justify left
    $plot element bind all <ButtonPress-1> [list    \
        PVIZ::Show_Mark $sim %W %x %y perf $spu     \
    ]
    $plot marker bind all <ButtonPress-1> {
        PVIZ::Hide_Mark %W [%W marker get current]
    }

    # Set up the element highlight and crosshair bindings
    PVIZ::Bind_Crosshairs $plot
    PVIZ::Bind_Legends $frm $plot

    # Create a layout for the frame
    blt::table $frm                 \
       $title        0,0  -cspan 7  \
       $plot         1,0  -cspan 7  \
       $t0           2,0            \
       $t1           2,1            \
       $t2           2,2            \
       $t3           2,3            \
       $t4           2,4            \
       $t5           2,5            \
       $t6           2,6

    # Return the frame name
    return $frm
}

#####################################################
#####################################################
##
## STATUS AND CONTROLS
##
#####################################################
#####################################################

####
# Marker summary text area
#
# \param    parent      Parent widget name
#
proc PVIZ::Marker_Summary_Area {parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.mark_summ -relief groove -borderwidth 2]

    # Instance the status labels
    set status_lbl [label ${frm}.status -textvariable PVIZ::Status_Marker_Summary  -width 80 -justify left -anchor w]
    set pc_lbl     [label ${frm}.pc     -textvariable PVIZ::PC_Marker_Summary      -width 80 -justify left -anchor w]
    set fetch0_lbl [label ${frm}.fetch0 -textvariable PVIZ::Fetch_0_Marker_Summary -width 80 -justify left -anchor w]
    set fetch1_lbl [label ${frm}.fetch1 -textvariable PVIZ::Fetch_1_Marker_Summary -width 80 -justify left -anchor w]
    set issue0_lbl [label ${frm}.issue0 -textvariable PVIZ::Issue_0_Marker_Summary -width 80 -justify left -anchor w]
    set issue1_lbl [label ${frm}.issue1 -textvariable PVIZ::Issue_1_Marker_Summary -width 80 -justify left -anchor w]
    set exec0_lbl  [label ${frm}.exec0  -textvariable PVIZ::Exec_0_Marker_Summary  -width 80 -justify left -anchor w]
    set exec1_lbl  [label ${frm}.exec1  -textvariable PVIZ::Exec_1_Marker_Summary  -width 80 -justify left -anchor w]

    # Create a layout for the frame
    blt::table $frm                 \
        $status_lbl     0,0         \
        $pc_lbl         1,0         \
        $fetch0_lbl     2,0         \
        $fetch1_lbl     3,0         \
        $issue0_lbl     4,0         \
        $issue1_lbl     5,0         \
        $exec0_lbl      6,0         \
        $exec1_lbl      7,0

    # Return the frame name
    return $frm
}

####
# Utility function to set the BE number. Check the validity of the
# input. If invalid, highlight in red and do nothing.
#
# \param    w       Widget name
# \param    sim     Simulator name
# \param    var     Linked variable name
#
proc PVIZ::Set_BE_Entry {w sim var} {
    upvar #0 $var txt
    if {$txt > [$sim pviz get be max]} {
        $w configure -foreground red -background black
    } else {
        $w configure -foreground black -background white
        $sim pviz set be $txt
        PVIZ::Set_Detail_Plot_Background $sim [$sim pviz get spu]
    }
}

####
# Utility function to set the SPU number. Check the validity of the
# input. If invalid, highlight in red and do nothing.
#
# \param    w       Widget name
# \param    sim     Simulator name
# \param    var     Linked variable name
#
proc PVIZ::Set_SPU_Entry {w sim var} {
    upvar #0 $var txt
    if {$txt > [$sim pviz get spu max]} {
        $w configure -foreground red -background black
    } else {
        $w configure -foreground black -background white
        $sim pviz set spu $txt
        PVIZ::Set_Detail_Plot_Background $sim $txt
    }
}

####
# Utility function to set the time delta. Check the validity of the
# input. If invalid, highlight in red and do nothing.
#
# \param    w       Widget name
# \param    sim     Simulator name
# \param    var     Linked variable name
#
proc PVIZ::Set_Time_Delta_Entry {w sim var} {
    upvar #0 $var txt
    if {$txt < 1} {
        $w configure -foreground red -background black
    } else {
        $w configure -foreground black -background white
        $sim pviz set delta $txt
    }
}

####
# Utility function to set the start/stop PVIZ state. Does nothing
# if already in the indicated state.
#
# \param    sim     Simulator name
# \param    var     Linked variable name
#
proc PVIZ::Start_Stop {sim var} {
    upvar #0 $var state
    switch -exact $state {
        "running"   {
            if {! [string equal [$sim pviz get state] $state]} {
                PVIZ::Set_Detail_Plot_Background $sim [$sim pviz get spu]
                $sim pviz run
            }
        }
        "stopped"   {
            if {! [string equal [$sim pviz get state] $state]} {
                $sim pviz stop
            }
        }
        default {
            puts "PVIZ::Start_Stop -- Unrecognized state '$state'"
        }
    }
}

####
# Utility function to set the overflow handler. Does nothing
# if already using the indicated handler
#
# \param    sim     Simulator name
# \param    var     Linked variable name
#
proc PVIZ::Overflow_Handler {sim var} {
    upvar #0 $var handler
    if {! [string equal [$sim pviz get overflow] $handler]} {
        $sim pviz set $handler
    }
}

####
# Instance the control panel
#
# \param    sim     Simulator name
# \param    parent  Parent widget name
#
proc PVIZ::Controls {sim parent} {
    # Frame for grouping purposes
    set frm [frame ${parent}.controls -relief groove -borderwidth 2]

    # SPU/BE number entry area. Typing a RETURN in these entry
    # widgets will invoke the proc to do the PVIZ set command
    set idx_frm [frame ${frm}.idx_frm -relief groove -borderwidth 2]

    global PVIZ::BE_Number
    set PVIZ::BE_Number [$sim pviz get be]
    set be_lbl [label ${idx_frm}.be_lbl -text "CBE"]
    set be_ent [entry ${idx_frm}.be_ent -textvariable PVIZ::BE_Number -relief sunken -width 2]
    bind $be_ent <Key> [list                                        \
        $be_ent configure -foreground white -background black       \
    ]
    bind $be_ent <Return> [list                         \
        PVIZ::Set_BE_Entry %W $sim PVIZ::BE_Number      \
    ]

    global PVIZ::SPU_Number
    set PVIZ::SPU_Number [$sim pviz get spu]
    set spu_lbl [label ${idx_frm}.spu_lbl -text "SPE"]
    set spu_ent [entry ${idx_frm}.spu_ent -textvariable PVIZ::SPU_Number -relief sunken -width 2]
    bind $spu_ent <Key> [list                                       \
        $spu_ent configure -foreground white -background black      \
    ]
    bind $spu_ent <Return> [list                       \
        PVIZ::Set_SPU_Entry %W $sim PVIZ::SPU_Number   \
    ]

    # Create a layout for the SPU/BE entry area
    blt::table $idx_frm     \
        $be_lbl     0,0     \
        $be_ent     0,1     \
        $spu_lbl    0,2     \
        $spu_ent    0,3


    # Run/Stop radio button area
    global PVIZ::Run
    set PVIZ::Run [$sim pviz get state]
    set run_frm [frame ${frm}.run_frm -relief groove -borderwidth 2]
    set run_start [radiobutton ${run_frm}.run -text "Run" -variable PVIZ::Run -value running \
        -command [list PVIZ::Start_Stop $sim PVIZ::Run]]
    set run_stop [radiobutton ${run_frm}.stop -text "Stop" -variable PVIZ::Run -value stopped \
        -command [list PVIZ::Start_Stop $sim PVIZ::Run]]

    blt::table $run_frm     \
        $run_start  0,0     \
        $run_stop   0,1

    # Overflow selection radio button area
    global PVIZ::Overflow
    set PVIZ::Overflow scroll
    set ovf_frm [frame ${frm}.ovf_frm -relief groove -borderwidth 2]
    set ovf_scroll [radiobutton ${ovf_frm}.scroll -text "Scroll" -variable PVIZ::Overflow -value scroll \
        -command [list PVIZ::Overflow_Handler $sim PVIZ::Overflow]]
    set ovf_compress [radiobutton ${ovf_frm}.compress -text "Compress" -variable PVIZ::Overflow -value compress \
        -command [list PVIZ::Overflow_Handler $sim PVIZ::Overflow]]

    blt::table $ovf_frm        \
        $ovf_scroll      0,0   \
        $ovf_compress    0,1

    # Time delta entry area. Typing a RETURN in this are will call the
    # TCL proc needed to invoke the PVIZ set command
    global PVIZ::Time_Delta
    set PVIZ::Time_Delta [$sim pviz get time delta]
    set dlt_frm [frame ${frm}.dlt_frm -relief groove -borderwidth 2]
    set dlt_lbl [label ${dlt_frm}.dlt_lbl -text "Time Delta"]
    set dlt_ent [entry ${dlt_frm}.dlt_ent -textvariable PVIZ::Time_Delta -relief sunken -width 10]
    bind $dlt_ent <Key> [list                                       \
        $dlt_ent configure -foreground white -background black      \
    ]
    bind $dlt_ent <Return> [list                                \
        PVIZ::Set_Time_Delta_Entry %W $sim PVIZ::Time_Delta     \
    ]

    blt::table $dlt_frm        \
        $dlt_lbl      0,0      \
        $dlt_ent      0,1

    # Set up the marker summary area
    set marker_frm [PVIZ::Marker_Summary_Area $frm]

    # Create the layout for the control panel
    blt::table $frm                         \
        $run_frm    0,0     $ovf_frm    0,1 \
        $idx_frm    1,0     $dlt_frm    1,1 \
        $marker_frm 2,0 -cspan 2

    # Return the control panel frame
    return $frm
}

#####################################################
#####################################################
##
## PVIZ CALLBACKS
##
#####################################################
#####################################################

####
# X-Axis rescale callback. PVIZ calls this when the X-Axis needs
# to be rescaled.
#
# \param    sim     Simulator name
# \param    min     Min X-Axis tick value
# \param    max     Max X-Axis tick value
# \param    delta   Tick delta value
# \param    raw_delta   Unscaled delta value
# \param    unit    Scale unit
#
proc PVIZ::Rescale_Callback {sim min max delta raw_delta unit_lbl} {
    global PVIZ::Time_Delta
    set PVIZ::Time_Delta $raw_delta

    #set bstep  [expr {$delta * 10}]
    #if {$raw_delta < 2} {
        #set bstep 1
    #} else {
        #set bstep 0
    #}

    set root ".pviz.tabs.detail"
    ${root}.perf.plot axis configure x -min $min -max $max -stepsize 0 -title "$unit_lbl"

    set children [list dma.chart fetch.chart issue.chart exec.chart exec_int.chart exec_float.chart]
    foreach {w} $children {
        ${root}.${w} axis configure x -min $min -max $max -stepsize 0 -title "$unit_lbl"
        ${root}.${w} configure -barwidth $delta
    }

    set root ".pviz.tabs.summary"
    for {set spu 0} {$spu < 8} {incr spu} {
        ${root}.summary_spu${spu}.plot axis configure x -min $min -max $max -stepsize 0 -title "$unit_lbl"
    }
}

####
# Update callback. PVIZ calls this whenever the Y-Axis data needs
# to be refreshed (as a result of having switched modes, having reset, etc.)
# We use this to update the marker summary area.
#
# \param    sim     Simulator name
# \param    cycle   Current cycle (SPU time domain)
#
proc PVIZ::Update_Callback {sim cycle} {
    array set data [$sim pviz get mark all]

    array set status_data $data(perf)
    global PVIZ::Status_Marker_Summary
    set PVIZ::Status_Marker_Summary "Cycle\t$cycle (cpn:$status_data(cpn), mode:$status_data(mode), count:$status_data(counting))"

    global PVIZ::PC_Marker_Summary
    set PVIZ::PC_Marker_Summary "PC\t\[addr=$status_data(addr), inst=$status_data(inst)\]->'$status_data(asm)'"

    array set typ_data $data(fetch)
    array set p0_data $typ_data(0)
    array set p1_data $typ_data(1)
    global PVIZ::Fetch_0_Marker_Summary
    global PVIZ::Fetch_1_Marker_Summary
    set PVIZ::Fetch_0_Marker_Summary "Fetch\t0:\[addr=$p0_data(addr), inst=$p0_data(inst)\]->'$p0_data(asm)'"
    set PVIZ::Fetch_1_Marker_Summary "\t1:\[addr=$p1_data(addr), inst=$p1_data(inst)\]->'$p1_data(asm)'"

    array set typ_data $data(issue)
    array set p0_data $typ_data(0)
    array set p1_data $typ_data(1)
    global PVIZ::Issue_0_Marker_Summary
    global PVIZ::Issue_1_Marker_Summary
    set PVIZ::Issue_0_Marker_Summary "Issue\t0:\[addr=$p0_data(addr), inst=$p0_data(inst)\]->'$p0_data(asm)'"
    set PVIZ::Issue_1_Marker_Summary "\t1:\[addr=$p1_data(addr), inst=$p1_data(inst)\]->'$p1_data(asm)'"

    array set typ_data $data(exec)
    array set p0_data $typ_data(0)
    array set p1_data $typ_data(1)
    global PVIZ::Exec_0_Marker_Summary
    global PVIZ::Exec_1_Marker_Summary
    set PVIZ::Exec_0_Marker_Summary "Exec\t0:\[addr=$p0_data(addr), inst=$p0_data(inst)\]->'$p0_data(asm)'"
    set PVIZ::Exec_1_Marker_Summary "\t1:\[addr=$p1_data(addr), inst=$p1_data(inst)\]->'$p1_data(asm)'"
}

####
# Detail element focus callback. PVIZ calls this whenever the SPU or BE number
# selected for the detail page is changed. We use this to remove any chart markers
# (since they'll be associate with another SPU and therefore stale).
#
# \param    sim     Simulator name
# \param    be      BE number selected
# \param    spu     SPU number selected
#
proc PVIZ::Detail_Element_Callback {sim be spu} {
    set root ".pviz.tabs.detail"
    set children [list perf.plot fetch.chart issue.chart exec.chart exec_int.chart exec_float.chart]

    foreach {w} $children {
        ${root}.${w} marker configure mrk -hide 1 -coords {}
    }
}

####
# SPU Enable state callback. PVIZ calls this whenever the SPU mode or stats collection
# changes. We use this to update the run/stop state toggles of PVIZ and to indicate whether
# data for an SPU is being collected
#
# \param    sim     Simulator name
# \param    be      Number of BE which changed
# \param    spu     Number of SPU which changed
# \param    enabled TRUE if enabled
# \param    running TRUE if PVIZ is running
proc PVIZ::SPU_State_Change_Callback {sim be spu enabled running} {
    global PVIZ::Run
    if {$running} {
        set PVIZ::Run "running"
	    if {$be == [$sim pviz get be]} {
			if {$spu == [$sim pviz get spu]} {
				PVIZ::Set_Detail_Plot_Background $sim $spu $enabled
			}
			PVIZ::Set_Summary_Plot_Background $sim all
		}
    } else {
        set PVIZ::Run "stopped"
		PVIZ::Set_Detail_Plot_Background $sim any false
		PVIZ::Set_Summary_Plot_Background $sim all false
    }
}

#####################################################
#####################################################
##
## PANE OPERATIONS
##
######################################################
#####################################################

####
# Utility function to set the detail pane plot backgrounds
# according to spu enable state
#
# \param    sim     Simulator name
# \param    spu     SPU number
# \param    enabled (Optional), explicit enable
#
proc PVIZ::Set_Detail_Plot_Background {sim {spu any} {enabled unknown}} {
    if {[string equal $spu "any"]} {
		set spu [$sim pviz get spu]
    }

    switch $enabled {
        unknown {
            set thawed [$sim pviz get enabled $spu]
        }
        default {
            set thawed $enabled
        }
    }

    set root ".pviz.tabs.detail"
    set children [list perf.plot dma.chart fetch.chart issue.chart exec.chart exec_int.chart exec_float.chart]
    foreach {w} $children {
        if {$thawed} {
            ${root}.${w} configure -plotbackground black
        } else {
            ${root}.${w} configure -plotbackground gray32
        }
    }
}

####
# Utility function to set the summary pane plot backgrounds
# according to spu enable state
#
# \param    sim     Simulator name
# \param    spu     (Optional), explicit spu number
# \param    enabled (Optional), explicit enable
#
proc PVIZ::Set_Summary_Plot_Background {sim {spu all} {enabled unknown}} {
    set root ".pviz.tabs.summary"

    switch $spu {
        all {
            set max [$sim pviz get spu max]
            for {set s 0} {$s <= $max} {incr s} {
				switch $enabled {
					unknown {
						set thawed [$sim pviz get enabled $s]
					}
					default {
						set thawed $enabled
					}
				}

                if {$thawed} {
                    ${root}.summary_spu${s}.plot configure -plotbackground black
                } else {
                    ${root}.summary_spu${s}.plot configure -plotbackground gray32
                }
            }
        }
        default {
            switch $enabled {
                unknown {
                    set thawed [$sim pviz get enabled $spu]
                }
                default {
                    set thawed $enabled
                }
            }

            if {$thawed} {
				${root}.summary_spu${spu}.plot configure -plotbackground black
            } else {
				${root}.summary_spu${spu}.plot configure -plotbackground gray32
            }
        }
    }
}

####
# Utility function to perform pane tab switching
#
# \param    sim     Simulator name
# \param    tab     Pane tab name
#
proc PVIZ::Tab_Switch {sim tab} {
    $sim pviz set mode $tab
    #switch $tab {
        #summary {
			#global PVIZ::Run
            #PVIZ::Set_Summary_Plot_Background $sim
        #}
    #}
}

proc PVIZ::Tab_Detail_Focus {sim spu} {
    global PVIZ::SPU_Number
    set PVIZ::SPU_Number $spu

    $sim pviz set mode detail
    $sim pviz set spu $spu
    PVIZ::Set_Detail_Plot_Background $sim $spu
    .pviz.tabs select 0
}

#####################################################
#####################################################
##
## TOPLEVEL OPERATIONS
##
######################################################
#####################################################

####
# Initialize the PVIZ GUI
#
# \param    sim     Simulator name
#
proc PVIZ::Init {sim} {
    toplevel .pviz

    # Set up the summary/detail page tabs
    set tabs [blt::tabset .pviz.tabs -relief sunken -borderwidth 2]
    $tabs insert 0 "Detail"
    $tabs insert 1 "Summary"

    # Set up the detail page widgets
    set detail_frm [frame ${tabs}.detail]
    set f_control [PVIZ::Controls $sim $detail_frm]
    set f_perf [PVIZ::Perf_Plot $sim $detail_frm]
    set f_dma [PVIZ::DMA_Chart $sim $detail_frm]
    set f_fetch [PVIZ::Fetch_Chart $sim $detail_frm]
    set f_issue [PVIZ::Issue_Chart $sim $detail_frm]
    set f_exec [PVIZ::Exec_Chart $sim $detail_frm]
    set f_exec_int [PVIZ::Exec_Int_Chart $sim $detail_frm]
    set f_exec_float [PVIZ::Exec_Float_Chart $sim $detail_frm]

    # Create a layout for the detail page
    blt::table $detail_frm                   \
    $f_perf     0,0     $f_control      0,1  \
    $f_dma      1,0     $f_exec         1,1  \
    $f_fetch    2,0     $f_exec_int     2,1  \
    $f_issue    3,0     $f_exec_float   3,1

    # Set up the summary page widgets
    set summary_frm [frame ${tabs}.summary]
    for {set i 0} {$i < 8} {incr i} {
        set spu${i} [PVIZ::Summary_Plot $sim $i $summary_frm]
    }

    # Create a layout for the summary page
    blt::table $summary_frm                  \
    $spu0       0,0     $spu1           0,1  \
    $spu2       1,0     $spu3           1,1  \
    $spu4       2,0     $spu5           2,1  \
    $spu6       3,0     $spu7           3,1

    # Set up PVIZ command bindings for the tabs
    $tabs tab configure "Detail" -window $detail_frm -command [list     \
        PVIZ::Tab_Switch $sim detail                                    \
    ]
    $tabs tab configure "Summary" -window $summary_frm  -command [list  \
        PVIZ::Tab_Switch $sim summary                                   \
    ]
    pack $tabs

    # Register callbacks with PVIZ
    $sim pviz set callback rescale PVIZ::Rescale_Callback
    $sim pviz set callback update PVIZ::Update_Callback
    $sim pviz set callback element PVIZ::Detail_Element_Callback
    $sim pviz set callback enable PVIZ::SPU_State_Change_Callback

    # Determine which mode we're in, and set the active page accordingly
    set mode [$sim pviz get mode]
    switch -exact $mode {
        "detail" {
            $tabs select 0
        }
        "summary" {
            $tabs select 1
        }
        default {
            puts "PVIZ::Init -- unexpected PVIZ mode '$mode'"
        }
    }

    # Set up to be notified when the window manager destroys, shows, or hides
    # the toplevel window
    bind .pviz <Destroy> [list  \
        PVIZ::Destroy $sim      \
    ]
    bind .pviz <Unmap> [list    \
        PVIZ::Visibility $sim   \
    ]

    # PVIZ window title
    wm title .pviz "Cell SPE Statistics Charts"
    # Set the initialized flag
    set PVIZ::initialized 1
}

####
# Shut down the PVIZ dialog. This is called when the dialog window is closed.
# We stop the PVIZ collector and de-register any callbacks.
#
# \param    sim     Simulator name
#
proc PVIZ::Destroy {sim} {
    bind .pviz <Destroy> {}
    $sim pviz stop
    $sim pviz set callback rescale
    $sim pviz set callback update
    $sim pviz set callback element
    $sim pviz set callback enable
    set PVIZ::initialized 0
}

####
# Hide the PVIZ dialog. This is called when the dialog window is unmapped.
# We stop the PVIZ collector
#
# \param    sim     Simulator name
#
proc PVIZ::Visibility {sim} {
    switch [wm state .pviz] {
        iconic {
            $sim pviz stop
            set PVIZ::Run [$sim pviz get state]
        }
    }
}

####
# Display the PVIZ dialog (if iconified or obscured in the window stacking).
# This will also initialize the GUI if not already done.
#
# \param    sim     Simulator name
#
proc PVIZ::Display {sim} {
    if {! $PVIZ::initialized} {
        PVIZ::Init $sim
    }

    wm deiconify .pviz
    raise .pviz
}
