#!/usr/bin/bltwish
#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003,2004
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  spu_prof.tcl
#------------------------------------------------------------------------

package require BLT
namespace import -force blt::*

proc Compute_SPU_Tag {machine benum spunum} {
    return "${machine}_be_${benum}_spu_${spunum}"
}
proc SPU_LS_Fields { machine mcmnum spunum } {
    return [split [$machine mcm $mcmnum spu $spunum display tcl_stats local_store_columns] " "]
}

###########################################################
# read_spu_data
#
#   This procedure actually queries the underlying
#   tcl interpreter to find out what statistics are available
#   associated with the local store.
#   It then creates a number of vectors filled with this data.
#   These vectors are then used to generate charts/graphs
#
proc read_spu_data { machine benum spunum } {
    set sputag [Compute_SPU_Tag ${machine} ${benum} ${spunum}]

    set column_index 0
    #
    # Get the list of local store fields available
    #
    foreach column [SPU_LS_Fields $machine $benum $spunum] {
        #
        # Create a global vector for this column of data
        #
        eval "global V_${sputag}_${column}"
        eval "vector create V_${sputag}_${column}"

        # Call back into the underlying interpreter asking for the data
        # associated with these columns
        eval "V_${sputag}_${column} set [list [$machine mcm ${benum} spu ${spunum} display tcl_stats local_store_indices $column_index]]"

        #
        # For the first column (could be any column
        # since they all have the same number of entries)
        # we want to create an X-axis vector that just has a sequentially
        # numbered list that we can use to plot against.
        # If we use the LSADDR column, it can create gaps in our graph
        # (like with a vector of address every 128 bytes, blt will leave
        # gaps thinking we may want to plot something later in those gaps)
        #
        if { $column_index == 0 } {
            set index 0
            eval "vector create V_${sputag}_XAXIS"
            eval "set veclist \$V_${sputag}_${column}(:)"
            foreach i $veclist {
                eval "V_${sputag}_XAXIS append $index"
                incr index
            }
        }

        incr column_index
    }

}


####################################################################
# GraphColor
#
#   For every graph we want to generate a stream of different colors
#   that the graphs use to plot each line.
#
set base_colorlist {blue red green purple orange}
proc GraphColor { tag } {
    global base_colorlist
    global colorlist
    global colorindex

    if { [llength [array names colorindex $tag]] <= 0 } {
        # Initialize the color setup for this tag
        set colorindex($tag) 0
    }

    set retval [lindex $base_colorlist $colorindex($tag)]
    set colorindex($tag) [expr $colorindex($tag) + 1]
    if { $colorindex($tag) >= [llength $base_colorlist] } {
        set colorindex($tag) 0
    }

    return $retval
}

######################################################################
# FormatXTickLabels & ComputeXTicks
#
#   TODO: BLT was having trouble with the major x-ticks (like there
#   were too many)  I just turned them off for now.   Should find
#   out how to add some number of them.
#
proc FormatXTickLabels {chart val} {
    global graphOptions

    return ""
}
proc ComputeXTicks {chart axis vector} {
    $chart axis configure $axis -minorticks {0}
    $chart axis configure $axis -majorticks {0}
}

######################################################################
#
# SPULS_Init &  SPULS_Refresh
#
# These procedures are used to tie the local store widget into the
# treesim structure used by Mambo's GUI.   It just basically calls the
# Machine_SPUInit function with the right parameters gleaned out of the
# overall Mambo GUI.
#
proc SPULS_Init { root _upstate args } {
    global SPULS_Sim SPULS_SPU SPULS

    upvar $_upstate initstate
    TclUtils::copyarray initstate SPULS $root

    set sim $SPULS($root,sim)
    set spu $SPULS($root,spunum)

    # Tuck this away for refresh
    set SPULS_Sim($root) $sim
    set SPULS_SPU($root) $spu

    Machine_SPUInit $sim $root 0 $spu

    set title [treesim::get_path_name $SPULS($root,widget) $SPULS($root,treeidx) 1 ]
    wm title $root "$title: Local Store Stats"

    return "SPULS_Refresh"
}
proc SPULS_Refresh { root args } {
    global SPULS_Sim SPULS_SPU SPULS

    set sim $SPULS($root,sim)
    set spu $SPULS($root,spunum)

    read_spu_data $sim 0 $spu

    update
}


######################################################################
#
# LS_Graph_Logscale
#
#   This procedure is called by a checkbox button to turn on/off
#   the logscale for the y-axis in the graphs
#
proc LS_Graph_Logscale {chart checkbutt sputag} {
    eval "global ${sputag}_logscale"
    eval "set logscale_state \$${sputag}_logscale"
    if { $logscale_state == 0 } {
        # When changing the logscale we occasionally get errors associated with the
        # current min and max.   If we reexecute the command it seems to work
        catch {$chart axis configure y -logscale false} foo
        $chart axis configure y -logscale false
    } else {
        # When changing the logscale we occasionally get errors associated with the
        # current min and max.   If we reexecute the command it seems to work
        catch {$chart axis configure y -logscale true} foo
        $chart axis configure y -logscale true
    }
}
proc ChangeGraphedData {chart machine benum spunum} {

    set graph_list ""
    set spu_ls_fields [SPU_LS_Fields $machine $benum $spunum]
    set sputag [Compute_SPU_Tag ${machine} ${benum} ${spunum}]
    foreach column [lrange $spu_ls_fields 1 end] {
        eval "global ${sputag}_${column}_toggle"
        eval "set tog_state \$${sputag}_${column}_toggle"
        if { $tog_state == 1 } {
            lappend graph_list $column
        }
    }
    $chart element show $graph_list
}

######################################################################
#
# Machine_SPUInit -- Initalize the monitor of the specified SPU's
#                       local store
#
proc Machine_SPUInit { machine mywin benum spunum } {
    global RefreshDelay

    set sputag [Compute_SPU_Tag ${machine} ${benum} ${spunum}]

    if { [llength $mywin] == 0 } {
        set winname ".${sputag}"

        #
        # If the window exists, just call the refresh call and return
        #
        if { [winfo exists $winname] } {
            return
        }

        #
        # Otherwise, we need to create a window for this machine's status info
        #
        toplevel $winname
    } else {
        set winname $mywin
    }

    read_spu_data $machine $benum $spunum

    set chart ${winname}.spuls_graph
    scrollbar ${winname}.scrollx -orient horizontal -command "${chart} axis view x"
    scrollbar ${winname}.scrolly -orient vertical -command "${chart} axis view y"
    barchart ${chart} -title "CBE:${benum} SPE:${spunum} Local Store" -height 200 -shadow white
    eval "global ${sputag}_logscale"
    eval "set ${sputag}_logscale 0"
    eval "checkbutton ${winname}.logscale -text \"Use Logarithmic Scale for Y-Axis\" -variable ${sputag}_logscale -command \"LS_Graph_Logscale ${chart} ${winname}.logscale ${sputag}\""

    ZoomSupport_Init ${chart} 0 0

    set spu_ls_fields [SPU_LS_Fields $machine $benum $spunum]
    #set x_axis [lindex $spu_ls_fields 0]
    set x_axis XAXIS
    ZoomSupport_Init ${chart} 0 0
    frame ${winname}.toggles
    set index 0
    foreach column [lrange $spu_ls_fields 1 end] {
        set xvals V_${sputag}_${x_axis}
        set yvals V_${sputag}_${column}

        # Toggle checkbox to choose whether to map data
        eval "global ${sputag}_${column}_toggle"
        eval "set ${sputag}_${column}_toggle 1"
        eval "checkbutton ${winname}.toggles.toggle_${column} -text \"Graph_${column}\" -variable ${sputag}_${column}_toggle -command \"ChangeGraphedData ${chart} ${machine} ${benum} ${spunum}\""
        eval "table ${winname}.toggles ${winname}.toggles.toggle_${column} 0,$index -fill x"
        incr index

        eval "$chart pen create $column -foreground [GraphColor ${sputag}_LS] -relief flat"
        eval "$chart element create $column -xdata $xvals -ydata $yvals -pen $column"
    }

    ComputeXTicks ${chart} x V_${sputag}_${x_axis}

    ${chart} axis configure x -title ${x_axis} -scrollcommand "${winname}.scrollx set" -command FormatXTickLabels
    ${chart} axis configure y -title "Bytes" -scrollcommand "${winname}.scrolly set"

    frame ${winname}.textboxes
    label ${winname}.textboxes.spuls_contents_label -justify center -text "SPU Local Store Contents:"
    frame ${winname}.textboxes.spuls_contents
    label ${winname}.textboxes.details_label -justify center -text "Graph Data Point Details:"
    text ${winname}.textboxes.details  -height 8 -font textbox_font
    label ${winname}.textboxes.helpbox_label -justify center -text "Helpful Hints:"
    text ${winname}.textboxes.helpbox  -height 8 -font textbox_font
    SPULS_Contents_Init ${winname}.textboxes.spuls_contents $machine $benum $spunum

    table ${winname}.textboxes \
        ${winname}.textboxes.details_label          0,0 -fill both \
        ${winname}.textboxes.helpbox_label          0,1 -fill both \
        ${winname}.textboxes.details                1,0 -fill both \
        ${winname}.textboxes.helpbox                1,1 -rowspan 3 -fill both \
        ${winname}.textboxes.spuls_contents_label   2,0 -fill both \
        ${winname}.textboxes.spuls_contents         3,0 -fill both

    table configure ${winname}.textboxes c0 c1 r0 r1 r2 r3 -resize both

    # Show data stacked
    ${chart} configure -barmode stacked
    global LSMarker
    set LSMarker(${chart}) [${chart} marker create line -coords {0 0 0 2000000} -outline yellow -linewidth 2]

    table $winname \
        ${winname}.spuls_graph      0,0 -fill both \
        ${winname}.scrolly          0,1 -fill y \
        ${winname}.scrollx          1,0 -fill x \
        ${winname}.toggles          2,0 -fill x \
        ${winname}.logscale         3,0 \
        ${winname}.textboxes        4,0 -fill x

    table configure $winname r0 r1 r2 r3 r4 c0 c1 c2 -resize none
    table configure $winname c0 r0 -resize both

    set x_axis_labels V_${sputag}_[lindex $spu_ls_fields 0]
    bind ${chart} <Shift-Motion>       "UpdateMotion  \%W \%x \%y $winname $machine $benum $spunum ${winname}.textboxes.spuls_contents"

    ${winname}.textboxes.helpbox insert end "X-Axis spans the entire SPU local store address range\n"
    ${winname}.textboxes.helpbox insert end "\n"
    ${winname}.textboxes.helpbox insert end "Graph Details:\n"
    ${winname}.textboxes.helpbox insert end "1) The detail sections below the graph contain more details of a single data point in the graph\n"
    ${winname}.textboxes.helpbox insert end "2) The yellow line on the graph highlights the single data point that is being studied\n"
    ${winname}.textboxes.helpbox insert end "3) Hold down shift key while moving mouse over the graph (near the x-axis) to move the Yellow-Line\n"
    ${winname}.textboxes.helpbox insert end "\n"
    ${winname}.textboxes.helpbox insert end "Zooming In/Out:\n"
    ${winname}.textboxes.helpbox insert end "1) Click (mouse-button-1) on graph, drag, and release to zoom in on a region on the graph\n"
    ${winname}.textboxes.helpbox insert end "2) Right-Click (mouse-button-3) on graph to zoom back out\n"
    update
}

######################################################################
#
# UpdateMotion
#   This procedure is bound to the event of someone moving the mouse
#   across the graph.   It detects where the mouse it and provides more
#   data.
#
proc UpdateMotion {chart x y win machine benum spunum spuls_contents} {

    set text ${win}.textboxes.details

    ${text} delete 0.0 end

    global element_list
    ${chart} element closest $x $y element_list

    # If the element_list contains an index field then update text with location information
    if { [lsearch [array names element_list] index] != -1 } {
        set x_index $element_list(index)

        # Now dump all the fields from the vectors for this index
        set spu_ls_fields [SPU_LS_Fields $machine $benum $spunum]
        set sputag [Compute_SPU_Tag ${machine} ${benum} ${spunum}]
        set lsaddr 0
        foreach column $spu_ls_fields {
            #${text} insert end [format "Local_Store_Address=0x%X (%.2f K)\n" [expr int($ls_addr)] [expr $ls_addr / 1024]]
            set vec V_${sputag}_${column}
            eval "global $vec"
            eval "set val [expr int(\$${vec}($x_index))]"
            if { [string compare "$column" "LSADDR"] == 0 } {
                eval "set val [format "0x%05X" $val]"
                set lsaddr $val
                eval "set diff [expr int(\$${vec}(1)) - int(\$${vec}(0))]"
            } else {
                eval "set val [format "%d" $val]"
            }
            eval "set min [expr int(\$${vec}(min))]"
            eval "set max [expr int(\$${vec}(max))]"
            ${text} insert end [format "%10s % 16s (Min=%10d, Max=%10d)\n" $column $val $min $max]
        }

        global LSMarker
        set max_y [lindex [$chart axis limits y] 1]
        ${chart} marker configure $LSMarker($chart) -coords {$x_index 0 $x_index $max_y}
        SPULS_Contents_Refocus $spuls_contents $lsaddr $diff
    }
}
#XXXXXXXXXXXXXXXXX SPUTRACKET>TCL

proc spu_break_point { sim benum spunum comment args } {
    set pc [$sim mcm $benum spu $spunum display pc]
    puts "SPU $spu $pc: break point ($comment)"
    simstop
}

proc SPULS_Contents_ToggleBreak {root sim benum spunum txtwidget y} {
    global SPULS_Contents_Start
    set cursor [$txtwidget index @0,$y]
    set line [ lindex [ split $cursor . ] 0 ]
    set addr [add64 $SPULS_Contents_Start($root) [expr ($line-1) * 4]]
    set hexaddr [format64 "0x%016llX" $addr]
    if { [string length [$sim trigger display spu $spunum pc addr $hexaddr]] > 0 } {
        puts "Clearing breakpoint at $hexaddr"
        $sim trigger clear spu $spunum pc $hexaddr
        textutils::untag_breakpoint $txtwidget $cursor
    } else {
        puts "Setting breakpoint at $hexaddr"
        $sim trigger set spu $spunum pc $hexaddr "spu_break_point $sim $benum $spunum Break_at_$hexaddr"
        textutils::tag_breakpoint $txtwidget $cursor
    }
}

######################################################################
#
# SPULS_Contents_Init -- Initalize the status window
#
# ------------------------------------------------------------------
proc SPULS_Contents_Init {root sim benum spunum} {

    global SPULS_Contents_End
    global SPULS_Contents_Length SPULS_Contents_Start
    global SPULS_Contents_UserLength SPULS_Contents_UserStart
    global SPULS_Contents

    set SPULS_Contents($root,sim) $sim
    set SPULS_Contents($root,benum) $benum
    set SPULS_Contents($root,spunum) $spunum

    set SPULS_Contents_Start($root) "0"
    set SPULS_Contents_UserStart($root) "0"
    set SPULS_Contents_End($root)   "-1"
    set SPULS_Contents_Length($root)     128
    set SPULS_Contents_UserLength($root) 128

    # starting address
    frame $root.start -borderwidth 2 -relief sunken
    label $root.start.label -text "Start Addr:" -pady 4
    entry $root.start.addr -textvariable SPULS_Contents_UserStart($root)
    pack $root.start.label $root.start.addr -side left
    bind $root.start.addr <KeyPress-Return> "SPULS_Contents_Refresh $root"


    # length and format
    frame $root.leng -borderwidth 2 -relief sunken
    label $root.leng.label -text "Length:" -padx 4
    entry $root.leng.addr -textvariable SPULS_Contents_UserLength($root) -width 8 -justify right
    bind $root.leng.addr <KeyPress-Return> "SPULS_Contents_Refresh $root"
    pack $root.leng.label $root.leng.addr -side left

    button $root.reload -text "Reload" -command "SPULS_Contents_Refresh $root"

    text $root.sputrk -width 70 -height 16 -yscrollcommand "$root.sputrksy set" -font textbox_font  -tabs "12c" -wrap none
    scrollbar $root.sputrksy -orient vert -command "$root.sputrk yview"
    bind $root.sputrk <Double-1> "SPULS_Contents_ToggleBreak $root $sim $benum $spunum %W %y"

    textutils::init_color_tags $root.sputrk

    table $root \
        $root.sputrk   0,0 -cspan 4 -fill both \
        $root.sputrksy 0,4 -cspan 1 -fill y \
        $root.start    1,0 -fill both \
        $root.leng     1,1 -fill both \
        $root.reload   1,2 -fill x
    table configure $root c* r* -resize none
    table configure $root r0 c0 -resize expand

    SPULS_Contents_Refresh $root

    return "SPULS_Contents_Refresh"
}

######################################################################
#
# SPULS_Contents_Refresh -- Refresh the contents of the status window
#
######################################################################
proc SPULS_Contents_Refresh { root {force_reload 0} args } {
    global SPULS_Contents_ SPULS_Contents_Start SPULS_Contents_End
    global SPULS_Contents_UserLength SPULS_Contents_UserStart SPULS_Contents_Length

    global SPULS_Contents

    set sim $SPULS_Contents($root,sim)
    set benum $SPULS_Contents($root,benum)
    set spunum $SPULS_Contents($root,spunum)

    if {$SPULS_Contents_UserLength($root) != $SPULS_Contents_Length($root)} {
       set SPULS_Contents_UserLength($root) [ TclUtils::get_larger_power2 $SPULS_Contents_UserLength($root)]
       set force_reload 1
    }

    if {$SPULS_Contents_UserStart($root) != $SPULS_Contents_Start($root)} {
       set force_reload 1
    }

    if {$force_reload} {
        set SPULS_Contents_Length($root) $SPULS_Contents_UserLength($root)
        set pc $SPULS_Contents_UserStart($root)
    }

    # See if we need to reload the contents
    if { $force_reload } {
            set SPULS_Contents_Start($root) [expr $pc & ~( $SPULS_Contents_Length($root) - 1 )]
            set SPULS_Contents_UserStart($root) [format "0x%X" $SPULS_Contents_Start($root)]
            set SPULS_Contents_End($root) [expr $SPULS_Contents_Start($root) + $SPULS_Contents_Length($root)]

            set addr $SPULS_Contents_Start($root)

            set bplist [$sim trigger display spu $spunum pc list ]
            $root.sputrk delete 1.0 end
            while { $addr < $SPULS_Contents_End($root) } {
                set value [$sim mcm $benum spu $spunum display memory $addr 4]
                set ascii [TclUtils::GenAscii $value]
                if { [lsearch $bplist [format "0x%08X" $addr] ] >= 0 } {
                    $root.sputrk insert end "B " breakpoint
                } else {
                    $root.sputrk insert end "  "
                }
                set line [format "%08X : %08X : %s : %s\t\t\n" $addr $value $ascii [$sim util spu_disasm $value]]
                $root.sputrk insert end $line
                incr addr 4
            }
            $root.sputrk delete insert-1c end       ;# delete last newline char
    } else {
        # We do need to run through and update any disasm line that has changed
        set addr $SPULS_Contents_Start($root)
        set linenum 1
        set bplist [$sim trigger display spu $spunum pc list ]
        while { $addr < $SPULS_Contents_End($root) } {
            set index $linenum.0
            set value [$sim mcm $benum spu $spunum display memory $addr 4]
            set oldvalue [$root.sputrk get "$index+13chars" "$index+13chars wordend" ]
            #puts "value=$value  oldvalue=$oldvalue "
            if { [expr $value != "0x$oldvalue"] } {
                set ascii [TclUtils::GenAscii $value]
                set line [format "%08X : %08X : %s : %s\t\t" $addr $value $ascii [$sim util spu_disasm $value]]
                $root.sputrk delete $linenum.2 $linenum.end
                $root.sputrk insert $linenum.2 $line
            }
            incr addr 4
            incr linenum
        }
    }
}

proc SPULS_Contents_Refocus { root start length } {
    global SPULS_Contents_UserLength SPULS_Contents_UserStart
    set SPULS_Contents_UserLength($root) $length
    set SPULS_Contents_UserStart($root) $start
    SPULS_Contents_Refresh $root 0 0
}
