#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004, 2007
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : spumodes.tcl
##
## Purpose    : GUI for SPU mode panel
##
#####################################################

namespace eval SPUModes {
    variable version        "1.0"
    variable spu_state
}
####
# Utility function to set the update SPU mode state. Does nothing
# if already in the indicated state.
#
# \param    sim     Simulator name
# \param    be      BE number
# \param    spu     SPU number
#
proc SPUModes::Update {sim be spu} {
    set state $SPUModes::spu_state($be,$spu)
    switch -exact $state {
        "pipeline"   {
            if {! [string equal [$sim mcm $be spu $spu display model] $state]} {
                $sim mcm $be spu $spu set model pipeline
            }
        }
        "instruction"   {
            if {! [string equal [$sim mcm $be spu $spu display model] $state]} {
                $sim mcm $be spu $spu set model instruction
            }
        }
        "fast"   {
            if {! [string equal [$sim mcm $be spu $spu display model] $state]} {
                $sim mcm $be spu $spu set model fast
            }
        }
        default {
            puts "SPUModes::Update -- Unrecognized state '$state' for BE:$be, SPU:$spu"
        }
    }
}

####
# Set all SPU's to the indicated mode
#
# \param    sim     Simulator name
# \param    mode    SPU mode
#
proc SPUModes::SetAll {sim be mode} {
    for {set i 0} {$i < 8} {incr i} {
        if {[catch {$sim mcm $be spu $i set model $mode} errs]} {
            puts "SPUModes::SetAll -- Could not set be $be spu $i to mode $mode ($errs)"
        }
    }
    SPUModes::Refresh $sim
}

#####################################################
#####################################################
##
## TOPLEVEL OPERATIONS
##
######################################################
#####################################################

####
# Update the SPU mode toggles to match their current state
# 
# \param    sim     Simulator name
#
proc SPUModes::Refresh { sim } {

    for {set be 0} {$be < [$sim display number_of_MCMs]} {incr be} {
        for {set i 0} {$i < 8} {incr i} {
            set SPUModes::spu_state($be,$i) [$sim mcm $be spu $i display model]
        }
    }
}


####
# Display the PerfModels dialog (if iconified or obscured in the window stacking).
# This will also initialize the GUI if not already done.
#
# \param    sim     Simulator name
#
proc SPUModes::Create {sim} {

    set widget .spumodes

    if { ![ winfo exists $widget ] } {

        # Create a new toplevel window / frame 
        toplevel $widget

        # Mark new window as transient and working for .
        wm transient $widget .

        # Frame for grouping purposes
        frame $widget.frm -relief flat -borderwidth 0

        for {set be 0} {$be < [$sim display number_of_MCMs]} {incr be} {
            # Create an SPU Modes dialog for each BE

            # Frame for grouping purposes
            set befrm [frame $widget.frm.be${be} -relief groove -borderwidth 2]

            # Area title
            label ${befrm}.lbl -text "BE:$be"
            grid ${befrm}.lbl

            # Individual SPU mode radio button area
            for {set i 0} {$i < 8} {incr i} {
                set SPUModes::spu_state($be,$i) [$sim mcm $be spu $i display model]
                set state_frm [frame ${befrm}.spu${i} -relief groove -borderwidth 2]
                set state_lbl [label ${state_frm}.lbl -text "SPU$i:"]
                set pipe_togl [radiobutton ${state_frm}.pipe -text "Pipe" -variable SPUModes::spu_state($be,$i) \
                                   -value pipeline -command [list SPUModes::Update $sim $be $i]]
                set inst_togl [radiobutton ${state_frm}.inst -text "Instruction" -variable SPUModes::spu_state($be,$i) \
                                   -value instruction -command [list SPUModes::Update $sim $be $i]]
                set fast_togl [radiobutton ${state_frm}.fast -text "Fast" -variable SPUModes::spu_state($be,$i) \
                                   -value fast -command [list SPUModes::Update $sim $be $i]]
                grid $state_lbl $pipe_togl $inst_togl $fast_togl -sticky ew
                grid $state_frm -sticky ew
            }

            # Global settings buttons
            set all_frm [frame ${befrm}.all -relief flat -borderwidth 0]
            set all_lbl [label ${all_frm}.lbl -text "All BE:$be"]
            set all_pipe [button ${all_frm}.pipe -text "Pipe" -command [list SPUModes::SetAll $sim $be pipeline]]
            set all_inst [button ${all_frm}.inst -text "Instruction" -command [list SPUModes::SetAll $sim $be instruction]]
            set all_fast [button ${all_frm}.fast -text "Fast" -command [list SPUModes::SetAll $sim $be fast]]
            grid $all_lbl $all_pipe $all_inst $all_fast -sticky ew
            grid $all_frm -sticky ew

            # Refresh button
            set fresh [button ${befrm}.fresh -text "Refresh" -command [list SPUModes::Refresh $sim]]
            grid $fresh -sticky ew

            grid $befrm -row 0 -column $be
        }

        grid $widget.frm

        # Set up to be notified when the window manager destroys, shows, or hides
        # the toplevel window
        bind $widget <Destroy> [list SPUModes::Destroy $sim]
        bind $widget <Map>     [list SPUModes::Map $sim]
    }

    wm deiconify [winfo toplevel $widget]
    raise $widget
}

####
# Shut down the SPUModes dialog. This is called when the dialog window is closed.
#
# \param    sim     Simulator name
#
proc SPUModes::Destroy {sim} {
    bind .spumodes <Destroy> {}
}

####
# Show the SPUModes dialog. This is called when the dialog window is mapped.
#
# \param    sim     Simulator name
#
proc SPUModes::Map {sim} {
    SPUModes::Refresh $sim
}
