#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003,2004
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  spuutils.tcl
#
#------------------------------------------------------------------------

namespace eval SPUUtils {}

#------------------------------------------------------------------------
# address translation using SPU SLBs
#------------------------------------------------------------------------
proc SPU-itranslate {sim cpunum addr up_error} {
     upvar $up_error translate_error
     return [SPU-dtranslate $sim $cpunum $addr translate_error]
}

proc SPU-dtranslate {sim cpunum addr up_error} {
    upvar $up_error translate_error

    set translate_error 0
    if {[catch {
          set rAddr [$sim spu $cpunum dtranslate $addr]
       } ] == 1 } then {
       set translate_error 1
       set rAddr $addr
    }

    return $rAddr
}

#------------------------------------------------------------------------
#  SPUUtils::spu_watch_go
#------------------------------------------------------------------------
proc SPUUtils::spu_watch_go { sim spu ls_addr size } {

    set current_ls_value [$sim spu $spu display memory $ls_addr $size]

    puts "SPU Watch Go (Started): SPU=$spu LS-Address=$ls_addr Size=$size Current-Value=$current_ls_value"

    set done 0
    set first_time 1
    while { $done != 1 } {
        # check if we want to stop the simulation
        if { ( $first_time == 0 ) && [string compare [string trim [simstop status]] "cyclecount"] != 0 } {
            # something external has stopped the cpu
            set done 1
        } else {
            set first_time 0
            $sim cycle 1
        }

        if { [string compare [$sim spu $spu display memory $ls_addr $size] $current_ls_value] != 0 } {
            # Value has changed
            set done 1
        }
    }
    set current_ls_value [$sim spu $spu display memory $ls_addr $size]
    puts "SPU Watch Go (Stopped): SPU=$spu LS-Address=$ls_addr Size=$size Current-Value=$current_ls_value"
}

######################################################################
#
# SPUStats_Init -- Initalize the SPU Statistics window
#
######################################################################
proc SPUStats_Init {root _upstate args } {
    global SPUStats

    upvar $_upstate  initstate 
    TclUtils::copyarray initstate SPUStats $root       ;# create SPUStats($root,x)
    set spucmd $SPUStats($root,spucmd)

    set title [treesim::get_path_name $SPUStats($root,widget) $SPUStats($root,treeidx) 1 ] 
    wm title $root "$title: Statistics" 

    frame $root.buttons -relief flat -borderwidth 0
    button $root.buttons.reset -text "Reset Stats" -command "$spucmd stats reset"
    pack $root.buttons.reset -side left
    pack $root.buttons -side bottom -fill x

    frame $root.stats -relief flat -borderwidth 0
    text $root.stats.text -font textbox_font -width 80 -height 50 -yscrollcommand "$root.stats.sy set" -wrap none
    scrollbar $root.stats.sy -orient vert -command "$root.stats.text yview"
    pack $root.stats.sy -side right -fill y
    pack $root.stats.text -side left -fill both -expand true
    pack $root.stats -side top -fill both

    SPUStats_Refresh $root

    return "SPUStats_Refresh"
}

proc SPUStats_Refresh { root args } {
    global SPUStats

    $root.stats.text delete 1.0 end
    set spucmd $SPUStats($root,spucmd)
    $root.stats.text insert end [$spucmd display statistics]
}

#------------------------------------------------------------------------
#  Additional SPU procedures
#------------------------------------------------------------------------

proc SelectSPUElfFile { _upstate args} {

    upvar $_upstate  initstate
    set spu $initstate(spunum)
    set spucmd $initstate(spucmd)

    set SPUElfFile(file_title) "Open SPU Elf Executable to load"
    set SPUElfFile(file_types) { { "All Files" { * } } }
    set SPUElfFile(file_pref)  [PrefManager::get_value SPUElfFile "" ]

    set filename [getfilename SPUElfFile]
    if { $filename != ""} {
       PrefManager::set_value SPUElfFile $filename
       LoadSonySPUFile $spucmd $spu $filename
    }
}

proc LoadSonySPUFile { spucmd spu args } {
    set filename [lindex $args 0]
    if {$filename != ""} {
        puts "LoadSonySPUFile called to load $filename into SPU#$spu\n"
        $spucmd load $filename
        UpdateAll 0
    }
}

#------------------------------------------------------------------------
# show spu triggers
# called by triggers/breakpoints window display
#------------------------------------------------------------------------
proc SPUUtils::show_triggers {triggers sim mcm prefix} {

    $triggers insert end "$prefix  SPU triggers"

    for {set spu 0} {$spu < 8} {incr spu} {
        set spuprefix "$prefix SPU $spu: "

        set trigger_list [split [$sim mcm $mcm trigger display memory spu $spu list] "\n"]
        if {[llength $trigger_list] != 0} { 
            $triggers insert end "$spuprefix  MEMORY triggers"
            foreach trig $trigger_list {
                $triggers insert end "$spuprefix    $trig "
            }
        }

        set trigger_list [$sim mcm $mcm trigger display spu $spu pc list]
        if {[llength $trigger_list] != 0} { 
            $triggers insert end "$spuprefix  PC triggers"
            foreach trig $trigger_list {
                set setting [$sim mcm $mcm trigger display spu $spu pc addr $trig]
                $triggers insert end "$spuprefix    $trig : \{$setting\}"
            }
        }

        set trigger_list [$sim mcm $mcm trigger display spu $spu channel list]
        if {[llength $trigger_list] != 0} { 
            $triggers insert end "$spuprefix  CHANNEL triggers"
            foreach trig $trigger_list {
                set setting [$sim mcm $mcm trigger display spu $spu channel $trig]
                set mval [format "0x%X" [lindex $setting 1]]
                set mmsk [format "0x%X" [lindex $setting 2]]
                $triggers insert end "$spuprefix    $trig : mask=$mmsk, val=$mval, proc=\{[lindex $setting 3]\}"
            }
        }
    }
}
