#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  checkpoint.tcl
#
#   There is one external function that should be called out of this file.
#   Namely:
#
#       proc Checkpoint::checkpoint { {fileprefix ckpt} {sim mysim} } {
#
#   This proc will cause a number of checkpoint files to be created that
#   can be used to restore a given simulation on an architectural level.
#
#   The main file that is generated is called "$fileprefix.ckpt.$count"
#       $fileprefix : is the first argument passed to the proc (default is "ckpt")
#       [pid]       : is the process id of the running simulation
#       $count      : is the sequence number of the number of times
#                       checkpoint is called for a given simulator instance.
#                       (0...N)
#   This main file will contain another single entry point called "restore"
#   which is a TCL procedure which will contain TCL code that will recreate
#   a simulation instance with the checkpointed architectural state.
#
#------------------------------------------------------------------------
namespace eval Checkpoint {

   variable version       "1.0"
   variable ckpt_count    0
}

#############################  Status #######################################
#
#   This proc is used throughout the checkpoint code to do the following:
#       1) It prints out what is currently being checkpointed
#       2) It puts a comment in the restore TCL output file specifying
#           the region of machine state that is being checkpointed/restored
#       3) It puts a print statement in the restore code to report
#           what is being restored when the restore function is called
#
proc Checkpoint::print_status { outfile state } {

    # Put a comment in the TCL checkpoint file
    puts $outfile "\n\t#\n\t# Restore the $state\n\t#"
    # Print a status msg as we are checkpointing
    puts "Saving $state..."
    # On the restore side, also print a message as we restore
    puts $outfile "\tputs \"Restoring $state...\""
}

#############################  Header #######################################
#
#   This proc will print the header for the restore function
#   along with the code to reconstruct the configuration object
#   used to create the current simulator.   It will also produce
#   the command to construct the new restored simulator
#
proc Checkpoint::checkpoint_header { outfile sim } {

    global sim_configs

    puts $outfile "proc checkpoint_restore \{ \{restore_sim $sim\} \} \{\n"

    # Save machine configuration and then create a machine based on it

    # Figure out what the default configuration is
    set default_config [display default_configure]

    puts $outfile "\tdefine dup $default_config restore_conf"

    # Now setup commands to build on top of restore configuration with the details found
    # in the current simulator
    array set sim_configs [$sim display config_tcl]
    foreach name [array names sim_configs] {
        puts $outfile "\trestore_conf config $name $sim_configs($name)"
    }

    set start_tick [string trimright [$sim display ticks] "t"]
    puts $outfile "\trestore_conf config start_tick $start_tick"
    puts $outfile "\tset_new_config_values restore_conf"
    puts $outfile "\tdefine machine restore_conf \$restore_sim\n"
    puts $outfile "\t\$restore_sim set restore_in_progress on"
    puts $outfile "\t\$restore_sim set instruction_count [$sim display instruction_count]"
}

#############################  Footer #######################################
#
proc Checkpoint::checkpoint_footer { outfile } {
    puts $outfile "\t\$restore_sim set restore_in_progress off"
    puts $outfile "\}\n";
}

#############################  PPC #######################################
#
#   This proc is called for each PowerPC CPU in the system
#
#   It will iterate through calling all the appropriate checkpoint
#   procs to save the whole state of the powerpc core including:
#       GPRs, FPRs, VMXRs, and any additional MMU state
#
proc Checkpoint::checkpoint_ppc { outfile sim cpu nthread} {

    global sim_configs

    # Checkpoint gpr regs
    print_status $outfile "GPR's for CPU $cpu, Thread $nthread"
    for { set i 0 } { $i < 32 } { incr i } {
        set gprval [$sim cpu $cpu thread $nthread display gpr $i]
        puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set gpr $i $gprval"
    }

    # Checkpoint fpr regs
    if { [Config::is_defined CONFIG_FLOATING_POINT ] } {
        print_status $outfile "FPR's for CPU $cpu Thread $nthread"
        for { set i 0 } { $i < 32 } { incr i } {
            set fprval [$sim cpu $cpu thread $nthread display fpr $i]
            puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set fpr $i $fprval"
        }
    }

    # Checkpoint vmxr regs
    set PPCCoreNumVMX 0
    if { [ Config::is_defined CONFIG_VMX32 ] } {
        set PPCCoreNumVMX 32
    }
    if { [ Config::is_defined CONFIG_VMX128 ] } {
        set PPCCoreNumVMX 128
    }

    if { $PPCCoreNumVMX > 0 } {
        print_status $outfile "VMXR's for CPU $cpu Thread $nthread"
        for { set i 0 } { $i < $PPCCoreNumVMX } { incr i } {
            set vmxrval [$sim cpu $cpu thread $nthread display vmxr $i]
            puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set vmxr $i $vmxrval"
        }
    }

    # Checkpoint spr regs
    print_status $outfile "SPR's for CPU $cpu Thread $nthread"
    foreach name [$sim cpu $cpu thread $nthread display spr list] {
        if { [string first "tlb" $name] != 0 } {
            puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set spr $name [$sim cpu $cpu thread $nthread display spr $name]"
        }
    }

    # Checkpoint MMU (SLBs TLBs)
    print_status $outfile "MMU for CPU $cpu Thread $nthread"

    print_status $outfile "MMU's IERAT for CPU $cpu Thread $nthread"
    for { set i 0 } { $i < $sim_configs(processor/number_of_iERAT_entries) } { incr i } {
        puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set ieratc $i [$sim cpu $cpu thread $nthread display ieratc $i]"
        for { set j 0 } { $j < $sim_configs(processor/iERAT_depth) } { incr j } {
            puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set ierate $i $j [$sim cpu $cpu thread $nthread display ierate $i $j]"
        }
    }

    print_status $outfile "MMU's DERAT for CPU $cpu Thread $nthread"
    for { set i 0 } { $i < $sim_configs(processor/number_of_dERAT_entries) } { incr i } {
        puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set deratc $i [$sim cpu $cpu thread $nthread display deratc $i]"
        for { set j 0 } { $j < $sim_configs(processor/dERAT_depth) } { incr j } {
            puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set derate $i $j [$sim cpu $cpu thread $nthread display derate $i $j]"
        }
    }

    print_status $outfile "MMU's SLBs for CPU $cpu Thread $nthread"
    for { set i 0 } { $i < $sim_configs(processor/number_of_SLB_Entries) } { incr i } {
        puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set slbentry $i [$sim cpu $cpu thread $nthread display slbentry $i]"
    }

    print_status $outfile "MMU's TLBs for CPU $cpu Thread $nthread"
    puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set tlbregs [$sim cpu $cpu thread $nthread display tlbregs]"
    puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set tlbinfo [$sim cpu $cpu thread $nthread display tlbinfo]"
    for { set i 0 } { $i < $sim_configs(processor/number_of_TLB_entries) } { incr i } {
        puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set tlbentry $i [$sim cpu $cpu thread $nthread display tlbentry $i]"
    }

    # Checkpoint other states
    puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set instruction_count [$sim cpu $cpu thread $nthread display instruction_count]"
    puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set instruction_seq [expr [$sim cpu $cpu thread $nthread display instruction_seq] - 1]"
    puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set timebase [$sim cpu $cpu thread $nthread display timebase]"
    puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set dec [$sim cpu $cpu thread $nthread display dec]"

    #Reset these, since they're being overwritten
    foreach name {"msr" "srr0" "srr1" "UPMC1" "UPMC2" "PMC1" "PMC2" "pc"} {
	puts $outfile "\t\$restore_sim cpu $cpu thread $nthread set spr $name [$sim cpu $cpu thread $nthread display spr $name]"
    }

    #Test registers (to make sure nothing was overwritten)
    print_status $outfile "Checking SPR's for CPU $cpu Thread $nthread"
    foreach name [$sim cpu $cpu thread $nthread display spr list] {
        if { [string first "tlb" $name] != 0 } {
	    set sprval [$sim cpu $cpu thread $nthread display spr $name]
            puts $outfile "\t\set newval \[\$restore_sim cpu $cpu thread $nthread display spr $name\]"
	    puts $outfile "\tif { \[string compare \"$sprval\" \$newval\] != 0} { puts \"Warning: $name != $sprval\" }"
        }
    }
}

#############################  BOGUS #######################################
#
#   Checkpoint the state of the bogus devices
#
#   PJB: For now we will do bogus disk
#
proc Checkpoint::checkpoint_bogus { outfile sim filename_prefix } {

    print_status $outfile "Bogus Disks"

    foreach disk [$sim bogus disk display] {

        array set dinfo $disk
		$sim bogus disk sync $dinfo(minor)
        puts "Checkpoint cow disk $filename_prefix $dinfo(minor)"

        # Only support newcow for now
        switch $dinfo(cow_flags)  {
            newcow {
                # Need to copy and save the COW file
                set new_cow_filename $filename_prefix.$dinfo(minor).cow_file
                file copy -force $dinfo(cow_path) $new_cow_filename
                #don't destroy the original cow file when we restore
                set use_me $new_cow_filename._\[pid\]
                puts $outfile "\t file copy $new_cow_filename $use_me"
                puts $outfile "\t\$restore_sim bogus disk init $dinfo(minor) $dinfo(path) cow $use_me 1024"
            }
            restcow - 
            cow {
                set image_filename $filename_prefix.$dinfo(minor).image
                set table_filename $dinfo(cow_path).table
                set new_cow_filename ${image_filename}_cow
                set new_cow_table_filename  $new_cow_filename.table
                file copy -force $dinfo(path) $image_filename
                file copy -force $dinfo(cow_path) $new_cow_filename
                file copy -force $table_filename $new_cow_table_filename
                puts $outfile "\t\$restore_sim bogus disk init $dinfo(minor) $image_filename restcow ${image_filename}_cow 16384\n"
            }
            default {
                puts "ERROR: unable to checkpoint bogus disk of type $dinfo(cow_flags)... only newcow|cow|restcow supported for now"
                exit -1
            }
        }
    }
}

#############################  CACHES #######################################
#
#   Checkpoint the state of caches.
#
#
proc Checkpoint::checkpoint_cache { outfile sim cache_filename } {

    # Checkpoint caches
    set cache_status [$sim cache display status]
    switch $cache_status {
	off {
	    puts "Caches disabled, omitting cache state from checkpoint file\n"
	}
	default {
	    print_status $outfile "Cache"
	    $sim cache checkpoint $cache_filename
	    puts $outfile "\t\$restore_sim cache restart $cache_filename"
	}
    }
}

#############################  MEMORY #######################################
#
#   Checkpoint the state of memory.
#
proc Checkpoint::checkpoint_memory { outfile sim memory_filename } {

    # Checkpoint memory
    print_status $outfile "Memory"
    set memory_size [$sim display memory_size]

    if { [$sim memory fwritecmp 0 $memory_size $memory_filename] != $memory_size } {
        puts $outfile "\tputs \"ERROR: bad checkpoint... failed to save memory\""
        puts "ERROR: failed to checkpoint memory to file $memory_filename"
        exit -1
    } else {
        puts $outfile "\tif { \[\$restore_sim memory freadcmp 0 $memory_size $memory_filename\] != $memory_size } {"
        puts $outfile "\t\tputs \"ERROR: failed to read $memory_size bytes from memory checkpoint file $memory_filename\""
        puts $outfile "\t\texit -1"
        puts $outfile "\t}"

    }
}

#############################  MEMORY #######################################
#
#   pseries NVRAM
#
proc Checkpoint::checkpoint_pseries_nvram { outfile sim filename } {
    
    # Checkpoint memory
    print_status $outfile "pseries_nvram"
    
    if { [file exists "pseries_nvram.bin"] } {
	puts "Checkpointing pseries_nvram.bin"
	file copy -force "pseries_nvram.bin" $filename
	puts $outfile "puts \"Overwriting existing pseries_nvram.bin file using restored file!\""
	puts $outfile "file copy -force $filename pseries_nvram.bin"
    } else {
	puts "Non-existant pseries_nvram.bin file, skipping...\n";
    }
}

#############################  MAIN #######################################
#
#   This is the main checkpoint routine
#
#   It will iterate through calling all the appropriate checkpoint
#   procs to save the whole state of the machine
#
proc Checkpoint::checkpoint { {fileprefix ckpt} {sim mysim} } {

    set count $Checkpoint::ckpt_count
    incr Checkpoint::ckpt_count

    #########################################
    #
    # Generate filename
    #
    set filename "$fileprefix.ckpt"

    puts "Checkpointing to file $filename.*"
    set outfile [open $filename.tcl "w"]
    if { $outfile < 0 } {
        puts "Unable to open file $filename.tcl"
    }

    #########################################
    #
    #
    # HEADER
    #
    checkpoint_header $outfile $sim

    #########################################
    #
    #
    # PowerPC CPUs
    #
    for { set cpu 0 } { $cpu < [$sim display cpus] } { incr cpu } {
        for { set nthread 0 } { $nthread < [$sim cpu $cpu display number_of_threads] } { incr nthread } {
            checkpoint_ppc $outfile $sim $cpu $nthread
        } 
    }

    # Now turn all the appropriate cores on
    print_status $outfile "CPU Run State"
    foreach i [$sim display tcl_cpulist] {
        array set cpu_state $i
        if { [string compare $cpu_state(state) "on"] == 0 } {
            puts $outfile "\t\$restore_sim $cpu_state(name) $cpu_state(number) config_on"
        }
    }

    #########################################
    #
    #
    # PIC
    #
    set type [$sim pic display type]
    if { [string equal $type "0"]} {
         print_status $outfile "PIC"
	 puts $outfile "\t\$restore_sim pic restart top 0 [$sim pic checkpoint top 0]"
	 for { set i 0 } { $i < [$sim pic display max_proc] } { incr i } {
             puts $outfile "\t\$restore_sim pic restart pps $i [$sim pic checkpoint pps $i]"
         }
         for { set i 0 } { $i < [$sim pic display max_ipi] } { incr i } {
             puts $outfile "\t\$restore_sim pic restart ipi $i [$sim pic checkpoint ipi $i]"
         }
         for { set i 0 } { $i < [$sim pic display max_gti] } { incr i } {
             puts $outfile "\t\$restore_sim pic restart gti $i [$sim pic checkpoint gti $i]"
         }
         for { set i 0 } { $i < [$sim pic display max_xts] } { incr i } {
             puts $outfile "\t\$restore_sim pic restart xts $i [$sim pic checkpoint xts $i]"
         }
    } elseif {[string equal $type "1"]} {
	print_status $outfile "PPCIC"
	puts $outfile "\t\$restore_sim pic restart top 0 [$sim pic checkpoint top 0 0 0] 0"
	for { set i 0 } { $i < [$sim pic display max_proc] } { incr i } {
	    puts $outfile "\t\$restore_sim pic restart mfrr $i [$sim pic checkpoint mfrr $i 0 0] 0"
	}
	for { set i 0 } { $i < [$sim pic display max_proc] } { incr i } {
	    puts $outfile "\t\$restore_sim pic restart pIndex $i [$sim pic checkpoint pIndex $i 0 0] 0"
	}
	for { set i 0 } { $i < [$sim pic display max_proc] } { incr i } {
	    puts $outfile "\t\$restore_sim pic restart cprr $i [$sim pic checkpoint cprr $i 0 0] 0"
	}
	for { set i 0 } { $i < [$sim pic display max_proc] } { incr i } {
	    for { set j 0 } { $j < [$sim pic display max_prio] } { incr j } {
		for { set k 0 } { $k < [$sim pic checkpoint num_ic_pending $i $j 0] } { incr k } {
		    set exists [$sim pic checkpoint ic_pending $i $j $k]
		    if {![string equal $exists "null"]} {
			puts $outfile "\t\$restore_sim pic restart ic_pending $i $exists $j"
 		    }
		}
	    }
	}
	for { set i 0 } { $i < [$sim pic display ic_src_num] } { incr i } {
	    set exists [$sim pic checkpoint ic_src $i 0 0]
	    if {![string equal $exists "null"]} {
		puts $outfile "\t\$restore_sim pic restart ic_src 0 $exists 0"
	    }
	}
    } else {
	print_status $outfile "Unknown PIC type, skipping checkpoint..."
    }

    #########################################
    #
    #
    # Bogus
    #
    checkpoint_bogus $outfile $sim "$filename.disk"

    #########################################
    #
    #
    # Caches
    #
    checkpoint_cache $outfile $sim "$filename.cache"

    #########################################
    #
    #
    # Memory
    #
    checkpoint_memory $outfile $sim "$filename.memory"

    #########################################
    #
    #
    # pseries NVRAM
    #
    
    checkpoint_pseries_nvram $outfile $sim "$filename.pseries_nvram"

    #########################################
    #
    #
    # Footer
    #
    checkpoint_footer $outfile

    close $outfile

    puts "Checkpoint complete!\n"
}

#############################################################
# Other useful tcl proc's 
#############################################################

proc reset_fast_mode_emat { { sim mysim } } {
  for {set i 0} {$i < [$sim display cpus]} {incr i} {
    $sim cpu $i set emat reset
  }
}

proc reset_reservations { { sim mysim } } {
  $sim mcm 0 set core_based_reservation [$sim mcm 0 display core_based_reservation]
}

proc reset_to_match_restore { { sim mysim } } {
  percssim_reset_stats $sim
  reset_fast_mode_emat $sim
  reset_reservations   $sim
}

