#!/bin/sh
# the next line restarts using wish\
exec wish "$0" "$@"

#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2004
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#       IBM Corp.
#  
#  gdbsniff.tcl
#       A sniffer used to peek at traffic to and from gdb.
#
#  See the Sniffer::Example function for an example of how to use this file
#------------------------------------------------------------------------

#------------------------------------------------------------------------
# Options this script can accept
#   -gdbport num        the port number that gdb will attach to
#   -tgtport num        the port number that the sniffer will attach to
#------------------------------------------------------------------------


#------------------------------------------------------------------------
# Sniffer namespace
#------------------------------------------------------------------------
namespace eval Sniffer {
    variable version        "v1.0"
    variable initialized      0
}


#------------------------------------------------------------------------
#  Sniffer::GetCmdLineVar
#     This function will parse the command line options. Certain options
#     may have two parameters, while others require one.
#------------------------------------------------------------------------
proc Sniffer::GetCmdLineVar { var } {

    variable orig_cmdline
    
    if { [ string index $var 0 ] != "-" } then {
        Trace "Invalid var($var) in Sniffer::GetCmdLineVar(). Must begin with a hyphen (-)."
        return
    }
   
    # Check for single or double parameter types.
    # Listed items are doubles, everything else is a single type.
    set single_type 0
    switch -- $var {
        "-gdbport" -
        "-tgtport" -
        "-topwin" { }
        default { set single_type 1 }
    }

    set cl_list [ split $orig_cmdline " " ]
    set index [ lsearch -exact $cl_list $var ]
    if { $index == "-1" } then { 
        return [ expr { $single_type ? "0" : "" } ]
    } elseif { $single_type } then {
        return 1
    } else {
        incr index
        return [ lindex $cl_list $index ]
    }
}

#------------------------------------------------------------------------
#  Sniffer::init
#     This function is called during startup, and is used to initialize 
#     some basic settings.
#------------------------------------------------------------------------
proc Sniffer::init { { argv "" } } {

    global tcl_platform
   
    variable hostlist         ""
    variable orig_cmdline     $argv
    variable initialized
    variable thisos
    variable ports
    variable options
    variable times
    
    if { $initialized } then {
        return
    }
   
    # Try to determine the operating system ( windows, aix, sunos or linux )
    regexp {[a-z]*} [string tolower $tcl_platform(os)] thisos
    
    # Set default options
    set options(topwin)          ""
    set options(trace)           1
    set options(showhexdata)     0
    set options(autoscroll)      1
    set options(loglevel)        1
    set options(skipcommon)      0
    set options(fwdtotgt)        1
    set options(fwdtogdb)        1
    set options(logfile)         ""
    set options(tracesep)        "\n=============================================="
    
    # Set default times
    set times(app_started)       [ clock format [ clock seconds ] ]
    set times(last_clock)        "unknown"
    
    # See if user has passed in gdb port. Default to 2346
    set ports(gdb) 2346
    set arg_port [ GetCmdLineVar "-gdbport" ]
    if { $arg_port != "" } then {
        if { [ string is digit $arg_port ] } then {
            set ports(gdb) $arg_port
        }  
    }
    
    # See if user has passed in target port. Default to 2345
    set ports(tgt) 2345
    set arg_port [ GetCmdLineVar "-tgtport" ]
    if { $arg_port != "" } then {
        if { [ string is digit $arg_port ] } then {
            set ports(tgt) $arg_port
        }  
    }
    
    # See if user has passed in the toplevel window. Default to ""
    set options(topwin) [ GetCmdLineVar "-topwin" ]
    
    catch {
        if { $options(logfile) != "" } then {
            set fileid [ open $options(logfile) w+ ]
            close $fileid
        }
    }
    
    # Default to a decent font
    font create def_font -family helvetica -size 10
    option add *font def_font
    
    set initialized 1
}

#------------------------------------------------------------------------
#  Sniffer::StartGUI
#     This function is called during startup, and is used to create
#     the graphical user interface.
#------------------------------------------------------------------------
proc Sniffer::StartGUI { } {

    variable thisos
    variable version
    variable ports
    variable options
    
    
    if { $options(topwin) != "" } then {
        toplevel $options(topwin)
        bind $options(topwin) <Destroy> { Sniffer::Quit }
        set winprefix ""
    } else {
        set winprefix "."
    }

    if { $thisos == "windows" } then {
        bind $winprefix$options(topwin) <Control-Key-F1> { console show }
    }
  
    frame $options(topwin).opt
        frame $options(topwin).opt.l -borderwidth 5 -relief ridge
            checkbutton $options(topwin).opt.l.fwdtotgt    -text "--> Forward from GDB to Target -->" -anchor w -variable Sniffer::options(fwdtotgt)
            checkbutton $options(topwin).opt.l.fwdtogdb     -text "<-- Forward from Target to GDB <--" -anchor w -variable Sniffer::options(fwdtogdb)
            checkbutton $options(topwin).opt.l.showtrace   -text "Capture trace messages"         -anchor w -variable Sniffer::options(trace)
            pack $options(topwin).opt.l.fwdtotgt $options(topwin).opt.l.fwdtogdb $options(topwin).opt.l.showtrace \
                 -anchor w -fill x -expand 1 -side top
      
        frame $options(topwin).opt.r -borderwidth 5 -relief ridge
            checkbutton $options(topwin).opt.r.showhexdata -text "Append hex data to messages"    -anchor w -variable Sniffer::options(showhexdata)
            checkbutton $options(topwin).opt.r.autoscroll  -text "Auto scroll trace window"       -anchor w -variable Sniffer::options(autoscroll)
            checkbutton $options(topwin).opt.r.skipcommon  -text "Skip common trace messages"     -anchor w -variable Sniffer::options(skipcommon)
            pack $options(topwin).opt.r.showhexdata $options(topwin).opt.r.autoscroll $options(topwin).opt.r.skipcommon \
                 -anchor w -fill x -expand 1 -side top
        
        pack $options(topwin).opt.l $options(topwin).opt.r -side left -fill x -expand 1
   
    frame $options(topwin).hull
        text $options(topwin).hull.txt \
            -yscrollcommand "$options(topwin).hull.v_scroll set" \
            -xscrollcommand "$options(topwin).hull.h_scroll set"
        scrollbar $options(topwin).hull.v_scroll -command "$options(topwin).hull.txt yview"
        scrollbar $options(topwin).hull.h_scroll -orient horizontal -command "$options(topwin).hull.txt xview"
      
        pack $options(topwin).hull.v_scroll -side right -fill y
        pack $options(topwin).hull.h_scroll -side bottom -fill x
        pack $options(topwin).hull.txt -side left -fill both -expand 1
   
    frame $options(topwin).toolbar -borderwidth 0
        button $options(topwin).toolbar.clear -text "Clear"  -width 10 -command "$options(topwin).hull.txt delete 0.0 end "
        button $options(topwin).toolbar.exit  -text "Exit"   -width 10 -command "Sniffer::Quit"
        pack $options(topwin).toolbar.exit $options(topwin).toolbar.clear -side right -padx 5 -pady 5


    pack $options(topwin).opt -side top -expand 0 -fill x
    pack $options(topwin).toolbar -side bottom -expand 0 -fill x
    pack $options(topwin).hull -fill both -expand 1
   
    # Init some more GUI stuff
    wm title $winprefix$options(topwin) "GDB Sniffer (GDB port: $ports(gdb)  Target port: $ports(tgt))"
    FixWindowSize $winprefix$options(topwin)
    wm geom . +200+200
    
    # Add some bindings
    bind . <Escape> "Sniffer::Quit"
   
}

#------------------------------------------------------------------------
# Sniffer::FixWindowSize - forces a window not not be resizeable
#        The 'wm resizable' feature has a strange bug       
#------------------------------------------------------------------------
proc Sniffer::FixWindowSize { win } {
   
    update idletasks
    set xlen [winfo width $win]
    set ylen [winfo height $win]
    wm minsize $win $xlen $ylen
}

#------------------------------------------------------------------------
#  Sniffer::ValidatePort
#     Make sure the port is set by the user. Otherwise display a message.
#------------------------------------------------------------------------
proc Sniffer::ValidatePort { name msg } {

    variable ports
    
    if { [ info exists ports($name) ] && $ports($name) > 0 } then {
        return 1
    } else {
        tk_messageBox -icon error -type ok -message $msg -title "Port not set correctly"
        return 0
    }
}


#------------------------------------------------------------------------
#  Sniffer::StartServer
#     This function is called during startup to start the server
#------------------------------------------------------------------------
proc Sniffer::StartServer { } {

    variable ports
    variable sockets
    
    
    ValidatePort gdb "Please provide the GDB port using the -gdbport option"
    ValidatePort tgt "Please provide the Target port using the -tgtport option"
    
    # Open the server socket if we can, otherwise show warning and exit.
    set temp_sock ""
    catch { set temp_sock [ socket -server Sniffer::TcpAccept $ports(gdb) ] }
    if { $temp_sock == "" } then {
        set    msg "This application cannot open a server socket on port $ports(gdb)."
        append msg "Either there is an existing listening socket already on this port, or "
        append msg "you do not have sufficient rights to create one.\n"
        append msg "This program will terminate.\n"
        tk_messageBox -icon error -type ok -message $msg -title "Failed to open listener socket on port $ports(gdb)"
        Sniffer::Quit
    }
}

#------------------------------------------------------------------------
#  Sniffer::TcpAccept
#     This function is called when a new socket request is made
#------------------------------------------------------------------------
proc Sniffer::TcpAccept { sock addr clientport } {
   
    variable ports
    variable sockets
    
    # GDB has connected with us, create the client socket and connect to the target (e.g. mambo)
    set sockets(gdb) $sock
    
    Trace "Host $addr has created a socket $sockets(gdb) on port $ports(gdb)"
    
    fileevent $sockets(gdb) readable "Sniffer::TcpRead $sockets(gdb) 0"
    fconfigure $sockets(gdb) -buffersize 10000 -buffering full -blocking false -translation binary
    
    # Connect to the target
    if { [ catch { set sockets(tgt) [ socket localhost $ports(tgt) ] } errmsg ] } then {
        
        set    msg "This application cannot open a socket on port $ports(tgt)\n"
        append msg "Please verify that you have started the target, and that it's listening\n"
        append msg "port is set to $ports(tgt). Here is the system error message:\n"
        append msg $errmsg
        
        tk_messageBox -icon error -type ok -message $msg -title "Failed to connect to target on port $ports(tgt)"
        
        # Close the GDB socket and get out of here
        Disconnect
        
        return
    }
    
    fileevent $sockets(tgt) readable "Sniffer::TcpRead $sockets(tgt) 1"
    fconfigure $sockets(tgt) -buffersize 10000 -buffering full -blocking false -translation binary
}

#------------------------------------------------------------------------
#  Sniffer::TcpRead
#     This function is called when data is sent to a socket
#------------------------------------------------------------------------
proc Sniffer::TcpRead { sock from_tgt } {

    variable sockets
    variable options
    
    if { [ lsearch -exact [ file channels ] $sock ] != -1 } then {
      
        if { [ eof $sock ] } then {
            Disconnect
            return
        }
      
        # Read in the data on the socket
        set data [ read -nonewline $sock ]
      
        if { $data != "" } then {
            
            if { $from_tgt } then {
                if { $options(showhexdata) } then {
                    Trace "\n<-- Relaying from Target to GDB: $data \n[ BinToHex $data ] $options(tracesep)"
                } else {
                    Trace "\n<-- Relaying from Target to GDB: $data $options(tracesep)"
                }
            catch {
               if { $options(fwdtogdb) } then {
                  puts -nonewline $sockets(gdb) $data
                  flush $sockets(gdb)
               }
            }
         } else {
            if { $options(showhexdata) } then {
               Trace "\n--> Relaying from GDB to Target: $data \n[ BinToHex $data ]"
            } else {
               Trace "\n--> Relaying from GDB to Target: $data"
            }
            catch {
               if { $options(fwdtotgt) } then {
                  puts -nonewline $sockets(tgt) $data
                  flush $sockets(tgt)
               }
            }
         }
      }
   } else {
      Trace "Invalid socket $sock (valid:[ file channels ] )"
   }

   update idletasks
}


#------------------------------------------------------------------------
#  Sniffer::BinToHex
#     Convert a binary stream to a list of hex numbers
#------------------------------------------------------------------------
proc Sniffer::BinToHex { bindata } {

   set hexchardata ""
   binary scan $bindata c* chardata
   return [ DecToHex $chardata ]
}

#------------------------------------------------------------------------
#  Sniffer::DecToHex
#     Convert a list of decimal numbers to a list of hex numbers
#------------------------------------------------------------------------
proc Sniffer::DecToHex { chardata } {

   set hexchardata ""
   foreach c $chardata {
      append hexchardata [ format "0x%x " [ expr $c & 0xFF ] ]
   }
   return $hexchardata
}

#------------------------------------------------------------------------
#  Sniffer::GetByteField
#     Given an input bytestream, extract a byte-subfield starting
#      at 'index' for 'length' number of bytes
#------------------------------------------------------------------------
proc Sniffer::GetByteField { data index length } {
   if { [ regexp ".{$index}(.{$length}).*" "$data" junk1 subfield junk2 ] } then {
      return $subfield
   } else {
      return {}
   }
}

#------------------------------------------------------------------------
#  Sniffer::TcpSend
#     Sends a datastream out a socket
#------------------------------------------------------------------------
proc Sniffer::TcpSend { sock data } {

   #::puts "<-- Sending data ([BinToHex $data]) to ($hostnamelist($sa($sock)))"
   Trace "<-- Sending data ([BinToHex $data])"
   ::tkcon_puts -nonewline $sock $data
   flush $sock
}

#------------------------------------------------------------------------
#  Sniffer::Trace
#------------------------------------------------------------------------
proc Sniffer::Trace { msg args } {

   variable options
   
   if { !$options(trace) } then {
      return
   }
   
   append msg $args
   
   # Not sure what to filter here. This script originally started with Riscwatch, which
   # had a lot of spammed messages and the filtering helped. This is a nop at the moment.
   if { $options(skipcommon) } then {
      if { [ string first "--> Relaying from GDB to Target: foo" $msg ] >= 0 } then {
         return
      } elseif { [ string first "<-- Relaying from Target to GDB: foo" $msg ] >= 0 } then {
         return
      }
   }
   
   $options(topwin).hull.txt insert end $msg
   if { $options(autoscroll) } then {
      $options(topwin).hull.txt see end
   }
   catch {
      if { $options(logfile) != "" } then {
         set fileid [ open $options(logfile) a+ ]
         ::puts -nonewline $fileid $msg
         close $fileid
      }
   }

}

#------------------------------------------------------------------------
#  Sniffer::Disconnect
#     Disconnect all sockets
#------------------------------------------------------------------------
proc Sniffer::Disconnect { } {

    variable sockets
    
    Trace "\nDisconnecting....\n"
    
    catch { close $sockets(gdb) }
    catch { close $sockets(tgt) }
    
    catch { set sockets(gdb) "" }
    catch { set sockets(tgt) "" }
}

#------------------------------------------------------------------------
# Sniffer::Quit
#------------------------------------------------------------------------
proc Sniffer::Quit {} {
    Disconnect
    exit
}

#------------------------------------------------------------------------
#  Sniffer::Example
#     This is a simple example of how to use this file.
#------------------------------------------------------------------------
proc Sniffer::Example { args } {

    eval Sniffer::init $args
    Sniffer::StartGUI
    Sniffer::StartServer
}

Sniffer::Example $argv

