#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  common_oo.tcl
#
#
#------------------------------------------------------------------------
#  This file allows a user to easily add object oriented information
#  to a tcl object using pure tcl.  So if you want to create a new
#  object or widget that can maintain instance information, use the
#  following template.
#
#  namespace eval MyClass {
#     variable _objectcount 0
#     variable data
#     source common_oo.tcl
#     namespace import OOCommon::*
#     proc Create { widget } {
#        variable _objectcount
#        variable data
#        incr _objectcount
#        set data($widget-instance) $_objectcount
#        setvar $widget widget $widget
#        ( Create, pack and bind a graphical widget here if desired )
#        # establish function that allows widget calls to get back into these functions
#        if { [ info command $widget ] != "" } then { rename $widget ::$widget:cmd }
#        proc ::$widget { command args } "return \[eval MyClass::\$command $widget \$args\]"
#        return $widget
#     }
#  }
#
#  Now you can create an object like this:
#     MyClass::Create .m
#
#  And then you can set and get variables from that object using getvar and setvar:
#     .m setvar name "Johnny"
#     .m getvar name    ( returns Johnny }
#     .m listvar        ( shows all instance variables for object .m )
#
#  Other functions include help, listglobal, setglobal, getglobal, getglobalname,
#  unsetvar and getvarname.
#
#  You can also make calls from the 'static' function. Just ensure that
#  if the function you call is instance related, that the corresponding widget
#  instance name is passed in as the first argument.
#     MyClass::setvar .m name "Johnny"    ( needs the .m instance widget info )
#     MyClass::listglobal                 ( does not need instance widget info )
#------------------------------------------------------------------------

# Do not allow sourcing from toplevel namespace
if { [ namespace current ] == "::" } then {
   return
}

#------------------------------------------------------------------------
# OOCommon::OOCommon
#
#------------------------------------------------------------------------
namespace eval OOCommon {

   variable version        "1.0"
   variable parent         [ namespace parent ]

   namespace export *

   # Set default global variables
   set [ subst $parent ]::data(global-enable_DebugMsg) 0

   #------------------------------------------------------------------------
   # OOCommon::help
   #     This function returns a list of namespace commands.
   #------------------------------------------------------------------------
   proc help { args } {

      variable parent
      ::puts "Commands:"
      foreach word [ lsort [ info commands [ subst $parent ]::*] ] {
         ::puts "\t$word"
      }
   }

   #------------------------------------------------------------------------
   # OOCommon::DebugMsg
   #     This function will print a debug message if the global variable
   #     'enable_DebugMsg' is 1
   #------------------------------------------------------------------------
   proc DebugMsg { args } {

      variable parent
      if { [ getglobal enable_DebugMsg ] } then {
         ::puts "[ subst $parent ]::\t$args"
      }
   }
}

#------------------------------------------------------------------------
# OOCommon::_checkinstance
#     Verify the existance of this instance. A return value of 0 indicates
#     an invalid instance.
#------------------------------------------------------------------------
proc OOCommon::_checkinstance { widget } {

   variable parent
   set var [ subst $parent ]::data($widget-instance)
   if { [ info exists $var ] } then {
      return [ subst $[ subst $var ] ]
   } else {
      return 0
   }
}


#------------------------------------------------------------------------
# OOCommon::setglobal
#     Set a global variable
#------------------------------------------------------------------------
proc OOCommon::setglobal { args } {

   variable parent
   switch [ llength $args ] {
      2 {
         # In this case, make sure the first arg is not a valid widget instance
         if { [ _checkinstance [ lindex $args 0 ] ] } then {
            return "Usage: setglobal var value"
         }
         set var    [ lindex $args 0 ]
         set value  [ lindex $args 1 ]
      }
      3 {
         set var    [ lindex $args 1 ]
         set value  [ lindex $args 2 ]
      }
      default { return "Usage: setglobal var value" }
   }
   set [ subst $parent ]::data(global-$var) $value
}

#------------------------------------------------------------------------
# OOCommon::getglobal
#     Get a global variable
#------------------------------------------------------------------------
proc OOCommon::getglobal { args } {

   variable parent
   switch [ llength $args ] {
      1 { set var [ lindex $args 0 ] }
      2 { set var [ lindex $args 1 ] }
      default { return "Usage: getglobal var" }
   }
   set var [ subst $parent ]::data(global-$var)
   if { [ info exists $var ] } then {
      return [ subst $[ subst $var ] ]
   }
   return ""
}

#------------------------------------------------------------------------
# OOCommon::getglobalname
#     Get a global variable's actual name in the data array
#------------------------------------------------------------------------
proc OOCommon::getglobalname { args } {

   variable parent
   switch [ llength $args ] {
      1 { set var [ lindex $args 0 ] }
      2 { set var [ lindex $args 1 ] }
      default { return "Usage: getglobalname var" }
   }
   set var [ subst $parent ]::data(global-$var)
   if { [ info exists $var ] } then {
      return $var
   }
   return ""
}

#------------------------------------------------------------------------
# OOCommon::listglobal
#     List current global variables
#------------------------------------------------------------------------
proc OOCommon::listglobal { args } {

   variable parent
   switch [ llength $args ] {
      0 {}
      1 {}
      default { return "Usage: listglobal" }
   }
   foreach { _item _value } [ array get [ subst $parent ]::data global-* ] {
      ::puts "[string trimleft [string trimleft $_item global] -] = $_value"
   }
}

#------------------------------------------------------------------------
# OOCommon::unsetvar
#     Set an instance variable
#------------------------------------------------------------------------
proc OOCommon::unsetvar { widget var } {

   variable parent
   set instance [ _checkinstance $widget ]
   #DebugMsg "unsetvar   instance = $instance   widget = $widget    var = $var"
   set var [ subst $parent ]::data($instance-$var)
   if { $instance } then {
      if { [ info exists $var ] } then {
        unset $var
      }
   }
}
#------------------------------------------------------------------------
# OOCommon::setvar
#     Set an instance variable
#------------------------------------------------------------------------
proc OOCommon::setvar { widget var value } {

   variable parent
   set instance [ _checkinstance $widget ]
   #DebugMsg "instance = $instance   widget = $widget    var = $var    value = $value"
   if { $instance } then {
      set [ subst $parent ]::data($instance-$var) $value
   }
}

#------------------------------------------------------------------------
# OOCommon::getvar
#     Get an instance variable
#------------------------------------------------------------------------
proc OOCommon::getvar { widget var } {

   variable parent
   set instance [ _checkinstance $widget ]
   if { $instance } then {
      set var [ subst $parent ]::data($instance-$var)
      if { [ info exists $var ] } then {
         return [ subst $[ subst $var ] ]
      }
   }
   return ""
}

#------------------------------------------------------------------------
# OOCommon::getvarname
#     Get an instance variable's actual name in the data array
#------------------------------------------------------------------------
proc OOCommon::getvarname { widget var } {

   variable parent
   set instance [ _checkinstance $widget ]
   if { $instance } then {
      set var [ subst $parent ]::data($instance-$var)
      if { [ info exists $var ] } then {
         return $var
      }
   }
   return ""
}

#------------------------------------------------------------------------
# OOCommon::listvar
#     List current instance variables
#------------------------------------------------------------------------
proc OOCommon::listvar { widget } {

   variable parent

   set instance [ _checkinstance $widget ]
   if { $instance } then {
      foreach { _item _value } [ array get [ subst $parent ]::data $instance-* ] {
         ::puts "[string trimleft $_item $instance-] = $_value"
      }
   }
}
