#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  control.tcl
#
#
#------------------------------------------------------------------------

namespace eval Control {

   variable version        "1.0"
   variable initialized    0
}


#------------------------------------------------------------------------
#  Control::init
#    We can only inject the settings after the generic machine is fixed, but
#    before the MCM is fixed for some changes to actually take. If the generic
#    machine is not yet fixed, schedule a callback to Control::inject_settings,
#    otherwise inject them immediately.
#------------------------------------------------------------------------
proc Control::init {} {

    variable initialized
    variable display_settings
    variable simulator_settings
    global gvars

    #puts "Control::init called"

    set sim $gvars(firstsim)
    if { !$initialized } then {
        # Read initial values from prefs file
        set display_settings(def_font) [ font configure default_font ]
        set display_settings(lb_font)  [ font configure listbox_font ]
        set display_settings(tb_font)  [ font configure textbox_font ]

        if { $sim != "" } then {
            #set simulator_settings(memsize)         [ get_pref sim/memsize  0x4000000 ]
            catch { set simulator_settings(gdbport) [ get_pref sim/gdb_port 2345 ] }
			Control::inject_simulator_settings 1
        }
		
        set initialized 1
    }
}


#------------------------------------------------------------------------
#  Control::create
#
#------------------------------------------------------------------------
proc Control::create { widget } {

    variable control_nb
    global gvars

    if { ![ winfo exists $widget ] } then {

        toplevel $widget

        wm transient $widget .

        # Intercept window close and call the exit handler
        set top_widget [ winfo toplevel $widget ]
        wm protocol $top_widget WM_DELETE_WINDOW "Control::_on_close $top_widget"
        bind $top_widget <Configure> "Control::_save_geometry $top_widget"
        bind $top_widget <Key-F5> "user_refresh"

        set geom [ get_pref win_geom/cpanel "650x250+100+100" ]

        #------------------------------------------------------------------------
        #  NP Control Panel
        #------------------------------------------------------------------------
        set cpanel [ frame $widget.control -borderwidth 2 -relief groove ]

        set control_nb [ Options::create $cpanel [ namespace current ] ]

        $control_nb insert end display -text "Display"
        $control_nb insert end simulator -text "Simulator"
        $control_nb raise display


        pack $cpanel -side top -fill both -expand yes

        wm geometry $top_widget $geom

    } else {

        wm deiconify [ winfo toplevel $widget ]
    }

    return $widget

}

#------------------------------------------------------------------------
# Control::_save_geometry
#------------------------------------------------------------------------
proc Control::_save_geometry { top_widget } {

   set_pref win_geom/cpanel [ winfo geometry $top_widget ]
}

#------------------------------------------------------------------------
# Control::-on_close
#------------------------------------------------------------------------
proc Control::-on_close { top_widget } {

   wm withdraw $top_widget
}


#------------------------------------------------------------------------
#  Control::_get_filename_dialog
#------------------------------------------------------------------------
proc Control::_get_filename_dialog { _var init_dir ftypes title } {

   upvar $_var var
   set var [ tk_getOpenFile -initialdir $init_dir -filetypes $ftypes -title $title ]
}

#------------------------------------------------------------------------
#  Control::get_option_value
#------------------------------------------------------------------------
proc Control::get_option_value { _array _index } {

   variable target_settings

   set rtn ""
   catch { set rtn [ subst $[subst $_array]($_index) ] }
   return $rtn
}

#------------------------------------------------------------------------
# Control::_compare_arrays
#   This compares two arrays and returns 0 if equal, 1 if not.
#   I previously used 'array get', but the return order of elements is not
#   guarateed, even for copies of the original array!
#------------------------------------------------------------------------
proc Control::_compare_arrays { _a1 _a2 } {

    upvar $_a1 a1
    upvar $_a2 a2

    set rtn 0
    foreach elem [ array names a1 ] {
        if { ![ string match $a1($elem) $a2($elem) ] } then {
            set rtn 1
            break
        }
    }
    return $rtn
}

#------------------------------------------------------------------------
#  Control::-createtab_display
#       Creates the 'display' control panel tab
#------------------------------------------------------------------------
proc Control::-createtab_display { {topf} } {

    variable control_nb
    variable display_settings
    variable temp_display_settings
    global gvars

    # Copy all of the settings to temp_settings array
    array set temp_display_settings [ array get display_settings ]

    # Font Selection group
    set def_fontf [ frame $topf.def_font -bd 2 -relief groove ]
        label $def_fontf.lab -text "Default Font:" -font system_font
        entry $def_fontf.val -textvariable ::Control::temp_display_settings(def_font) \
                         -width 60 -state disabled \
                         -font system_font
        button $def_fontf.brw -text "Select Font" -command "Control::_pickfont def_font" -font system_font
        pack $def_fontf.lab -side left -padx 5
        pack $def_fontf.brw $def_fontf.val -side right -padx 5

    set lb_fontf [ frame $topf.lb_font -bd 2 -relief groove ]
        label $lb_fontf.lab -text "ListBox Font:" -font system_font
        entry $lb_fontf.val -textvariable ::Control::temp_display_settings(lb_font) \
                         -width 60 -state disabled \
                         -font system_font
        button $lb_fontf.brw -text "Select Font" -command "Control::_pickfont lb_font" -font system_font
        pack $lb_fontf.lab -side left -padx 5
        pack $lb_fontf.brw $lb_fontf.val -side right -padx 5

    set tb_fontf [ frame $topf.tb_font -bd 2 -relief groove ]
        label $tb_fontf.lab -text "TextBox Font:" -font system_font
        entry $tb_fontf.val -textvariable ::Control::temp_display_settings(tb_font) \
                         -width 60 -state disabled \
                         -font system_font
        button $tb_fontf.brw -text "Select Font" -command "Control::_pickfont tb_font" -font system_font
        pack $tb_fontf.lab -side left -padx 5
        pack $tb_fontf.brw $tb_fontf.val -side right -padx 5


    # Pack all groups
    pack $def_fontf $lb_fontf $tb_fontf -side top -anchor w -fill both -expand 1
    pack $topf -side top -fill both -expand yes

}

#------------------------------------------------------------------------
#  Control::-on_ok_display
#------------------------------------------------------------------------
proc Control::-on_ok_display { } {

   variable display_settings
   variable temp_display_settings

   # If any of the 'display settings' changed, update the system.
   if { [ _compare_arrays display_settings temp_display_settings ] } then {

      # Copy all of the temp_settings to the settings array
      array set display_settings [ array get temp_display_settings ]

      inject_display_settings

      # Update the prefs
      set_pref fonts/default $display_settings(def_font)
      set_pref fonts/listbox $display_settings(lb_font)
      set_pref fonts/textbox $display_settings(tb_font)
   }
}

#------------------------------------------------------------------------
#  Control::-on_cancel_display
#------------------------------------------------------------------------
proc Control::-on_cancel_display { } {

   variable display_settings
   variable temp_display_settings

   array set temp_display_settings [ array get display_settings ]
}

#------------------------------------------------------------------------
#  Control::inject_display_settings
#       Injects various settings back into the simulator.
#------------------------------------------------------------------------
proc Control::inject_display_settings { } {

    variable display_settings

    # Inject the font information back into the GUI
    eval font configure default_font $display_settings(def_font)
    eval font configure listbox_font $display_settings(lb_font)
    eval font configure textbox_font $display_settings(tb_font)
    #HierSim_Refresh .simlist

}

#------------------------------------------------------------------------
#  Control::_pickfont
#       Opens the font dialog box
#------------------------------------------------------------------------
proc Control::_pickfont { tgt_font } {

    variable control_nb
    variable temp_display_settings

    $control_nb freeze
    Fonts::create .pickfont $control_nb $tgt_font
    vwait Fonts::trigger

    if { $Fonts::ok_pushed } then {
        set temp_display_settings($tgt_font) [ font configure sample_font ]
    }
    $control_nb thaw
}

#------------------------------------------------------------------------
#  Control::-createtab_simulator
#       Creates the 'simulator' control panel tab
#------------------------------------------------------------------------
proc Control::-createtab_simulator { {topf} } {

    variable control_nb
    variable simulator_settings
    variable temp_simulator_settings
    global gvars

    # Copy all of the settings to temp_settings array
    array set temp_simulator_settings [ array get simulator_settings ]

    # Memory size option
    #set memf [ frame $topf.memf -bd 2 -relief groove ]
    #    label $memf.lab -text "Memory Size (MB):" -font system_font
    #    entry $memf.val -textvariable ::Control::temp_simulator_settings(memsize) \
    #                     -width 20 -font system_font
    #    pack $memf.lab $memf.val -side left -padx 5

    # GDB port option
    set portf [ frame $topf.portf -bd 2 -relief groove ]
        label $portf.lab -text "GDB port:" -font system_font
        entry $portf.val -textvariable ::Control::temp_simulator_settings(gdbport) \
                         -width 20 -font system_font
        pack $portf.lab $portf.val -side left -padx 5


    # Pack all groups
    pack $portf -side top -anchor w -fill both -expand 1
    pack $topf -side top -fill both -expand 1 -anchor w

}

#------------------------------------------------------------------------
#  Control::-on_ok_simulator
#------------------------------------------------------------------------
proc Control::-on_ok_simulator { } {

   variable simulator_settings
   variable temp_simulator_settings

   # If any of the 'simulator settings' changed, update the system.
   if { [ _compare_arrays simulator_settings temp_simulator_settings ] } then {

      # Copy all of the temp_settings to the settings array
      array set simulator_settings [ array get temp_simulator_settings ]

      inject_simulator_settings 0

      # Update the prefs
      #set_pref sim/memsize      $simulator_settings(memsize)
      set_pref sim/gdb_port     $simulator_settings(gdbport)
   }
}

#------------------------------------------------------------------------
#  Control::-on_cancel_simulator
#------------------------------------------------------------------------
proc Control::-on_cancel_simulator { } {

   variable simulator_settings
   variable temp_simulator_settings

   array set temp_simulator_settings [ array get simulator_settings ]
}

#------------------------------------------------------------------------
#  Control::inject_simulator_settings
#       Injects various settings back into the simulator.
#------------------------------------------------------------------------
proc Control::inject_simulator_settings { bootstrap } {

    variable simulator_settings

    #puts "Control::inject_simulator_settings called"

    set errcnt 0
    foreach sim [ display machines ] {
        #incr errcnt [ catch { $sim config memory_size      $simulator_settings(memsize)  } ]
        incr errcnt [ catch { $sim set gdb_port    $simulator_settings(gdbport) } ]
    }
    if { $errcnt > 0 } then {
        after 300 "Control::_display_fixed_warning $bootstrap"
    }
}

#------------------------------------------------------------------------
#  Control::_display_fixed_warning
#------------------------------------------------------------------------
proc Control::_display_fixed_warning { bootstrap } {

    set msg    "Warning: Simulator already fixed! Could not alter simulator settings.\n\n"
    if { $bootstrap } then {
        append msg "\tPlease edit your .gmambo.tcl script and move any image loading commands\n"
        append msg "\tso they occur after the command to start the GUI.\n"
    } else {
        append msg "\tPlease restart the simulator to apply changes.\n"
    }
    tk_messageBox -icon warning -title "Simulator Fixed..." -type ok -message $msg
}

