#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  controls.tcl
#
#
#------------------------------------------------------------------------
set running 0

namespace eval Controls {

   variable version        "1.0"
   variable initialized    0
   variable sim_based_buttons_enabled 1
   variable sim_based_buttons ""

   variable sim_list ""
}

package require BWidget 1.7

proc Controls::NewSimBasedButton { buttonname } {

    ListPushUnique Controls::sim_based_buttons $buttonname
}

proc Controls::DisableSimBasedButtons { } {

    if { ! $Controls::sim_based_buttons_enabled } {
        return;
    }

    foreach buttonname $Controls::sim_based_buttons {
        $buttonname configure -state disabled
    }

    set Controls::sim_based_buttons_enabled 0
}

proc Controls::EnableSimBasedButtons { } {
    if { $Controls::sim_based_buttons_enabled } {
        return;
    }

    foreach buttonname $Controls::sim_based_buttons {
        $buttonname configure -state normal
    }

    set Controls::sim_based_buttons_enabled 1
}

#=========================================================
# Controls::CommaInsert
#    Put punctuating commas into numbers, so that large integers
#    can be parsed by eye. 1000000000  becomes 1,000,000,000
#=========================================================
proc Controls::CommaInsert { in } {

    set out ""
    set sep ""
    # Remove any non-numeric characters
    regsub -all {[a-zA-Z]} $in {} result
    set in $result

    while { [ string length $in ] > 3 } {
        set tail [ string range $in end-2 end ]
        set out "$tail$sep$out"
        set sep ","
        set in   [ string range $in 0 end-3 ]
    }
    set out "$in$sep$out"
    return $out
}

#=========================================================
# Controls::Find_Table_Hole
#
# This is actually just a generic routine to find a hole
# in a table to place a new thing
#=========================================================
proc Controls::Find_Table_Hole { table } {
    set max_row 10
    set max_col 10

    for { set r 0 } { $r < $max_row } { incr r } {
        for { set c 0 } { $c < $max_col } { incr c } {
            set idx($r,$c) ""
        }
    }
    for { set r 0 } { $r < $max_row } { incr r } {
        if { [catch { set row_data($r) [table search $table -start r$r] } errs] } {
            set max_row $r
        } 
    }
    for { set c 0 } { $c < $max_col } { incr c } {
        if { [catch { set col_data($c) [table search $table -start c$c] } errs] } {
            set max_col $c
        } else {
            foreach slave $col_data($c) {
                set found 0
                for { set r 0 } { !$found && ( $r < $max_row ) } { incr r } {
                    if { [lsearch $row_data($r) $slave] >= 0 } {
                        set idx($r,$c) $slave
                        set found 1
                    }
                }
                if { !$found } {
                    puts "Could not find home for $slave"
                }
            }
        }
    }
    for { set r 0 } { $r < $max_row } { incr r } {
        for { set c 0 } { $c < $max_col } { incr c } {
            if { [string compare "" $idx($r,$c)] == 0 } {
                return "$r,$c"
            }
        }
    }
}

#=========================================================
# Controls::Select_GDB_MCM
#=========================================================
proc Controls::Select_GDB_MCM { } {

    variable currentSim
    variable gdbmcm
    
    _validateSim

    # Only set the gdb mcm if not already set
    if { $gdbmcm != "" } then {
        return
    }

    # If only one mcm, then don't ask, just pick mcm 0
    if { [ $currentSim display number_of_MCMs ] == 1 } then {
        set gdbmcm 0
        return
    }

    # ask which mcm to target for gdb
    set win .selgdb
    toplevel $win
    wm transient $win .
    TclUtils::PlaceWindow . $win 200 50
    wm title $win "Select GDB Target"
    label $win.msg -text "\nMultiple chips were detected in this configuration.\nPlease select one for GDB to attach to.\n\n"
    pack $win.msg -side top
    for { set i 0 } { $i < [ $currentSim display number_of_MCMs ] } { incr i } {
        button $win.sel-$i -text "GDB will attach to MCM $i" -command "set Controls::gdbmcm $i"
        pack $win.sel-$i -side top
    }
    wm protocol $win WM_DELETE_WINDOW "set Controls::gdbmcm cancel"
    vwait Controls::gdbmcm
    destroy $win
}



#=========================================================
# Controls::Toggle_GDBService
#=========================================================
proc Controls::Toggle_GDBService { } {

    variable widget
    variable currentSim
    variable gdbmcm
    
    _validateSim

    if { $gdbmcm == "" } then {
        Select_GDB_MCM
        if { $gdbmcm == "cancel" } then {
            set gdbmcm ""
            return
        }
    }

    if { [ $currentSim mcm $gdbmcm debugger is_connected ] } then {
        $currentSim mcm $gdbmcm debugger cancel
        set gdbmcm ""
        simstop
    } else {
        set seconds 1000
        puts "Waiting for $seconds seconds for debugger to attach (Hit Control-C to stop waiting)..."
        $widget.buttons.gdb configure -text "Waiting for GDB..."
        $currentSim mcm $gdbmcm debugger wait $seconds
    }
}

#=========================================================
# Controls::UpdateGUI
#=========================================================
proc Controls::UpdateGUI { } {

    mambo_update 1
    update

    Controls::Refresh
}

#=========================================================
# Controls::UpdateStepDomain
#=========================================================
proc Controls::UpdateStepDomain {} {

    variable currentSim
    variable domain

    _validateSim

    set result [$currentSim display time_domain_for_step]
    if { [regexp {step cycle time domain is (.*)} $result all dom] } {
        set domain $dom
    } else {
        set domain "???"
    }
}

#=========================================================
# Controls::DomainMenuSelect
#=========================================================
proc Controls::DomainMenuSelect {} {

    variable currentSim
    variable cycles
    variable domain
    variable domain_cbox

    _validateSim

    if { [string is space $domain] } {
        set domain "cpu"
    }

    set lst [$domain_cbox cget -values]
    if { [lsearch -exact $lst $domain] != -1 } {
        $currentSim set time_domain_for_display $domain
        $currentSim set time_domain_for_step $domain
        set cycles [ CommaInsert [ $currentSim display cycles ] ]
        $domain_cbox configure -foreground black
    } else {
        $domain_cbox configure -foreground red
    }
}

#=========================================================
# Controls::PopulateDomainMenu
#=========================================================
proc Controls::PopulateDomainMenu {} {

    variable currentSim
    variable domain_cbox
    variable domain_menu_populated

    _validateSim

    array set tdl [$currentSim tdomain list]
    $domain_cbox configure -values [array names tdl]
    set domain_menu_populated 1
}

#=========================================================
# Controls::AdvanceCycle
#=========================================================
proc Controls::AdvanceCycle { } {

    variable currentSim
    variable cycleAdvance
    
    _validateSim

    #$widget.buttons.go  configure -state disabled
    #$widget.buttons.adv configure -state disabled
    after idle "Controls::Refresh"
    $currentSim cycle $cycleAdvance

    #UpdateGUI
}

#=========================================================
# Controls::Go
#=========================================================
proc Controls::Go { } {

    variable currentSim
    
    _validateSim

    #$widget.buttons.go  configure -state disabled
    #$widget.buttons.adv configure -state disabled
    after idle "Controls::Refresh"
    Hooks::callback "Go"

    # Don't die when we break out of go early
    catch {$currentSim go} errors

    #UpdateGUI
}

#=========================================================
# Controls::Add_Machine_Specific_Buttons
#     Add machine-specific controls to the status window
#=========================================================
proc Controls::Add_Machine_Specific_Buttons {} {
    
    variable widget
    
    if { [ catch { Machine_Button_Defs $widget.buttons } errmsg ] } then {
        # If it is not defined for a machine we should not say anything
        #puts "\tError sourcing Machine_Button_Defs: $errmsg"
        #if {[regexp -nocase "invalid command name" $errmsg] != 0} {
        #    puts "\ttclIndex may be out of date"
        #}
    }
}

#=========================================================
# Controls::Init
#     Initalize the status window
#=========================================================
proc Controls::Init { _widget } {

    variable widget $_widget
    variable cycles
    variable domain 
    variable domain_cbox
    variable domain_menu_populated 0
    variable cycleAdvance
    variable currentSim ""
    variable gdbmcm ""

    set ccf [frame $widget.cycCountFrame]
    set domain_cbox [ComboBox ${ccf}.cycCountSel -autocomplete 1 -hottrack 1 -expand tab -textvariable Controls::domain -modifycmd Controls::DomainMenuSelect -command Controls::DomainMenuSelect]
    set ccl [label ${ccf}.cycCountLabel -justify right -text "Cycles:"]
    set ccc [label ${ccf}.cycCount      -justify left  -text "0" -textvariable Controls::cycles]
    grid $domain_cbox $ccl $ccc

    scale $widget.cycleAdvance -orient horizontal -showvalue true \
        -from 1 -to 100000000 -variable Controls::cycleAdvance -bigincrement 1000

    $widget.cycleAdvance set 1
    label $widget.advanceValLabel -text "Advance Cycle Amount :"
    entry $widget.advanceVal -textvariable Controls::cycleAdvance

    frame $widget.buttons -relief groove -bd 3
        button $widget.buttons.adv  -text "Advance Cycle"        -command "Controls::AdvanceCycle"
        Controls::NewSimBasedButton $widget.buttons.adv
        button $widget.buttons.go   -text "Go"                   -command "Controls::Go"
        Controls::NewSimBasedButton $widget.buttons.go
        button $widget.buttons.stop -text "Stop"                 -command "simstop"
        button $widget.buttons.exit -text "Exit"                 -command "GUI::shutdown"
        button $widget.buttons.gdb  -text "Service GDB"          -command "Controls::Toggle_GDBService" -width 22
        Controls::NewSimBasedButton $widget.buttons.gdb
        button $widget.buttons.upd  -text "Update GUI"           -command "Controls::UpdateGUI"
        button $widget.buttons.trig -text "Triggers/Breakpoints" -command "Controls::TriggerWindow"
        Controls::NewSimBasedButton $widget.buttons.trig
        button $widget.buttons.dbg  -text "Debug Controls"       -command "DebugControls::toggle_window"
        button $widget.buttons.opt  -text "Options"              -command "Control::create .control"
        button $widget.buttons.emit -text "Emitters"             -command "Emitters::create .emitters"
        Controls::NewSimBasedButton $widget.buttons.emit
        button $widget.buttons.simmode -text "Mode"              -command { SimMode::Create $Controls::currentSim }
        Controls::NewSimBasedButton $widget.buttons.simmode
        button $widget.buttons.proctree -text "Process-Tree"     -state disabled -command { Machine_ProcessTreeInit $Controls::currentSim {} }
        button $widget.buttons.proctreestats -text "Process-Tree-Stats"     -state disabled -command { Machine_ProcessTreeStats_Init $Controls::currentSim {} }
        button $widget.buttons.trackall -text "Track All PCs"    -command { TrackAll::create .trackall $Controls::currentSim }
        Controls::NewSimBasedButton $widget.buttons.trackall
        button $widget.buttons.eventlog -text "Event Log"        -command { EventLog::Init $Controls::currentSim }
        Controls::NewSimBasedButton $widget.buttons.eventlog

        set non_emitter [catch {simemit list} non_emitter]

        if {$non_emitter} {
           $widget.buttons.emit configure -state disabled
        }

        table $widget.buttons \
            $widget.buttons.adv         0,0 -cspan 1 -fill x \
            $widget.buttons.go          0,1 -cspan 1 -fill x \
            $widget.buttons.stop        0,2 -cspan 1 -fill x \
            $widget.buttons.gdb         0,3 -cspan 1 -fill x \
            $widget.buttons.trig        1,0 -cspan 1 -fill x \
            $widget.buttons.upd         1,1 -cspan 1 -fill x \
            $widget.buttons.dbg         1,2 -cspan 1 -fill x \
            $widget.buttons.opt         1,3 -cspan 1 -fill x \
            $widget.buttons.emit        2,0 -cspan 1 -fill x \
            $widget.buttons.simmode     2,1 -cspan 1 -fill x \
            $widget.buttons.proctree    3,0 -cspan 1 -fill x \
            $widget.buttons.proctreestats    3,1 -cspan 1 -fill x \
            $widget.buttons.trackall    3,2 -cspan 1 -fill x \
            $widget.buttons.eventlog    3,3 -cspan 1 -fill x \
            $widget.buttons.exit        4,3 -cspan 1 -fill x

        Add_Machine_Specific_Buttons
        #Controls::Machine_Specific $widget.buttons 3 4

    table $widget \
        $widget.cycCountFrame           0,0 -cspan 3 -fill both     \
        $widget.advanceValLabel         1,0 -cspan 1 -fill x -anchor sw \
        $widget.advanceVal              1,1 -cspan 1 -fill x -anchor sw \
        $widget.cycleAdvance            1,2 -cspan 1 -fill x \
        $widget.buttons                 2,0 -cspan 3 -fill both

    table configure $widget c2 r4 -resize expand

    _validateSim
    
}

#=========================================================
# Controls::Refresh
#     Refresh the contents of the status window
#=========================================================
proc Controls::Refresh { } {

    variable widget
    variable cycles
    variable currentSim
    variable domain_menu_populated
    variable gdbmcm

    _validateSim

    if { [info commands "$currentSim"] == "" } {
        # Simulator does not exist
        return
    }

    if { [winfo exists .trackall] == 1 } {
        TrackAll::Refresh .trackall 
    }

    if { ! $domain_menu_populated } {
        Controls::PopulateDomainMenu
    }

    # Update the Simulation Mode state
    if { [winfo exists .simmode] == 1 } {
        SimMode::Refresh $Controls::currentSim
    }

    if { $gdbmcm != "" } {
        if { ![ $currentSim mcm $gdbmcm debugger is_connected ] } then {
            set gdbmcm ""
        }
    }

    if { $gdbmcm == "" } then {

        # GDB is not controlling the simulator
        $widget.buttons.gdb configure -text "Service GDB"

        set stopstatus [string trim [simstop status]]
        if { [string compare $stopstatus "not-stopped"] == 0 } {
            # The simulator is running
            $widget.buttons.adv   configure -state disabled
            $widget.buttons.go    configure -state disabled
            $widget.buttons.stop  configure -state normal
            $widget.buttons.gdb   configure -state disabled

        } else {
            # The simulator is stopped
            $widget.buttons.go    configure -state normal
            $widget.buttons.adv   configure -state normal
            $widget.buttons.stop  configure -state normal
            $widget.buttons.gdb   configure -state normal
        }

    } else {
        # GDB is controlling the simulator
        $widget.buttons.gdb configure -text "Disconnect from GDB"
        $widget.buttons.adv     configure -state disabled
        $widget.buttons.go      configure -state disabled
        $widget.buttons.stop    configure -state disabled
    }

    array set ptinfo [$currentSim processtree display all]
    if { $ptinfo(processes) > 1 } {
        $widget.buttons.proctree configure -state normal 
        $widget.buttons.proctreestats configure -state normal 
    }

    set cycles [ CommaInsert [ $currentSim display cycles ] ]
    Controls::UpdateStepDomain 

    catch { Machine_Refresh $currentSim } errmsg
}

#=========================================================
# Controls::TriggerWindow
#=========================================================
proc Controls::TriggerWindow {} {

    variable currentSim
    
    _validateSim

    Trigger_Window $currentSim
}

#=========================================================
# Controls::_validateSim
#     Make sure currentSim is set. The new config stuff seems to confuse this.
#=========================================================
proc Controls::_validateSim {} {

    variable currentSim

    if { [string compare "$Controls::sim_list" [display machines]] != 0 } {
        set Controls::sim_list [ display machines ]
        # The simulators changed.   Make sure the sim is still in the list
        if { [lsearch -exact $Controls::sim_list $Controls::currentSim] == -1 } {
            set currentSim ""
        }
    }
    
    if { $currentSim == "" } then {
        set Controls::sim_list [ display machines ]
        set currentSim [ lindex $Controls::sim_list 0 ]

        if { $currentSim == "" } then {
            DisableSimBasedButtons
        } else {
            EnableSimBasedButtons
        }
    } else {
            EnableSimBasedButtons
    }
}

