#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  debug_controls.tcl
#
#
#------------------------------------------------------------------------


namespace eval DebugControls {
   variable version        "1.0"
}

package require BWidget 1.7

# --------------------------------------------------------------
# DebugControls::toggle_window
#   The window is created once, and then hidden/unhidden when needed.
# --------------------------------------------------------------
proc DebugControls::toggle_window {} {

    global gvars
    variable top .debugcontrols

    if { [ winfo exists $top ] } then {
        import_simulator
        if { [ winfo ismapped $top ] } then {
            set_pref "win_geom/debug_controls" [ winfo geometry $top ] 
            wm withdraw $top
        } else {
            wm deiconify $top
        }
    } else {
        toplevel $top
        set geom [ get_pref "win_geom/debug_controls" "not_set"  ]
        if {$geom != "not_set"} { 
            wm geometry $top $geom
        }

        init
    }
}



# --------------------------------------------------------------
# DebugControls::init -- Create a list of controls for all the
#   debug options. Each selectable button is a debug message class.
#   A collection of like classes are grouped into a category.
#   We want to display the grouped classes together, but provide some
#   autoplacement of these sub-frames that looks halfway decent.
# --------------------------------------------------------------
proc DebugControls::init {} {

    variable top
    variable state

    import_simulator

    set num_cols 6
    set hull [ frame $top.hull ]
        # Get the list of categories
        set categories [ lsort [ simdebug category listformat ] ]
        # Create a fixed number of 'columns' to place each category in
        for { set col_num 0 } { $col_num < $num_cols } { incr col_num } {
            set col_frame$col_num [ frame $hull.col$col_num -borderwidth 2 -relief raised ]
        }

        foreach cat $categories {
            set _catsize($cat) [ llength [ simdebug category $cat listformat ] ]
        }
        TclUtils::array_flip _catsize catsize unique

        set col_num         0
        set fill_direction  1
        # For each category, create a LabelFrame with the category as the title
        foreach index [ lsort -real -decreasing [ array names catsize ] ] {
            set cat $catsize($index)
            set col_frame [ subst $[ subst {col_frame$col_num} ] ]
            set catframe [ LabelFrame $col_frame.[ TclUtils::safename $cat ] -text "$cat" -side top -borderwidth 2 -relief sunken -anchor center -pady 5 ]
                set catframe_user [ $catframe getframe ]
                # Inside each category frame, create a checkbutton for each class in that category
                foreach class [ lsort [ simdebug category $cat listformat ] ] {
                    set lab [ checkbutton $catframe_user.[ TclUtils::safename $class ] -text "$class" -anchor w -variable DebugControls::state($class) ]
                    pack $lab -expand 1 -fill x
                }
                pack $catframe -expand 1 -fill x
            # Calculate the next column to use
            incr col_num $fill_direction
            if { $col_num == $num_cols } then {
                set fill_direction -1
                incr col_num $fill_direction
            } elseif { $col_num == -1 } then {
                set fill_direction 1
                incr col_num $fill_direction
            }
        }
        # Pack the column frames last, and from left to right
        for { set col_num 0 } { $col_num < $num_cols } { incr col_num } {
            set col_frame [ subst $[ subst {col_frame$col_num} ] ]
            pack $col_frame -side left -anchor n -padx 3 -pady 3
        }

    pack $hull

    set tray [frame $top.tray]
    button $tray.prefs -text "   Prefs   " -width 10 -pady 3 -padx 3 -command "DebugControls::import_prefs"
    button $tray.fresh -text "  Refresh  " -width 10 -pady 3 -padx 3 -command "DebugControls::import_simulator"
    button $tray.apply -text "   Apply   " -width 10 -pady 3 -padx 3 -command "DebugControls::export_simulator"
    button $tray.done  -text "   Close   " -width 10 -pady 3 -padx 3 -command "DebugControls::toggle_window"
    grid $tray.prefs $tray.fresh $tray.apply $tray.done -pady 10

    pack $tray
}

# --------------------------------------------------------------
# DebugControls::import_prefs
#   Imports the debug message preferences into the simulator
# --------------------------------------------------------------
proc DebugControls::import_prefs {} {

    variable state

    foreach cat [ simdebug category listformat ] {
        foreach class [ simdebug category $cat listformat ] {
            set state($class) [ get_pref debug_msg/$class [ string trim [ simdebug status $class ] ] ]
        }
    }
}


# --------------------------------------------------------------
# DebugControls::import_simulator
#   Imports the current simulator values into the 'state' array.
# --------------------------------------------------------------
proc DebugControls::import_simulator {} {

    variable state

    foreach cat [ simdebug category listformat ] {
        foreach class [ simdebug category $cat listformat ] {
            set state($class) [ string trim [ simdebug status $class ] ]
        }
    }
}

# --------------------------------------------------------------
# DebugControls::export_simulator
#   Exports the current state array values into the simulator.
#   Update the prefs as we go along.
# --------------------------------------------------------------
proc DebugControls::export_simulator {} {

    variable state

    foreach cat [ simdebug category listformat ] {
        foreach class [ simdebug category $cat listformat ] {
             simdebug set $class $state($class)
             set_pref debug_msg/$class $state($class)
        }
    }
}

# --------------------------------------------------------------
# DebugControls::refresh
#   This is just an alias for import_simulator
# --------------------------------------------------------------
proc DebugControls::refresh { args } {

    variable top .debugcontrols

    if { [ winfo exists $top ] } then {
       eval import_simulator $args
    }
}

