#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#       IBM Corp.
#
#------------------------------------------------------------------------

namespace eval Dialog {
    variable dialoginfo
}

#------------------------------------------------------------------------
# Dialog::create
#        Creates a generic dialog box. This proc now takes as a main input,
#        an array passed by reference. This array must contain a (top) member,
#        but all other members are optional. Current members include:
#
#      top -  the toplevel window name to check for and or create
#   parent -  the toplevel window that is the owner of $top. $top
#             will become a transient of $parent, unless parent does not exist,
#             in which case the transient call will be skipped.
#    title -  the title to appear in $top
#     geom -  the geometry values to apply to $top
#
# usage:
#     Dialog::create member_array_name args
# example:
#     array set dlgInfo "top .mytop title {My Top Window} geom 200x500+100+100"
#     Dialog::create dlgInfo -borderwidth 5
#------------------------------------------------------------------------
proc Dialog::create { _dlgInfo args } {

   variable dialoginfo
   upvar $_dlgInfo dlgInfo

   # Set the defaults, for the cases when a user has not specified a member
   set parent     "."
   set geom       "none"
   set title      ""

   set top $dlgInfo(top)
   catch { set parent $dlgInfo(parent) }
   catch { set geom   $dlgInfo(geom) }
   catch { set title  $dlgInfo(title) }

	if { [ winfo exists $top ] } then {
		switch -- [ wm state $top ] {
			normal {
				# Raise a buried window
				raise $top
			}
			withdrawn -
			iconified {
				# Open and restore geometry
				wm deiconify $top
				catch {
               wm geometry $top $dialoginfo(geo,$top)
               focus $top
            }
			}
		}
		return 0
	} else {
      eval toplevel $top $args
      if { [ winfo exists $parent ] } then {
         # make the dialog part of the input window
         wm transient $top [ winfo toplevel $parent ]
      }

      if { $geom != "none" } then {
         wm geom $top $geom
      }
      wm title $top $title
      return 1
	}
}

#------------------------------------------------------------------------
# Dialog::wait
#        Waits
#        example: Dialog::wait .topwin {.button_to_focus_on}
#------------------------------------------------------------------------
proc Dialog::wait {top varName {focus {}}} {
	upvar $varName var

   # Poke the variable if the user nukes the window
	bind $top <Destroy> [list set $varName $var]

	# Grab focus for the dialog
	if {[string length $focus] == 0} {
		set focus $top
	}
	set old [focus -displayof $top]
	focus $focus
	if {[winfo viewable $top] != 1} then {
      catch {tkwait visibility $top}
   }
   catch {grab $top}

	# Wait for the dialog to complete
	tkwait variable $varName
	catch {grab release $top}
	catch {focus $old}
}

#------------------------------------------------------------------------
# Dialog::dismiss
#        Closes a dialog box created with Dialog::create
#        example: Dialog::dismiss .topwin
#------------------------------------------------------------------------
proc Dialog::dismiss {top {geom_var none}} {
	variable dialoginfo
	# Save current size and position
	catch {
		# window may have been deleted
		set dialoginfo(geo,$top) [wm geometry $top]
      if {$geom_var != "none"} then {
         upvar $geom_var temp
         set temp [wm geometry $top]
         # only save the position of the window, depending on version window size may change
         # and not display correctly when prefs are used
         set index [string first "+" $temp]
         set tmp [string range $temp $index end]
         set temp $tmp
      }
		wm withdraw $top
	}
}

