#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  emitter.tcl
#
#
#------------------------------------------------------------------------

namespace eval Emitters {

   variable version        "1.0"
   variable initialized    0
}

#------------------------------------------------------------------------
#  Emitters::init
#------------------------------------------------------------------------
proc Emitters::init {} {

    variable initialized
    variable writer_settings
    variable reader_settings
    variable expect_setting
    variable emitter_binary
    variable readers_started

    global gvars

    set sim $gvars(firstsim)

    if { !$initialized } then {

       Hooks::register "Go" "Emitters::start_readers"
       # Use the simemit command to get the list of writers than can be enabled/disabled
       if { [catch {array set emitter_names [simemit list]} errors] } {

           set emitter_binary 0
           return
       }
       set emitter_binary 1
       # Read initial values from prefs file
       foreach d [lsort [array names emitter_names]] {
           set writer_settings($d) [get_pref emitters/writers/$d 0]
       }

       set expect_setting [get_pref emitters/expect_setting 0]

       # Get the readers to be launched from the prefs file
       set reader_settings(reader) [get_pref emitters/readers/reader 0]
       set reader_settings(sstats_live) [get_pref emitters/readers/sstats_live 0]
       set reader_settings(erat_reader) [get_pref emitters/readers/erat_reader 0]


       set readers_started(reader) 0
       set readers_started(sstats_live) 0

       if { $sim != "" } then {
           inject_writer_settings
       }

       set initialized 1
    }
}


#------------------------------------------------------------------------
#  Emitter::create
#
#------------------------------------------------------------------------
proc Emitters::create { widget } {

    variable control_nb
    global gvars

    if { ![ winfo exists $widget ] } then {

        toplevel $widget

        wm transient $widget .

        # Intercept window close and call the exit handler
        set top_widget [ winfo toplevel $widget ]
        wm protocol $top_widget WM_DELETE_WINDOW "Emitters::_on_close $top_widget"
        bind $top_widget <Configure> "Emitters::_save_geometry $top_widget"
        bind $top_widget <Key-F5> "user_refresh"

        set geom [ get_pref win_geom/cpanel "800x600+50+50" ]

        #------------------------------------------------------------------------
        #  Emitter Panel
        #------------------------------------------------------------------------
        set cpanel [ frame $widget.emitter -borderwidth 2 -relief groove ]

        set emitter_nb [ Options::create $cpanel [ namespace current ] ]

        $emitter_nb insert end writers -text "Writers"
        $emitter_nb insert end readers -text "Readers"
        $emitter_nb raise writers


        pack $cpanel -side top -fill both -expand yes

        wm geometry $top_widget $geom

    } else {

        wm deiconify [ winfo toplevel $widget ]
    }

    return $widget

}

#------------------------------------------------------------------------
# Emitters::_save_geometry
#------------------------------------------------------------------------
proc Emitters::_save_geometry { top_widget } {

   set_pref win_geom/cpanel [ winfo geometry $top_widget ]
}

#------------------------------------------------------------------------
# Emitters::-on_close
#------------------------------------------------------------------------
proc Emitters::-on_close { top_widget } {

   wm withdraw $top_widget
}



#------------------------------------------------------------------------
# Emitters::_compare_arrays
#   This compares two arrays and returns 0 if equal, 1 if not.
#   I previously used 'array get', but the return order of elements is not
#   guarateed, even for copies of the original array!
#------------------------------------------------------------------------
proc Emitters::_compare_arrays { _a1 _a2 } {

    upvar $_a1 a1
    upvar $_a2 a2

    set rtn 0
    foreach elem [ array names a1 ] {
        if { ![ string match $a1($elem) $a2($elem) ] } then {
            set rtn 1
            break
        }
    }
    return $rtn
}

#------------------------------------------------------------------------
#  Emitters::-createtab_writer
#       Creates the 'writers' emitter panel tab
#------------------------------------------------------------------------
proc Emitters::-createtab_writers { {topf} } {

    variable emitter_nb
    variable writer_settings
    variable temp_writer_settings
    variable expect_setting
    variable temp_expect_setting

    global gvars

    # Copy all of the settings to temp_settings array
    array set temp_writer_settings [ array get writer_settings ]

    set writersf [ frame $topf.writersf -bd 2 -relief groove ]
    set maxcols 5



    set row 0
    set col 1


    set entries [llength [array names temp_writer_settings]]
    set maxrows [expr $entries / $maxcols + 1]
    foreach d [lsort [array names temp_writer_settings]] {

        checkbutton $topf.w$d -anchor nw -justify left -text "$d" \
             -variable Emitters::temp_writer_settings($d)

        table $topf $topf.w$d $row,$col -anchor w -fill x
        if { $row > $maxrows } {
            incr col
            set row 0
        } else {
            incr row
        }
    }

}

#------------------------------------------------------------------------
#  Emitters::-on_ok_writers
#------------------------------------------------------------------------
proc Emitters::-on_ok_writers { } {

   variable writer_settings
   variable temp_writer_settings
   variable expect_setting

   # If any of the 'writer settings' changed, update the system.
   if { [ _compare_arrays writer_settings temp_writer_settings ] } then {
      # Copy all of the temp_settings to the settings array
      array set writer_settings [ array get temp_writer_settings ]


      inject_writer_settings

      # Update the prefs
      foreach d [lsort [array names temp_writer_settings]] {
          #set prefname [string map {" " "_"} $d]
          set_pref emitters/writers/$d $temp_writer_settings($d)
      }
   }
}

#------------------------------------------------------------------------
#  Emitters::-on_cancel_display
#------------------------------------------------------------------------
proc Emitters::-on_cancel_writers { } {

   variable writer_settings
   variable temp_writer_settings

   array set temp_writer_settings [ array get writer_settings ]
}

#------------------------------------------------------------------------
#  Emitters::inject_writer_settings
#       Injects various settings back into the simulator.
#------------------------------------------------------------------------
proc Emitters::inject_writer_settings { } {

    variable writer_settings
    variable shm_allocated
    variable expect_setting


    foreach d [lsort [array names writer_settings]]  {
        simemit set $d $writer_settings($d)
    }

}

#------------------------------------------------------------------------
#  Emitters::-createtab_simulator
#       Creates the 'readers' emitter panel tab
#------------------------------------------------------------------------
proc Emitters::-createtab_readers { {topf} } {

    variable emitter_nb
    variable reader_settings
    variable temp_reader_settings
    variable expect_setting
    variable temp_expect_setting

    global gvars

    # Copy all of the settings to temp_settings array
    array set temp_reader_settings [ array get reader_settings ]
    set temp_expect_setting $expect_setting

    set reader [ frame $topf.readerf -bd 2 -relief groove ]
    checkbutton $reader.button -variable Emitters::temp_reader_settings(reader) \
         -text "Reader (simple \"printf\" of emitter tags)" -font system_font

    pack $reader.button -side left -padx 5

    set stripstats_live [ frame $topf.stats_live -bd 2 -relief groove ]
    #label $stripstats_live.lab

    checkbutton $stripstats_live.button -variable Emitters::temp_reader_settings(sstats_live) \
        -text "Stripstats Live" -font system_font


    pack $stripstats_live.button -side left -padx 5

    set erat_reader [ frame $topf.erat_reader -bd 2 -relief groove ]
    checkbutton $erat_reader.button -variable Emitters::temp_reader_settings(erat_reader)  \
       -state disabled  -text "Erat Reader" -font system_font

    pack $erat_reader.button -side left -padx 5


    set expectf [ frame $topf.expectf -bd 2 -relief groove ]

    label $expectf.lab -text "Wait for selected  readers to attach?"
    checkbutton $expectf.button -variable Emitters::expect_setting -text yes
    pack $expectf.lab $expectf.button -side left -padx 5

    pack $reader $stripstats_live $erat_reader $expectf -side top -anchor w -fill both -expand 1
    pack $topf -side top -fill both -expand 1 -anchor w

}

#------------------------------------------------------------------------
#  Emitters::-on_ok_readers
#------------------------------------------------------------------------
proc Emitters::-on_ok_readers { } {

   variable reader_settings
   variable temp_reader_settings
   variable expect_setting

   # If any of the 'readers settings' changed, update the system.
   if { [ _compare_arrays reader_settings temp_reader_settings ] } then {

      # Copy all of the temp_settings to the settings array
      array set reader_settings [ array get temp_reader_settings ]


      # Update the prefs
      set_pref emitters/readers/reader      $reader_settings(reader)
      set_pref emitters/readers/sstats_live $reader_settings(sstats_live)
      set_pref emitters/readers/erat_reader $reader_settings(erat_reader)

      set_pref emitters/expect_setting      $expect_setting

   }

}

#------------------------------------------------------------------------
#  Emitters::-on_cancel_readers
#------------------------------------------------------------------------
proc Emitters::-on_cancel_readers { } {

   variable reader_settings
   variable temp_reader_settings
   variable expect_setting
   variable temp_expect_setting

   array set temp_reader_settings [ array get reader_settings ]
   set temp_expect_setting $expect_setting
}

proc Emitters::CheckEmitSettings {} {
        # Search to see if anyone has turned on any emitter records
        if { [catch {array set emitter_names [simemit list]} errors] } {
            #puts "Error on array set"
            return 0
        }
        foreach d [lsort [array names emitter_names]] {
            if { $emitter_names($d) != 0 } {
                #puts "simemit $d is turned on"
                return 1
            }
        }
        # Nothing is turned on
        #puts "No emit records are turned on"
        return 0
}

proc Emitters::start_readers {} {
    variable reader_settings
    variable readers_started
    variable emitter_binary
    global env

    if {!$emitter_binary} {
        return
    }

    if { [CheckEmitSettings] == 0 } {
        return
    }

    simemit init
    set wait_for_readers 0

    if {!$readers_started(reader)} {
       if {$reader_settings(reader)} {
          puts "Starting emitter: $env(EXEC_DIR)/emitter/reader"
          ereader start $env(EXEC_DIR)/emitter/reader [pid]
          set readers_started(reader) 1
          incr wait_for_readers
       }
    } else {
       #puts "Emitter reader \"reader\" already started"
    }

    if {!$readers_started(sstats_live)} {

       if {$reader_settings(sstats_live)} {
           puts "Starting emitter: $env(EXEC_DIR)/emitter/stripstats_live"
           set readers_started(sstats_live) 1
           ereader start $env(EXEC_DIR)/emitter/stripstats_live [pid] $env(PROFILER_DIR)/live-graph
           incr wait_for_readers
       }
    } else {
       #puts "Emitter reader \"stripstats_live already started"
    }

    if {$wait_for_readers} {
       ereader expect $wait_for_readers
    }
    set wait_for_readers 0

}
