#!/usr/bin/wish
#!wish

package require BLT
namespace import -force blt::*

namespace eval EventLog {
    variable active_events
    variable settings
    variable winnames
    variable canvas_fonts
    variable cycles
    variable components
    variable component_events
    variable component_event_start
    variable component_event_stop
    variable component_stackedevent_list
    variable component_stringevent_list
    variable cycletimes
    variable histogram_set

    set settings(initialized)   0
    set settings(cycle_refresh_rate) 10000000

    font create canvas_fonts(tics) -family helvetica -size 8
    font create canvas_fonts(component) -family helvetica -size 10
    font create canvas_fonts(event) -family helvetica -size 8
}

proc EventLog::NewCycle {sim cycle} {
    variable cycles

    if { $cycles($sim,first) == -1 } {
        set cycles($sim,first) $cycle
        set cycles($sim,last) $cycle
    } else {
        if { [compare64 $cycle $cycles($sim,last)] > 0 } {
            set cycles($sim,last) $cycle
        }
    }
}

################################################################
#
#   EventStart & EventStop
#
################################################################
proc EventLog::EventTagLookup {sim component event base_tag} {
    variable component_event_tags

    set matching_list [lsearch -all -inline -glob $component_event_tags($sim,$component,$event) "${base_tag}*"]

    if { [llength $matching_list] == 0 } {
        puts "ERROR ERROR ERROR: EventTagLookup: unable to find last tag for $sim $component $event $base_tag ($component_event_tags($sim,$component,$event)"
        return "UNKNOWN"
    }

    return [lindex $matching_list [expr [llength $matching_list] - 1]]
}

proc EventLog::EventStart {sim component tag event cycle} {
    variable active_events
    variable components
    variable component_events
    variable component_event_tags
    variable component_event_start

    ListPushUnique EventLog::components($sim) $component
    ListPushUnique EventLog::component_events($sim,$component) $event

    incr active_events($sim)

    # Make sure we have a unique tag name for the list
    set uniq_tag "${tag}_${cycle}"

    if { [info exists component_event_tags($sim,$component,$event)] == 0 } {
        set component_event_tags($sim,$component,$event) $uniq_tag
    } else {
        if { [lsearch -exact $component_event_tags($sim,$component,$event) $uniq_tag] != -1 } {
            puts "ERROR ERROR: non unique tag for $sim $component $event"
        }
        lappend component_event_tags($sim,$component,$event) $uniq_tag
    }
    
    set component_event_start($sim,$component,$event,$uniq_tag) $cycle
}
proc EventLog::EventStop {sim component tag event cycle} {
    variable active_events
    variable component_event_stop

    incr active_events($sim) -1
    set uniq_tag [EventTagLookup $sim $component $event $tag]
    set component_event_stop($sim,$component,$event,$uniq_tag) $cycle
}

################################################################
#
#   EventOnce
#
################################################################
proc EventLog::EventOnce {sim component event cycle} {
    variable components
    variable component_events
    variable component_stackedevent_list

    ListPushUnique EventLog::components($sim) $component
    ListPushUnique EventLog::component_events($sim,$component) $event

    if { [info exists component_stackedevent_list($sim,$component,$event)] == 0 } {
        set component_stackedevent_list($sim,$component,$event) $cycle
    } else {
        lappend component_stackedevent_list($sim,$component,$event) $cycle
    }
}

################################################################
#
#   EventString
#
################################################################
proc EventLog::EventString {sim component event cycle str} {
    variable components
    variable component_events
    variable component_stringevent_list

    ListPushUnique EventLog::components($sim) $component
    ListPushUnique EventLog::component_events($sim,$component) $event

    lappend component_stringevent_list($sim,$component,$event) "$cycle $str"
}

################################################################
#
#   Channel_Stall_Start & Channel_Stall_Stop
#
################################################################
proc EventLog::Channel_Stall_Start {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d channel %d cycle %s" spunum channel cycles

    if { [catch "EventStart $sim SPU${spunum} none CHANNEL_STALL $cycles" errs] != 0 } {
        puts "Channel_Stall_Start: EventStart failed with errors: $errs"
        return
    }
    set cycletimes(SPU$spunum,channel_stall_start) $cycles
    NewCycle $sim $cycles
    $winnames($sim).log.listing insert end "$cycles : CHANNEL_STALL_START #$spunum channel $channel\n"

    #if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
    #    puts "Channel_Stall_Start: Refresh failed with errors: $errs"
    #    return
    #}
}

proc EventLog::Channel_Stall_Stop {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d cycle %s" spunum cycles

    EventStop $sim SPU${spunum} none CHANNEL_STALL $cycles
    NewCycle $sim $cycles

    if { [info exists cycletimes(SPU$spunum,channel_stall_start)] } {
        set diff [sub64 $cycles $cycletimes(SPU$spunum,channel_stall_start)] 
    } else {
        set diff "?"
    }
    $winnames($sim).log.listing insert end "$cycles : CHANNEL_STALL_STOP #$spunum (DIFF: $diff cycles)\n"

    #if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
    #    puts "Refresh had errors :  $errs"
    #}
}

################################################################
#
#   SPE_LIST_DMA Start & Stop
#
################################################################
proc EventLog::SPE_LIST_DMA_Start {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d action %s list_entries %s list_addr %s cycle %s" spunum action list_entries list_addr cycles
    set tag "${action}_${list_entries}_${list_addr}"
    if { [catch "EventStart $sim SPU${spunum} $tag DMA_LIST $cycles" errs] != 0 } {
        puts "SPE_LIST_DMA_Start: EventStart failed with errors: $errs"
        return
    }
    set cycletimes(SPU$spunum,dma_list_start,$action,$list_addr,$list_entries) $cycles
    NewCycle $sim $cycles
    $winnames($sim).log.listing insert end "$cycles : SPE_LIST_DMA_START #$spunum ACTION:$action LIST_ADDR:$list_addr LIST_ENTRIES:$list_entries\n"
}

proc EventLog::SPE_LIST_DMA_Stop {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d action %s list_entries %s list_addr %s cycle %s" spunum action list_entries list_addr cycles
    set tag "${action}_${list_entries}_${list_addr}"
    EventStop $sim SPU${spunum} $tag DMA_LIST $cycles
    NewCycle $sim $cycles

    if { [info exists cycletimes(SPU$spunum,dma_list_start,$action,$list_addr,$list_entries)] } {
        set diff [sub64 $cycles $cycletimes(SPU$spunum,dma_list_start,$action,$list_addr,$list_entries)] 
    } else {
        set diff "?"
    }
    $winnames($sim).log.listing insert end "$cycles : SPE_LIST_DMA_STOP #$spunum ACTION:$action LIST_ADDR:$list_addr LIST_ENTRIES:$list_entries (DIFF: $diff cycles)\n"
}

################################################################
#
#   SPE_DMA Start & Stop
#
################################################################
proc EventLog::SPE_DMA_Start {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d action %s size %s EA %s LS %s cycle %s" spunum action size ea ls cycles

    set tag "${action}_${ea}_${ls}_${size}"

    if { [catch "EventStart $sim SPU${spunum} $tag DMA $cycles" errs] != 0 } {
        puts "SPE_DMA_Start: EventStart failed with errors: $errs"
        return
    }
    set cycletimes(SPU$spunum,dma_start,$ea,$ls,$action,$size) $cycles
    NewCycle $sim $cycles
    $winnames($sim).log.listing insert end "$cycles : SPE_DMA_START #$spunum ACTION:$action EA:$ea LS:$ls SIZE:$size\n"
}

proc EventLog::SPE_DMA_Stop {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d action %s size %s EA %s LS %s cycle %s" spunum action size ea ls cycles
    set tag "${action}_${ea}_${ls}_${size}"

    if { [catch "EventStop $sim SPU${spunum} $tag DMA $cycles" errs] } {
        puts "SPE_DMA_Stop: EventStop errs is $errs"
    }
    NewCycle $sim $cycles

    if { [info exists cycletimes(SPU$spunum,dma_start,$ea,$ls,$action,$size)] } {
        set diff [sub64 $cycles $cycletimes(SPU$spunum,dma_start,$ea,$ls,$action,$size)] 
    } else {
        set diff "?"
    }
    $winnames($sim).log.listing insert end "$cycles : SPE_DMA_STOP #$spunum ACTION:$action EA:$ea LS:$ls SIZE:$size (DIFF: $diff cycles)\n"
}

################################################################
#
#   SPE DMAQ_Suspend_Xlate Start & Stop
#
################################################################
proc EventLog::DMAQ_Suspend_Xlate_Start {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d cycle %s" spunum cycles

    if { [catch "EventStart $sim SPU${spunum} none DMAQ_SUSPEND_XLATE $cycles" errs] != 0 } {
        puts "SPU_Start: EventStart failed with errors: $errs"
        return
    }
    set cycletimes(SPU$spunum,dmaq_suspend_xlate_start) $cycles
    NewCycle $sim $cycles
    $winnames($sim).log.listing insert end "$cycles : DMAQ_SUSPEND_XLATE_START #$spunum\n"

    #if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
    #    puts "SPU_Start: Refresh failed with errors: $errs"
    #    return
    #}
}

proc EventLog::DMAQ_Suspend_Xlate_Stop {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d cycle %s" spunum cycles

    EventStop $sim SPU${spunum} none DMAQ_SUSPEND_XLATE $cycles
    NewCycle $sim $cycles

    if { [info exists cycletimes(SPU$spunum,start)] } {
        set diff [sub64 $cycles $cycletimes(SPU$spunum,dmaq_suspend_xlate_start)] 
    } else {
        set diff "?"
    }
    $winnames($sim).log.listing insert end "$cycles : DMAQ_SUSPEND_XLATE_STOP  #$spunum (DIFF: $diff cycles)\n"

    #if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
    #    puts "Refresh had errors :  $errs"
    #}
}

################################################################
#
#   SPU Start & Stop
#
################################################################
proc EventLog::SPU_Start {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d cycle %s" spunum cycles

    if { [catch "EventStart $sim SPU${spunum} none RUN $cycles" errs] != 0 } {
        puts "SPU_Start: EventStart failed with errors: $errs"
        return
    }
    set cycletimes(SPU$spunum,start) $cycles
    NewCycle $sim $cycles
    $winnames($sim).log.listing insert end "$cycles : SPU_START #$spunum\n"

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "SPU_Start: Refresh failed with errors: $errs"
        return
    }
}

proc EventLog::SPU_Stop {sim args} {
    variable winnames
    variable cycletimes

    scan [lindex $args 0] " spu %d cycle %s" spunum cycles

    EventStop $sim SPU${spunum} none RUN $cycles
    NewCycle $sim $cycles

    if { [info exists cycletimes(SPU$spunum,start)] } {
        set diff [sub64 $cycles $cycletimes(SPU$spunum,start)] 
    } else {
        set diff "?"
    }
    $winnames($sim).log.listing insert end "$cycles : SPU_STOP  #$spunum (DIFF: $diff cycles)\n"

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "Refresh had errors :  $errs"
    }
}

################################################################
#
#   SPE XLate Fault
#
################################################################
proc EventLog::SPE_XLate_Fault {sim args} {
    variable winnames

    scan [lindex $args 0] " spu %d ea %s cycle %s" spunum ea cycles
    EventOnce $sim SPU${spunum} SPE_XLATE_FAULT $cycles
    NewCycle $sim $cycles
    $winnames($sim).log.listing insert end "$cycles : SPE_XLATE_FAULT #$spunum EA $ea\n"
}

proc EventLog::KernelExec { sim args } {
    variable winnames

    scan [lindex $args 0] "address %s mcm %s cpu %s thread %s" addr mcm cpu thread
    set r3 [$sim mcm $mcm cpu $cpu:$thread display gpr 3]
    set cycle [$sim display cycles]

    NewCycle $sim $cycle

    if { [catch "set pa [$sim mcm $mcm cpu $cpu:$thread dtranslate $r3]" errs] == 0 } {
        if { [catch "set execname [$sim mcm $mcm memory display $pa STRING]" errs] == 0 } {
            if { [catch "EventString $sim PROCESS EXEC $cycle $execname" errs] } {
                puts "EventString had errors: $errs"
                exit
            }
            $winnames($sim).log.listing insert end "$cycle : PROCESS EXEC CPU#$cpu:$thread EXECNAME $execname\n"
        } else {
            puts "KernelExec failed with $errs"
        }
    } else {
        puts "KernelExec failed to translate $r3"
    }

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "EventString: Refresh failed with errors: $errs"
        return
    }
}

proc EventLog::Set_Kernel_Func_Trigger { sim function_name command } {
    global vmlinux_filename
    if { [catch "set entry_info [list [object address $vmlinux_filename $function_name]]" errs] == 0 } {
        if { [catch "array set func_info [list $entry_info]" errs] == 0 } {
            $sim trigger set pc $func_info(FUNC_START) $command
        } else {
            puts "Set_Kernel_Func_Trigger failed with $errs"
        }
    } else {
        puts "Set_Kernel_Func_Trigger object call failed with $errs"
    }
}

proc EventLog::RegisterIfAvailable { sim assoc value } {
    
    set assoc_triggers [$sim trigger display assoc avail]

    if { [lsearch -exact $assoc_triggers $assoc] != -1 } {

        $sim trigger set assoc $assoc $value

    }
}

proc EventLog::Set_Cell_Triggers { sim } {
    
    RegisterIfAvailable $sim SPE_DMA_START "EventLog::SPE_DMA_Start $sim"
    RegisterIfAvailable $sim SPE_DMA_STOP "EventLog::SPE_DMA_Stop $sim"
    RegisterIfAvailable $sim SPE_LIST_DMA_START "EventLog::SPE_LIST_DMA_Start $sim"
    RegisterIfAvailable $sim SPE_LIST_DMA_STOP "EventLog::SPE_LIST_DMA_Stop $sim"
    RegisterIfAvailable $sim SPU_START "EventLog::SPU_Start $sim"
    RegisterIfAvailable $sim SPU_STOP "EventLog::SPU_Stop $sim"
    RegisterIfAvailable $sim SPE_XLATE_FAULT "EventLog::SPE_XLate_Fault $sim"
    RegisterIfAvailable $sim SPE_CHANNEL_STALL_START "EventLog::Channel_Stall_Start $sim"
    RegisterIfAvailable $sim SPE_CHANNEL_STALL_STOP "EventLog::Channel_Stall_Stop $sim"
    RegisterIfAvailable $sim SPE_DMAQ_SUSPEND_XLATE_START "EventLog::DMAQ_Suspend_Xlate_Start $sim"
    RegisterIfAvailable $sim SPE_DMAQ_SUSPEND_XLATE_STOP "EventLog::DMAQ_Suspend_Xlate_Stop $sim"

    # Try to get exec name from these points.... r3 should point to the filename
    # .compat_do_execve
    # .do_execve
    # .open_exec I believe causes too many hits
    catch {
        Set_Kernel_Func_Trigger $sim ".compat_do_execve" "EventLog::KernelExec $sim"
        Set_Kernel_Func_Trigger $sim ".do_execve" "EventLog::KernelExec $sim"
    } errs
}

proc EventLog::CreateTextbox { win sim } {

    # Create Log Window
    frame ${win}.log
    label ${win}.log.label -justify center -text "Log Trace:"
    scrollbar ${win}.log.listx -orient horiz -command "${win}.log.listing xview"
    scrollbar ${win}.log.listy -orient vert -command "${win}.log.listing yview"
    text ${win}.log.listing -width 50 -height 15 -font textbox_font -xscrollcommand "${win}.log.listx set" -yscrollcommand "${win}.log.listy set" -tabs "12c" -wrap none
    table ${win}.log \
          ${win}.log.label      0,0 -fill x \
          ${win}.log.listing    1,0 -fill both \
          ${win}.log.listy      1,1 -fill y \
          ${win}.log.listx      2,0 -fill x
    table configure ${win}.log c* r* -resize none
    table configure ${win}.log r1 c0 -resize both

    return ${win}.log
}

proc EventLog::Convert_Graph2Cycle { sim graphx } {
    variable canvas_settings
    variable cycles

    set cycle_range [sub64 [CycleEnd $sim] [CycleStart $sim]]
    set x_range [expr $canvas_settings($sim,x) - $canvas_settings($sim,graph_start_x)]

    set graphx [expr int($graphx)]
    if { $graphx < $canvas_settings($sim,graph_start_x) } {
        #set graphx $canvas_settings($sim,graph_start_x)
        set graphx 0
    } else {
        set graphx [expr $graphx - $canvas_settings($sim,graph_start_x)]
    }

    set cycle [expr "round( double($graphx) * (double(${cycle_range}) / double(${x_range})) )"]
    set cycle [add64 $cycle [CycleStart $sim]]

    return $cycle
}

proc EventLog::Convert_Cycle2Graph { sim cycle } {
    variable canvas_settings
    variable cycles

    set cycle_start [CycleStart $sim]
    set cycle_end [CycleEnd $sim]

    if { [compare64 $cycle $cycle_start] < 0 } {
        return $canvas_settings($sim,graph_start_x)
    }
    if { [compare64 $cycle $cycle_end] > 0 } {
        return $canvas_settings($sim,x)
    }
    set cycle_range [sub64 $cycle_end $cycle_start]
    set x_range [expr $canvas_settings($sim,x) - $canvas_settings($sim,graph_start_x)]
    set cycles_per_x [expr "double(${cycle_range}) / double(${x_range})"]
#puts "total x is $canvas_settings($sim,x)"
#puts "graphx_start is $canvas_settings($sim,graph_start_x)"
#puts "CycleStart is $cycle_start"
#puts "CycleEnd is $cycle_end"
#puts "cycles_per_x is $cycles_per_x"
#puts "cycle is $cycle"
    if { $cycles_per_x == 0 } {
        set x_coord 0
    } else {
        set x_coord [sub64 $cycle $cycle_start]
        set x_coord [expr "round( double($x_coord) / double($cycles_per_x) )"]
    }
    set x_coord [expr $x_coord + $canvas_settings($sim,graph_start_x)]

    if { $x_coord > $canvas_settings($sim,x) } {
        puts "ERROR ERROR ERROR: x_coord $x_coord exceeds graph x-length $canvas_settings($sim,x)"
        puts "ERROR ERROR ERROR: x_coord $x_coord exceeds graph x-length $canvas_settings($sim,x)"
        puts "ERROR ERROR ERROR: x_coord $x_coord exceeds graph x-length $canvas_settings($sim,x)"
        puts "ERROR ERROR ERROR: x_coord $x_coord exceeds graph x-length $canvas_settings($sim,x)"
        set x_coord $canvas_settings($sim,x)
    }
    return $x_coord
}

proc EventLog::SelectEvent { sim x y } {
    variable canvas_settings

    set cvs $canvas_settings($sim,canvas)

    set events [$cvs find closest $x $y]

    foreach event $events {
        puts "Event ID is $event"
        $cvs itemconfigure $event -fill orange
    }
}

proc EventLog::Motion { sim x } {
    variable canvas_settings

    set cvs $canvas_settings($sim,canvas)
    $cvs delete marker_line
    set cx [$cvs canvasx $x]
    #set cx $x

    set canvas_settings($sim,marker_x) [format64 "%lld" [Convert_Graph2Cycle $sim $cx]]
    set x [Convert_Cycle2Graph $sim $canvas_settings($sim,marker_x)]
    set y $canvas_settings($sim,graph_start_y)

    $cvs create line $x $y $x $canvas_settings($sim,graph_end_y) -fill blue -tag {marker_line event_line}

}

proc EventLog::Clip_End { sim } {
    variable canvas_settings

    set canvas_settings($sim,clip_end) $canvas_settings($sim,marker_x)

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "Clip_End: Refresh failed with errors: $errs"
        return
    }
}

proc EventLog::Unclip_End { sim } {
    variable canvas_settings

    set canvas_settings($sim,clip_end) 0

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "Unclip_End: Refresh failed with errors: $errs"
        return
    }
}

proc EventLog::Clip_Start { sim } {
    variable canvas_settings

    set canvas_settings($sim,clip_start) $canvas_settings($sim,marker_x)

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "Clip_Start: Refresh failed with errors: $errs"
        return
    }
}

proc EventLog::Unclip_Start { sim } {
    variable canvas_settings

    set canvas_settings($sim,clip_start) 0

    if { [catch "EventLog::Refresh $sim nothing" errs] != 0 } {
        puts "Unclip_Start: Refresh failed with errors: $errs"
        return
    }
}

proc EventLog::CreateCanvas { win sim } {

    variable canvas_settings
    variable cycles

    set canvas_settings($sim,x) 1280
    set canvas_settings($sim,y) 2048
    set canvas_settings($sim,tics_per_x) 8
    set canvas_settings($sim,graph_start_x) 185
    set canvas_settings($sim,graph_start_y) 25
    set canvas_settings($sim,graph_end_y) $canvas_settings($sim,y)
    set canvas_settings($sim,canvas) ${win}.activity.cvs
    set canvas_settings($sim,clip_start) 0
    set canvas_settings($sim,clip_end) 0
    set canvas_settings($sim,marker_x) [Convert_Graph2Cycle $sim $canvas_settings($sim,graph_start_x)]

    # Create Log Window
    frame ${win}.activity
    label ${win}.activity.label -justify center -text "Activity Chart:"
    scrollbar ${win}.activity.listx -orient horiz -command "${win}.activity.cvs xview"
    scrollbar ${win}.activity.listy -orient vert -command "${win}.activity.cvs yview"
    set cvs $canvas_settings($sim,canvas)
    canvas ${cvs} -width 400 -height 400 \
        -xscrollcommand "${win}.activity.listx set" \
        -yscrollcommand "${win}.activity.listy set" \
        -scrollregion "0 0 $canvas_settings($sim,x) $canvas_settings($sim,y)"

    frame ${win}.activity.controls
    label ${win}.activity.controls.marker_label -justify center -text "Marker Cycle:"
    entry ${win}.activity.controls.marker_cycle -width 20 -relief sunken -textvariable EventLog::canvas_settings($sim,marker_x)
    button ${win}.activity.controls.unclip_start -text " Unclip Start " -command "EventLog::Unclip_Start $sim"
    button ${win}.activity.controls.start_at -text " Clip Start at Marker " -command "EventLog::Clip_Start $sim"
    button ${win}.activity.controls.unclip_end -text " Unclip End " -command "EventLog::Unclip_End $sim"
    button ${win}.activity.controls.end_at -text " Clip End at Marker " -command "EventLog::Clip_End $sim"

    table ${win}.activity.controls \
          ${win}.activity.controls.marker_label     0,0 -fill x \
          ${win}.activity.controls.marker_cycle     0,1 -fill x \
          ${win}.activity.controls.start_at         0,2 -fill x \
          ${win}.activity.controls.unclip_start     0,3 -fill x \
          ${win}.activity.controls.end_at           0,4 -fill x \
          ${win}.activity.controls.unclip_end       0,5 -fill x 
    table configure ${win}.activity.controls c* r* -resize none

    bind $cvs <Shift-Motion> "EventLog::Motion $sim %x"
    ${cvs} bind event_line <Button-1> "EventLog::SelectEvent $sim %x %y"

    table ${win}.activity \
          ${win}.activity.label      0,0 -fill x \
          ${cvs}                     1,0 -fill both \
          ${win}.activity.listy      1,1 -fill y \
          ${win}.activity.listx      2,0 -fill x \
          ${win}.activity.controls   3,0 -cspan 2 -fill x 
    table configure ${win}.activity c* r* -resize none
    table configure ${win}.activity r1 c0 -resize both

    return ${win}.activity
}

proc EventLog::CycleStart { sim } {
    variable cycles
    variable canvas_settings

    set cycle_start $canvas_settings($sim,clip_start)
    if { $cycle_start == 0 } {
        set cycle_start $cycles($sim,first)

        if { $cycle_start == -1 } {
            return 0
        }
    }
    return $cycle_start
}

proc EventLog::CycleEnd { sim } {
    variable cycles
    variable canvas_settings

    set cycle_end $canvas_settings($sim,clip_end)
    if { $cycle_end == 0 } {
        set cycle_end $cycles($sim,last)
    }
    return $cycle_end
}

proc EventLog::CyclesInMega { cycles } {
    set mega [div64 $cycles 1000000]

    return "[format64 "%lld" $mega]M"
}

proc EventLog::CycleRangeFitsOnGraph { sim begin end } {
    set graph_cycle_start [CycleStart $sim]
    set graph_cycle_end [CycleEnd $sim]

    #
    #   If both begin and end are before the start
    #   or
    #   If both begin and end are after the end
    #   then the range is not in the current view
    #
    if { ( ( ([compare64 $begin $graph_cycle_start] < 0) && ([compare64 $end $graph_cycle_start] < 0 ) ) || 
           ( ([compare64 $begin $graph_cycle_end] > 0) && ([compare64 $end $graph_cycle_end] > 0 ) ) ) } {
         return 0
    } else {
         return 1
    }
}

proc EventLog::CycleFitsOnGraph { sim cycle } {
    if { ([compare64 $cycle [CycleStart $sim]] < 0) || 
         ([compare64 $cycle [CycleEnd $sim]] > 0 ) } {
         return 0
    } else {
         return 1
    }
}

################################################################
#
#   SPE XLate Fault
#
################################################################
proc EventLog::Refresh { sim args } {
    variable cycles
    variable canvas_settings
    variable winnames
    variable settings
    variable components
    variable component_events
    variable active_events
    variable component_event_tags
    variable component_event_start
    variable component_event_stop
    variable component_stackedevent_list
    variable component_stringevent_list
    variable canvas_fonts
    variable histogram_set

    set y_index 0

    set cvs $winnames($sim).activity.cvs
    $cvs delete all

    incr y_index 10

    #
    # If we still have active events, we should update the latest end cycle time
    #
    if { $active_events($sim) > 0 } {
        NewCycle $sim [$sim display cycles]
    }

    set cycle_start [CycleStart $sim]
    set cycle_end [CycleEnd $sim]

    # Create the tick marks along x (or at least start and stop)
    set cycles_total [sub64 $cycle_end $cycle_start]
    set cycles_per_tick [div64 $cycles_total $canvas_settings($sim,tics_per_x)]
    if { $cycles_per_tick > 100 } {
        for { set tick $cycle_start } { $tick <= $cycle_end } { set tick [add64 $tick $cycles_per_tick] } {
            $cvs create text [Convert_Cycle2Graph $sim $tick] $y_index -anchor nw -justify left -text "[EventLog::CyclesInMega $tick]" -tag event_line -font canvas_fonts(tics)
        }
    }

    incr y_index 15
    set graph_y_start $y_index
    set count 0
    foreach component $components($sim) {
        set component_start_y $y_index
        incr y_index 10
        set x_index 0
        $cvs create text $x_index $y_index -anchor nw -justify left -text $component -font canvas_fonts(component)
        incr x_index $canvas_settings($sim,graph_start_x)

        foreach event $component_events($sim,$component) {
            $cvs create text [expr $x_index -5] $y_index -anchor ne -justify right -text $event  -font canvas_fonts(event)

            #
            # In case this is a histogram set, clear the histogram
            #
            set histogram_set($sim) ""
            for { set i 0 } { $i <= $canvas_settings($sim,x) } { incr i } {
                set histogram($i) 0
            }

            #
            #   component_stringevent_list has strings associated with a cycle
            #
            if { [info exists component_stringevent_list($sim,$component,$event)] } {
                set y_advance 0
                foreach event_info $component_stringevent_list($sim,$component,$event) {
                    if { [CycleFitsOnGraph $sim [lindex $event_info 0]] } {
                        # We have a start stop pair
                        # Convert event cycle to x coord
                        set x_coord [EventLog::Convert_Cycle2Graph $sim [lindex $event_info 0]]
                        set str [lindex $event_info 1]
                        $cvs create text $x_coord [expr $y_index + $y_advance] -anchor nw -justify left -text "$str" -tag event_line -font canvas_fonts(tics)
                        incr y_advance 10 
                        if { $y_advance > 80 } {
                            set y_advance 0
                        }
                    }
                }
                incr y_index 90
            }

            #
            #   component_event_tags includes start-stop events
            #
            if { [info exists component_event_tags($sim,$component,$event)] } {

                incr y_index 5
                set y_advance 0

                foreach tag $component_event_tags($sim,$component,$event) {

                    # Get start cycle for this event
                    set cycle0 $component_event_start($sim,$component,$event,$tag)

                    # Get stop cycle for this event, or current time if still active
                    if { [info exists component_event_stop($sim,$component,$event,$tag)] } {
                        set cycle1 $component_event_stop($sim,$component,$event,$tag)
                    } else {
                        set cycle1 [$sim display cycles]
                    }

                    if { [CycleRangeFitsOnGraph $sim $cycle0 $cycle1] } {
                        # We have a start stop pair
                        # Convert event cycle to x coord
                        set x_coord0 [EventLog::Convert_Cycle2Graph $sim $cycle0]
                        set x_coord1 [EventLog::Convert_Cycle2Graph $sim $cycle1]
                        $cvs create line $x_coord0 [expr $y_index + $y_advance] $x_coord1 [expr $y_index + $y_advance] -tag event_line
                        incr y_advance 2 
                        if { $y_advance > 20 } {
                            set y_advance 0
                        }
                    }
                }
                incr y_index 20
            }

            #
            #   component_stackedevent_list has single events which get added up
            #       at each point in the graph
            #
            if { [info exists component_stackedevent_list($sim,$component,$event)] } {

                incr y_index 10 
                set y_advance 0

                foreach event_info $component_stackedevent_list($sim,$component,$event) {

                    set cycle [lindex $event_info 0]

                    if { [CycleFitsOnGraph $sim $cycle] } {

                        set x_coord [EventLog::Convert_Cycle2Graph $sim $cycle]

                        ListPushUnique EventLog::histogram_set($sim) $x_coord
                        if { [info exists histogram($x_coord)] } {
                            incr histogram($x_coord)
                        } else {
                            set histogram($x_coord) 0
                        }
                    }
                }

                foreach entry $histogram_set($sim) {
                    $cvs create line $entry $y_index $entry [expr $y_index - ( 2 * $histogram($entry))] -tag event_line -fill red
                }
            }
        }

        ################################################################
        # Now draw the bounding box for this diagram
        ################################################################
        incr y_index 5
        incr count
        if { ( $count % 2 ) != 0 } {
            $cvs create rectangle 0 $component_start_y $canvas_settings($sim,x) $y_index -tags bbox -fill gray
            $cvs lower bbox
        }
    }

    set canvas_settings($sim,graph_end_y) $y_index
    ################################################################
    # Now a line separating the x labels from the graph
    ################################################################
    $cvs create line $canvas_settings($sim,graph_start_x) $graph_y_start $canvas_settings($sim,graph_start_x) $y_index -fill black -tag bbox
    $cvs lower bbox

    ################################################################
    # Now draw the marker line that can move with the Shift-Motion
    # of the mouse
    ################################################################
    set x [Convert_Cycle2Graph $sim $canvas_settings($sim,marker_x)]
    set y $canvas_settings($sim,graph_start_y)

    $cvs create line $x $y $x $canvas_settings($sim,graph_end_y) -fill blue -tag {marker_line event_line}
}

proc EventLog::Init { sim } {

    variable winnames
    variable settings
    variable cycles
    variable components
    variable component_events
    variable active_events

    set winnames($sim) .eventlog_${sim}


    if { [winfo exists $winnames($sim)] } {
        raise $winnames($sim)
        return
    }
    toplevel $winnames($sim)

    if { $settings(initialized) != 1 } {
        set components($sim) ""
        set cycles($sim,first) -1
        set cycles($sim,last) 0
        set active_events($sim) 0
        EventLog::Set_Cell_Triggers $sim

        set settings(initialized) 1
    }

    table $winnames($sim) \
        [EventLog::CreateCanvas $winnames($sim) $sim] 0,0 -fill both \
        [EventLog::CreateTextbox $winnames($sim) $sim] 1,0 -fill both

    table configure $winnames($sim) c* r* -resize none
    table configure $winnames($sim) c0 r0 r1 -resize both

    EventLog::Refresh $sim
}
