#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2004
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  flyover.tcl
#------------------------------------------------------------------------

#
#  There is one flyover window containing the value of the word under it.
#  The contents of the window is determined by the window under it.
#

namespace eval flyover {
    variable flywin
    variable mpos
    variable clients   
    variable refresh_interval 100

    array set clients "cnt 0 "

    # mouse position info 
    array set mpos "win .  x 0 y 0 last_idx 0.0" 

    # flyover window info
    array set flywin "
       last_scan    0.0
       txt          \" \"
       value        \" \"
       text_up       0
    "
}

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
proc flyover::init {} {

    # Initialize text window
    if { ![ winfo exists .flyover ] }  then {
       #toplevel .flyover -cursor top_left_arrow       ;# create toplevel window on first use
       toplevel .flyover -cursor arrow       ;# create toplevel window on first use
       wm withdraw .flyover                           ;# Hide until positioned.
       label .flyover.help -justify left\
          -textvariable flyover::flywin(txt) \
          -bg lightyellow -fg black -borderwidth 1 -relief solid
       entry .flyover.value -textvariable flyover::flywin(value)
       pack .flyover.help -ipadx 20

       # I am not sure why this needs to be done, but if you are
       # using CDE Window Manager and telnet to a Linux box, the
       # wm override causes the window to be underneath everyone else.
       # This bit of voodoo seems to make it work correctly.
       wm deiconify .flyover
       wm override .flyover 1
       wm withdraw .flyover
    }
}

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
proc flyover::register {win scancmd} {
    variable clients    

    flyover::init

    lappend clients(winlist) $win
    set clients(scancmd,$win) $scancmd

    bind $win <Motion>       "+flyover::TrackMouse %W %x %y"
    bind $win <Button-3>     "+flyover::grab"
    bind $win <Destroy>      "+flyover::unregister $win"
    incr clients(cnt)

    if {$clients(cnt) == 1} { 
        flyover::refresh
    }
}

#------------------------------------------------------------------------
# this gets called when the window is destroyed
#------------------------------------------------------------------------
proc flyover::unregister {win} {
    variable clients    

    # find name in window list and remove it
    set idx [lsearch $clients(winlist) $win]
    if {$idx != -1} {
        set clients(winlist) [lreplace $clients(winlist) $idx $idx]
        unset clients(scancmd,$win)
        incr clients(cnt) -1
    }
    if {$clients(cnt) <= 0} { 
        variable mpos
        set mpos(win) .
        
    }
}

#------------------------------------------------------------------------
# flyover::TrackMouse
# the current mouse position gets saved whenever it moves.
# since it moves a lot, it gets processed periodically by refresh
#------------------------------------------------------------------------
proc flyover::TrackMouse { win x y } {
   variable mpos

   set mpos(win) $win
   set mpos(x) $x
   set mpos(y) $y
}

#------------------------------------------------------------------------
# flyover::grab
#     Grabs the last flyover text value and stores it in the selection
#     buffer (aka the clipboard) for pasting.
#------------------------------------------------------------------------
proc flyover::grab { } {
   # Update the hidden value widget (used for auto-selection of last flyover)
   .flyover.value selection clear
   .flyover.value selection range 0 end
}

#------------------------------------------------------------------------
# flyover::set_text
#------------------------------------------------------------------------
proc flyover::set_text { msg val } {
   variable flywin

   # Set the global textvariables for the flyover window
   append msg $val
   set flywin(txt) $msg
   set flywin(value)  $val

   # Position window a little bit down and to right of cursor
   set x [expr [ winfo pointerx . ] + 2 ]
   set y [expr [ winfo pointery . ] + 8 ]
   wm geometry  .flyover "+$x+$y"
   wm deiconify .flyover
   set flywin(text_up) 1
}

#------------------------------------------------------------------------
# flyover::erase_text
#------------------------------------------------------------------------
proc flyover::erase_text {} {
   variable flywin

   wm withdraw .flyover
   set flywin(text_up) 0
}

#------------------------------------------------------------------------
# flyover::scantext
#------------------------------------------------------------------------
proc flyover::scantext { win startidx } {
   variable clients
   variable flywin

   # See if this scan was done before
   if { $flywin(last_scan) == $startidx } then {
      return
   }
   set flywin(last_scan) $startidx
                  
   # call client proc to do the scan. use eval since client cmd may already include args
   set newtxt [eval $clients(scancmd,$win) $win $startidx ]

  # client can set text directly using set_text or return text as a 2 element list 
   if { $newtxt != {} } then {
       set txt [lindex $newtxt 0]
       set val [lindex $newtxt 1]
       set_text $txt $val
   }
}


#------------------------------------------------------------------------
# flyover::refresh
#------------------------------------------------------------------------
proc flyover::refresh {} {

   variable flywin
   variable mpos
   variable clients
   variable refresh_interval

   if {$mpos(win) != "." && 
       [winfo exists $mpos(win)] } {
      set startidx [$mpos(win) index "@$mpos(x),$mpos(y) wordstart" ]

      if { $startidx == $mpos(last_idx) } then {
         if { !$flywin(text_up) } then {
            scantext $mpos(win) $startidx
         }
      } else {
         if { $flywin(text_up) } then {
            flyover::erase_text
         }
         set mpos(last_idx) $startidx
      }

   }
   if {$clients(cnt) > 0} { 
       after $refresh_interval "flyover::refresh"
   } else {
       if { $flywin(text_up) } then {
          flyover::erase_text
       }
   }
}
