#------------------------------------------------------------------------
#
#       IBM Software
#       Licensed Material - Property of IBM
#       (C) Copyright International Business Machines Corp. 2004
#       All Rights Reserved.
#
#       U.S. Government Users Restricted Rights - Use, duplication or
#       disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  fpregs.tcl
#
#------------------------------------------------------------------------

#------------------------------------------------------------------------
#
# FPRegs_Init -- Initalize the FP registers window
#
#  |
#  |FPR 0 -3.14            0xXXXXXXXXXXXXXXXX     12 1.234567E9      0xXXXXXXXXXXXXXXXX     24 2.71828E-7      0xXXXXXXXXXXXXXXXX
#  |    1 2.58003          0xXXXXXXXXXXXXXXXX     13 1.701           0xXXXXXXXXXXXXXXXX     25 9.876543        0xXXXXXXXXXXXXXXXX
#  |      ...
#  |   11 3.6              0xXXXXXXXXXXXXXXXX     23 0               0xXXXXXXXXXXXXXXXX     35 8.7732          0xXXXXXXXXXXXXXXXX
#  |
#
#------------------------------------------------------------------------
proc FPRegs_Init {root _upstate args} {
    global FPRegs

    upvar $_upstate  initstate
    TclUtils::copyarray initstate FPRegs $root      ;# create FPRegs($root,x)
    set cpucmd $FPRegs($root,cpucmd)
    set nfpr [$cpucmd display nfpr]

    # Display FP regs in a set of columns.
    # Compute number of rows/columns based on the number of FPRs.
    # Register numbers run down the columns; last column may not be full.
    # (For now, hard-code the number of columns at 3.)
    set cols 3
    set rows [expr ($nfpr + $cols - 1) / $cols]

    # Tuck these data away so that Refresh can use them
    set FPRegs($root,nfpr) $nfpr
    set FPRegs($root,rows) $rows
    set FPRegs($root,cols) $cols

    # Text widget with FP register numbers and values.
    # We use a read-only text widget so we can control the colors of various parts.
    set t [text $root.fprvals -width [expr (42 * $cols) + 2] -height [expr $rows + 2] -font listbox_font -wrap none]

    # Define two tags -- one for changed reg contents, one for contents that haven't changed
    $t tag configure changed -background yellow
    $t tag configure same    -background [$t cget -background]

    table $root $t 0,0 -fill y
    table configure $root c0 -resize none

    set title [treesim::get_path_name $FPRegs($root,widget) $FPRegs($root,treeidx) 1 ]
    wm title $root "$title: FPRs"

    FPRegs_Refresh $root 1

    return "FPRegs_Refresh"
}

######################################################################
#
# FPRegs_Refresh -- Refresh the contents of the status window
#
######################################################################
proc FPRegs_Refresh { root force_refresh args } {
    global FPRegs FPRegsVals

    set nfpr $FPRegs($root,nfpr)
    set cols $FPRegs($root,cols)
    set rows $FPRegs($root,rows)
    set cpucmd $FPRegs($root,cpucmd)

    set ic [$cpucmd display instruction_count]

    set first_time 0
    if { [array names FPRegsVals $root,0 ] == "" } {
        set first_time 1
    }

    if {!$first_time &&
        !$force_refresh &&
         [string compare $FPRegs($root,ic)  $ic] == 0} {
        return
    }
    set FPRegs($root,ic) $ic

    # Fetch new values for the registers
    for { set i 0 } { $i < $nfpr } { incr i } {
        set v [$cpucmd display fpr $i]
        set fpv [$cpucmd display fpr_as_fp $i]
        set newvals($i) [format "%s %s" $fpv $v]
        if {$first_time} {
            set FPRegsVals($root,$i) $newvals($i)
        }
    }

    # Allow modifications to the text widget
    set t "$root.fprvals"
    $t configure -state normal

    # Delete all the existing text
    $t delete 1.0 end

    # Skip a line to leave some vertical space
    $t insert end "\n"

    # Insert reg numbers and contents by row
    # Tag changed reg values so they'll display with a different background color
    for {set row 0} {$row < $rows} {incr row} {
        for {set col 0} {$col < $cols} {incr col} {
            set regnum [expr ($col * $rows) + $row]
            if {$regnum < $nfpr} {
                if {$regnum == 0} {
                    $t insert end "FPR 0 " {same}
                } else {
                    $t insert end [format "%5d " $regnum] {same}
                }
                if {[string compare $FPRegsVals($root,$regnum) $newvals($regnum)] != 0} {
                    $t insert end $newvals($regnum) {changed}
                    set FPRegsVals($root,$regnum) $newvals($regnum)
                } else {
                    $t insert end $newvals($regnum) {same}
                }
                $t insert end " " {same}
            }
        }
        $t insert end "\n"
    }

    # Disallow changes again
    $t configure -state disabled

}
