#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2004
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  pctracker.tcl
#------------------------------------------------------------------------

namespace eval pctrack {
   variable flyover
   set flyover(initialized) 0
}

#------------------------------------------------------------------------
#
# PCTrack_Init -- Initalize the status window
#
#------------------------------------------------------------------------
proc PCTrack_Init {root  _upstate args} {
    global PCTrackSim PCTrackCPU PCTrackStart PCTrackEnd pctrkmenutitle
    global PCTrackUserLength PCTrackUserStart PCTrackLength PCTrackObjectFile
    global PCTrack PCTrackFirstTime

    upvar $_upstate  initstate
    TclUtils::copyarray initstate PCTrack $root      ;# create PCTrack($root,x)
    set sim $PCTrack($root,sim)
    set cpu $PCTrack($root,cpunum)
    set cpucmd $PCTrack($root,cpucmd)
    set mcmcmd $PCTrack($root,mcmcmd)

    set PCTrackSim($root) $sim
    set PCTrackCPU($root) $cpu
    set pc [$cpucmd display spr pc]
    set PCTrackStart($root)     0
    set PCTrackUserStart($root) 0
    set PCTrackLength($root)     64
    set PCTrackUserLength($root) 64
    set PCTrackEnd($root) 0
    set PCTrackObjectFile($root) ""
    set PCTrackFirstTime($root) 1

    # starting address
    frame $root.start -borderwidth 2 -relief sunken
    label $root.start.label -text "Start Addr:" -pady 4
    entry $root.start.addr -textvariable PCTrackUserStart($root)
    pack $root.start.label $root.start.addr -side left
    bind $root.start.addr <KeyPress-Return> "PCTrack_Refresh $root 1"

    # length and format
    frame $root.leng -borderwidth 2 -relief sunken
    label $root.leng.label -text "Length:" -padx 4
    entry $root.leng.addr -textvariable PCTrackUserLength($root) -width 8 -justify right
    bind $root.leng.addr <KeyPress-Return> "PCTrack_Refresh $root 1"
    pack $root.leng.label $root.leng.addr -side left

    # object file association
    frame $root.objfile -borderwidth 2 -relief sunken
    label $root.objfile.label -text "Object File:" -pady 4
    entry $root.objfile.name -textvariable PCTrackObjectFile($root) -state disabled
    pack $root.objfile.label $root.objfile.name -side left

    frame $root.buttons -borderwidth 2 -relief sunken

    button $root.buttons.reload -text "Reload" -command "PCTrack_Refresh $root 1"
    button $root.buttons.close -text "Close" -command "destroy $root"
    button $root.buttons.step -text "Step" -command "$cpucmd step 1"
    button $root.buttons.go -text "Go" -command "$sim go"

    pack $root.buttons.reload $root.buttons.close $root.buttons.step $root.buttons.go -side left

    text $root.pctrk -width 70 -height 14 -yscrollcommand "$root.pctrksy set" -font textbox_font -tabs "12c" -wrap none
    scrollbar $root.pctrksy -orient vert -command "$root.pctrk yview"
    bind $root.pctrk <Double-ButtonRelease-1> "PCTrack_ToggleBreak $root $mcmcmd %W %y"

    textutils::init_color_tags $root.pctrk

    table $root \
        $root.pctrk 0,0 -cspan 3 -fill both \
        $root.pctrksy 0,3 -cspan 1 -fill y \
        $root.start    1,0 -fill both \
        $root.leng     1,1 -fill both \
        $root.objfile  1,2 -fill x    \
        $root.buttons   2,0 -cspan 3 -fill x    
    table configure $root c* r* -resize none
    table configure $root r0 c0 -resize expand

    set title [treesim::get_path_name $PCTrack($root,widget) $PCTrack($root,treeidx) 1 ]
    wm title $root "$title: PC Tracker"

    PCTrack_Refresh $root 1

    pctrack::flyover_init $root $cpucmd

    return "PCTrack_Refresh"
}

proc search_break_list { bplist addr } {
    foreach break_addr $bplist {
        if { [compare64 $break_addr $addr] == 0 } {
            return 1
        }
    }
    return 0
}

#------------------------------------------------------------------------
#
# PCTrack_Refresh -- Refresh the contents of the status window
#
#------------------------------------------------------------------------
proc PCTrack_Refresh { root force_reload args } {
    global PCTrackSim PCTrackCPU PCTrackStart PCTrackEnd
    global PCTrackUserLength PCTrackUserStart PCTrackLength
    global PCTrack PCTrackObjectFile PCTrackFirstTime

    set sim $PCTrackSim($root)
    set cpu $PCTrackCPU($root)
    set cpucmd $PCTrack($root,cpucmd)
    set mcmcmd $PCTrack($root,mcmcmd)
    set pc [$cpucmd display spr pc]


    # We ignore the force_reload argument since we test for the need
    set force_reload 0

    set user_force_reload 0

    if {$PCTrackUserLength($root) != $PCTrackLength($root)} {
       set PCTrackUserLength($root) [ TclUtils::get_larger_power2 $PCTrackUserLength($root)]
       set user_force_reload 1
    }

    if {[sub64 $PCTrackUserStart($root) $PCTrackStart($root)] != 0} {
       set user_force_reload 1
    }

    if {$user_force_reload} {
        set PCTrackLength($root) $PCTrackUserLength($root)
        set pc $PCTrackUserStart($root)
        set force_reload 1
    }

    set iMem "$root.pctrk"
    set msr [$cpucmd display spr msr]
    set mask [expr ~($PCTrackLength($root)-1)]

    # See if we need to reload the contents
    if { $PCTrackFirstTime($root) ||
         [ lessthan_u64 $pc $PCTrackStart($root)] ||
         [ lessthan_u64 $PCTrackEnd($root) $pc  ] ||
         $force_reload } {
        set PCTrackFirstTime($root) 0
        set PCTrackStart($root) [and64 $pc $mask]
        set PCTrackEnd($root) [add64 $PCTrackStart($root) [expr $PCTrackLength($root) - 1]]
        set PCTrackUserStart($root) [format64 "0x%016llX" $PCTrackStart($root)]

        set addr $PCTrackStart($root)

        # 
        # Try to find out if we have symbols loaded in one of the object files
        #
        set object_files [object match_address_to_files $addr]
        set PCTrackObjectFile($root) ""
        foreach file $object_files {
            # While the EA may have matched we want to 
            # check the instruction found in memory to that found
            # in the file.   If that matches, we will assume this is
            # the correct object file
            set matching 1
            while { ( $matching == 1 ) && [lessthan_u64 $addr $PCTrackEnd($root)] } {
                if {[catch {set rAddr [$cpucmd itranslate $addr]}] == 1} then {
                    set translate_error 1
                    set rAddr $addr
                } else {
                    set translate_error 0
                }

                if {$translate_error ||
                    [catch {
                        set value [$mcmcmd memory display $rAddr 4]
                        set disasm [$sim util ppc_disasm $value $addr]
                        }] == 1} then {
                    set value -1
                    set disasm "????????"
                    set matching 0
                } else {
                    #
                    # If the value found in memory and that found in the file don't match, this is the wrong file
                    #
                    if { [catch { set file_value [lindex [object inst_data $file $addr $addr] 1] }] == 1 } {
                        set matching 0
                    } else {
                        if { [compare64 $value $file_value] != 0 } {
                            set matching 0
                        }
                    }
                }
                incr64 addr 4
            }
            if { $matching == 1 } {
                set PCTrackObjectFile($root) $file
            }
        }

        set addr $PCTrackStart($root)
        set ea_bplist [$mcmcmd trigger display pc list]
        set ra_bplist [$mcmcmd trigger display real_pc list]

        $iMem delete 1.0 end
        while { [lessthan_u64 $addr $PCTrackEnd($root)] } {
            if {[catch {set rAddr [$cpucmd itranslate $addr]}] == 1} then {
                set translate_error 1
                set rAddr $addr
            } else {
                set translate_error 0
            }
            set addr_label ""

            if {$translate_error ||
                [catch {
                    set value [$mcmcmd memory display $rAddr 4]
                    set disasm [$sim util ppc_disasm $value $addr]
                    }] == 1} then {
                set value -1
                set disasm "????????"
            } else {
                if { [llength $PCTrackObjectFile($root)] > 0 } {
                    set linfo [object label $PCTrackObjectFile($root) $addr]
                    if { [llength $linfo] > 0 } {
                        array set label_info $linfo
                        set addr_label $label_info(LABEL)
                    }
                }
            }

            set ascii [TclUtils::GenAscii $value]
            set addr64 [format64 "%016llX" $addr]
            if { ( [search_break_list $ea_bplist 0x$addr64] == 1 ) || ( [search_break_list $ra_bplist $rAddr] == 1 ) } {
                $iMem insert end "B " breakpoint
            } else {
                $iMem insert end "  "
            }

            # Add the label
            if { [llength $PCTrackObjectFile($root)] > 0 } {
                $iMem insert end [format "%30s " $addr_label]
            }

            set line [format "%s : %s : %08X : %s : %s\t\t\n" $addr64 $rAddr $value $ascii $disasm]
            $iMem insert end $line
            incr64 addr 4
        }
        $iMem delete insert-1c end       ;# delete last newline char
    } else {
        textutils::untag_pcline $iMem $PCTrack($root,last_pccursor)
    }

    # Now highlight the current pc
    set pc_index [sub64 $pc $PCTrackStart($root)]
    set pc_index [expr ($pc_index/4) + 1]
    textutils::tag_pcline $iMem $pc_index.0
    $iMem see $pc_index.0
    set PCTrack($root,last_pccursor) $pc_index.0
}

#------------------------------------------------------------------------
# This is called to set/remove breakpoints.
# bound to double-click event
#------------------------------------------------------------------------
proc PCTrack_ToggleBreak {root mcmcmd txtwidget y} {
    global PCTrackStart
    set cursor [$txtwidget index @0,$y]
    set line [ lindex [ split $cursor . ] 0 ]
    set addr [add64 $PCTrackStart($root) [expr ($line-1) * 4]]
    set hexaddr [format64 "0x%016llX" $addr]
    if { [string length [$mcmcmd trigger display pc addr $hexaddr]] > 0 } {
        puts "Clearing breakpoint at $hexaddr"
        $mcmcmd trigger clear pc $hexaddr
        $mcmcmd trigger clear real_pc $hexaddr
        textutils::untag_breakpoint $txtwidget $cursor
    } else {
        puts "Setting breakpoint at $hexaddr"
        $mcmcmd trigger set pc $hexaddr "break_point $mcmcmd EA_Break_at_$hexaddr"
        $mcmcmd trigger set real_pc $hexaddr "break_point $mcmcmd RA_Break_at_$hexaddr"
        textutils::tag_breakpoint $txtwidget $cursor
    }
}

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
proc pctrack::flyover_init {root cpucmd} {
   variable flyover

   if { !$flyover(initialized) } {

      set flyover(initialized) 1

      # Setup the tracking variables array (gprs and fprs)
      for { set i 0 } { $i <= 31 } { incr i } {
         set flyover(r$i) "display gpr $i"
         set flyover(f$i) "display fpr $i"
         set flyover(v$i) "display vmxr $i"
         #set flyover(R[ format %02d $i]) "display gpr $i"
         #set flyover(F[ format %02d $i]) "display fpr $i"
      }

      set sprlist [$cpucmd display spr list]

      foreach spr $sprlist {
         set flyover($spr) "display spr $spr"
         #set flyover([ string toupper $spr ]) "display spr $spr"
      }
   }

   flyover::register $root.pctrk "pctrack::flyover_handler $root $cpucmd"
}

#------------------------------------------------------------------------
#
#------------------------------------------------------------------------
proc pctrack::flyover_handler { root cpucmd txtwin startidx } {
   global PCTrack
   variable flyover

   set rtn [list]

   set chr [ $txtwin get $startidx ]
   switch -glob -- $chr {
      [A-Za-z]  {
                   # id/keyword
                   set txt [ $txtwin get $startidx "$startidx wordend" ]

                   # Now we have the target word under the mouse, see if it in our list of
                   # trackable variables
                   set cmd [ lindex [ array get flyover $txt ] 1 ]
                   if { $cmd != "" } then {
                      set txt "$txt = "
                      set val [eval $cpucmd $cmd]
                      set rtn [list $txt $val]
                   }
                }
      [0-9]     {
                    #
                    #  If over a number, look for $+number or number(reg)
                    #
                    set prevchar [$txtwin get "$startidx-2chars" ]
                    if {$prevchar == "$"} {
                        set startidx [$txtwin index "$startidx-2chars" ]
                        set rtn [ textutils::branch_target $txtwin $startidx ]
                    } else {
                        set endidx [ $txtwin index "$startidx wordend" ]
                        set nextchar [$txtwin get "$endidx" ]
                        set mcmcmd $PCTrack($root,mcmcmd)
                        if {$nextchar == "("} {
                            set rtn [ pctrack::memory_address $mcmcmd $cpucmd $txtwin $startidx $endidx]
                        }
                    }
                }
      "$"       {
                    # $-num, $+num
                    set rtn [ textutils::branch_target $txtwin $startidx ]
                }
      "-"       -
      "+"       {
                    # $-num, $+num
                    set prevchar [$txtwin get "$startidx-1chars" ]
                    if {$prevchar == "$"} {
                        set startidx [$txtwin index "$startidx-1chars" ]
                        set rtn [ textutils::branch_target $txtwin $startidx ]
                    }
                }

   }

   return $rtn
}

#-----------------------------------------------------------------------------
# text at startidx is a number
# calculate num(reg)
#-----------------------------------------------------------------------------
proc pctrack::memory_address { mcmcmd cpucmd txtwin startidx numendidx } {

   set regidx [$txtwin index "$numendidx+2chars" ]
   set regendidx [$txtwin index "$regidx wordend" ]
   set reg    [$txtwin get $regidx $regendidx ]
   set sign   [$txtwin get "$startidx-1chars" ]
   if {$sign == "-"} {
       set startidx [$txtwin index "$startidx-1chars" ]
   }
   set offset [$txtwin get $startidx $numendidx ]
   set txt    [$txtwin get $startidx "$regendidx+1chars" ]
   set addrs  [ get_indirect_addresses $cpucmd $offset $reg ]
   set effaddr  [format64 "%llX" [lindex $addrs 0]]
   set realaddr  [lindex $addrs 1]
   set data "?"
   if {$realaddr != "?"} {
       catch {
           set data  [string trim [ $mcmcmd memory display $realaddr 4 4 ] ]
       }
   }

   return [list "$txt=>\n" "$effaddr: $realaddr: $data" ]
}


#------------------------------------------------------------------
#
#  input is a sting of the form offset(Rn)
#  offset can be of the form 0x1234 or 1234 or -1234
#
#------------------------------------------------------------------
proc get_indirect_addr_text {cpucmd addrstring} {

   set addrs {}
   set ok [regexp -nocase "(\[-\[:alnum:]]+)\[(]R(\[\[:digit:]]+)\[)]"  $addr  Match offset regnum]
   if {$ok} {
       set addrs [ get_indirect_addr $cpucmd $offset $regnum ]
   }
   return $addrs
}

#------------------------------------------------------------------
#
#------------------------------------------------------------------
proc get_indirect_addresses {cpucmd offset regnum } {

   if {$regnum == 0} {
       set regval 0
   } else {
       set regval [$cpucmd display gpr $regnum]
   }
   set effaddr [add64 $regval $offset]

   if {[catch {
         set realaddr [$cpucmd dtranslate $effaddr]
      } ] == 1 } then {
      set realaddr "?"
   }

   return "$effaddr $realaddr"
}
