#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  prefmgr.tcl
#
#
#------------------------------------------------------------------------

namespace eval PrefManager {
   variable version        "v1.0"
   variable refnames
   variable pref_filename  ""
}

#------------------------------------------------------------------------
#  PrefManager::set_value
#     Sets the value of the target preference. If the refname
#     is not found in the PrefManager's current list, it is added.
#
#  Parameters:
#     refname  - the preference's internal reference name
#     value    - the value to set
#------------------------------------------------------------------------
proc PrefManager::set_value { refname value } {

   variable current_values

   if { ![ _is_registered $refname ]  } then {
      _register $refname
   }
   set current_values($refname) "$value"
}

#------------------------------------------------------------------------
#  PrefManager::get_value
#     Returns the current value of the target preference. If the refname
#     is not found in the PrefManager's current list, just return the
#     default value, but do not add the refname to the list. This will
#     keep only those settings that are different from the defaults.
#
#  Parameters:
#     refname       - the preference's internal reference name
#     default_value - the default value to use if the refname is not found
#------------------------------------------------------------------------
proc PrefManager::get_value { refname default_value } {

   variable current_values

   if { [ _is_registered $refname ]  } then {
      return $current_values($refname)
   } else {
      return $default_value
   }
}

#------------------------------------------------------------------------
#  PrefManager::load_file
#
#  Parameters:
#     pref_filename - the name of the prefs file to source in
#------------------------------------------------------------------------
proc PrefManager::load_file { { _fname "" } } {

   variable pref_filename
   variable current_values

   if { $_fname != "" } then {
      set pref_filename $_fname
   }

   if { $pref_filename == "" } then {
      return
   }

   if { [ file exists $pref_filename ] } then {
      if { [ catch { source $pref_filename } errmsg ] } then {
         puts stderr "$errmsg"
      }
   } else {
      #puts stderr "File not found: $pref_filename"
   }
}

#------------------------------------------------------------------------
#  PrefManager::save_file
#     Returns 1 if the save was successful, 0 otherwise.
#
#  Parameters:
#     pref_filename - the name of the prefs file to save
#------------------------------------------------------------------------
proc PrefManager::save_file { { _fname "" } } {

    variable refnames
    variable pref_filename
    variable current_values

    #puts "Saving file $pref_filename"

    if { $_fname != "" } then {
      set pref_filename $_fname
    }

    if { $pref_filename == "" } then {
      return
    }

    # See if the directory exists. If not, try and create it.
    if { ![ file exists [ file dirname $pref_filename ] ] } then {
        if { [ catch { file mkdir [ file dirname $pref_filename ] } ] } then {
            puts stderr "$::errorInfo"
            return 0
        }
    }

    # See if we can write to the current pref filename
    if { [ catch { set fhandle [ open $pref_filename w+ ] } ] } then {
      puts stderr "$::errorInfo"
      return 0
    }

    puts $fhandle "#------------------------------------------------------------------"
    puts $fhandle "#    IBM Software                                                    "
    puts $fhandle "#    Licensed Material - Property of	IBM                            "
    puts $fhandle "#    (C) Copyright International Business Machines Corp. 2003        "
    puts $fhandle "#    All Rights Reserved.                                            "
    puts $fhandle "#                                                                  "
    puts $fhandle "#    U.S. Government	Users Restricted Rights - Use, duplication or "
    puts $fhandle "#    disclosure restricted by GSA ADP Schedule Contract with         "
    puts $fhandle "#    IBM Corp.                                                  "
    puts $fhandle "#                                                                  "
    puts $fhandle "#------------------------------------------------------------------"
    puts $fhandle "#  Filename: $pref_filename"
    puts $fhandle "#   Created: [ clock format [ clock seconds ] ]"
    puts $fhandle "#     This is a automatically generated preference file.           "
    puts $fhandle "#     Do not modify.                                               "
    puts $fhandle "#------------------------------------------------------------------"

    foreach refname [lsort [array names refnames]] {
      puts $fhandle "PrefManager::set_value [list $refname] \"$current_values($refname)\""
    }

    close $fhandle

    return 1
}

#------------------------------------------------------------------------
#  PrefManager::names
#     Returns a list containing the names of all of the elements in the array that match pattern 
#
#  Parameters:
#     pattern - a search pattern for use with array names
#------------------------------------------------------------------------
proc PrefManager::names { { pattern "_no_pattern_" } } {

   variable refnames

   if { $pattern == "_no_pattern_" } then {
      return [array names refnames]
   }

   return [array names refnames $pattern ]
}

#------------------------------------------------------------------------
#  PrefManager::clear
#     Clears all current_values and refnames
#------------------------------------------------------------------------
proc PrefManager::clear { } {

   variable refnames
   variable current_values

   array unset current_values
   array unset refnames 

}

#------------------------------------------------------------------------
#  PrefManager::_register
#     Adds a refname to the list of current refnames
#
#  Parameters:
#     refname - the preference's internal reference name
#------------------------------------------------------------------------
proc PrefManager::_register { refname } {

   variable refnames

   if { ![ _is_registered $refname ]  } then {
      set refnames($refname) 1
   }
}


#------------------------------------------------------------------------
#  PrefManager::_is_registered
#     Return 1 if the refname is registered, 0 otherwise
#
#  Parameters:
#     refname - the preference's internal reference name
#------------------------------------------------------------------------
proc PrefManager::_is_registered { refname } {
   variable refnames
   return [ info exists refnames($refname) ]
}

#------------------------------------------------------------------------
#  PrefManager::bootstrap
#     This function is called immediately
#      script, npx.bs2.  It loads nptk.users to determine which
#      user specific preference file to load.
#------------------------------------------------------------------------
proc PrefManager::bootstrap { } {

   global env tcl_platform
   variable pref_filename

   # Try to determine the operating system ( windows, aix, sunos or linux )
   regexp {[a-z]*} [ string tolower $tcl_platform(os) ] thisos

   switch $thisos {
      linux   { set pref_filename "$env(HOME)/.mambo/.prefs.tcl" }
      windows { set pref_filename "$env(APPDATA)/mambo/prefs.tcl" }
      default { return }
   }

   clear
   load_file
}


#------------------------------------------------------------------------
# set window geometry to nearest 8 pixel grid point         
# to keep windows from growing/moving a few pixels on each restart 
#------------------------------------------------------------------------
proc adjust_win_geom {geom} {

    # geom looks like "WxH+X+Y" ; convert x and + to spaces and use foreach to set 
    foreach {W H X Y} [ string map { x " " + " " } $geom ] break;   ;# geom -> W,H,X,Y

    set W [expr (($W + 7)/8)*8 ]
    set H [expr (($H + 7)/8)*8 ]
    set X [expr ($X/8)*8 ]    ;# round down on position 
    set Y [expr ($Y/8)*8 ]

    #puts "geom $geom -> ${W}x$H+$X+$Y"
    return ${W}x$H+$X+$Y
}


#------------------------------------------------------------------------
#  get_pref
#     Simple wrapper for PrefManager::get_value
#------------------------------------------------------------------------
proc get_pref { refname default_value } {
   return [ PrefManager::get_value $refname $default_value ]
}

#------------------------------------------------------------------------
#  set_pref
#     Simple wrapper for PrefManager::set_value
#------------------------------------------------------------------------
proc set_pref { refname value } {
   if {[string range $refname 0 7] == "win_geom"} { 
      set value [adjust_win_geom $value]
   }
   return [ PrefManager::set_value $refname $value ]
}

#------------------------------------------------------------------------
#  pref_names
#     Simple wrapper for PrefManager::names
#------------------------------------------------------------------------
proc pref_names { pattern } {
   return [ PrefManager::names $pattern ]
}
