#!/usr/bin/bltwish
#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2003,2004
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  process_tree.tcl
#------------------------------------------------------------------------
proc Compute_ProcessTree_Tag {machine} {
    return "processtree_${machine}"
}

package require BLT
namespace import -force blt::*

proc add_stat_info_to_tree { prefix tree stat_string } {

    if { [string length $stat_string] < 1 } {
        return
    }

    eval "array set stat_info $stat_string"

    set stat_root $prefix

    foreach name [array names stat_info] {
        $tree insert end "${stat_root}^$name=$stat_info($name)"
    }
}
proc add_file_info_to_tree { machine prefix tree file_string pid chart } {

    if { [string length $file_string] < 1 } {
        return
    }

    eval "array set file_info $file_string"

    set ea_end [format64 "0x%016llX" [add64 $file_info(ea) $file_info(length)]]
    set file_root "${prefix}^Filename:$file_info(filename) EA:$file_info(ea) to EA:$ea_end"
    $tree insert end $file_root -opencommand "UpdateFileGraph $machine $pid $file_info(file_index) $chart"

    #puts "$prefix : NEW FILE _________________________"

    foreach name [array names file_info] {
        #puts "$prefix $name=$file_info($name)"
        $tree insert end "${file_root}^$name=$file_info($name)"
    }
}
proc add_process_info_to_tree { machine prefix tree process_string chart } {

    global GlobalProcessInfo
    global GlobalProcessFileInfo

    if { [string length $process_string] < 1 } {
        return
    }

    eval "array set pt_info $process_string"

    #puts "$prefix : NEW PROCESS :::::::::::::::::::::"
    set process_root "${prefix}^$pt_info(pid) : $pt_info(execname) ($pt_info(state))"
    set GlobalProcessInfo($pt_info(pid)) $process_string
    set pid_idx [$tree insert end $process_root]
    $tree open $pid_idx

    foreach name [array names pt_info] {
        if { ([string compare $name "pid"] != 0) &&
             ([string compare $name "ppid"] != 0) &&
             ([string compare $name "stats"] != 0) &&
             ([string compare $name "state"] != 0) &&
             ([string compare $name "execname"] != 0) } {
            if { [string compare $name "children"] == 0 } {
                set children $pt_info(children)
            } else {
                if { [string compare $name "files"] == 0 } {
                    set files $pt_info(files)
                } else {
                    #puts "$prefix $name=$pt_info($name)"
                    $tree insert end "${process_root}^$name=$pt_info($name)"
                }
            }
        }
    }
    eval "set GlobalProcessFileInfo($pt_info(pid)) [list $files]"
    #puts "$prefix FILES:"

    set stats_root "${process_root}^STATS"
    add_stat_info_to_tree $stats_root $tree [list $pt_info(stats)]

    set file_root "${process_root}^FILES"
    foreach file $GlobalProcessFileInfo($pt_info(pid)) {
        add_file_info_to_tree $machine $file_root $tree [list $file] $pt_info(pid) $chart
    }

    eval "set child_list [list $children]"
    #set child_root "${process_root}^CHILDREN"
    set child_root "${process_root}"
    foreach child $child_list {
        add_process_info_to_tree $machine $child_root $tree [list $child] $chart
    }
}

proc UpdateFileGraph { machine pid file_index base_chart_name } {

    global FileGraphInfo

    array set FileGraphInfo [$machine processtree display tcl_stats proc_file_info $pid $file_index]

    set inst_chart ${base_chart_name}_inst
    set cache_chart ${base_chart_name}_cache

    global FileGraphTag
    set FileGraphTag "${machine}_${pid}_${file_index}_filedata"

    # Create the Vectors
    foreach column $FileGraphInfo(columns) {
        set vecname V_${FileGraphTag}_${column}
        eval "global $vecname"

        # If the vector already exists, destroy it
        if { [llength [vector names ::${vecname}]] > 0 } {
            eval "vector destroy ${vecname}"
        }
        eval "vector create ${vecname}"
        eval "${vecname} set [list $FileGraphInfo($column)]"
    }

    # First delete all existing elements
    foreach CHART [list $inst_chart $cache_chart] {
        if { [llength [$CHART element names]] > 0 } {

            foreach el [$CHART element names] {
                $CHART element delete $el
            }
        }
    }

    foreach column $FileGraphInfo(columns) {
        if { ([string compare $column "ADDR"] != 0) &&
             ([string compare $column "INST"] != 0) &&
             ([string compare $column "INST_COUNT"] != 0) } {

            set xvals V_${FileGraphTag}_ADDR
            set yvals V_${FileGraphTag}_${column}

            if { [string first "CACHE" $column] == -1 } {
                # Instruction Graph
                if { [llength [$inst_chart pen names $column]] == 0 } {
                    $inst_chart pen create $column -foreground [GraphColor ${FileGraphTag}] -relief flat
                }
                if { [string first "ICOUNT" $column] != -1 } {
                    $inst_chart element create $column -mapy y2 -xdata $xvals -ydata $yvals -pen $column
                } else {
                    $inst_chart element create $column -xdata $xvals -ydata $yvals -pen $column
                }
            } else {
                # Cache Graph
                if { [llength [$cache_chart pen names $column]] == 0 } {
                    $cache_chart pen create $column -foreground [GraphColor ${FileGraphTag}] -relief flat
                }
                $cache_chart element create $column -xdata $xvals -ydata $yvals -pen $column
            }
        }
    }
}

###########################################################
# read_process_tree_data
#
#   This procedure actually queries the underlying
#   tcl interpreter to find out the state of the process tree
#
proc read_process_tree_data { machine tree chart } {
    global CurrentPID

    set tag [Compute_ProcessTree_Tag ${machine}]

    array set pt_info [$machine processtree display all]

    set CurrentPID($tag) $pt_info(current_pid)

    add_process_info_to_tree $machine "" $tree $pt_info(process_tree) $chart
}


######################################################################
#
# Machine_ProcessTreeInit -- Initalize the process tree structure
#
proc Machine_ProcessTreeInit { machine mywin } {
    global RefreshDelay

    set tag [Compute_ProcessTree_Tag ${machine}]

    if { [llength $mywin] == 0 } {
        set winname ".${tag}"

        #
        # If the window exists, just call the refresh call and return
        #
        if { [winfo exists $winname] } {
            return
        }

        #
        # Otherwise, we need to create a window for this machine's status info
        #
        toplevel $winname
    } else {
        set winname $mywin
    }

    set tree ${winname}.processtree
    hierbox $tree -separator "^" -autocreate true -scrollmode listbox -xscrollcommand "$winname.scrollx set" -yscrollcommand "$winname.scrolly set" -width 400
    scrollbar $winname.scrollx -orient horizontal -command "$tree xview"
    scrollbar $winname.scrolly -orient vertical -command "$tree yview"

    set base_chart_name ${winname}.file_graph
    set inst_chart ${base_chart_name}_inst
    set cache_chart ${base_chart_name}_cache

    read_process_tree_data $machine $tree $base_chart_name

    scrollbar ${winname}.inst_scrollx -orient horizontal -command "${inst_chart} axis view x"
    scrollbar ${winname}.inst_scrolly -orient vertical -command "${inst_chart} axis view y"
    barchart ${inst_chart} -title "Statistics Across File: " -height 200 -shadow white
    ZoomSupport_Init ${inst_chart} 0 0
    ${inst_chart} axis configure x -title "EA Offset" -scrollcommand "${winname}.inst_scrollx set"
    ${inst_chart} axis configure y -title "Counts" -scrollcommand "${winname}.inst_scrolly set"
    ${inst_chart} axis configure y2 -title "Counts" -hide false
    ${inst_chart} legend configure -position bottom
    ${inst_chart} configure -barmode aligned

    # Don't mess with xticks for now...
    $inst_chart axis configure x -minorticks {0}
    $inst_chart axis configure x -majorticks {0}

    scrollbar ${winname}.cache_scrollx -orient horizontal -command "${cache_chart} axis view x"
    scrollbar ${winname}.cache_scrolly -orient vertical -command "${cache_chart} axis view y"
    barchart ${cache_chart} -title "Statistics Across File: " -height 200 -shadow white
    ZoomSupport_Init ${cache_chart} 0 0
    ${cache_chart} axis configure x -title "EA Offset" -scrollcommand "${winname}.cache_scrollx set"
    ${cache_chart} axis configure y -title "Counts" -scrollcommand "${winname}.cache_scrolly set"
    ${cache_chart} legend configure -position bottom
    ${cache_chart} configure -barmode stacked

    # Don't mess with xticks for now...
    $cache_chart axis configure x -minorticks {0}
    $cache_chart axis configure x -majorticks {0}

    set chart_list [list ${inst_chart} ${cache_chart}]
    global PT_Marker
    global Func_Marker
    global Func_Name_Marker
    foreach ch $chart_list {
        #${ch} configure -barmode stacked
        set PT_Marker(${ch}) [${ch} marker create text -coords {0 Inf} -text "0x0" -fill "" -outline red]
        set Func_Marker(${ch}) [${ch} marker create polygon -under true -coords {0 0 0 0 0 Inf 0 Inf} -fill yellow -outline black]
        set ymax [expr [string range [lindex [${ch} axis limits y] 1] 0 end-2] * 0.75]
        set Func_Name_Marker(${ch}) [${ch} marker create text -coords {0 $ymax} -text "" -fill "" -outline red]
        bind ${ch} <Shift-Motion>       "UpdateMotion_File  \%W \%x \%y \"$chart_list\" $winname $machine ${winname}.textboxes.assembly"
        ${ch} legend configure -position bottom
    }

    frame ${winname}.textboxes
    label ${winname}.textboxes.assembly_label -justify center -text "Assembly:"
    text ${winname}.textboxes.assembly -width 50 -height 15 -font textbox_font
    #label ${winname}.textboxes.helpbox_label -justify center -text "Helpful Hints:"
    #text ${winname}.textboxes.helpbox  -width 10 -height 15 -font textbox_font

    table ${winname}.textboxes \
        ${winname}.textboxes.assembly_label 0,0 -fill both \
        ${winname}.textboxes.assembly       1,0 -fill both
        #${winname}.textboxes.helpbox_label  0,1 -fill both
        #${winname}.textboxes.helpbox        1,1 -fill both

    table $winname \
        ${tree} 0,0 -rowspan 5 -fill both \
        ${winname}.scrolly 0,1 -rowspan 5 -fill y \
        ${inst_chart} 0,2 -fill both \
        ${winname}.inst_scrolly 0,3 -fill y \
        ${winname}.inst_scrollx 1,2 -fill x \
        ${cache_chart} 2,2 -fill both \
        ${winname}.cache_scrolly 2,3 -fill y \
        ${winname}.cache_scrollx 3,2 -fill x \
        ${winname}.scrollx 5,0 -fill x  \
        ${winname}.textboxes 4,2 -fill x

    table configure $winname c0 r0 r1 r2 r3 r4 r5 c1 c2 c3 -resize none
    table configure $winname r0 r2 c2 -resize both

    update
}

proc UpdateMotion_File {chart x y chart_list win machine assembly_contents} {

    set text ${assembly_contents}

    ${text} delete 0.0 end

    global element_list
    ${chart} element closest $x $y element_list

    # If the element_list contains an index field then update text with location information
    if { [lsearch [array names element_list] index] != -1 } {
        set x_index $element_list(index)

        global FileGraphTag
        global FileGraphInfo

        # Get the EA offset at this index
        set addrvec V_${FileGraphTag}_ADDR
        eval "global $addrvec"
        eval "set addrvecval \$${addrvec}($x_index)"
        set point_addr [format64 "%lld" [string range $addrvecval 0 end-2]]
        set assembly(addr) [format64 "0x%016llX" [add64 $FileGraphInfo(ea) $point_addr]]

        # Get the encoding for this instruction
        set instvec V_${FileGraphTag}_INST
        eval "global $instvec"
        eval "set instvecval \$${instvec}($x_index)"
        set assembly(encoding) [format64 "0x%08llX" [string range $instvecval 0 end-2]]

        set label ""
        set func_start $assembly(addr)
        set func_end $assembly(addr)
        set func_name "Unknown"

        set filename [object load $FileGraphInfo(filename)]
        if { [string length $filename] > 0 } {
            set labelinfo [object label $FileGraphInfo(filename) $assembly(addr)]
            #puts "Label query for ea:$assembly(addr) got: $labelinfo"
            if { [string first "LABEL" $labelinfo] != -1 } {
                array set asm_info $labelinfo
                set label $asm_info(LABEL)
                set func_start $asm_info(FUNC_START)
                set func_end $asm_info(FUNC_END)
                set func_name $asm_info(FUNC)
            }
        }
        if { [string length $label] > 0 } {
            # Iterate through the EA instruction range for this procedure
            # and update the text window.   If we don't have data for it in the
            # the Vector, pull it from the object file itself.   We should also verify the
            # encoding in our vector and that found in the object file, match.
        } else {
            # No function data so just show the one instruction
            ${text} insert end [format "%20s %18s %s %s\n" $label $assembly(addr) $assembly(encoding) [$machine util ppc_disasm $assembly(encoding) $assembly(addr)] ]
        }

        global PT_Marker
        global Func_Marker
        global Func_Name_Marker
        foreach ch $chart_list {
            set label [format64 "0x%llX" [add64 $FileGraphInfo(ea) $point_addr]]
            ${ch} marker configure $PT_Marker($ch) -coords {$point_addr Inf} -text $label
            set func_start_index [sub64 $func_start $FileGraphInfo(ea)]
            set func_end_index [sub64 $func_end $FileGraphInfo(ea)]
            ${ch} marker configure $Func_Marker($ch) -coords {$func_start_index 0 $func_start_index Inf $func_end_index Inf $func_end_index 0}
            set ymax [expr [string range [lindex [${ch} axis limits y] 1] 0 end-2] * 0.75]
            ${ch} marker configure $Func_Name_Marker($ch) -coords {$point_addr $ymax} -text $func_name
        }
    }
}
