#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2007
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  simmode.tcl
#------------------------------------------------------------------------

namespace eval SimMode {
    variable version        "1.0"
}

#------------------------------------------------------------------------
# Update the simulator mode text
# 
# \param    sim     Simulator name
#------------------------------------------------------------------------
proc SimMode::Refresh { sim } {

    set widget .simmode

    if { [ winfo exists $widget ] } {
        $widget.lbl config -text "[string trim [$sim display mode]]"
    }
}

#------------------------------------------------------------------------
# Switch simulator modes.
#
# \param    sim     Simulator name
# \param    mode    Mode to switch to
#------------------------------------------------------------------------
#
proc SimMode::SwitchMode { sim mode } {

    $sim mode $mode

    Controls::UpdateGUI
}

#------------------------------------------------------------------------
# Display the SimMode dialog (if iconified or obscured in the window stacking).
# This will also initialize the GUI if not already done.
#
# \param    sim     Simulator name
#------------------------------------------------------------------------
proc SimMode::Create { sim } {

    set widget .simmode

    if { ![ winfo exists $widget ] } {

        # Create a new toplevel window / frame 
        toplevel $widget

        # Mark new window as transient and working for .
        wm transient $widget .

        set geom [ get_pref win_geom/cpanel "800x600+80+80" ]

        label $widget.lbl -width 60 -text "[string trim [$sim display mode]]"
        pack $widget.lbl -fill x

        # Frame for grouping purposes
        frame $widget.buttons -relief flat -borderwidth 0

        # Mode switch buttons
        button $widget.buttons.fast   -text "Fast Mode" -command "SimMode::SwitchMode $sim fastest"
        grid $widget.buttons.fast -row 0 -column 0 -sticky ew

        button $widget.buttons.simple -text "Simple Mode" -command "SimMode::SwitchMode $sim simple"
        grid $widget.buttons.simple -row 0 -column 1 -sticky ew

        button $widget.buttons.cycle  -text "Cycle Mode" -command "SimMode::SwitchMode $sim cycle"
        grid $widget.buttons.cycle -row 0 -column 2 -sticky ew

        grid columnconfigure $widget.buttons "0 1 2" -uniform all -weight 1
        pack $widget.buttons -side bottom -fill x

        set top_widget [ winfo toplevel $widget ]

        set geom [ get_pref win_geom/simmode "480x60+400+300" ]
        wm geometry $top_widget $geom

        # Intercept window close and call the exit handler
        wm protocol $top_widget WM_DELETE_WINDOW "SimMode::_on_close $top_widget"
        bind $top_widget <Configure> "SimMode::_save_geometry $top_widget"
        bind $top_widget <Key-F5> "user_refresh"

        # Set up to be notified when the window manager destroys, shows, or hides
        # the toplevel window
        bind $widget <Destroy> [list SimMode::Destroy $sim]
        bind $widget <Map>     [list SimMode::Map $sim]
    }

    wm deiconify [winfo toplevel $widget]
    raise $widget
}

#------------------------------------------------------------------------
# SimMode::_save_geometry
#------------------------------------------------------------------------
proc SimMode::_save_geometry { top_widget } {

   set_pref win_geom/simmode [ winfo geometry $top_widget ]
}

#------------------------------------------------------------------------
# SimMode::_on_close
#------------------------------------------------------------------------
proc SimMode::_on_close { top_widget } {
   wm withdraw $top_widget
}

#------------------------------------------------------------------------
# Shut down the SimMode dialog. This is called when the dialog window is closed.
#
# \param    sim     Simulator name
#------------------------------------------------------------------------
proc SimMode::Destroy {sim} {
    bind .simmode <Destroy> {}
}

#------------------------------------------------------------------------
# Show the SimMode dialog. This is called when the dialog window is mapped.
#
# \param    sim     Simulator name
#------------------------------------------------------------------------
proc SimMode::Map {sim} {
    SimMode::Refresh $sim
}
