#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2003
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  tcl_utils.tcl
#
#  Common generic tcl and tk functions
#------------------------------------------------------------------------

namespace eval TclUtils {
   variable version        "1.0"
}

# --------------------------------------------------------------
# ScrollBoth and ScrollSet
#   These procedure are used to control two listboxes together.
# --------------------------------------------------------------
proc TclUtils::ScrollBoth { lb1 lb2 args } {
    eval ${lb1} yview $args
    eval ${lb2} yview $args
}

proc TclUtils::ScrollbarSet { sb lb1 lb2 x y } {
    $sb set $x $y
    $lb1 yview moveto $x
    $lb2 yview moveto $x
}

# --------------------------------------------------------------
# safename
#   Takes a string as an input and converts it to a string that
#   can be used as widget names.
# --------------------------------------------------------------
proc TclUtils::safename { in } {

    return [ string tolower [ string map { { } {_} {.} {_} {:} {_}} $in ] ]
}


# --------------------------------------------------------------
# array_flip
#   Takes an array and flips the keys and values, returning a new
#   array with the flipped values.  Keys are returned as values
#   and values as keys.  If a mode is specified, it has the following
#   effects:
#       unique - add an ever incrementing suffix to the key starting
#                at oldkey.0000. This not only makes each key unique,
#                but allows for easy sorting when two or more 'oldkeys'
#                are identical.
# --------------------------------------------------------------
proc TclUtils::array_flip { _in _out {mode none} } {

    upvar $_in  in
    upvar $_out out

    set unique [ expr [ string compare $mode "unique" ] ? 0 : 1 ]

    if { ![ array exists in ]  } then {
        error "TclUtils::array_flip was called with a value that is not an array"
        return
    }

    array unset out
    set usuffix 0
    foreach { key val } [ array get in ] {
        if { $unique } then {
            set newkey $val.[ format %04d $usuffix ]
            incr usuffix
            set out($newkey) $key
        } else {
            if { [ info exists out($val) ] } then {
                puts "Error: key value already exists: $val"
                error "TclUtils::array_flip was called with a non usable input array"
            }
            set out($val) $key
        }
    }
}


# ------------------------------------------------------------------
# return list of words in B not in A
# ------------------------------------------------------------------
proc TclUtils::addedwords {strA strB} {
   set added {}
   foreach word $strA {
       set wordsA($word) 1
   }
   foreach word $strB {
       if {![info exists wordsA($word)]} { 
          lappend added $word
       }
   }
   return $added;
}

# ------------------------------------------------------------------
# return list of words in A not in B
# ------------------------------------------------------------------
proc TclUtils::removedwords {strA strB} {
   set removed {}
   foreach word $strB {
       set wordsB($word) 1
   }
   foreach word $strA {
       if {![info exists wordsB($word)]} { 
          lappend removed $word
       }
   }
   return $removed;
}

# ------------------------------------------------------------------------
# print error 
# ------------------------------------------------------------------------
proc TclUtils::pe {}  {
   global errorInfo
   puts $errorInfo
}

# ------------------------------------------------------------------------
# This proc simply gets the contents of a file and returns
# it in a tcl variable
# ------------------------------------------------------------------------
proc TclUtils::get_file {filename} {
  set fd [open $filename "r"]
  set data [read $fd]
  close $fd
  return $data
}

# ------------------------------------------------------------------------
# write list to file
# ------------------------------------------------------------------------
proc TclUtils::put_file {filename output_lines} {
  set fd [open $filename "w"]
  foreach line $output_lines {
      puts $fd "$line"
  }
  close $fd
}

# ------------------------------------------------------------------------
# like parray for lists
# ------------------------------------------------------------------------
proc TclUtils::plist {lines} {
  foreach line $lines { puts "$line" }
}

# ------------------------------------------------------------------------
# Used with sort_tclIndex
# ------------------------------------------------------------------------
proc TclUtils::count_comment_lines {input_lines} {
    set num_comments 0
    foreach line $input_lines {
        if {[string index $line 0] == "#"} { 
            incr num_comments
        }  else {
            break;
        }
    }
    return $num_comments;
}

# ------------------------------------------------------------------------
# This sorts a tclIndex file.  
# ------------------------------------------------------------------------
proc TclUtils::sort_tclIndex {infname outfname} {
   set input_text [ get_file $infname ]
   set input_lines [split $input_text "\n"]
   set num_comments [count_comment_lines $input_lines ]

   set comments [lrange $input_lines 0 [expr $num_comments - 1]]
   set rest     [lsort [lrange $input_lines $num_comments end ] ]
   set output_lines [concat $comments $rest]

   put_file $outfname $output_lines
}

# ------------------------------------------------------------------
# copyarray - copy array with a single element name to a new
# array with an pair of element names ( test(x) -> test2(x,id) )
# ------------------------------------------------------------------
proc TclUtils::copyarray {_upinarray _upoutarray instance_idx } {
   upvar $_upinarray   inarray
   upvar $_upoutarray  outarray
   foreach name [array names inarray] {
       set outarray($instance_idx,$name) $inarray($name)
   }
}

# ------------------------------------------------------------------
# generate ascci representation for input 4 byte value
# ------------------------------------------------------------------
proc TclUtils::GenAscii { value } {
	set v0 [expr ( ($value >> 24) & 0xff )]
	set v1 [expr ( ($value >> 16) & 0xff )]
	set v2 [expr ( ($value >> 8) & 0xff )]
	set v3 [expr ( ($value >> 0) & 0xff )]
	if { ( $v0 < 0x20 ) || ( $v0 > 0x7e ) } { set v0 [expr 0x2a + 0] }
	if { ( $v1 < 0x20 ) || ( $v1 > 0x7e ) } { set v1 [expr 0x2a + 0] }
	if { ( $v2 < 0x20 ) || ( $v2 > 0x7e ) } { set v2 [expr 0x2a + 0] }
	if { ( $v3 < 0x20 ) || ( $v3 > 0x7e ) } { set v3 [expr 0x2a + 0] }

	return [format "%c%c%c%c" $v0 $v1 $v2 $v3]
}

# ------------------------------------------------------------------
# since the window len can be any number, get the next bigger power of 2 for buffer size
# ------------------------------------------------------------------
proc TclUtils::get_larger_power2 {winlen} {
    set bits_needed [expr log($winlen)/log(2) ]
    set int_bits [expr int($bits_needed)]
    if {$int_bits != $bits_needed} { 
       incr int_bits
    }
    return [expr (1<<$int_bits) ]
}

# --------------------------------------------------------------
# PlaceWindow
#   Takes two window paths and sets the x/y position of one relative
#   to the other. This can be used to popup a custom dialog box on top
#   of a 'parent' window.
# --------------------------------------------------------------
proc TclUtils::PlaceWindow { parent child {x 25} {y 25} } {

    set xp [ winfo x $parent ]
    set yp [ winfo y $parent ]
    wm geom $child "+[ expr $xp + $x]+[ expr $yp + $y ]"
}
