#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2004
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  treesim.tcl
#
#------------------------------------------------------------------------

package require BLT
namespace import -force blt::table

namespace eval treesim {
    variable siminfo
    variable cmds

    #  Note: The cmd procedures can be replaced by changing
    #  the associated variable in the treesim::cmds array.
    #  more treesim cmds can be added if needed.
    #  These are the only ones that are used as of now.
    set cmds(add_cpu_entries)  "treesim::add_cpu_entries"
    set cmds(add_mcm_multiple) "treesim::add_mcm_multiple"
    set cmds(refresh_cpu)      "treesim::refresh_cpu"

    set siminfo(reopened_windows) 0

    namespace export *
}

#------------------------------------------------------------------
#  return number of mcms in machine
#------------------------------------------------------------------
proc treesim::get_num_mcms {sim} {
    return [$sim display number_of_MCMs]
}

#------------------------------------------------------------------
#  return path of given tree index.  tree index should exist!
#  updirs will chop off names at the end
#------------------------------------------------------------------
proc treesim::get_path_name {widget treeidx {updirs 0 } } {

    set treename [$widget get -full 0]
    set start [expr [string length $treename] + 1 ]
    set path [$widget get -full $treeidx]

    while {$updirs > 0} {
       set path [file dirname $path]
       incr updirs -1
    }

    return [string range $path $start end ]
}

proc treesim::get_sim_default_entries {} {
}

proc treesim::is_shared_mode {sim} {
    set attach [$sim display mcm_attachment]
    set idx [string first "shared" $attach]
    return [expr $idx == 0]
}

proc treesim::get_machine_default_entries {sim} {

   set machine_entries [list]
   if {[is_shared_mode $sim]} {
      set machine_entries [concat $machine_entries [get_memory_default_entries]]
   }
   #puts "mach_entries=$machine_entries"
   return $machine_entries;
}

proc treesim::get_memory_default_entries {} {
    set mem_entries [ list "MemoryMap" managed "MemMap_Init"            \
                           "SystemMemory/sysmem_0" managed "SystemMemory_Init" ]

    # This will allow adding new system memory entries to the tree
    # the %%n will be replaced by the inserted position number
    lappend mem_entries "SystemMemory/New_Display" add2tree [list "sysmem_%%n" managed "SystemMemory_Init" ]

    #puts "mem_entries=$mem_entries"
  return $mem_entries
}

proc treesim::get_mcm_default_entries {sim} {
    set mcm_entries [ list "Load-Elf-App"  dialog   "SelectElfFile 1" \
                           "Load-Elf-Kernel" dialog "SelectElfFile 0" ]

    if {![is_shared_mode $sim]} {
       set mcm_entries [concat $mcm_entries [get_memory_default_entries]]
    }

    if { [ Config::is_defined "MCONFIG_STATS" ] } then {
        lappend mcm_entries  "Stats" managed "Stats_Init %P" \
                             "Pipeline" managed "Pipeline_Init"
    }
    #puts "mcm_entries=$mcm_entries"
  return $mcm_entries
}

proc treesim::get_ppc_default_entries {} {
    set ppc_entries [ list "PCTrack"  managed "PCTrack_Init" \
                           "PPCCore"  managed "PPCCore_Init" \
                           "PPCStack" managed "PPCStack_Init" \
                           "GPRegs"   managed "GPRegs_Init"  \
                           "FPRegs"   managed "FPRegs_Init" ]

    if { [ Config::is_defined "CONFIG_SLB" ] } then {
       lappend ppc_entries "PPC_Xlate" managed "PPC_xlate::Init"
    } else {
        if { [ Config::is_defined "CONFIG_PPC_64BIT" ] } {
            lappend ppc_entries "Addressing" managed "Embedded_xlate::Init"
        } else {
            lappend ppc_entries "Addressing" managed "PPC4xx_xlate::Init"
        }
    }
    return $ppc_entries
}


#------------------------------------------------------------------
# Create the tree with a root node called "Machines"
#
#------------------------------------------------------------------
proc treesim::init { win } {
    variable siminfo

    #puts "treesim init $win "

    blt::tree create myTree

    blt::treeview $win.tree -tree myTree -separator "/" -autocreate true -scrollmode listbox -yscrollcommand "$win.scroll set"
    scrollbar $win.scroll -command "$win.tree yview"

    $win.tree entry configure root -label "Machines"

    # create a hidden column for saving opened window names
    $win.tree column insert end "winname" -hide yes

    blt::table $win \
        $win.tree        0,0  -fill both -anchor w \
        $win.scroll     0,1  -fill y
    blt::table configure $win c1 -resize none

    # the treesim_cmds_init command is in both the default and machine specific areas
    # the tcl path will find the machine specific one first to override the default treesim::cmds array
    treesim_cmds_init

    set siminfo(added) ""
    treesim::refresh $win

    return $win.tree
}


#------------------------------------------------------------------
# Update the tree with the simulators
#
#------------------------------------------------------------------
proc treesim::refresh { win } {
    variable siminfo

    #puts "treesim refresh $win "
    set widget $win.tree
    set simlist [display machines]

    if {$simlist != $siminfo(added)} {
        sim_addremove $widget

        # after first sim has been added, reopen any windows in use
        if {!$siminfo(reopened_windows)} {
           set siminfo(reopened_windows) 1
           reopen_windows_on_startup $widget
        }
    }

    foreach sim $simlist {
        treesim::refresh_sim $widget $sim $siminfo($sim,idx)
    }
}

#------------------------------------------------------------------
# call open for each element of the input path
#------------------------------------------------------------------
proc treesim::open_window_path {widget myTree path} {
   set nodes [split $path "/" ]
   set idx 0
   foreach node $nodes {
       set idx [$myTree findchild $idx $node]
       if {$idx == -1} {
          break;
       }
       $widget open $idx
   }
}

#------------------------------------------------------------------
# open a list of windows given tree paths
#------------------------------------------------------------------
proc treesim::open_windows {widget winlist} {
    set myTree [$widget cget -tree]
    foreach path $winlist {
        open_window_path $widget $myTree $path
    }
    after 1000  { focus -force .; raise . }
}

#------------------------------------------------------------------
#
#------------------------------------------------------------------
proc treesim::reopen_windows_on_startup {widget} {
   # Restore latest open windows
   set reopen [ get_pref open_windows/reopen_on_startup 1 ]
   if {$reopen} {
       set winlist [ get_pref open_windows/latest "" ]
       treesim::open_windows $widget $winlist
   }
}

#------------------------------------------------------------------
# Add/remove simulators from tree
#------------------------------------------------------------------
proc treesim::sim_addremove { widget } {
    variable siminfo

    set simlist [display machines]
    #puts "simlist =$simlist treelist = $siminfo(added)"
    set sims_added    [TclUtils::addedwords    $siminfo(added) $simlist ]
    set sims_removed  [TclUtils::removedwords  $siminfo(added) $simlist ]

    set siminfo(added) $simlist

    # remove any deleted sims from the tree
    foreach sim $sims_removed {
        treesim::remove_child_windows $widget $siminfo($sim,idx)
        treesim::remove_sim $widget $sim $siminfo($sim,idx)
        unset siminfo($sim,idx)
    }

    # add any new sims to the tree
    foreach sim $sims_added {
        set simidx1 [treesim::add_sim $widget $sim]
        set siminfo($sim,idx) $simidx1
    }
}


#------------------------------------------------------------------
# when the fonts change, change it for all the tree text
#------------------------------------------------------------------
proc treesim::font_changed { win } {

    set widget $win.tree

    # Update the font type if it has changed
    foreach entry [ $widget find ] {
        $widget entry configure $entry -labelfont default_font
    }
}

#------------------------------------------------------------------
# remove deleted sim from tree
#------------------------------------------------------------------
proc treesim::remove_sim {widget sim simidx} {
    #puts "removing sim $sim"
    $widget delete $simidx
}

#------------------------------------------------------------------
#  Add new simulator to tree
#------------------------------------------------------------------
proc treesim::add_sim {widget sim} {

    #puts "adding sim $sim"

    set simindex3 [$widget insert end /$sim]
    $widget tag add "SIM" $simindex3
    $widget tag add $sim $simindex3

    # remember sim data
    set dataname [initdata $widget $simindex3 ]
    adddata $dataname sim $sim name $sim treeidx $simindex3

    treesim::add_sim_entries $widget $simindex3

    # if machine exists, add it
    treesim::add_machine $widget $sim $simindex3

    return $simindex3
}

#------------------------------------------------------------------
#  Add new simulator machine to tree
#------------------------------------------------------------------
proc treesim::add_machine {widget sim simidx5} {

    #puts "adding machine for sim $sim"

    set num_mcms [$sim display number_of_MCMs]

    if {$num_mcms == 1} {
        treesim::add_mcm_single $widget $sim $simidx5 0
    } else {
        $treesim::cmds(add_mcm_multiple) $widget $sim $simidx5 $num_mcms "MCM"
    }

    treesim::add_machine_entries $widget $sim $simidx5

    return "";
}

#------------------------------------------------------------------
#  Add more than 1 mcms to tree
#------------------------------------------------------------------
proc treesim::add_mcm_multiple { widget sim simidx6 num_mcms mcmname } {

    #puts "adding multiple ($num_mcms) mcms"

    set mcm_idxs [list]
    for {set mcm 0} {$mcm < $num_mcms} {incr mcm} {
        #puts "adding MCM $mcm"

        set mcmidx [$widget insert -at $simidx6 $mcm ${mcmname}_$mcm]
        set dataname [initdata $widget $mcmidx $simidx6]
        adddata $dataname treeidx $mcmidx

        lappend mcm_idxs $mcmidx

        # if mcm exists, add it
        treesim::add_mcm_single $widget $sim $mcmidx $mcm
    }

    set dataname [treedataname $widget $simidx6]
    adddata $dataname mcms $mcm_idxs

    $widget open $simidx6
}

#------------------------------------------------------------------
#  Add new mcm to tree
#------------------------------------------------------------------
proc treesim::add_mcm_single {widget sim parentidx mcmnum} {

    #puts "adding mcm for $widget $sim $parentidx $mcmnum ok"


    $widget tag add "MCM" $parentidx
    set mcm_cmd [alias create_unique $sim mcm $mcmnum ]
    set dataname [treedataname $widget $parentidx]
    adddata $dataname mcmcmd $mcm_cmd mcmnum $mcmnum

    set cpu_idxs [treesim::add_cpus $widget $mcm_cmd $parentidx]

    treesim::add_mcm_entries $widget $sim $parentidx

    adddata $dataname cpus $cpu_idxs
    $widget open $parentidx

    return
}

#------------------------------------------------------------------
#
#------------------------------------------------------------------
proc treesim::add_sim_entries {widget simidx7} {
    set sim_entries [ treesim::get_sim_default_entries ]
    add_children $widget $simidx7 $sim_entries
}

#------------------------------------------------------------------
#
#------------------------------------------------------------------
proc treesim::add_machine_entries {widget sim machineidx} {
    set machine_entries [ treesim::get_machine_default_entries $sim]
    add_children $widget $machineidx $machine_entries
}

#------------------------------------------------------------------
#
#------------------------------------------------------------------
proc treesim::add_mcm_entries {widget sim mcmidx} {
    set mcm_entries [ treesim::get_mcm_default_entries $sim ]
    add_children $widget $mcmidx $mcm_entries
}

#------------------------------------------------------------------
# refresh simulator tree view
#------------------------------------------------------------------
proc treesim::refresh_sim {widget sim simidx} {
    #puts "refreshing sim $sim"

    # find the mcms under this simulator
    set myTree [$widget cget -tree]
    set mcms [$myTree find $simidx -tag MCM ]

       foreach mcmidx $mcms {
          treesim::refresh_mcm $widget $mcmidx
       }
    }

#------------------------------------------------------------------
# refresh mcm tree view
#------------------------------------------------------------------
proc treesim::refresh_mcm {widget mcmidx} {
    #puts "refreshing mcm $mcmidx"

    array set current [getdata $widget $mcmidx]

    if {[array names current cpus] == ""} {
       return;
    }

    # Now run through and color the processors as active (green), waiting (yellow), 
    # not active (red)
    # First color all the entries red for inactive
    foreach cpuidx $current(cpus) {
        $widget entry configure $cpuidx -foreground darkred
        set cpu [$widget entry cget $cpuidx -label]
        set cpu_idx($cpu)    $cpuidx               ;# keep array of indexes by name
        set cpuname($cpuidx) $cpu                  ;# keep array of names by index
    }

    # Now color all the active entries green
    set activelist [$current(mcmcmd) display runlist]
    foreach cpu $activelist {
        if {[llength [array names cpu_idx $cpu]] != 0} {
            $widget entry configure $cpu_idx($cpu) -foreground darkgreen
        } else {
            puts "treesim::refresh_mcm -- WARNING: runlist $cpu not found in cpu_idx"
        }
    }

    # Now color all the stalled entries yellow
    set waitlist [$current(mcmcmd) display waitlist]
    foreach cpu $waitlist {
        if {[llength [array names cpu_idx $cpu]] != 0} {
            $widget entry configure $cpu_idx($cpu) -foreground darkorange
        } else {
            puts "treesim::refresh_mcm -- WARNING: waitlist $cpu not found in cpu_idx"
        }
    }

    # Now let any open cpus refresh
    foreach cpuidx $current(cpus) {
        if {[$widget entry isopen $cpuidx]} {
            $treesim::cmds(refresh_cpu) $widget $cpuidx $cpuname($cpuidx)
        }
    }
}

proc treesim::treedataname {widget idx} {
  return ::treesim::data($widget.$idx)
}

#------------------------------------------------------------------
# get data for given idx as a list
#------------------------------------------------------------------
proc treesim::getdata {widget idx} {
    set dataname [treedataname $widget $idx]
    return [set $dataname]
}

#------------------------------------------------------------------
# add data to idx list
#------------------------------------------------------------------
proc treesim::adddata {dataname args } {
    eval lappend $dataname $args        ;# add input list to data list
    array set current [set $dataname]   ;# convert data from list to array
    set $dataname [array get current]   ;# and back to a list to remove duplicates
}

#------------------------------------------------------------------
# init current idx's data with parent's data
# each entry usually copies its parent's data.
# if there are data values not needed by the children, add the data
# the children don't need after the children are created
#------------------------------------------------------------------
proc treesim::initdata {widget idx {parentidx {}} } {

    set dataname [treedataname $widget $idx]

    if {$parentidx == ""} {
        set $dataname [list widget $widget]           ;# init with widget name
    } else {
        set $dataname  [getdata $widget $parentidx]   ;# init with parent info
    }
    return $dataname
}

#------------------------------------------------------------------
# add all cpus for this sim to the tree
#------------------------------------------------------------------
proc treesim::add_cpus {widget mcm parentidx} {

    #puts "add cpus for mcm $mcm idx $parentidx "

    set cpuentries {}
    set insertpos 0

    # add all cpus
    foreach cpu [$mcm display cpulist] {
        set idx [$widget insert -at $parentidx $insertpos $cpu]
        incr insertpos
        $widget tag add "CPU" $idx

        # remember some stuff for this cpu
        set dataname  [initdata $widget $idx $parentidx ]  ;# init with parent info
        adddata $dataname treeidx $idx name $cpu

        $widget entry configure $idx -button on \
                                     -opencommand "treesim::open_cpu $widget %# $cpu"

        lappend cpuentries $idx
    }

    return $cpuentries
}

#------------------------------------------------------------------
# This is called to do a "lazy" add of cpu entries
#------------------------------------------------------------------
proc treesim::open_cpu {widget cpuidx cpu} {

    #puts "open cpu for cpu $cpu idx $cpuidx "

    # redirect any more open commands to the refresh command
    $widget entry configure $cpuidx -opencommand [list   \
            $treesim::cmds(refresh_cpu) $widget $cpuidx $cpu \
            ]

    $treesim::cmds(add_cpu_entries) $widget $cpuidx $cpu
}



#------------------------------------------------------------------
# Add entries under PPC node
#
# parentcmd is a tcl command that will take CPU as an argument
#                such as 'mysim' or an alias
#------------------------------------------------------------------
proc treesim::add_ppc_entries {widget cpuidx cpu parentcmd } {

    set dataname [treedataname $widget $cpuidx]

    set cpunum [string range $cpu 3 end ]   ;# assumes something like PPC0:1
    set cpucmd [alias create_unique $parentcmd cpu $cpunum]

    adddata $dataname cpuname $cpu cpucmd $cpucmd cpunum $cpunum name $cpu

    set ppc_entries [ get_ppc_default_entries ]
    add_children $widget $cpuidx $ppc_entries
}

#------------------------------------------------------------------
# default cpu adder - assumes only PPC cpus
#------------------------------------------------------------------
proc treesim::add_cpu_entries {widget cpuidx cpu} {
    # read current data to get mcmcmd
    array set current [getdata $widget $cpuidx]
    set parentcmd  $current(mcmcmd)
    add_ppc_entries $widget $cpuidx $cpu $parentcmd
}


#------------------------------------------------------------------
# default cpu refresh - assumes only PPC cpus
#------------------------------------------------------------------
proc treesim::refresh_cpu {widget cpuidx cpu} {
}

#------------------------------------------------------------------
#  add 'child' nodes for sim or CPUs
#------------------------------------------------------------------
proc treesim::add_children {widget parentidx nameandcmdlist} {

    #puts "add children idx $parentidx "

    foreach {name cmdtype opencmd} $nameandcmdlist {
       eval add_child $widget $parentidx $name $cmdtype $opencmd
    }
}

#------------------------------------------------------------------
#  add 'sibling' nodes
#------------------------------------------------------------------
proc treesim::add_sibling { widget sibling_idx args } {
   set nameandcmdlist $args
   #puts "add sibling of idx $sibling_idx , command $nameandcmdlist"

   set parentidx [$widget index -at $sibling_idx parent ]
   set entrycnt  [$widget entry size $parentidx ]
   set insertpos [expr $entrycnt - 1 ]

   foreach {name cmdtype opencmd} $nameandcmdlist {
      # if the name has a %n in it, convert it to a number
      set percent_n [string first "%n" $name]
      if {$percent_n != -1} {
         set name [string range $name 0 [expr $percent_n - 1]]$insertpos[string range $name [expr $percent_n+2] end ]
      }
      set idx [$widget insert -at $parentidx $insertpos $name]
      incr insertpos
      eval init_node $widget $idx $sibling_idx $name $cmdtype $opencmd $args
   }
   return $idx
}

#------------------------------------------------------------------
#  add 'child' node
#------------------------------------------------------------------
proc treesim::add_child {widget parentidx name cmdtype opencmd args } {

    #puts "add child idx $parentidx "
    set idx [$widget insert -at $parentidx end $name]
    eval init_node $widget $idx $parentidx $name $cmdtype $opencmd $args

    return $idx
}

#------------------------------------------------------------------
#  init node
#------------------------------------------------------------------
proc treesim::init_node {widget idx template_idx name cmdtype opencmd args } {

    # remember some stuff for this node
    set dataname  [initdata $widget $idx $template_idx ]    ;# init with parent/sibling info

    set name [lindex [split $name "/"] end]                 ;# ignore any prefix path entries (eg, SystemMemory)

    adddata $dataname treeidx $idx cmdtype $cmdtype name $name

    array set current [set $dataname]                       ;# convert data from list to array
    set $dataname [array get current]                       ;# and back to a list to remove duplicates

    $widget entry configure $idx -opencommand "treesim::open_child $widget %# $opencmd $args"

    return $idx
}

#------------------------------------------------------------------
# Call child's opencmd.  If child is managed, only allow 1 window per child
# and keep track of the open windows for calling refresh.
#
# If not managed, just call opencmd and pass in current data array with args
#------------------------------------------------------------------
proc treesim::open_child {widget idx opencmd args } {

    #puts "open child for idx $idx "

    array set current [getdata $widget $idx]


    switch $current(cmdtype) {
        "managed" {
            set path [get_path_name $widget $idx ]
            set win [winmgr::open_managed_window $path $idx $opencmd current $args]
            # save window name in tree data
            $widget entry configure $idx -data [list winname $win]
        }
        "add2tree" {
            set newidx [eval add_sibling $widget $idx $opencmd $args ]
            $widget open $newidx
        }
        default {
            # call simple open command (not managed)
            eval $opencmd current $args
        }
    }

    # once an entry is 'open' the opencommand has no effect.
    # close the entry so the user doesn't have to double-click twice to reopen it
    $widget close $idx
}

#------------------------------------------------------------------
# remove child windows for a deleted sim
#------------------------------------------------------------------
proc treesim::remove_child_windows {widget simidx} {

    # search tree under simidx for windows that have been opened
    set myTree [$widget cget -tree]
    set winidxs [$myTree find $simidx -glob * -key winname  ]

    foreach idx $winidxs {
        array set windata [$widget entry cget $idx -data ]

        if { [winfo exists $windata(winname)] } {
           destroy $windata(winname)
           unset [treedataname $widget $idx]
        }
    }
}
