#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : buffer.tcl
##
## Purpose    : TVU Buffer GUI constructs
##
## CVS INFO:
##
## $Id: buffer.tcl,v 1.1 2006/03/05 03:18:10 mkistler Exp $
##
#####################################################
#####################################################

namespace eval TVU_Buffer {
    variable version        "1.0"
    variable declared
    variable data
}

package provide TVU_Buffer $TVU_Buffer::version
package require TVU_Cell

####
# Create a new buffer widget
#
# \param    parent      Parent widget
# \param    lay_name    Name of layout
# \param    name        Name of this buffer (must be unique)
# \param    tobj        TVU object to associate
# \param    loc         TVU location to associate (Widget title)
# \param    slots       Number of slots in the buffer
#
proc TVU_Buffer::Create {parent lay_name name tobj loc slots} {
    set full_name ${lay_name}_${name}
    if {[llength [array names TVU_Buffer::declared $full_name]] != 0} {
        puts "TVU_Buffer::Create -- Buffer to create '$full_name' already exists"
        return -code error
    }

    # Frame for grouping purposes
    set frm [frame ${parent}.${full_name}]
    set flbl [label ${frm}.title -text "$loc"]

    set flst [frame ${frm}.list]
    set cvas [canvas ${flst}.cvas -yscrollcommand [list ${flst}.ybar set] \
                -width 475 -height 100  \
                -scrollregion [list 0 0 0 [expr $slots * 25]] \
                -relief raised -confine true -yscrollincrement 25]
    set ybar [scrollbar ${flst}.ybar -orient vertical -command [list $cvas yview]]
    grid $cvas $ybar -sticky news
    grid rowconfigure $flst 0 -weight 1
    grid columnconfigure $flst 0 -weight 1

    grid $flbl -sticky ew
    grid $flst -sticky news

    # Create the slots
    for {set i 0} {$i < $slots} {incr i 1} {
        set i_frm [TVU_Cell::Create $cvas $lay_name $full_name $i $i $tobj [list $loc $i]]
        $cvas create window 0 [expr $i * 25 + 10] -anchor w -window $i_frm
    }

    # Save relevant data
    set TVU_Buffer::declared($full_name) 1
    set TVU_Buffer::data(tvu,$full_name) $tobj
    set TVU_Buffer::data(parent,$full_name) $parent
    set TVU_Buffer::data(widget,$full_name) $frm
    set TVU_Buffer::data(loc,$full_name) $loc
    set TVU_Buffer::data(slots,$full_name) $slots

    return $frm
}

####
# Create a new buffer widget by duplicating a previously created widget
#
# \param    pname   Name of the parent buffer
# \param    name    Name of this buffer (must be unique)
# \param    dup     Name of the buffer to duplicate (must be unique)
# \param    loc     TVU location to associate (Widget title)
#
proc TVU_Buffer::Dup {pname name dup loc} {
    set full_name ${pname}_${name}
    set dup_name ${pname}_${dup}
    if {[llength [array names TVU_Buffer::declared $dup_name]] == 0} {
        puts "TVU_Buffer::Dup -- Buffer to duplicate '$dup_name' is unknown"
        return -code error
    }

    set frm [TVU_Buffer::Create $TVU_Buffer::data(parent,$dup_name) $pname $name $TVU_Buffer::data(tvu,$dup_name) $loc $TVU_Buffer::data(slots,$dup_name)]
    return $frm
}

####
# Clear a buffer widget
#
# \param    name    Name of the buffer to clear
#
proc TVU_Buffer::Clear {name} {
    # Clear the slots
    for {set i 0} {$i < $TVU_Buffer::data(slots,$name)} {incr i 1} {
        TVU_Cell::Clear $name $i
    }
}

####
# Clear all matching buffers
#
# \param    pat     Pattern to match buffer names
#
proc TVU_Buffer::ClearMatching {pat} {
    # Clear matching buffers
    foreach p [array names TVU_Buffer::declared $pat] {
        TVU_Buffer::Clear $p
    }
}
