#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : cell.tcl
##
## Purpose    : TVU Location cell GUI constructs
##
## CVS INFO:
##
## $Id: cell.tcl,v 1.1 2006/03/05 03:18:10 mkistler Exp $
##
#####################################################
#####################################################

namespace eval TVU_Cell {
    variable version        "1.0"
    variable declared
    variable data
    variable hilite
}

package provide TVU_Cell $TVU_Cell::version

####
# Perform the cell hilite operation
#
# \param    clist       Hilite color list {foreground background}
# \param    lay_name    Name of layout
# \param    trid        Transaction ID
#
proc TVU_Cell::SetHilite {clist lay_name trid} {
    set TVU_Cell::hilite($lay_name,$trid) $clist
}

####
# Return the current highlight
#
# \param    lay_name    Name of layout
# \param    trid        Transaction ID
#
proc TVU_Cell::GetHilite {lay_name trid} {
    set trid_list [split $trid .]
    for {set i 0} {$i < [llength $trid_list]} {incr i 1} {
        set t [join [lrange $trid_list 0 end-$i] .]
        if { [llength [array names TVU_Cell::hilite "$lay_name,$t"]] != 0 } {
            return $TVU_Cell::hilite($lay_name,$t)
        }
    }
    return [list white black]
}

####
# Perform the cell menu hilite operation
#
# \param    color       Hilite color
# \param    lay_name    Name of layout
# \param    full_name   Name of cell
#
proc TVU_Cell::MenuHilite {color lay_name full_name} {
    set trid $TVU_Cell::data(trid,$full_name)
    TVU_Cell::SetHilite [list $color black] $lay_name $trid
    $TVU_Cell::data(itxt,$full_name) configure -foreground $color -background black
}

####
# Create a new cell operation menu
#
# \param    parent      Parent widget
# \param    lay_name    Name of layout
# \param    full_name   Name of cell
#
proc TVU_Cell::MenuCreate {parent lay_name full_name} {
    set mnu [menu ${parent}.mnu]
    set color_list [list yellow orange red green cyan magenta brown]
    foreach color $color_list {
        $mnu add command -label Hilite -foreground $color -background black -command [list TVU_Cell::MenuHilite $color $lay_name $full_name]
    }
    $mnu add command -label None -foreground white -background black -command [list TVU_Cell::MenuHilite white $lay_name $full_name]
    return $mnu
}

####
# Create a new cell widget
#
# \param    parent      Parent widget
# \param    lay_name    Layout name (must be unique)
# \param    pname       Parent name (must be unique)
# \param    name        Name of this cell
# \param    title       Title of this cell
# \param    tobj        TVU object to associate with this cell
# \param    loc         TVU location list to associate with this cell
#
proc TVU_Cell::Create {parent lay_name pname name title tobj loc} {
    set lower_name [string tolower $name]
    set full_name ${pname}_${lower_name}
    if {[llength [array names TVU_Cell::declared $full_name]] != 0} {
        puts "TVU_Cell::Create -- Cell to create '$full_name' already exists"
        return -code error
    }

    # Frame for grouping purposes
    set frm [frame ${parent}.${full_name}]
    set lbl [label ${frm}.lbl -text "$title:" -width 8 -justify right -anchor e]

    set i_var ${full_name}_i
    set i_frm [frame ${frm}.ifrm -relief groove -borderwidth 2]
    set i_txt [label ${i_frm}.txt -textvariable $i_var -width 30 -justify left -anchor w -foreground white -background black]
    grid $i_txt

    set x_var ${full_name}_x
    set x_frm [frame ${frm}.xfrm -relief groove -borderwidth 2]
    set x_txt [label ${x_frm}.txt -textvariable $x_var -width 30 -justify left -anchor w -foreground white -background black]
    grid $x_txt

    grid $lbl $i_frm $x_frm -sticky ew

    # Set up the TVU object association
    set c_cmd [list TVU_Cell::Set $lay_name $full_name %T% $i_var \{%I%\} $x_var \{%X%\}]
    if { [catch {$tobj map def $loc $c_cmd} errs] } {
        puts "TVU_Cell::Create -- $errs"
    }

    # Create the selection menu if not already done
    set mnu [TVU_Cell::MenuCreate $i_txt $lay_name $full_name]

    # Set up button bindings
    bind $i_txt <Button-3> [list                 \
        tk_popup $mnu %X %Y                      \
    ]

    # Save relevant data
    set TVU_Cell::declared($full_name) 1
    set TVU_Cell::data(parent,$full_name) $parent
    set TVU_Cell::data(widget,$full_name) $frm
    set TVU_Cell::data(selmenu,$full_name) $mnu
    set TVU_Cell::data(ivar,$full_name) $i_var
    set TVU_Cell::data(itxt,$full_name) $i_txt
    set TVU_Cell::data(xvar,$full_name) $x_var
    set TVU_Cell::data(xtxt,$full_name) $x_txt
    set TVU_Cell::data(tvu,$full_name) $tobj
    set TVU_Cell::data(loc,$full_name) $loc
    set TVU_Cell::data(trid,$full_name) ""

    return $frm
}

####
# Set values in a cell widget
#
# \param    lay_name    Layout name
# \param    full_name   Name of cell
# \param    id          TRID string
# \param    ivar        Info field cell label variable
# \param    ivalue      Info field text
# \param    xvar        Text field cell label variable
# \param    xvalue      Text field text
#
proc TVU_Cell::Set {lay_name full_name id ivar ivalue xvar xvalue} {
    upvar #0 $ivar info_var
    upvar #0 $xvar text_var

    set info_var $ivalue
    set text_var $xvalue
    set TVU_Cell::data(trid,$full_name) $id
    set hlite [TVU_Cell::GetHilite $lay_name $id]
    $TVU_Cell::data(itxt,$full_name) configure -foreground [lindex $hlite 0] -background [lindex $hlite 1]
}

####
# Clear a cell widget
#
# \param    pname   Parent widget name
# \param    name    Cell widget name
#
proc TVU_Cell::Clear {pname name} {
    set lower_name [string tolower $name]
    set full_name ${pname}_${lower_name}
    if {[llength [array names TVU_Cell::declared $full_name]] == 0} {
        puts "TVU_Cell::Clear -- Cell to clear '$full_name' does not exist"
        return -code error
    }

    upvar #0 $TVU_Cell::data(ivar,$full_name) info_var
    upvar #0 $TVU_Cell::data(xvar,$full_name) text_var
    set info_var ""
    set text_var ""
    $TVU_Cell::data(itxt,$full_name) configure -foreground white -background black
}
