#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2005
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  window.tcl
#
#
#------------------------------------------------------------------------
namespace eval Window {
}

#--------------------------------------------------------------------------
# Window::init_menu
#
#--------------------------------------------------------------------------
proc Window::init_menu {m} {
    variable reopen

    set reopen [ get_pref open_windows/reopen_on_startup 1 ]
    $m add checkbutton -label "Reopen Windows on Restart" -command Window::_reopen_checkbox -variable Window::reopen
    $m add command -label "Close All Windows" -command Window::_close_all
    $m add command -label "Save Window Setup As..." -command Window::_get_save_name

    set m2 [menu $m.loadwincfg -tearoff 0 ]
    $m2 configure -postcommand "Window::_load_window_postcmd $m2"
    $m add cascade -menu $m2 -label "Load Window Setup"

    # only need to call this proc once
    $m configure -postcommand ""
}

#--------------------------------------------------------------------------
# Window::_reopen_checkbox
#   called when the menu checkbox changes
#--------------------------------------------------------------------------
proc Window::_reopen_checkbox {} {
    variable reopen
    set_pref open_windows/reopen_on_startup $reopen
}

#--------------------------------------------------------------------------
# Window::_close_all
#   close all open windows
#--------------------------------------------------------------------------
proc Window::_close_all {} {

   # close all the windows owned by the manager
   winmgr::close_windows

   # close any other toplevel windows
   foreach win [winfo children .] {
       switch $win {
           .mambo    -
           .menubar  -
           .#menubar {} ;# leave these !

           default {
               destroy $win
           }
       }
   }
}

#--------------------------------------------------------------------------
# Window::_get_save_name
#   open dialog to get a name for saving the current window configuration
#--------------------------------------------------------------------------
proc Window::_get_save_name {} {
   variable save_info

   set dlg .windowsavedlg
   set saveentry $dlg.topf.name.entry

   set title "Save Window Configurations"
   set geom [ get_pref "win_geom/windowsavedialog" "none"  ]

   # Create the dialog
   array set dlgInfo "top $dlg title {$title} geom $geom"
   if { [ Dialog::create dlgInfo -borderwidth 5 ] } then {

      set topf [frame $dlg.topf]

      set name [frame $topf.name -bd 2]

      label $name.label -text "Save Description"
      entry $name.entry -width 20 -relief sunken -textvariable Window::save_info(name)

      set Window::save_info(name) "New"

      pack $name.label $name.entry -side top -anchor w
      pack $name -padx 10 -anchor w

      frame $topf.buttons
      button $topf.buttons.set     -text "   Ok  " -command "set Window::save_info(trigger) set_pushed " -anchor w
      button $topf.buttons.cancel -text "  Cancel " -command "set Window::save_info(trigger) cancel_pushed"
      pack $topf.buttons.set $topf.buttons.cancel  -pady 5 -padx 10 -side left -fill x
      pack $topf.buttons -side top -padx 10 -pady 10

      pack $topf -fill both -pady 5 -padx 5 -side left -side top

      bind $saveentry <Return> "$topf.buttons.set invoke"
      bind $topf <Return> "$topf.buttons.set invoke"
      bind $topf <Escape> "$topf.buttons.cancel invoke"
   }

   $saveentry selection range 0 end
   $saveentry icursor end

   # reset the trigger variable
   set Window::save_info(trigger) "reset"

   # Save settings in case user cancels later
   array set saved_window_save_info [array get Window::save_info]

   # Wait for user to press the Ok or Cancel button, which will alter the trigger variable.
   # Set the text entry widget in focus
   Dialog::wait $dlg Window::save_info(trigger) $saveentry

   set_pref "win_geom/windowsavedialog" [winfo geometry $dlg ] 

   # Close the dialog box
   Dialog::dismiss $dlg

   if { $Window::save_info(trigger) == "set_pushed" } then {
       Window::_save_setup $save_info(name)
   } else {
       # restore original settings
       array set Window::save_info  [array get saved_window_save_info]
   }
}

#------------------------------------------------------------------
# Window::_save_setup
#   use name to save open windows paths and geometries
#------------------------------------------------------------------
proc Window::_save_setup {name} {

   set openpaths [ winmgr::get_open_window_paths]
   set openinfo [ winmgr::get_open_window_info ]

   # openwins contains both window name path info
   foreach {win path} $openinfo {
       set_pref "$name/win_geom/$path" [ winfo geometry $win ]
   }
   # add name to list of saved setups
   set prevsaved [get_pref "open_windows/saved_window_names" "" ]
   set idx [lsearch $prevsaved $name]
   if {$idx != -1} {
       set prevsaved [lreplace $prevsaved $idx $idx]    ;# if it already exists, remove it
   }
   set_pref "open_windows/saved_window_names" [linsert $prevsaved 0 $name ]
}

#---------------------------------------------------------
#  Window::_load_window_postcmd
#    
#---------------------------------------------------------
proc Window::_load_window_postcmd  {menu} {

   set prevsaved [get_pref "open_windows/saved_window_names" "" ]
   $menu delete 0 end
   foreach name $prevsaved {
       $menu add command -label "$name" -command "Window::_load_setup $name" 
   }
}

#---------------------------------------------------------
# Window::_load_setup
#     Reopen windows using the saved window configuration
#---------------------------------------------------------
proc Window::_load_setup {name} {
   set saved_paths [PrefManager::names $name/win_geom/* ]

   set pathidx [string length "$name/win_geom/" ]

   # get geometry settings for all the saved paths
   foreach path $saved_paths {
       set winpath [string range $path $pathidx end ]
       set reload($winpath) [get_pref $path "not_set" ]
   }

   # close saved windows that are already open
   set openinfo [ winmgr::get_open_window_info ]
   foreach {win path} $openinfo {
       if { [info exists reload($path)] } { 
           destroy $win
       }
   }

   set reload_list [list]
   # set window geometries using saved settings
   foreach path [array names reload] {
       set_pref "win_geom/$path" $reload($path)
       lappend reload_list $path
   }

   # reopen saved windows
   treesim::open_windows .mambo.simlist.tree $reload_list
}
