#------------------------------------------------------------------------
#
#	IBM Software
#	Licensed Material - Property of	IBM
#	(C) Copyright International Business Machines Corp. 2004
#	All Rights Reserved.
#
#	U.S. Government	Users Restricted Rights - Use, duplication or
#	disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#  winmgr.tcl
#
#
#------------------------------------------------------------------------

namespace eval winmgr {
    variable managed_window_list {}
    variable managed_window_array 
}

#------------------------------------------------------------------
# keep track of windows being opened
#------------------------------------------------------------------
proc winmgr::open_managed_window { path treeidx win_init_cmd _upcurrent args } {

    upvar $_upcurrent current 

	variable managed_window_list

    # create window name from init command 
    set winname .[ TclUtils::safename ${win_init_cmd}_$treeidx  ]


	# If the window exists, just raise it
	if { [lsearch -exact $managed_window_list $winname] != -1 } {
		raise $winname
		return
	}

    # create window and add it to the list
    toplevel $winname

    # call init command to open the window
    # return value is name of refresh command
    set refresh_cmd [eval $win_init_cmd  $winname current $args]

    if {$refresh_cmd != ""} { 
        add_managed_window $winname $refresh_cmd $treeidx $path
    } else {
        puts "Error: return value from $win_init_cmd should be the name of the refresh command."
    }

    # move and resize window per user
    set geom [ get_pref "win_geom/$path" "not_set"  ]
    if {$geom != "not_set"} { 
        wm geometry $winname $geom
    }

    return $winname
}

#------------------------------------------------------------------
# add already opened window to managed list
#------------------------------------------------------------------
proc winmgr::add_managed_window {winname refcmd treeidx path} {
	variable managed_window_list
    variable managed_window_array

    lappend managed_window_list $winname
    bind $winname <Destroy> "winmgr::close_managed_window $winname"
    bind $winname <Configure> "winmgr::_save_geometry $winname"
    set managed_window_array(refresh_cmd,$winname) $refcmd
    set managed_window_array(tree_idx,$winname) $treeidx
    set managed_window_array(path,$winname) $path
}

#------------------------------------------------------------------
# keep track of windows being closed
#------------------------------------------------------------------
proc winmgr::close_managed_window { name } {
	variable managed_window_list
    variable managed_window_array

    # find name in window list and remove it
    set idx [lsearch $managed_window_list $name]
    if {$idx != -1} {
        set managed_window_list [lreplace $managed_window_list $idx $idx]
        unset managed_window_array(refresh_cmd,$name)
        unset managed_window_array(tree_idx,$name)
        unset managed_window_array(path,$name)
    }
}

#------------------------------------------------------------------
# call refresh command for each managed window 
#------------------------------------------------------------------
proc winmgr::refresh_managed_windows { {force 0} } {
	variable managed_window_list
    variable managed_window_array

	foreach winname $managed_window_list {
        if {[winfo exists $winname]} { 
            if { [catch {$managed_window_array(refresh_cmd,$winname) $winname $force} refresh_error] } {

                puts "ERROR: Refresh of $winname got error of $refresh_error"
            }
        }
	}
}

#------------------------------------------------------------------------
# winmgr::_save_geometry
#------------------------------------------------------------------------
proc winmgr::_save_geometry { name } {
    variable managed_window_array

    set path $managed_window_array(path,$name)
    set_pref "win_geom/$path" [ winfo geometry $name ]
}

#------------------------------------------------------------------------
# winmgr::get_open_window_paths
#
# find the windows that are open
# return a list of the tree paths used when creating those windows
#------------------------------------------------------------------------
proc winmgr::get_open_window_paths { } {
    variable managed_window_array

    set open_tree_paths [list]

    set openwins [array names managed_window_array path,*]

	foreach idx $openwins {
        lappend open_tree_paths $managed_window_array($idx) 
	}

    return $open_tree_paths
}

#------------------------------------------------------------------------
# winmgr::get_open_window_info
#
# return a list of the windows and tree paths 
#------------------------------------------------------------------------
proc winmgr::get_open_window_info { } {
    variable managed_window_list
    variable managed_window_array

    set openwins [list]
	foreach winname $managed_window_list {
        lappend openwins $winname $managed_window_array(path,$winname)
	}
    return $openwins
}


#------------------------------------------------------------------------
# winmgr::close_windows
#
# close all managed windows 
#------------------------------------------------------------------------
proc winmgr::close_windows {} {
	variable managed_window_list

	foreach winname $managed_window_list {
        destroy $winname
	}
}


