#------------------------------------------------------------------------
#
#   IBM Software
#   Licensed Material - Property of IBM
#   (C) Copyright International Business Machines Corp. 2005
#   All Rights Reserved.
#
#   U.S. Government Users Restricted Rights - Use, duplication or
#   disclosure restricted by GSA ADP Schedule Contract with
#   IBM Corp.
#
#------------------------------------------------------------------------
#   linux.tcl
#------------------------------------------------------------------------

namespace eval Linux_Utils {
    variable version            "1.0"
    variable vmlinux_log_line   ""
}

package provide Linux_Utils $Linux_Utils::version

# ----------------------------------------------------------------------
# advance simulator cycles to align with next epoch.
# epoch value must be a power of 10
# ----------------------------------------------------------------------
proc cycle_align { epoch {sim mysim} {mcm 0} } {
   set curcycle  [$sim display cycles ]
   set epochlen  [expr [string length $epoch ] - 2 ]
   set curoffset [string range $curcycle end-$epochlen end ]      ;# get least signficant digits
   set curoffset [string trimleft $curoffset 0 ]                  ;# get rid of leading 0's
   set delaycycle [expr $epoch - $curoffset ]

   if {[simstop status] == "not-stopped"} {
       $sim mcm $mcm trigger set cycle [add64 $curcycle $delaycycle] { simstop }
   } else {
       $sim cycle $delaycycle
   }
}

proc Linux_showed_prompt { sim mcm args } {

    array set triginfo $args

    # get rid of trigger
    $sim mcm $mcm trigger clear console $triginfo(match)

    simstop
}

proc Linux_init_done { sim mcm args } {

    array set triginfo $args

    # get rid of trigger
    $sim mcm $mcm trigger clear console $triginfo(match)

    $sim mcm $mcm trigger set console "#" "Linux_showed_prompt $sim $mcm"
}

proc Fast_Linux_init_done { sim mcm args } {
    eval Linux_init_done $sim $mcm $args
    $sim set fast off
}

proc wait_for_linux_prompt { { mcm 0 } { sim mysim } } {

    $sim mcm $mcm trigger set console "INIT: Entering runlevel" "Linux_init_done $sim $mcm"
    $sim go
}

proc wait_for_linux_pound_sign { { mcm 0 } { sim mysim } } {
    wait_for_linux_prompt $mcm $sim
}

proc wait_for_just_linux_pound_sign { { mcm 0 } { sim mysim } } {

    $sim mcm $mcm trigger set console "#" "Linux_showed_prompt $sim $mcm"
    $sim go
}

proc fastest_wait_for_linux_prompt { { mcm 0 } { sim mysim } } {
     $sim mcm $mcm trigger set console "INIT: Entering runlevel" "Fast_Linux_init_done $sim $mcm"
     $sim mode fastest
     $sim go
}

proc fast_wait_for_linux_prompt { { mcm 0 } { sim mysim } } {
     $sim mcm $mcm trigger set console "INIT: Entering runlevel" "Fast_Linux_init_done $sim $mcm"
     $sim set fast on
     $sim go
}

proc turbo_wait_for_linux_prompt { { mcm 0 } { sim mysim } } {
     $sim mcm $mcm trigger set console "INIT: Entering runlevel" "Fast_Linux_init_done $sim $mcm"
     $sim set turbo on
     $sim go
}

proc kernel_panic_called {sim args} {

    simstop
    if { [catch "array set arginfo $args" errs] == 0 } {
        set mcm $arginfo(mcm)
        set cpu $arginfo(cpu)
        set thread $arginfo(thread)
    } else {
        set mcm 0
        set cpu 0
        set thread 0
    }
    set r3 [$sim mcm $mcm cpu $cpu:$thread display gpr 3]
    set cycle [$sim display cycles]

    # Get string pointed to by R3
    if { [catch "set pa [$sim cpu $cpu:$thread dtranslate $r3]" errs] == 0 } {
        if { [catch "set panic_msg \"[$sim memory display $pa STRING]\"" errs] != 0 } {
            puts "errors : $errs"
        }
    }
    set pc [$sim mcm $mcm cpu $cpu:$thread display spr pc]
    set ra [get_imem_ra $sim $pc $cpu:$thread]
    puts "\nWARNING: MCM=$mcm CPU=$cpu:$thread PC=$pc: (real-addr:$ra) VMLINUX PANIC CALLED with MSG: ($panic_msg)\n"
    #mcm_display_gprs $sim $mcm $cpu
    simstop
}

set vmlinux_filename "vmlinux-not-loaded"

proc vmlinux_loaded {sim filename} {
    
    global vmlinux_filename 

    set vmlinux_filename $filename
    #
    #   We want to set a breakpoint on .panic
    #   Extract information on symbol from object file and set a breakpoint
    #
    set function_name .panic
    if { [catch "set entry_info [list [object address $vmlinux_filename $function_name]]" errs] == 0 } {
        if { [catch "array set func_info [list $entry_info]" errs] == 0 } {
            if { [info exists func_info(FUNC_START)] } {
                #eval "$sim trigger set pc $func_info(FUNC_START) \"kernel_panic_called $sim\""
            #} else {
            #    puts "Cannot set trigger for kernel panics -- symbol .panic not found"
            }
        } else {
            puts "vmlinux_loaded failed with $errs"
        }
    } else {
        puts "vmlinux_loaded object call failed with $errs"
    }
}


#
# 32-bit value to ascii
#
proc vmlinux_ascii32 { value } {
	set s [format "%c%c%c%c" [expr ( ($value >> 24) & 0xff )] \
	                         [expr ( ($value >> 16) & 0xff )] \
                             [expr ( ($value >> 8) & 0xff )] \
	                         [expr ( $value & 0xff )]]

    while {! [string is graph -failindex bad $s]} {
        set s [string replace $s $bad $bad "."]
    }
	return $s
}

#
# Display the accumulated vmlinux log line ($Linux_Utils::vmlinux_log_line)
#
proc vmlinux_log_dump {args} {
    if {[string length $Linux_Utils::vmlinux_log_line] != 0} {
        puts "LOG: '$Linux_Utils::vmlinux_log_line'"
        set Linux_Utils::vmlinux_log_line ""
    }
}

#
# Capture vmlinux kernel log calls in $Linux_Utils::vmlinux_log_line
#
proc vmlinux_log_capture {{cpu 0} {thread 0} {sim mysim}} {
    set gpr [$sim cpu $cpu thread $thread display gpr 3]
    set char [string range $gpr [expr [string length $gpr] - 2] end]
    if {[string equal $char "0A"]} {
        vmlinux_log_dump
    } else {
        set letter [string index [vmlinux_ascii32 $gpr] end]
        set Linux_Utils::vmlinux_log_line "${Linux_Utils::vmlinux_log_line}${letter}"
    } 
}

#
# Print kernel calls set up via vmlinux_entries (below). Treats ".emit_log_char" specially
# to show log strings
#
proc vmlinux_call_trigger {sim args} {
    set call [lindex $args 0]
    array set p [lindex $args 1]
    if {[string equal $call ".emit_log_char"]} {
        vmlinux_log_capture $p(cpu) $p(thread) $sim
    } else {
        if {[string equal $call ".__delay"]} {
            vmlinux_log_dump
        }
        puts -nonewline "KERNEL: $call, $p(address) {"
        for {set i 3} {$i < 6} {incr i} {
            set v [$sim cpu $p(cpu) thread $p(thread) display gpr $i]
            set ascii [vmlinux_ascii32 $v]
            puts -nonewline "R$i:$v|$ascii,"
        }
        puts "}"
    }
}

#
# Set PC triggers for all entry points in an objdump file given
# Recognizes lines like the following:
#
# c000000000020d00 <.__delay>:
#
proc vmlinux_trace {odfile {trig vmlinux_call_trigger} {sim mysim}} {
    if {[catch {open $odfile r} chan]} {
        puts "vmlinux_trace: Can't open '$odfile' for reading"
        return -code error
    }
    puts "vmlinux_trace: Scanning file '$odfile'..."
    
    set trig_count 0
    while {[gets $chan line] >= 0} {
        if {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+<([^>]+)>:} $line match addr entry_pt]} {
            $sim trigger set pc 0x$addr [list $trig $sim "$entry_pt"]
            incr trig_count
        }
    }
    close $chan
    puts "vmlinux_trace: $trig_count entry points registered from '$odfile'"
}

#
# Print kernel log calls set up via vmlinux_log (below)
#
proc vmlinux_log_trigger {sim args} {
    array set p [lindex $args 0]
    vmlinux_log_capture $p(cpu) $p(thread) $sim
}

#
# Set PC triggers for the .emit_log_char entry points in an objdump file given
# Recognizes lines like the following:
#
# c000000000020d00 <.emit_log_char>:
# c000000000020d40 <.__delay>:
#
# Associates log dump with the delay entry point, in case we delay without
# seeing an end-of-line
#
proc vmlinux_log {odfile {sim mysim}} {
    if {[catch {open $odfile r} chan]} {
        puts "vmlinux_log: Can't open '$odfile' for reading"
        return -code error
    }
    puts "vmlinux_log: Scanning file '$odfile'..."

    while {[gets $chan line] >= 0} {
        if {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+<\.emit_log_char>:} $line match addr]} {
            $sim trigger set pc 0x$addr [list vmlinux_log_trigger $sim]
            puts "vmlinux_log: found .emit_log_char in '$odfile' at address $addr"
        }
        if {[regexp -nocase {^([cC][0-9a-fA-F]+)\s+<\.__delay>:} $line match addr]} {
            $sim trigger set pc 0x$addr [list vmlinux_log_dump]
            puts "vmlinux_log: found .__delay in '$odfile' at address $addr"
        }
    }
    close $chan
}

