##############################################
#
# Widget to allow a user to choose any
# number of items from a "source" list,
# adding them to or removing them from a "target"
# list.  The items in the target list can be
# rearranged as desired.
#
##############################################

option add *Entry.background white
option add *Listbox.background white

# Global variables

array set chooseList {
    callBack ""          ;# proc to invoke when done
    sourceList ""        ;# the list of items from which to choose
    targetList ""        ;# the list of items chosen
    axisList ""          ;# the list of axes (Left, Right) for chosen elements
    top                  ;# top level window for choose-from-list
}

#############################################
#
#  ChooseFromList -- construct the dialog window
#
#############################################

proc ChooseFromList {graph titleText fromList toList axisList callBack} {
    global chooseList

    set top $graph.top
    set chooseList(top) $top
    toplevel $top

    set chooseList(sourceList) $fromList
    set chooseList(targetList) $toList
    set chooseList(axisList)   $axisList
    set chooseList(callBack)   $callBack

    #############################################
    #   Title at top
    #############################################

    label $top.title -text "$titleText"

    #############################################
    #   "OK" and "Cancel" buttons
    #############################################

    frame $top.buttons
    button $top.ok -text OK \
	    -borderwidth 4 \
	    -background grey90 \
	    -activebackground grey95 \
	    -command {ReturnTargetList}
    button $top.cancel -text Cancel \
	    -command {$chooseList(callBack) {} {}; destroy $chooseList(top)}
    table $top.buttons \
	    0,0 $top.ok -padx {0 10} \
	    0,1 $top.cancel

    ##################################################
    #   Source listbox
    ##################################################

    frame $top.source
    label $top.source.label -text "Available:"
    listbox $top.source.list -width 30 -height 12 \
	    -xscrollcommand "$chooseList(top).source.xsb set" \
	    -yscrollcommand "$chooseList(top).source.ysb set" \
	    -selectmode extended
    scrollbar $top.source.xsb -orient horizontal \
	    -command "$chooseList(top).source.list xview"
    scrollbar $top.source.ysb -orient vertical \
	    -command "$chooseList(top).source.list yview"
    table $top.source \
	0,0 $top.source.label -anchor w -padx {15 0} \
	1,0 $top.source.list  -anchor w -padx {15 0} \
	1,1 $top.source.ysb   -anchor n -fill y \
	2,0 $top.source.xsb   -anchor e -padx {15 0} -fill x

    ##################################################
    #   Target listbox
    ##################################################

    frame $top.target
    label $top.target.label -text "Selected:"
    listbox $top.target.axes -width 5 -height 12 \
	    -yscrollcommand "$chooseList(top).target.ysb set" \
	    -selectmode single
    listbox $top.target.list -width 30 -height 12 \
	    -xscrollcommand "$chooseList(top).target.xsb set" \
	    -yscrollcommand "$chooseList(top).target.ysb set" \
	    -selectmode extended
    scrollbar $top.target.xsb -orient horizontal \
	    -command "$chooseList(top).target.list xview"
    scrollbar $top.target.ysb -orient vertical \
	    -command ScrollYTargetLists
    table $top.target \
	0,0 $top.target.label -anchor w -padx {15 0} \
	1,0 $top.target.axes  -anchor w -padx {35 0} \
	1,1 $top.target.list  -anchor w -padx {0 0}  \
	1,2 $top.target.ysb   -anchor n -fill y \
	2,1 $top.target.xsb   -anchor e -padx {0 0} -fill x

    bind $top <Key-Up>     {MoveTargetSelection up}
    bind $top <Key-Down>   {MoveTargetSelection down}
    bind $top <Key-Delete> {RemoveFromTarget}

    ##################################################
    #   Table arranging it all
    ##################################################

    table $top \
	0,0 $top.title -pady {5 15} -cspan 2 \
	1,0 $top.source -anchor nw \
	1,1 $top.target -anchor nw \
	2,0 $top.buttons -anchor w -padx {15} -pady {0 15}

    #table configure $top c2 r3 -resize none

    wm protocol $top WM_DELETE_WINDOW { destroy $top }

    #################################################
    #   Set the initial contents based on the
    #   global variables for source and target lists
    #################################################

    $top.source.list delete 0 end
    eval [concat $top.source.list insert 0 $chooseList(sourceList)]

    $top.target.axes delete 0 end
    eval [concat $top.target.axes insert 0 $chooseList(axisList)]

    $top.target.list delete 0 end
    eval [concat $top.target.list insert 0 $chooseList(targetList)]

    bind $top.source.list <Double-Button-1> {AddToTarget}
    bind $top.target.axes <Button-1> {ChangeAxis %y}
    bind $top.target.list <Double-Button-1> {RemoveFromTarget}
    bind $top <KeyPress-Return> {ReturnTargetList}
}

##################################################
#
#   ReturnTargetList
#
##################################################

proc ReturnTargetList {} {
    global chooseList

    set chooseList(axisList) [$chooseList(top).target.axes get 0 end]
    set chooseList(targetList) [$chooseList(top).target.list get 0 end]
    $chooseList(callBack) $chooseList(targetList) $chooseList(axisList)
    destroy $chooseList(top)
}

##################################################
#
#   Vertically scroll the target listboxes
#   (axes, column names) together
#
##################################################

proc ScrollYTargetLists {args} {
    global chooseList

    $chooseList(top).target.axes yview $args
    $chooseList(top).target.list yview $args
}


##################################################
#
#   Double-click on a source list element
#
#   Add that element to the end of the target list.
#   If any others are currently selected, add them too.
#   If any of the selected items are already in the
#   target list, just ignore them.
#
##################################################

proc AddToTarget {} {
    global chooseList

    set targetList [$chooseList(top).target.list get 0 end]

    foreach n [$chooseList(top).source.list curselection] {
	set elem [$chooseList(top).source.list get $n]
	if {[lsearch -exact $targetList $elem] < 0} {
	    $chooseList(top).target.axes insert end Left
	    $chooseList(top).target.list insert end $elem
	}
    }
}

##################################################
#
#   Click on an axis list element
#
#   Change from Left <--> Right
#
##################################################

proc ChangeAxis {ycoord} {
    global chooseList

    set sel [$chooseList(top).target.axes nearest $ycoord]
    set current [$chooseList(top).target.axes get $sel]
    if {$current == "Left"} {
	set new "Right"
    } else {
	set new "Left"
    }
    $chooseList(top).target.axes delete $sel $sel
    $chooseList(top).target.axes insert $sel $new
}

##################################################
#
#   Double-click on a target list element
#
#   Remove it (and any others currently selected)
#
##################################################

proc RemoveFromTarget {} {
    global chooseList

    set sel [lsort -decreasing [$chooseList(top).target.list curselection]]
    foreach n $sel {
	$chooseList(top).target.axes delete $n
	$chooseList(top).target.list delete $n
    }
}

##################################################
#
#   MoveTargetSelection  up or down
#
##################################################

proc MoveTargetSelection {dir} {
    global chooseList

    set v $chooseList(top).target.axes
    set w $chooseList(top).target.list

    set sel [$w curselection]

    # If nothing selected, just return

    if {$sel == ""} {
	return
    }

    # Make sure that the selected element(s) are contiguous

    set prev -1
    foreach n $sel {
	if {$prev < 0} {
	    set prev $n
	} else {
	    incr prev
	    if {$prev != $n} {
		return      ;# discontiguous
	    }
	}
    }

    if {$dir == "up"} {

	# Move a block of entries up one position by:
	#   Fetching the element just preceeding the block
	#   Deleting that element from the list
	#   Adding it back after the block

	set first [lindex $sel 0]
	if {$first == 0} {
	    return            ;# block is already at top
	}
	set len [llength $sel]
	set prev [expr $first - 1]
	set temp [$v get $prev]
	$v delete $prev
	$v insert [expr $first + $len - 1] $temp
	set temp [$w get $prev]
	$w delete $prev
	$w insert [expr $first + $len - 1] $temp

    } else {

	# Move a block of entries down one position by:
	#   Fetching the element just after the block
	#   Deleting that element from the list
	#   Adding it back before the block

	set first [lindex $sel 0]
	set len [llength $sel]
	set next [expr $first + $len]
	set end [$w index end]   ;# number of entries in list
	if {$next == $end} {
	    return             ;# block is already at end
	}
	set temp [$v get $next]
	$v delete $next
	$v insert $first $temp
	set temp [$w get $next]
	$w delete $next
	$w insert $first $temp
    }
}
