#!/usr/bin/wish

##############################################
#
# Widget to allow a user to specify
# the name of a file to be opened.
#
##############################################

option add *Entry.background white
option add *Listbox.background white

# Global variables

array set selectFile {
    callBack ""          ;# proc to invoke when done
    pattern "*"          ;# file names must match this
    currentDirectory {}  ;# currently displayed directory,
                          #   as a list of path components
    displayedFileName "" ;# displayed path, as modified by user
    top                  ;# top level window for file select
}

#############################################
#
#  GetFileName -- construct the dialog window
#
#############################################

proc GetFileName {titleText startingDir matchPattern callBack} {
    global selectFile

    set top .top
    set selectFile(top) $top
    toplevel $top

    #############################################
    #   Title at top
    #############################################

    label $top.title -text "$titleText"

    #############################################
    #   File name
    #############################################

    frame $top.name
    label $top.name.label -text "File:"
    entry $top.name.name -width 65 \
	    -textvariable selectFile(displayedFileName)
    table $top.name \
	    0,0 $top.name.label -anchor w -padx {15 15} \
	    1,0 $top.name.name  -anchor w -padx {15 15}
    bind $top.name.name <KeyPress-Return> \
	    {ProcessEnteredPathName}

    #############################################
    #   Match pattern
    #############################################

    frame $top.pattern -borderwidth 0
    label $top.pattern.label -text "Match pattern:"
    entry $top.pattern.pattern -width 15 \
	    -textvariable selectFile(pattern)
    table $top.pattern \
	    0,0 $top.pattern.label -anchor w \
	    1,0 $top.pattern.pattern -anchor w
    bind $top.pattern.pattern <KeyPress-Return> \
	    {RegenerateDisplay}

    #############################################
    #   "Open" and "Cancel" buttons
    #############################################

    frame $top.buttons
    button $top.open -text Open \
	    -borderwidth 4 \
	    -background grey90 \
	    -activebackground grey95 \
	    -command ProcessEnteredPathName
    button $top.cancel -text Cancel \
	    -command {
                destroy $selectFile(top)
	        $selectFile(callBack) {} {} {}
            }
    table $top.buttons \
	    0,0 $top.open -padx {0 10} \
	    0,1 $top.cancel

    ##################################################
    #   Directory listbox
    ##################################################

    frame $top.dirs
    label $top.dirs.label -text "Directories:"
    listbox $top.dirs.list -width 30 -height 12 \
	    -xscrollcommand "$selectFile(top).dirs.xsb set" \
	    -yscrollcommand "$selectFile(top).dirs.ysb set" \
	    -selectmode single
    scrollbar $top.dirs.xsb -orient horizontal \
	    -command "$selectFile(top).dirs.list xview"
    scrollbar $top.dirs.ysb -orient vertical \
	    -command "$selectFile(top).dirs.list yview"
    table $top.dirs \
	0,0 $top.dirs.label -anchor w -padx {15 0} \
	1,0 $top.dirs.list  -anchor w -padx {15 0} \
	1,1 $top.dirs.ysb   -anchor n -fill y \
	2,0 $top.dirs.xsb   -anchor e -padx {15 0} -fill x

    ##################################################
    #   Files listbox
    ##################################################

    frame $top.files
    label $top.files.label -text "Files:"
    listbox $top.files.list -width 30 -height 12 \
	    -xscrollcommand "$selectFile(top).files.xsb set" \
	    -yscrollcommand "$selectFile(top).files.ysb set" \
	    -selectmode single
    scrollbar $top.files.xsb -orient horizontal \
	    -command "$selectFile(top).files.list xview"
    scrollbar $top.files.ysb -orient vertical \
	    -command "$selectFile(top).files.list yview"
    table $top.files \
	0,0 $top.files.label -anchor w -padx {15 0} \
	1,0 $top.files.list  -anchor w -padx {15 0} \
	1,1 $top.files.ysb   -anchor n -fill y \
	2,0 $top.files.xsb   -anchor e -padx {15 0} -fill x

    ##################################################
    #   Table arranging it all
    ##################################################

    table $top \
	0,0 $top.title -pady {5 15} -cspan 2 \
	1,0 $top.name  -pady {0 15} -cspan 2 \
	2,0 $top.dirs -anchor nw \
	2,1 $top.files -anchor nw \
	3,1 $top.pattern -anchor nw -padx {15} \
	4,0 $top.buttons -anchor w -padx {15} -pady {0 15}

    #table configure $top c2 r3 -resize none

    wm protocol $top WM_DELETE_WINDOW { destroy $top }

    #################################################
    #   Set the initial contents based on the
    #   global variables for file name and pattern
    #################################################

    if [file isdirectory $startingDir] {
	set startDir $startingDir
    } else {
	set startDir [file dirname $startingDir]
    }
    set selectFile(pattern) $matchPattern
    set selectFile(callBack) $callBack

    # Resolve "../" and "./" entries in the starting path
    set startDir [exec /bin/sh -c "(cd $startDir; pwd)"]

    set selectFile(currentDirectory) [file split $startDir]
    set selectFile(displayedFileName) [AddTrailingSlash $startDir]

    PopulateDirectoryList $top.dirs.list $startDir
    PopulateFilesList $top.files.list $startDir $selectFile(pattern)

    bind $top.dirs.list <Double-Button-1> {SelectDirectory %y}
    bind $top.files.list <Double-Button-1> {SelectFile %y}
    bind $top <KeyPress-Return> {ProcessEnteredPathName}
    bind $top.files.list <ButtonRelease-1> +{SingleClickFile}
}

###############################################
#
#   Populate directory list
#
###############################################

proc PopulateDirectoryList {w directory} {

    # discard currently-displayed list
    $w delete 0 end

    # directory list starts with "..", unless at root
    if {$directory != "/"} {
	$w insert end "../"
    }

    # create new list of directories
    set files [lsort [glob -nocomplain $directory/*]]
    foreach i $files {
	if {[file isdirectory $i]} {
	    $w insert end "[file tail $i]/"
	}
    }
}

###############################################
#
#   Populate files list
#
###############################################

proc PopulateFilesList {w directory pattern} {

    # discard currently-displayed list
    $w delete 0 end

    # if user has deleted the pattern, show all files
    if {$pattern == ""} {
	set pattern "*"
    }

    # create new list of files
    set files [lsort [glob -nocomplain $directory/$pattern]]
    foreach i $files {
	if {[file isfile $i]} {
	    $w insert end [file tail $i]
	}
    }
}

##################################################
#
#   Double-click on a directory
#       -- Make that directory current in $top.dirs.list
#       -- Re-populate $top.files.list with its files
#
##################################################

proc SelectDirectory {ycoord} {
    global selectFile

    # Get the selected directory name; remove trailing "/"
    set num [$selectFile(top).dirs.list nearest $ycoord]
    set val [RemoveTrailingSlash [$selectFile(top).dirs.list get $num]]

    if {$val == ".."} {
	# Selected entry is ".." -- go up one level
	set selectFile(currentDirectory) \
		[lrange $selectFile(currentDirectory) 0 \
		[expr [llength $selectFile(currentDirectory)] - 2]]
    } else {
	# Selected entryis not ".." -- append it to the current path
	lappend selectFile(currentDirectory) $val
    }
    RegenerateDisplay
}

####################################################
#
#   Double-click on a file
#
#   Construct full path name, and pass it back.
#
####################################################

proc SelectFile {ycoord} {
    global selectFile

    # SingleClickFile has already processed the selected file
    # name and constructed the full path in
    # selectFile(displayedFileName).

    # All we have to do is invoke ProcessEnteredPathName

    ProcessEnteredPathName
}

###########################################################
#
#   Single-click on a file (in addition to normal binding)
#
#   Construct full path name, place it in the
#   file name entry widget
#
###########################################################

proc SingleClickFile {} {
    global selectFile

    flush stdout
    set sel [$selectFile(top).files.list curselection]
    if {$sel == ""} {return}
    set path [concat $selectFile(currentDirectory) \
	    [$selectFile(top).files.list get [lindex $sel 0]]]
    set selectFile(displayedFileName) \
	    [eval [concat file join $path]]
}

###################################################
#
#  Enter key pressed while in the full-path-name
#  entry widget.
#
#  If the path name specifies a file, return that as the
#  value specified.  If not, figure out which directory
#  is specified, and set the listboxes to display that
#  directory and its files.
#
###################################################

proc ProcessEnteredPathName {} {
    global selectFile

    set top $selectFile(top)
    set fullPath $selectFile(displayedFileName)
    set readable [file readable $fullPath]
    set isfile [file isfile $fullPath]

    if {$readable} {
	if {$isfile} {
	    set dirName [file dirname $fullPath]
	    destroy $selectFile(top)
	    $selectFile(callBack) $fullPath $dirName $selectFile(pattern)
	    return
	}
	set selectFile(currentDirectory) [file split $fullPath]
	RegenerateDisplay
    } else {
	ErrorDialog "File $fullPath not found." $top
    }
}

###################################################
#
#   Regenerate the display, after changes to
#     the controlling variables (pattern, etc.)
#
###################################################

proc RegenerateDisplay {} {
    global selectFile

    # Construct current directory path as a string
    set currentDirectory [AddTrailingSlash [eval \
	    [concat file join $selectFile(currentDirectory)]]]
    set selectFile(displayedFileName) $currentDirectory

    PopulateDirectoryList $selectFile(top).dirs.list \
	    $currentDirectory
    PopulateFilesList $selectFile(top).files.list \
	    $currentDirectory $selectFile(pattern)
}

#######################################################
#
#  Utility procs
#
#######################################################

proc RemoveTrailingSlash {name} {
    if {[string range $name end end] == "/"} {
	return [string range $name 0 \
		[expr [string length $name] - 2]]
    } else {
	return $name
    }
}

proc AddTrailingSlash {name} {
    if {[string range $name end end] == "/"} {
	return $name
    } else {
	return "$name/"
    }
}

