#####################################################
#####################################################
## Licensed Materials - Property of IBM.
## (C) Copyright IBM Corporation 2004
## All Rights Reserved.
##
## US Government Users Restricted Rights -
## Use, duplication or disclosure restricted by
## GSA ADP Schedule Contract with IBM Corporation.
##
## Filename   : Step.tcl
##
## Purpose    : Enhanced Mambo step functions
##
## CVS INFO:
##
## $Id$
##
#####################################################
#####################################################

namespace eval Step {
    variable version            "1.0"
    variable user_mode_icount   0
    variable user_mode_begin    0
    variable priv_mode_icount   0
    variable priv_mode_begin    0
    variable mode               0
    variable cpu                0
    variable thread             0
    variable trig_id
}

package provide Step $Step::version

####
# Return the current mode, given and msr value
#
proc Step::get_mode { msr } {
    if {[and64 $msr 0x4000]} {
        # User mode
        return 1
    } else {
        # Supervisor mode
        return 0
    }
}

####
# Set the msr change trigger
#
proc Step::msr_trigger { sim cpu thread args } {
    if {[llength [array names Step::trig_id $cpu,$thread]] == 0} {
        set Step::trig_id($cpu,$thread) [eval $sim cpu $cpu thread $thread trigger spr msr change $args]
    } else {
        eval $sim cpu $cpu thread $thread trigger spr msr id $Step::trig_id($cpu,$thread) change $args
    }
}

####
# Reset all icount state variables
#
proc Step::icount_reset { sim cpu thread } {
    if {($cpu != $Step::cpu) || ($thread != $Step::thread)} {
        Step::msr_trigger $sim $Step::cpu $Step::thread
        set Step::cpu $cpu
        set Step::thread $thread
    }

    set icount [$sim cpu $Step::cpu thread $Step::thread display instruction_count]
    set msr    [$sim cpu $Step::cpu thread $Step::thread display spr msr]

    set Step::user_mode_icount 0
    set Step::user_mode_begin $icount
    set Step::priv_mode_icount 0
    set Step::priv_mode_begin $icount
    set Step::mode [Step::get_mode $msr]
}

####
# Determine the privilege and user mode instruction
# counts made since the last call to this function.
# Accumulate total counts in user_mode_icount and
# priv_mode_icount
#
proc Step::icount_tally { sim msr } {
    set icount [$sim cpu $Step::cpu thread $Step::thread display instruction_count]
    set mode [Step::get_mode $msr]


    if {$Step::mode} {
        # Accumulate user mode count
        set Step::user_mode_icount [add64 $Step::user_mode_icount [sub64 $icount $Step::user_mode_begin]]
        set Step::user_mode_begin $icount
    } else {
        # Accumulate supervisor mode count
        set Step::priv_mode_icount [add64 $Step::priv_mode_icount [sub64 $icount $Step::priv_mode_begin]]
        set Step::priv_mode_begin $icount
    }

    if { $mode && !$Step::mode } {
        # Transition to user mode
        set Step::user_mode_begin $icount
    } elseif {!$mode && $Step::mode } {
        # Transition to supervisor mode
        set Step::priv_mode_begin $icount
    }

    set Step::mode $mode
}

####
# Display the current user and privilege mode icount
#
proc Step::icount_display { {sim mysim} } {
    Step::icount_tally $sim [$sim cpu $Step::cpu thread $Step::thread display spr msr]
    puts [format "*** ICOUNT: User=%d, Priv=%d" $Step::user_mode_icount $Step::priv_mode_icount]
}

####
# Callback function for msr register change trigger. This
# will in turn call $func
#
proc Step::priv_toggle { func args } {
    array set p $args
    if {[catch {eval $func $p(val)} result]} {
        puts "Step::priv_toggle: '$func $p(val)' could not be evaluated"
        puts "Step::priv_toggle: result was '$result'"
        exit
    }
}

####
# Simstop if the msr privilege mode matches the given value
#
proc Step::stop_on_mode { sim mode msr } {
    if {[Step::get_mode $msr] == $mode} {
        puts "Step::stop_on_mode: Reached mode $mode"
        simstop
    }
}

####
# Run until we are in the given mode
#
proc Step::run_until_mode { sim mode } {
    # If not already in desired mode...
    if {[Step::get_mode [$sim cpu $Step::cpu thread $Step::thread display spr msr]] != $mode} {
        Step::msr_trigger $sim $Step::cpu $Step::thread [list Step::priv_toggle [list Step::stop_on_mode $sim $mode]] 0x4000
        puts "Step::run_until_mode: Advancing cpu $Step::cpu thread $Step::thread to mode $mode..."
        $sim go
    }
}

####
# User-mode instruction step
#
proc Step::ustep { count {cpu 0} {thread 0} {sim mysim} } {
    Step::icount_reset $sim $cpu $thread

    while {[lessthan_u64 $Step::user_mode_icount $count]} {
        Step::run_until_mode $sim 1
        Step::icount_tally $sim [$sim cpu $Step::cpu thread $Step::thread display spr msr]

        if {[lessthan_u64 $Step::user_mode_icount $count]} {
            Step::msr_trigger $sim $cpu $thread [list Step::priv_toggle [list Step::icount_tally $sim]] 0x4000

            $sim cpu $Step::cpu thread $Step::thread step [sub64 $count $Step::user_mode_icount]
            Step::icount_tally $sim [$sim cpu $Step::cpu thread $Step::thread display spr msr]
        }
    }
    Step::icount_display $sim
}

####
# Privilege-mode instruction step
#
proc Step::kstep { count {cpu 0} {thread 0} {sim mysim} } {
    Step::icount_reset $sim $cpu $thread

    while {[lessthan_u64 $Step::priv_mode_icount $count]} {
        Step::run_until_mode $sim 0
        Step::icount_tally $sim [$sim cpu $Step::cpu thread $Step::thread display spr msr]

        if {[lessthan_u64 $Step::priv_mode_icount $count]} {
            Step::msr_trigger $sim $cpu $thread [list Step::priv_toggle [list Step::icount_tally $sim]] 0x4000

            $sim cpu $Step::cpu thread $Step::thread step [sub64 $count $Step::priv_mode_icount]
            Step::icount_tally $sim [$sim cpu $Step::cpu thread $Step::thread display spr msr]
        }
    }
    Step::icount_display $sim
}
