/* ********************************************************
 *
 * Licensed Materials - Property of IBM.
 * (C) Copyright IBM Corporation 2004, 2005
 * All Rights Reserved.
 *
 * US Government Users Restricted Rights -
 * Use, duplication or disclosure restricted by
 * GSA ADP Schedule Contract with IBM Corporation.
 *
 * Filename : seqdma.c
 *
 * Purpose  : Nanobenchmark to measure performance of sequential DMAs
 *
 * CVS INFO:
 *
 * $Id: daxpy.c,v 1.2 2006/02/14 05:26:49 mkistler Exp $
 *
 ******************************************************** */

#include <spu_mfcio.h>
#include "profile.h"

/* DMA size in bytes -- must be a multiple of 16, less than 16K */
#define DMASIZE 4096
#define DMASIZE_ELEMENTS (DMASIZE / sizeof(double) )
#define DMASIZE_QUADWORDS (DMASIZE >> 4)

/* Vector size in elements -- should be a multiple of (NSPUS*DMASIZE_ELEMENTS) */
#define VECSIZE_ELEMENTS (1L << 15) 
#define VECSIZE ((VECSIZE_ELEMENTS) * sizeof(double))

struct main_memory {
    /* Allocate main arrays(in global memory) */
    double x[VECSIZE_ELEMENTS] __attribute__ ((aligned (16)));
    double y[VECSIZE_ELEMENTS] __attribute__ ((aligned (16)));
};

/* Define base address for structures in main memory */
#define BASEPTR  ((struct main_memory *) 0x100000LL)

int main(void) {
    int i, j;
    int my_start, my_len;
    double a = 3.1415626;
    unsigned long xp, yp;
    unsigned int tag = 31, tid = 0, rid = 0;

    /* Local copy of chunks of real storage */
    double x_buf[DMASIZE_ELEMENTS] __attribute__ ((aligned (16)));
    double y_buf[DMASIZE_ELEMENTS] __attribute__ ((aligned (16)));

    my_start = 0;
    my_len = VECSIZE_ELEMENTS;

    xp = (unsigned long) (BASEPTR->x);
    yp = (unsigned long) (BASEPTR->y);

    prof_cp31();  /* Stop the perf stats */
    prof_cp0();   /* Clear the perf stats */
    for (i=0; i<my_len; i+=DMASIZE_ELEMENTS) {
        
        vector double *xv = (vector double *) x_buf, *yv = (vector double *) y_buf;
        vector double av = spu_splats( a );

        /* DMA read segments of x and y into local storage */
        mfc_get(x_buf, xp, DMASIZE, tag, tid, rid);
        mfc_get(y_buf, yp, DMASIZE, tag, tid, rid);
        mfc_write_tag_mask(1<<tag); /* Set tag mask */
        mfc_read_tag_status_all(); /* Wait for DMA to complete */

        prof_cp30();  /* Start the perf stats */
        /* Perform the computation: x = a*x + y */
        for (j=0; j<DMASIZE_QUADWORDS; j++) {
            xv[j] = spu_madd(xv[j], av, yv[j]);
        }
        prof_cp31();  /* Stop the perf stats */

        /* DMA write segment of x from local storage back to memory */
        mfc_put(x_buf, xp, DMASIZE, tag, tid, rid);
        mfc_write_tag_mask(1<<tag); /* Set tag mask */
        mfc_read_tag_status_all(); /* Wait for DMA to complete */

        xp += DMASIZE;
        yp += DMASIZE;
    }

    return 0;
}
